//////////////////////////////////////////////////////////////////
// (c) Copyright 2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Utilities/OhmmsInfo.h"
#include "Particle/ParticleSet.h"
#include "Message/Communicate.h"
#include "Particle/BulkSystemBuilder.h"
#include "Particle/NNEngine.h"
#include "Particle/NNRegGridEngine.h"
#include "ParticleIO/XMLParticleIO.h"
#include "Utilities/SpeciesCollection.h"
using namespace OHMMS;

/** example mdsim.cpp
 * a sample code to show how to use OhmmsParticle class
 *
 * The code does:
 * - create a set of Particle
 * - create Torque, a ParticleAttrib of 3-D vector type
 * - create MassTensor, a ParticleAttrib of 3-D tensor type
 * - initialize the configuration of a particle with command-line options
 * - evaluate the nearest-neighbor list with an interaction radius of 3 AA
 * - add an attribute and species to SpeciesCollection
 * - modify the particle attributes, Torque and MassTensor
 * - save the current configuration in a XML file 
 */
int main(int argc, char** argv) {

  Controller->initialize(argc,argv);
  OhmmsInfo ohmms(argc,argv,Controller->mycontext());

  if(argc<2) {
    CreateBulkSystem::printUsage(cout);
    return 1;
  }

  //====================================================
  //create a Particle Set
  //====================================================
  ParticleSet sys;

  //====================================================
  //create ParticleAttribute for torque
  //====================================================
  ParticleSet::ParticlePos_t Torque(ParticleTags::postype_tag,"torque");
  typedef ParticleSet::Tensor_t Tensor_t;
  typedef ParticleSet::Scalar_t Scalar_t;
  ParticleAttrib<Tensor_t> MassTensor(ParticleTags::tensortype_tag,"mz");
  sys.addAttribute(Torque);
  sys.addAttribute(MassTensor);

  //====================================================
  //initialize OhmmsParticle using the options
  //====================================================
  CreateBulkSystem::apply(&sys,argc,argv);
   
  //====================================================
  //get pair list with a negative interaction radius
  //====================================================
  sys.getPairlist(3.0);
  sys.getNNEngine()->get(cout);

  //====================================================
  //Get a reference to SpeciesBase to access mass
  //====================================================
  SpeciesBase& species = *(SpeciesCollection::getSpecies());

  //====================================================
  //Add C(arbon) and Si(licon) to SpeciesCollection
  //Typically, species are added by an input configuration file or interatomic potentials
  //====================================================
  SpeciesCollection::addSpecies("C");
  SpeciesCollection::addSpecies("Si");

  //save the index of an  attribute with the name of "mass"
  int imass = SpeciesCollection::addAttrib("mass");
  for(int i=0; i<species.getTotalNum(); i++) species(imass,i) = 12.0;

  ///assign a diagonal tensor as the inverse of mass
  for(int i=0; i<sys.getLocalNum(); i++) {
    Scalar_t massinv = 1.0/species(imass,sys.IonID[i]);
    MassTensor[i].diagonal(massinv);
  }

  ///Save the configuration of the particle set in a xml-file
  XMLSaveParticle ptclsave(sys);
  ptclsave.reset("bulk");
  ptclsave.report(0);

  return 0;
}  
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: mdsim.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
  
