//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#if defined(ENABLE_XMGR)
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <grace_np.h>
#endif
#include <map>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
using namespace std;

#include "Analysis/DynamicsAnalysis.h"
#include "Propagator/PropagatorBase.h"

namespace OHMMS {

  void AddDynamicsAnalysis(map<int,DynamicsAnalysis*>& m, int ana, 
			   int stride, bool useplot) {

    map<int,DynamicsAnalysis*>::iterator it = m.find(ana);

    if(it == m.end()) { // a newtool is needed
      DynamicsAnalysis* newtool  = NULL;
      switch(ana) {
      case(report_energies):
	newtool = new EnergiesVsTime(stride,useplot);
	break;
      case(report_gofr):
	newtool = new GofR(stride,useplot);
	break;
      default:
	break;
      }
      if(newtool) {
	m.insert(std::make_pair(ana,newtool));
      }
    } else {
      // overwrite the parameters of the existing plotting devices
      (*it).second->Stride = stride;
      (*it).second->UsePlot = useplot;
    }
  }

  /////////////////////////////////////////////////////////////
  // member functions of EnergiesVsTime
  /////////////////////////////////////////////////////////////
  void EnergiesVsTime::put(istream& ) {

  }

  bool EnergiesVsTime::Update(PropagatorBase* WorkingSystem, int iter) {

    //local static data: nobody knows this and can use
    static string fname;

    if(Stride>0) {

      //check if time to update
      if(Stride != 1 && iter%Stride != 0) return true;
#if defined(ENABLE_XMGR)
      if(UsePlot) {
	if(iter == 0) {
	  oneovern = 1.0/static_cast<double>(WorkingSystem->getPtcl()->getTotalNum());
	  // Start xmgr with a buffer size of 2048 and open the pipe 
	  if (GraceOpen(2048) == -1) {
	    fprintf (stderr, "Can't run xmgr. \n");
	    return EXIT_FAILURE;
	  }
	  // setting the page
	  GracePrintf("g0 on");
	  GracePrintf("with g0");
	  GracePrintf("view xmin 0.15");
	  GracePrintf("view xmax 0.85");
	  GracePrintf("view ymin 0.7");
	  GracePrintf("view ymax 0.925");
	  GracePrintf("subtitle \"Simulation of %s, Nions = %d\"",
		      WorkingSystem->getPtcl()->objName(), 
		      WorkingSystem->getPtcl()->getTotalNum());
	  GracePrintf("subtitle size 1.2");
	  GracePrintf("yaxis label \"Energy/particle (eV)\"");	  
	  GracePrintf("xaxis  ticklabel off");
	  GracePrintf("g1 on");
	  GracePrintf("with g1");
	  GracePrintf("view xmin 0.15");
	  GracePrintf("view xmax 0.85");
	  GracePrintf("view ymin 0.45");
	  GracePrintf("view ymax 0.675");
	  GracePrintf("yaxis  label \"Temperature (K)\"");	  
	  GracePrintf("xaxis  label \" Time (fsec)\"");
	  GracePrintf ("g0.s0 line linestyle 1");
	  GracePrintf ("g0.s0 line color 1");
	  GracePrintf ("g0.s0 line linewidth 2.0");
	  GracePrintf ("g0.s1 line linestyle 1");
	  GracePrintf ("g0.s1 line color 2");
	  GracePrintf ("g0.s1 line linewidth 2.0");
	  GracePrintf ("g0.s2 line linestyle 1");
	  GracePrintf ("g0.s2 line color 3");
	  GracePrintf ("g0.s2 line linewidth 2.0");
	  GracePrintf ("g1.s0 line linestyle 1");
	  GracePrintf ("g1.s0 line color 1");
	  GracePrintf ("g1.s0 line linewidth 2.0");
	} else {
	  stringstream s;
	  double x,y1,y2,y3,y4;
	  WorkingSystem->report(s);
	  s >> x >> y1 >> y2 >> y3 >> y4;
	  GracePrintf ("g0.s0 point %f, %f",x, y1*oneovern);
	  GracePrintf ("g0.s1 point %f, %f",x, y2*oneovern);
	  GracePrintf ("g0.s2 point %f, %f",x, y3*oneovern);
	  GracePrintf ("g1.s0 point %f, %f",x, y4);
	  GracePrintf("with g1");
	  GracePrintf ("autoscale");
	  GracePrintf ("redraw");
	  GracePrintf("with g0");
	  GracePrintf ("autoscale");
	  GracePrintf ("redraw");
	}
      }
#endif
      if(iter == 0) {
	fname = WorkingSystem->getPtcl()->objName();
	fname.append(".log");
	std::ofstream fout(fname.c_str());
	WorkingSystem->report(fout);
	fout.flush();
      } else {
	std::ofstream fout(fname.c_str(), ios::app);
	WorkingSystem->report(fout);
	fout.flush();
	}
    }
    return true;
  }

  /////////////////////////////////////////////////////////////
  // member functions of GofR
  /////////////////////////////////////////////////////////////
  void GofR::put(istream& ) {

  }

  double MaxRadius(const ParticleLayout_t& a) {
    SingleParticlePos_t u1(0.5,0.0,0.0);
    SingleParticlePos_t u2(0.0,0.5,0.0);
    SingleParticlePos_t u3(0.0,0.0,0.5);
    double rmax = a.Dot(u1,u1);
    rmax = max(rmax, a.Dot(u2,u2));
    rmax = max(rmax, a.Dot(u3,u3));
    rmax = max(rmax, a.Dot(u1,u2));
    rmax = max(rmax, a.Dot(u1,u3));
    rmax = max(rmax, a.Dot(u2,u3));
    return sqrt(3*rmax);
  }

  bool GofR::Update(PropagatorBase* WorkingSystem, int iter) {

    //local static data: nobody knows and can use

    static string fname;
    static double rmax;
    if(Stride>0) {
      if(Stride != 1 && iter%Stride != 0) return true;

      if(iter == 0) {
	rmax = MaxRadius(WorkingSystem->getPtcl()->Lattice);
      }

      ///////////////////////////////////
      //\todo Move to Particle Utilities
      //calculate g(r): very generic to OHMMS::ParticleBase
      ///////////////////////////////////
      const Particle_t& ptcl = *WorkingSystem->getPtcl();
      int n = 500;
      double dr = rmax/static_cast<double>(n);
      double drinv = 1.0/dr;
      vector<double> y(n+100,0.0);
      register int inunit = ptcl.R.InUnit;
      register SingleParticlePos_t drij;
      
      if(gofrsum.empty()) {
	gofrsum.resize(n+100,0.0);
      }

      if(ptcl.R.InUnit) {
	for(int i=0; i<ptcl.getTotalNum(); i++) {
	  for(int j=i+1; j<ptcl.getTotalNum(); j++) {
	    drij = ptcl.Lattice.BConds.apply(ptcl.R[i]-ptcl.R[j]);
	    double sep = sqrt(ptcl.Lattice.Dot(drij,drij));
	    y[static_cast<int>(sep*drinv)]+= 1;
	  }
	}
      } else {
	for(int i=0; i<ptcl.getTotalNum(); i++) {
	  for(int j=i+1; j<ptcl.getTotalNum(); j++) {
	    drij = ptcl.Lattice.BConds.apply(ptcl.Lattice.toUnit(ptcl.R[i]-ptcl.R[j]));
	    double sep = sqrt(ptcl.Lattice.Dot(drij,drij));
	    y[static_cast<int>(sep*drinv)]+= 1;
	  }
	}
      }
      //accumulate gofr for average
      for(int i=0; i<n; i++) {
	gofrsum[i] += y[i];	
      }

      ///////////////////////////////////
      //done calculating g(r)
      ///////////////////////////////////
      int num = iter/Stride; 
      double norm = 2.0/static_cast<double>(ptcl.getTotalNum());
      double avgfac = norm/static_cast<double>(num);

#if defined(ENABLE_XMGR)
      if(UsePlot) {//if useplot is used, only write to xmgr
	if(iter == 0) {
	  // Start xmgr with a buffer size of 2048 and open the pipe 
	  if(!GraceIsOpen()) {
	    if (GraceOpen(2048) == -1) {
	      fprintf (stderr, "Can't run xmgr. \n");
	      return EXIT_FAILURE;
	    }
	  }
	  // setting the page
	  GracePrintf("g2 on");
	  GracePrintf("with g2");
	  GracePrintf("view xmin 0.15");
	  GracePrintf("view xmax 0.85");
	  GracePrintf("view ymin 0.125");
	  GracePrintf("view ymax 0.375");
	  GracePrintf("yaxis label \"time averaged g(r) \"");
	  GracePrintf("xaxis label \"r A\"");	  
	} else {
	  double r0 = 0.0;
	  for(int i=0; i<n; i++) {
	    GracePrintf ("g2.s%d point %f, %f", num, r0, gofrsum[i]*avgfac);
	    r0 += dr;
	  }
	  GracePrintf("with g2");
	  GracePrintf ("autoscale");
	  GracePrintf ("redraw");
	}
      }
#endif
      if(iter == 0) {
	fname = WorkingSystem->getPtcl()->objName();
	fname.append(".gofr");
	std::ofstream fout(fname.c_str());
	fout << "# Pair-correlation Maximum radius rmax = " << rmax << " nbin " 
	     <<  n << endl;
      } else {
	std::ofstream fout(fname.c_str(), ios::app);
	double r0 = 0.0;
	for(int i=0; i<n; i++) {
	  fout<< r0 << " " << y[i]*norm << " " << gofrsum[i]*avgfac << endl;
	  r0 += dr;
	}
	fout << endl;
      }

    }
    return true;
  }

}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: DynamicsAnalysis.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
