//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifndef OHMMS_OHMMSELEMENTSET_H
#define OHMMS_OHMMSELEMENTSET_H

/*** A manager class of a set of OhmmsElementBase*'s
 *
 * It has OhmmsElementBase*'s which are mapped to a xml node and provides
 * name- or index-based getNode function to access each OhmmsElementBase.
 */
#include <map>
#include <string>
#include <iostream>
#include "OhmmsData/OhmmsElementBase.h"

namespace OHMMS {

  /*!\enum MainKeys 
   *\brief Enumerations to access the default set of OhmmsElementBase* contained in OhmmsElementSet::Nodes. 
   *
   * Five default OhmmsElementBases are added to the Nodes by the constructor.
   * The default Nodes, ProjectData, ParticleSets, WorkingPropagator, IonicPotential
   * and RandomNumberControl are simple in that they are independent of other Nodes.
   * They are responsible for creating basic objects that are used by higher-level
   * OhmmsElementBase types, such as MDSim.
   */
  enum {Key_project=0, //!< ProjectData
	Key_random,    //!< RandomNumberControl
	Key_particle,  //!< ParticleSets
	Key_ensemble,  //!< WorkingPropagator
	Key_potential, //!< IonicPotential
	Key_playout,   //!< ParticleLayout
  };

  class OhmmsElementSet: public OhmmsElementBase {

  public:

    typedef std::vector<OhmmsElementBase*>::size_type size_type;
    typedef std::map<std::string,size_type>           KeyMap_t;

    ///constructor
    OhmmsElementSet();

    ///destructor
    ~OhmmsElementSet();

    ///write to a ostream
    bool get(std::ostream& ) const;

    ///read from istream
    bool put(std::istream& );

    ///read from an xmlNode 
    bool put(xmlNodePtr cur);

    ///reset member data
    void reset();

    /** add an OhmmsElementBase* to a vector container.
     *@param o an OhmmsElementBase* to be added
     *
     * The node is added, if its name is not in conflict with the existing keys.
     * An OhmmsElementBase among Nodes can be accessed by an index or a name.
     */
    void addNode(OhmmsElementBase* o, int id=-1);

    /** get the pointer to a node with a name
     *@param aname a name/tag
     *@return OhmmsElementBase*, if aname is found among the Keys. 
     *
     *When no matching name is found among the Keys, NULL is returned
     */
    OhmmsElementBase* getNode(const std::string& aname);

    /** get the pointer to a node with an index
     *@param i the index of the node
     *@return i-th OhmmsElementBase if i is less than the size of the Nodes
     */
    inline OhmmsElementBase* getNode(size_type i) {
      if(i<Nodes.size()) 
	return Nodes[i];
      else 
	return NULL;
    }

    /** get the pointer to a node with an index
     *@param i the index of the node
     *@return i-th OhmmsElementBase if i is less than the size of the Nodes
     */
    inline OhmmsElementBase* operator[](size_type i) {
      return Nodes[i];
    }

    /// return the size of nodes
    inline size_type size() const { return Nodes.size();}

  private:

    /** Copy constructor
     * 
     * Disabled. No good will come out by allowing the copy of this object.
     */
    OhmmsElementSet(const OhmmsElementSet& ) {}


    /*** Keys connecting the name (xml tags) to an OhmmsElementBase* of Nodes.
     *
     * Instead of using map<string, OhmmsElementBase*>, an auxilary Keys are used.
     */
    KeyMap_t Keys;

    /** A list of OhmmsElementBase* that are statically added by the constructor.
     *
     * An application that owns this object.
     * stl::vector is used to enable random access to the Nodes element.
     */
    std::vector<OhmmsElementBase*> Nodes;       
  };
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: OhmmsElementSet.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
