//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <string>
#include <vector>
#include <deque>
#include <utility>
#include <iostream>
#include <fstream>
using namespace std;

#include "AppBase/ParticleSets.h"
#include "AppBase/TransformParticle.h"
#include "Particle/NNEngine.h"
#include "ParticleIO/ParticleLayoutIO.h"
#include "ParticleIO/ParticleParser.h"
#include "ParticleIO/RecordParticle.h"
#include "Utilities/OhmmsInfo.h"
#include "AppBase/ApplyConstraints.h"
namespace OHMMS {

  ParticleSets::ParticleSets(const char* aname): 
    OhmmsElementBase(aname), 
    Counter(0),
    ActivePtcl(NULL),
    m_cur(NULL),
    RestartConfig("xml")
  { 
  }

  void ParticleSets::reset() {

    if(m_cur) {
      xmlFreeNodeList(m_cur->xmlChildrenNode);
      m_cur->xmlChildrenNode = NULL;
      xmlNodePtr 
	anode =xmlNewChild(m_cur,m_cur->ns,(const xmlChar*)"particleset",NULL);
      string newfile(get()->getRoot());
      newfile.append(".restart.");
      newfile.append(RestartConfig);
      xmlNewProp(anode,(const xmlChar*)"src",(const xmlChar*)(newfile.c_str()));
    }
  }  

  ParticleSets::~ParticleSets() {
    PtclSetList_t::iterator it = NewPtclList.begin();    
    while(it != NewPtclList.end()) { delete (*it++);}
  }

  void 
  ParticleSets::setActivePtclSet(Particle_t* anew) {
    addPtclSet(anew);
    ActivePtcl = anew;
  }

  void ParticleSets::addPtclSet(Particle_t* anew) {
    std::map<string,Particle_t*>::iterator it = PtclCollection.find(anew->getName());
    if(it == PtclCollection.end()) {
      PtclCollection[anew->getName()] = anew;
    } else {
      if(anew != (*it).second) {
	WARNMSG("\tParticleCollection has the a particle set with the same name. Remove the exisiting one")
        delete (*it).second;
	(*it).second = anew;
      }
    }
  }

  ParticleSets::Particle_t* ParticleSets::get() {
    return ActivePtcl;
  }

  ParticleSets::Particle_t* ParticleSets::get(const string& aname) {
    std::map<string,Particle_t*>::iterator it = PtclCollection.find(aname);
    if(it == PtclCollection.end()) {
      //WARNMSG("\tParticleSets::get(" << aname <<") cannot find the particle set. Return ActivePtcl.") 
      return NULL;
    } else {
      return (*it).second;
    }
  }

  bool ParticleSets::remove(Particle_t* apt) {

    PtclSetList_t::iterator it = NewPtclList.begin();    
    while(apt && it != NewPtclList.end()) {
      if(apt == (*it)) {
	NewPtclList.erase(it);
	WARNMSG("\t" << apt->getName() << " is going to be destroyed.")
	delete apt; apt = NULL;
      }
      it++;
    }
    if(apt) { 
      ERRORMSG("\t" << apt->getName() << " cannot be removed.")
      return false;
    } else {
      return true;
    }
  }

  string ParticleSets::setPtclSetName() {
    char x[8];//always generate a name
    sprintf(x,"ptcl%d",Counter++); 
    return x;
  }

  void ParticleSets::clear() {
    if(ActivePtcl)  {

      PtclCollection.clear();
      PtclCollection[ActivePtcl->getName()] = ActivePtcl;

      PtclSetList_t::iterator it = NewPtclList.begin();    
      while(it != NewPtclList.end()) {
	if((*it) != ActivePtcl) {
	  delete (*it);
	} 
	it++;
      }
      NewPtclList.clear();
      NewPtclList.push_back(ActivePtcl);
    } else {
      ERRORMSG("\tNo ActivePtcl is set. Nothing is removed")
    }
  }

  void ParticleSets::clearAll() {
    PtclSetList_t::iterator it = NewPtclList.begin();    
    while(it != NewPtclList.end()) {
      delete (*it);it++;
    }
    NewPtclList.clear();     
    ActivePtcl = NULL;
  }

#if defined(HAVE_LIBXML2)
//   extern 
//   bool 
//   TransformParticleFactory(ParticleSets::ParticleSetList_t& pin,
// 		           ParticleSets::Particle_t& pout, 
// 			   xmlNodePtr cur);

  bool ParticleSets::put(xmlNodePtr cur) {

    m_cur = cur;
    cur = cur->xmlChildrenNode;
    ParticleLayout_t lat;
    bool gotalattice = false;

    vector<xmlNodePtr> builder, layout, freeze;
    vector<string> layout_target;
    xmlNodePtr record_ptr = NULL;
    while (cur != NULL) {

      string cname((const char*)(cur->name));
      ///get the /particlesets/[particleset|unitcell|transform|particlelayout]/@name
      //and use it as the particle name
      string pname("a");
      if(xmlHasProp(cur, (const xmlChar *) "name")) {
	pname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
      }
      //the particle to work on
      Particle_t* aptcl = get(pname);
      if(cname == "particleset") {
	if(!aptcl) {//need to create
	  aptcl = new Particle_t;
	  NewPtclList.push_back(aptcl);
	  pname = setPtclSetName();
	}
	ParticleParser aHandle(*aptcl);
	aHandle.put(cur);
	addPtclSet(aptcl);
	LOGMSG("\tThe active ParticleSet is set to " << pname)
	ActivePtcl = aptcl;
      } else if (cname == "UnitCell" || cname == "unitcell") {
	LatticeParser lat(Lattice);
	lat.put(cur);
	gotalattice =true;
      } else if(cname == "transform" || cname == "TransformParticle") {
	builder.push_back(cur);
      } else if(cname == "particlelayout") {
        ERRORMSG("particleset/particlelayout is invalid")
        //layout.push_back(cur);        
	//layout_target.push_back(pname);
      } else if(cname == "record") {
	record_ptr = cur;
      } else if(cname == "freeze") {
	freeze.push_back(cur);
      }
      cur = cur->next;
    }
    
    if(gotalattice) {
      PtclSetMap_t::iterator it = PtclCollection.begin();
      while(it != PtclCollection.end()) {
	WARNMSG("\tGlobal unit cell is set to " << (*it).second->getName())
	(*it).second->Lattice = Lattice; it++;
      }
    }

    ///transformation is required
    vector<xmlNodePtr>::iterator cur_ptr = builder.begin();
    while(cur_ptr != builder.end()) {
      transformParticle(*cur_ptr); cur_ptr++;
    }

    /** ParticleSet does not handle the layout directly.
    ///set the layout
    cur_ptr = layout.begin();
    int ilayout=0;
    if(cur_ptr != layout.end()) {
      if(layout_target[ilayout] == "a") {//default, active particle set
	if(ActivePtcl) {
	  ActivePtcl->setNNEngine(NNEngineBuilder::create(*cur_ptr));
	}
      } else if(layout_target[ilayout] == "all") {
	PtclSetMap_t::iterator it = PtclCollection.begin();
	while(it != PtclCollection.end()) {
	  (*it).second->setNNEngine(NNEngineBuilder::create(*cur_ptr));
	}
      } else {
	Particle_t* aptcl = get(layout_target[ilayout]);
	if(aptcl) 
	  aptcl->setNNEngine(NNEngineBuilder::create(*cur_ptr));
      }
      ilayout++; cur_ptr++;
    }
    */

    for(int i=0; i<freeze.size(); i++) {
      FreezeParticles fixptcls(*this);
      fixptcls.put(freeze[i]);
    }

    if(record_ptr) {
      vector<RecordProperty*> alist;
      if(createRecordProperty(record_ptr,alist)) {
	for(int i=0; i<alist.size(); i++) {
	  alist[i]->reset("test");
	  alist[i]->report(1);
	  alist[i]->finalize();
	  delete alist[i];
	}
      }
    }


    return true;
  }

  bool
  ParticleSets::createRecordProperty(xmlNodePtr cur,
				     vector<RecordProperty*>& records){
    
    bool dumpit = false;
    bool all = false;
    string aform,pname("none");

    vector<Particle_t*> torecord;
    ///process attributes: type or format
    xmlAttrPtr att = cur->properties;
    while(att != NULL) {
      string aname((const char*)(att->name));
      string vname((const char*)(att->children->content));
      if(aname == "condition" || aname== "type") {
	if(vname == "overwrite" || vname == "restart") dumpit = true;
      } else if(aname == "format") {
	aform =vname;
      } else if(aname == "target") {
	if(vname == "all") 
	  all = true;
	else {
	  Particle_t* atry = get(vname);
	  if(atry) //add a particle set with the name
	    torecord.push_back(atry);
	}
      }
      att = att->next;
    }

    if(all) {
      PtclSetMap_t::iterator it = PtclCollection.begin();
      while(it != PtclCollection.end()) {
	torecord.push_back((*it).second); it++;
      }
    } 

    if(ActivePtcl && torecord.empty()) torecord.push_back(ActivePtcl);

    bool success = false;
    for(int i=0; i<torecord.size(); i++) {
      RecordProperty *arecord = NULL;
      if(dumpit) {
	if(aform.size()) {
	  ///@warning hdf5 is called within xml file now.
	  ///@todo hdf5 should handle everything by itself
	  if(aform == "hdf5")  aform = "h5";//used to be xml
	  RestartConfig = aform;
	}
	arecord = new SaveParticle(*(torecord[i]));
      } else {
	arecord = new RecordParticle(*(torecord[i]));
      }
      if(arecord) {
	arecord->put(cur);
	records.push_back(arecord);
	success = true;
      }
    }
    return success;
  }
#endif
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 253 $   $Date: 2005-01-14 11:40:00 -0600 (Fri, 14 Jan 2005) $
 * $Id: ParticleSets.cpp 253 2005-01-14 17:40:00Z jnkim $ 
 ***************************************************************************/
