//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////

/*!\author Jeongnim Kim
 Based on kptgen.f for space group search and kpoint generation.
 Input data are not processed by the member functions.
 Instead, user objects or programs provide 
 a) Particle_t: lattice + particle configuration for findSpaceGroup
 b) MC indices and origin or Kpoints with weigths

      program kptgen
      c This program, kptgen.f and related subroutines, takes input unit cell
      c and atomic coordinate data and some control parameters and deduces
      c the space group and Bravais lattice and computes a special k point
      c Monkhorst-Pack type Brillouin zone quadrature grid.
      
      c It needs a single file name as input, the name of the input file.
      c The rest of the input comes from the input file, described below.
      c Output is sent to two hard-coded files: symmetry.summary and 
      c abinit.data.  The first is detailed analysis of crystal symmetry,
      c the second is a section of a file which would be readable by ABINIT.
      
      c Most recent revisions by Douglas C. Allan, Corning Incorporated, Mar 1995.
      c Updated for ABINIT, XG, February 1999.
      c This program was in public domain and presumably remains so.
      c Not sure of its history, see comments in subroutines for some indication.
      
      c I left some vestiges of old routines in place, even though I did not 
      c modify or upgrade them, assuming they might possibly be useful to others
      c who further upgrade this package.  dca
*/

#include "CrystalSymmetry/KPointBase.h"
#include "ParticleBase/ParticleUtility.h"

namespace OHMMS {
  // declaration of fortran subroutines provided kptgen program
#ifdef ADD_
#define pointg  pointg_
#define group1  group1_
#define symgen  symgen_
#define symatm  symatm_
#define cpointg cpointg_
#define spchek  spchek_
#define speck  speck_
#endif

  extern "C" {

    void pointg(const int&, int& ibrav,int& inpgd,int* is,int& nsym,int &iinv);

    void group1 (const int&, double* a1, double* a2, double* a3, const int&,
		 int* ty,const int& ,double* atom,
		 double* b1,double *b2,double *b3,
		 int& ihg,int& ihc,int & symorp,int &iinv,int &nsym,
		 int* ind,double* tau, int* f0,double * r);

    void symgen(double* a1,double* a2,double *a3, double* b1,double* b2,double *b3,
		int& nsym,double* rot, int* is);

    void symatm(int & natom, double* xred,int *f0, int& nsym,int *is, 
		double* , double* xredsym, double*);

    void cpointg(double* b1,double* b2,double* b3,
		 const int& ntype,int &natom,int *nkatom,double* atom,
		 int &nsym, double* rot,double* tnons);            

    void spchek(double* a1,double* a2,double* a3,int & mshdim,
		double* rsq,int* rvec,int& nkpt,double* kpt,double* wtk, 
		int& nsym,int* is,int& irmag,double& rmag,double& smag);

    void speck(const int&, const int&,
	       int* nk,double* gk0, double* b1,double *b2,double* b3,
	       int& nsym,int& iinv, double* ck, const int & mkpt,int & nkpt, 
	       int* is, int& ihc0, int& nbrav,
	       int* symlat);
  }

  ////////////////////////////////////////////////////////////
  // return true if |a-b|_i < smallnumber
  ////////////////////////////////////////////////////////////
  template<class Pos_t>
  inline bool similar(const Pos_t& a, const Pos_t& b) {
    typedef typename Pos_t::Type_t Type_t;
    return (fabs(a[0]-b[0]) < numeric_limits<Type_t>::epsilon() && 
	    fabs(a[1]-b[1]) < numeric_limits<Type_t>::epsilon() && 
	    fabs(a[2]-b[2]) < numeric_limits<Type_t>::epsilon()); 
  }

  template<class Pos_t>
  inline bool inverse(const Pos_t& a, const Pos_t& b) {
    typedef typename Pos_t::Type_t Type_t;
    return (fabs(a[0]+b[0]) < numeric_limits<Type_t>::epsilon() && 
	    fabs(a[1]+b[1]) < numeric_limits<Type_t>::epsilon() && 
	    fabs(a[2]+b[2]) < numeric_limits<Type_t>::epsilon()); 
  }


  //!< constructor
  KPointBase::KPointBase() {

    Reduced = true;
    Distributed = false;
    kdist.push_back(pair<int,int>(0,1));
    ID = 0;

    hgroup[1]="triclinic";
    hgroup[2]="monclinic";
    hgroup[3]="orthorhombic";
    hgroup[4]="tetragonal";
    hgroup[5]="cubic";
    hgroup[6]="trigonal";
    hgroup[7]="hexagonal";
    ref_ = NULL;
  }

  //!< destructor
  KPointBase::~KPointBase() {

  }

  void KPointBase::addKpt(const SingleParticlePos_t& kin, Scalar_t wgt){
    K.push_back(kin);
    W.push_back(wgt);
  }

  /** Computing special point grid based on the group operation
   *@param nk Monhorst-Pack indices
   *@param gk0 origin for this grid
   */
  void KPointBase::buildKpoint(SingleParticleIndex_t& nk, SingleParticlePos_t& gk0) {

    const int unit00 = 10;
    const int unit01 = 11;
    int mkpt = 2*nk[0]*nk[1]*nk[2]; //cannot be larger than n1*n2*n3
    vector<Scalar_t> ck(mkpt*4);

    SingleParticlePos_t b1 = LatRef.b(0);
    SingleParticlePos_t b2 = LatRef.b(1); 
    SingleParticlePos_t b3 = LatRef.b(2);

    speck(unit00, unit01, 
	  &(nk[0]), &(gk0[0]), &(b1[0]), &(b2[0]), &(b3[0]),
	  NumSym, HasInversion, &(ck[0]), mkpt,NumKpt,
	  &(is[0][0][0]),CubicGroup,NumBravis,&(symlat[0][0][0]));

    K.resize(NumKpt);
    W.resize(NumKpt);

    int ik=0;
    for(int i=0; i<NumKpt; i++) {
      K[i][0] = ck[ik++];
      K[i][1] = ck[ik++];
      K[i][2] = ck[ik++];
      W[i]    = ck[ik++];
    } 

    // number of Irreducible k-point is identical to the current numkpt
    NumIrrKpt = NumKpt; 
    partition(1);
  }

  void 
  KPointBase::partition(int n) {

    kdist.resize(n);
    if(n==1) {
      kdist[0].first = 0;
      kdist[0].second = K.size();
      Distributed = false;
    }  else {
      int nkpt_per_node = K.size()/n;
      int dn = K.size()%n;
      kdist[0].first = 0;
      for(int i=0; i<n; i++) {
	if(i < dn) 
	  kdist[i].second = kdist[i].first+nkpt_per_node+1;
	else
	  kdist[i].second = kdist[i].first+nkpt_per_node;
	if(i < n-1) {
	  kdist[i+1].first = kdist[i].second;
	}
      }
      Distributed = true;
    }

    //    kdist.resize(n+1);
    //    kdist[0] = 0;
    //    if(n==1) kdist[1] = K.size();
    //    else {
    //      int nkpt_per_node = K.size()/n;
    //      int dn = K.size()%n;
    //      for(int i=0; i<n; i++) {
    //        if(i < dn)
    //  	kdist[i+1] = kdist[i]+nkpt_per_node+1;
    //        else
    //  	kdist[i+1] = kdist[i]+nkpt_per_node;
    //      }
    //    }
  }

  /** Assign the input k-point and weight data to KPointBase
    *@param kin input recip translations
    *@param win weights
    */
  void 
  KPointBase::buildKpoint(vector<SingleParticlePos_t>& kin, 
			  vector<Scalar_t>& win){
    //copying input k points in terms of recip translations, and weights
    NumKpt = kin.size();
    K = kin;
    W = win;

    // \todo Add reduction operations based on the symmetry
    // number of Irreducible k-point is identical to the current numkpt
    NumIrrKpt = NumKpt;
    partition(1);
  }

  //void KPointBase::findSpaceGroup(Particle_t& ptcl, ostream& fout) {
  void 
  KPointBase::findSpaceGroup(Particle_t& ptcl) {

    LatRef.copy(ptcl.Lattice); // make copy of the lattice of this particle
    // ptcl.Lattice is defined in Lattice/CrystalLattice.(h, cpp)
    // lattice vectors and reciprocal vectors are already calculated
    // ptlc.Lattice also provides functions to convert particle position
    // from the Cartesian units to the unit-cell units vice versa.
    SingleParticlePos_t a1 = ptcl.Lattice.a(0);
    SingleParticlePos_t a2 = ptcl.Lattice.a(1);
    SingleParticlePos_t a3 = ptcl.Lattice.a(2);
    SingleParticlePos_t b1 = ptcl.Lattice.b(0);
    SingleParticlePos_t b2 = ptcl.Lattice.b(1); 
    SingleParticlePos_t b3 = ptcl.Lattice.b(2);

    const int unit00 = 10;
    int ihg, ihc, symorp;

    int natom = ptcl.getTotalNum(); // number of particles
    double r[D][D][49];             // temporary rotation

    tau.resize(48);
    tnons.resize(48);
    rot.resize(48);
    xredsym.resize(natom);

    // maxmimum number of species
    const int ntype = 10;
    vector<int> nkatom(ntype,0);
    for(int i=0; i<natom; i++) {
      nkatom[ptcl.GroupID[i]]++; // accumulate the number of particles per species
    }

    // temporary position in cartesian and unit-cell units
    // ptcl.R can be either in unit or cartesian
    Particle_t::ParticlePos_t atom(natom), xred(natom);

    //ptcl.Lattice.convert2Unit(ptcl.R, xred);
    //ptcl.Lattice.convert2Cart(xred, atom);
    ptcl.convert2Unit(ptcl.R, xred);
    ptcl.convert2Cart(xred, atom);

    //adjust the origin to avoid numerical problems and ease symmetry calculation
    Scalar_t rmin = 10000; int icorner=-1;
    for(int i=0; i<xred.size(); i++) {
      Scalar_t d2=dot(atom[i],atom[i]);
      if(d2<rmin) { rmin = d2; icorner=i;}
    }
    if(icorner>-1) {
      SingleParticlePos_t corner_cart = atom[icorner];
      SingleParticlePos_t corner_red =  ptcl.Lattice.toUnit(atom[icorner]);
      atom -= corner_cart;
      xred -= corner_red;
    }

    vector<int> f0(natom*48);
    //ibrav = 0;
    //inpgd = 0;
    BravisLatticeType = 0; // ibrav
    PointGroup = 0;        // inpgd

    group1 (unit00,&(a1[0]),&(a2[0]),&(a3[0]),
	    natom,&(ptcl.GroupID[0]),natom,&(atom[0][0]),
	    &(b1[0]),&(b2[0]),&(b3[0]),
	    HolohedralGroup,CubicGroup, Symmorphic, HasInversion, NumSym, 
	    ind, &(tau[0][0]),&(f0[0]),&(r[0][0][0]));

    for(int isym=0; isym<NumSym; isym++) {
      int ii=ind[isym]-1; // careful: c <-> fortran array base differences
      //  Get nonsymmorphic translations in dimensional cartesian coordinates
      tnons[isym] = ptcl.Lattice.toCart(tau[isym]);
      for(int i=0; i<D; i++) {
	for(int j=0; j<D; j++) {
	  rot[isym](i,j)= r[j][i][ii];
	}
      }
    }

    /*
      if(fout) {
      // Recall: the nonsymmmorphic translations tau are in terms of (a1,a2,a3)
      fout << "nsym from point1 " << NumSym << endl;
      for(int isym=0; isym<NumSym; isym++) {
      int ii=ind[isym]-1;
      fout << setw(19) << setprecision(2) << r[0][0][ii] 
      << setw(10) << r[1][0][ii] 
      << setw(10) << r[2][0][ii] 
      << setw(34) << setprecision(5) << tau[isym][0] << endl;
      fout << " r(" << setw(2) << ii << ") = " 
      << setw(10) << setprecision(2) << r[0][1][ii] 
      << setw(10) << r[1][1][ii]  
      << setw(10) << r[2][1][ii]
      << "          tau(" << setw(2) << ii << ") = " 
      << setw(14) << setprecision(5) <<tau[isym][1] << endl;
      fout << setw(19) << setprecision(2) << r[0][2][ii] 
      << setw(10) << r[1][2][ii] 
      << setw(10) << r[2][2][ii] 
      << setw(34) << setprecision(5) << tau[isym][2] << endl;
      }
      }
    */

    symgen(&(a1[0]),&(a2[0]),&(a3[0]),&(b1[0]),&(b2[0]),&(b3[0]),
	   NumSym,&(rot[0](0,0)),&(is[0][0][0]));
  
    /*
      if(fout) {
      fout << " Reciprocal space rotation matrices and translations "
      << endl << NumSym << " " << Symmorphic << endl;
      for(int isym=0; isym<NumSym; isym++) {
      for(int i=0; i<D; i++) {
      for(int j=0; j<D; j++) {
      fout << setw(5) << is[isym][i][j];
      }
      }
      fout << tau[isym] << endl;
      }
      }
    */

    symatm(natom,&(xred[0][0]),&(f0[0]),NumSym,&(is[0][0][0]),
	   &(tau[0][0]),&(xredsym[0][0]),&(tau[0][0]));

    /*
      if(fout) {
      fout << " bravais lattice index --> " << BravisLatticeType << endl 
      << " point group index --> " << PointGroup << endl
      << " no. of point group elements in scf--> " << NumSym << endl;

      }
    */

    // check whether the point group symmetry is correct or not
    cpointg(&(b1[0]),&(b2[0]),&(b3[0]),
	    ntype,natom,&(nkatom[0]),&(atom[0][0]),
	    NumSym,&(rot[0](0,0)),&(tnons[0][0]));

    // Determine the lattice point group (no atoms) needed for k point stuff
    // Note that this ignores any actual atoms, it only considers
    // the point group symmetries of the lattice by itself
    // Place a single atom at the origin as a place holder
    int natom0 = 1;
    int nat0 = 1;
    double atom0[] = {0.0, 0.0, 0.0};
    double tau0[48][D];
    int ty0 =1;
    int iinv0;

    group1 (unit00,&(a1[0]),&(a2[0]),&(a3[0]),
	    natom0,&ty0, nat0, atom0, &(b1[0]), &(b2[0]), &(b3[0]), 
	    HolohedralGroup, CubicGroup, Symmorphic,
	    iinv0,NumBravis,
	    ind,tau0[0],&(f0[0]),&(r[0][0][0]));

    //  nbrav is returned as the number of elements in the point group
    //  iinv0=0 means no inversion; >0 means inversion is present
    //  r(3,3,ind(i)),i=1,nbrav are the point group operations

    if(iinv0 == 0) {
      cerr <<" kptgen error: every Bravais lattice must have inversion symmetry."
	   << endl;
      return;
    }

    //for(int ib=0; ib<nbrav; ib++) {
    for(int ib=0; ib<NumBravis; ib++) {
      for(int i=0; i<D; i++) {
	for(int j=0; j<D; j++) {
	  rot[ib](i,j) = r[j][i][ind[ib]-1];
	}
      }
    }

    symgen(&(a1[0]),&(a2[0]),&(a3[0]),
	   &(b1[0]),&(b2[0]),&(b3[0]),
	   NumBravis,&(rot[0](0,0)),&(symlat[0][0][0]));
 
  }  

  void KPointBase::extend2FBZ() {

    if(Kmode ==  kpt_band || !Reduced || rot.empty()) return;

    vector<vector<SingleParticlePos_t> *> kfull;
    vector<double> wtmp(NumKpt);
    int nkptadd = 0;
    for(int ikpt=0; ikpt<NumKpt; ikpt++) {
      SingleParticlePos_t k0 = K[ikpt];
      double kmag=dot(k0,k0);
      kfull.push_back(new  vector<SingleParticlePos_t>);
      if(kmag<numeric_limits<double>::epsilon()) { //gamma point
        kfull[ikpt]->push_back(k0);
      } else {
      for(int isym=0; isym<NumSym; isym++) {
	SingleParticlePos_t k1(static_cast<double>(is[isym][0][0])*k0[0]+
			       static_cast<double>(is[isym][1][0])*k0[1]+
			       static_cast<double>(is[isym][2][0])*k0[2],
			       static_cast<double>(is[isym][0][1])*k0[0]+
			       static_cast<double>(is[isym][1][1])*k0[1]+
			       static_cast<double>(is[isym][2][1])*k0[2],
			       static_cast<double>(is[isym][0][2])*k0[0]+
			       static_cast<double>(is[isym][1][2])*k0[1]+
			       static_cast<double>(is[isym][2][2])*k0[2]);

	fold2BZ(k1);

	bool newk = true;
	int jkpt=0;
	while(newk && jkpt < kfull[ikpt]->size()) {

	  if(similar(kfull[ikpt]->operator[](jkpt),k1)) newk = false;

	  //if(newk && HasInversion) {//new k, check if it is -k
	    if(inverse(kfull[ikpt]->operator[](jkpt),k1)) newk = false;
	  //}
	  jkpt++;
	}
	if(newk) 	kfull[ikpt]->push_back(k1);
      }
      }
      nkptadd += (kfull[ikpt]->size()-1); 
      wtmp[ikpt] = W[ikpt]/static_cast<double>(kfull[ikpt]->size());
    }

    SingleParticlePos_t pzero;

    // will append to the list
    K.insert(K.end(),nkptadd,pzero);
    W.insert(W.end(),nkptadd,1.0);
    int nk = NumKpt;
    for(int ikpt=0; ikpt<NumKpt; ikpt++) {
      W[ikpt] = wtmp[ikpt];
      for(int jkpt=1; jkpt < kfull[ikpt]->size(); jkpt++, nk++) {
	K[nk] = (*kfull[ikpt])[jkpt];
	W[nk] = wtmp[ikpt];
      }
      delete kfull[ikpt]; // clean up this
    }

    NumKpt = K.size();
    partition(kdist.size());
    Reduced = false; // not a reduced k-point
  }

  ////////////////////////////////////////////////////////////
  // fold a vector in a BZ
  ////////////////////////////////////////////////////////////
  void KPointBase::fold2BZ(SingleParticlePos_t& pos) const{

    int alpha[3], alpha0[3], alpha1[3];

    const double MIN_ERROR = 1e-6;
 
    for (int i=0; i<3; ++i) {
      // Shift into the interval [-1, 1] (carefully!)
      if (pos[i] > 0)
	pos[i] -= static_cast<double>(static_cast<int>(pos[i] + 1e-4));
      else
	pos[i] -= static_cast<double>(static_cast<int>(pos[i] - 1e-4));
      // Determine the shifts we want to try.
      if (fabs(pos[i]) > MIN_ERROR) {
	if (pos[i] > 0)
	  {alpha0[i] = -1; alpha1[i] = 0;}
	else
	  {alpha0[i] = 0; alpha1[i] = 1;}
      }
      else
	{alpha0[i] = 0; alpha1[i] = 0;}  // Assume no shift.
      alpha[i] = alpha0[i];
    }

    // Calculate our magnitude.
    SingleParticlePos_t kcart = LatRef.toUnit(pos);
    double min_magn = dot(kcart,kcart);
    double magn_error = min_magn*MIN_ERROR;
    int better;  // True if we've constructed a "better" kpoint.

    SingleParticlePos_t k0 = pos;
    do {
      SingleParticlePos_t q = k0;
      q[0] += static_cast<double>(alpha[0]);
      q[1] += static_cast<double>(alpha[1]);
      q[2] += static_cast<double>(alpha[2]);

      kcart = LatRef.toUnit(q);    // Calc it's magnitude.
      double test_magn = dot(kcart,kcart);

      if ((test_magn + magn_error) < min_magn)
	better = -1;    // If it's clearly better, choose it.
      else
	if ((test_magn - magn_error) < min_magn) {
	  // It's a boundary k-point, so we have to think more.
	  // This is a _really_ squirrely piece of code.  It checks
	  // the multiple possibilities one by one.
	  better = (q[0] > pos[0]);
	  if (fabs(q[0]-pos[0]) < MIN_ERROR) {
	    better = (q[1] > pos[1]);
	    if (fabs(q[1]-pos[1]) < MIN_ERROR)
	      better = (q[2] > pos[2]);
	  }
	}
	else
	  better = 0;  // We haven't gotten a better one.
      if (better) {
	min_magn = test_magn;
	pos = q;
      }

      // Increment counters.
      ++alpha[0];
      if (alpha[0]>alpha1[0]) {alpha[0] = alpha0[0]; ++alpha[1];}
      if (alpha[1]>alpha1[1]) {alpha[1] = alpha0[1]; ++alpha[2];}
    } while (alpha[2]<=alpha1[2]);
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: KPointBase.cpp 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
