//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#ifndef OHMMS_SPACEGROUP_KPOINT_H
#define OHMMS_SPACEGROUP_KPOINT_H
#include "ohmms-config.h"
#include <utility>
#include <vector>
#include <string>
#include <map>
#include "Particle/ParticleSet.h"
#include "OhmmsData/OhmmsElementBase.h"

namespace OHMMS {

  /** KPointBase class to determine group properties of a system and make a kpoint-set
   *
   * Using routines provied by Abinit.
   * The names of the datamembers may use those in f77 routines.
   * \var HolohedralGroup: point group of the primitive lattice, 
   * holohedral group number:
   *  1 stands for triclinic system
   *  2 stands for monoclinic system
   *  3 stands for orthorhombic system
   *  4 stands for tetragonal system
   *  5 stands for cubic system
   *  6 stands for trigonal system
   *  7 stands for hexagonal system
   * (holohedry is the point group of the lattice)
   */
  struct KPointBase: public OhmmsElementBase {
    
    typedef ParticleSet                       Particle_t;
    typedef ParticleSet::size_type            size_type;
    typedef Particle_t::ParticleLayout_t      ParticleLayout_t;
    typedef Particle_t::Scalar_t              Scalar_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::Tensor_t              Tensor_t;
    
    enum {D = OHMMS_DIM};
    enum KPOINT_MODE {kpt_mp, kpt_list, kpt_band};
    
    //////////////////////////////
    // data member <-> kptgen variable name
    //////////////////////////////
    /// ibrav, the type of Bravis Lattice
    int BravisLatticeType; 
    /// iinv, indicator if the supercell has an inversion symmetry
    int HasInversion;     
    /// inpgd, index for the point group
    int PointGroup;     
    ///ihg, point group of the primitive lattice, holohedral group number
    int HolohedralGroup;  
    /// symorp, 1 for symmorphic; 0 for nonsymmorphic
    int Symmorphic;     
    /// ihc, 1 = Cubic; 0 = Hexagonal 
    int CubicGroup;       
    /// nsym, the number of symmetry operations
    int NumSym;           
    /// nbrav, not sure what this is supposed to be
    int NumBravis;        
    /// nkpt, the number of total K-point
    int NumKpt;          
    /// the number of irreducible K-points
    int NumIrrKpt;         
    /// the type of k-point set: auto | list | band
    KPOINT_MODE Kmode;
    /// true if the active K-points are the irreducible set.
    bool Reduced;         

    ///////////////////////////////
    // symmetry tensor related: used internally
    ///////////////////////////////
    /// lattice for this SpaceGroup and k-points
    ParticleLayout_t LatRef;
    ///symmetries of the crystal in the basis 
    int is[48][D][D];      
    ///symmetry matrices of lattice 
    int symlat[48][D][D];
    ///symmetry indices
    int ind[48];

    /** nonprimitive translations (in the case of nonsymmor-phic groups), in the unit-cell unit. 
     *
     * tau[n] is the translation vector connected with the n-th element
     * of the point group (i.e. with the rotation number ind[n]-1 ).
     */
    vector<SingleParticlePos_t> tau;

    /** nonprimitive translations (in the case of nonsymmor-phic groups), in the Carteisn unit. 
     *
     * tnons[n] is the translation vector connected with the n-th element
     * of the point group (i.e. with the rotation number ind[n]-1 ).
     */
    vector<SingleParticlePos_t> tnons;

    /// tensors for the rotation operations in the Cartesian unit.
    vector<Tensor_t>            rot;  

    /// local copy of the particle coordinates, ParticleBase::curR could be used instead.
    vector<SingleParticlePos_t> xredsym;   

    /** Array of k-points in the reciprocal unit
     *
     * [0,NumIrrKpt)  = irreducible k-points
     * [NumIrrKpt,NumKpt) = the rest
     * It is possible to have NumIrrKpt == NumKpt
     */
    vector<SingleParticlePos_t> K; // kpoint

    /** Array of weights for the corresponding k-point
     *
     *Indentical index scheme to K
     */
    vector<Scalar_t>            W; // weight

    ///if true, the k-points are distributed over the processors
    bool Distributed;

    ///id
    int ID;

    ///manager of a distributed k-points
    vector<pair<int,int> > kdist;

    KPointBase();
    ~KPointBase();
    
    //void findSpaceGroup(Particle_t&, ostream&);
    void findSpaceGroup(Particle_t&);
    
    void buildKpoint(SingleParticleIndex_t& nk, SingleParticlePos_t& gk0);
    void buildKpoint(vector<SingleParticlePos_t>& kin, vector<Scalar_t>& win);
    
    /// add a kpoint with a weight
    void addKpt(const SingleParticlePos_t& kin, Scalar_t wgt);
    
    ///\todo Need to change this for MPI parallelization of k-points
    inline size_type size() const { return K.size();}
    inline bool distributed() const { return Distributed;}
    inline int begin(int pe=0) const { return kdist[pe].first;}
    inline int end(int pe=0) const { return kdist[pe].second;}
    
    void partition(int n);
    
    //!< Returns an irreducible k-point in Reciprocal unit
    inline SingleParticlePos_t k_unit(size_type i) const { return K[i];}
    
    //!< Returns k-point in Cartesian unit
    inline SingleParticlePos_t k_cart(size_type i) const { 
      return LatRef.k_cart(K[i]);
    }
    
    void fold2BZ(SingleParticlePos_t& pos) const;
    
    void extend2FBZ();
    
    void print(ostream&, int i=0) const;
    
    mutable std::map<int,string> hgroup;
    
    //@{implementation of the virtual functions of OhmmsElementBase
    bool get(std::ostream& ) const;
    bool put(std::istream& );
    bool put(xmlNodePtr cur);
    void reset();
    void begin_node(std::ostream& os) const;
    void end_node(std::ostream& os) const;
    //@}

    void putSimpleList(xmlNodePtr cur, int n);
    Particle_t* ref_;
  };
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: KPointBase.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
