      subroutine listmg(iut,gr,iplace)

c Linked list to manipulate lists of 3-dimensional vectors gr

 
c     mesh maintains a list of vectors for placement and/or lookup
c
c     additional entry points: remove ..... remove vector from list
c                              garbag ..... was vector removed ?
c
c     iut ...... output unit
c     gr(3) .... vector
c     iplace .... on input:  -2 means: initialize  the list (and return)
c                            -1 means: find gr in the list
c                             0 means: add  gr to the list
c                            >0 means: return gr number iplace
c                 on output: the position assigned to gr
c                            (=0 if gr is not in the list)

      implicit none

      integer iut,iplace
      double precision gr(3)

c The maximum number of distinct vectors which can be handled is 
c equal to nlist/nwords; beyond that the arrays get overrun (this is
c tested for below, code will stop with message).

      integer nhash, nlist, nwords
      parameter (nhash = 4000, nlist = 100000, nwords = 4)

c     tol was originally 1.d-5
c     tol sets the tolerance for equality of components of gr
      double precision tol
      parameter (tol=1.d-5)

      integer list(nhash+nlist)
      double precision rlist(nhash+nlist)

      integer nil
      parameter (nil = 0)
      integer mlist

c     h1, h2, h3 are chosen for form a hash function
      double precision h1, h2, h3
      parameter (h1=0.7890d0, h2=0.6810d0, h3=0.5811d0)

      integer igarb0,ipoint,nstore,ihash,i,j,istore,igarbg
      double precision rhash

      save rlist,list,igarb0,istore,nstore,mlist
c
      data igarb0/0/, igarbg/0/
 
      if (iplace .le. -2) then
c      initialization call
        mlist = nhash + nlist
        do i = 1, mlist
         list(i) = nil
        enddo
        istore = nhash + 1
        nstore = 0
        mlist = nhash + nlist
        return
      endif

      if (iplace .gt. 0) then
c      Return the actual vector stored at iplace
       ipoint = (iplace-1) * nwords + nhash + 1
       if (ipoint .ge. istore) go to 307
       do i = 1, 3
        gr(i) = rlist(ipoint + i)
       enddo
       return
      endif

c     From now on, iplace is -1 (find gr in list) or 0 (add gr to list)
 
c     Compute the hash function
 
      rhash = (h1 * gr(1) + h2 * gr(2) + h3 * gr(3) + 1.0d-6) * nhash
 
      ihash = int(abs(rhash))
      ihash = mod(ihash,nhash) + 1
      if (ihash .gt. nhash ) go to 307
c     ihash is in [1,nhash]
 
c     search for linked list 
 
      ipoint = list(ihash)
      if (ipoint .eq. 0) go to 304
 
      do 101 i = 1, nstore + 1

       if (ipoint .eq. 0) then
c       as soon as an empty slot is found (ipoint=0), jump out of 101 loop
        go to 304
       endif
 
c      compare gr with entries already in list

       if (abs(gr(1)-rlist(ipoint+1)).le.tol .and.
     &     abs(gr(2)-rlist(ipoint+2)).le.tol .and.
     &     abs(gr(3)-rlist(ipoint+3)).le.tol ) then

c       gr was found in the list
 
        if (iplace .eq. 0) then
c        gr was to be added to the list but is found already there
         return
        else
c        iplace was input as -1 => wanted to find gr in list; return iplace
         iplace = (ipoint - (nhash + 1)) / nwords + 1
         return
        endif
       endif
 
c      gr not found at ipoint, jump to next element in the list
 
       ihash = ipoint
       ipoint = list(ihash)
101   continue


c     list too long
 
      write(*,401) nstore
401   format(/,'listmg error: search for list is too long,',/,
     &  'choose a better hash function, nstore=',i12)
      stop
 
304   continue
 
      if (iplace .eq. 0) go to 305
 
c     iplace = -1 search for gr unsuccessful
c     To land here, iplace is -1 and ipoint is 0; ipoint=0 means
c     no vector stored at ipoint

      iplace = 0
      return
 
c     iplace = 0, add gr to the list
 
305   continue

c     To land here, ipoint is 0 and iplace is 0; in this situation,
c     gr is to be added to the list (iplace=0) and there is no vector
c     already stored there (ipoint=0)

      if (ihash .gt. mlist) then
       write(*,1100) ihash, mlist
 1100  format(/,'listmg error: ihash=',i12,/,'exceeds dimension ',
     &  'nhash+nlist=mlist=',i12,/,'raise nhash or nlist ',
     &  '(usually nlist) in listmg and recompile.',/)
       stop
      endif
      if (istore .gt. mlist) then
       write(*,1110) istore, mlist
 1110  format(/,'listmg error: istore=',i12,/,' exceeds dimension ',
     &  'nhash+nlist=mlist=',i12,/,'raise nhash or nlist ',
     &  '(usually nlist) in listmg and recompile.',/)
       stop
      endif

c     Add gr to the list: the element at ihash is set to point at istore
c     (initialized to nhash+1) and the element at istore is set to 0
      list(ihash) = istore
      list(istore) = 0
      nstore = nstore + 1
c
      do i = 1, 3
       rlist(istore+i) = gr(i)
      enddo

      istore = istore + nwords
c     iplace is returned as the list placement of gr
c     (this goes in counting order as vectors are added: 1,2,3,...)
      iplace = (istore - (nhash + 1)) / nwords
      return

c     error --> beyond the list
 
307   continue
      write(iut,402) iplace
402   format(/' subroutine listmg *** warning ***'/
     &       ,' iplace = ', i5,' is beyond the lists -- gr set to '
     &       ,' 1.0e38')
      do i = 1, 3
       gr(i) = 1.0d38
      enddo
      return
 
c-----------------------------------------------------------------------
 
c     entry point for removing a wavevector
 
c     input:   gr
c     output:  iplace ..... 1 if gr was removed
c                           0 if gr was not removed (gr not in the
c                             linked lists)
 
      entry remove (iut,gr,iplace)
 
c     Compute hash function
 
      rhash = (h1 * gr(1) + h2 * gr(2) + h3 * gr(3) + 1.0d-6) * nhash
 
      ihash = int(abs(rhash))
      ihash = mod(ihash,nhash) + 1
 
c     search for gr in linked list
 
      ipoint = list(ihash)
 
      do 220 i = 1,100
 
c       list exhausted
        if (ipoint .eq. nil) then
 
c         gr was not found in the mesh:
 
          iplace = 0
          return
        endif
 
c       compare gr with this element
 
        do j = 1,3
          if (abs(gr(j) - rlist(ipoint+j)) .gt. tol) goto 215
        enddo
 
c       gr located, now remove it from the list:
 
        list(ihash) = list(ipoint)
c       list(ihash) now points to the next element in the list,
c       and the present gr has become garbage.
 
c       add gr to the list of garbage:
 
        if (igarb0 .eq. 0) then
c         start up the garbage list:
          igarb0 = ipoint
        else
          list(igarbg) = ipoint
        endif
        igarbg       = ipoint
        list(igarbg) = nil
        iplace = 1
 
        return
 
c       next element of list
  215   continue
        ihash = ipoint
        ipoint = list(ihash)
  220 continue
 
c     list too long
 
      write (*,126)
126   format(/,'subroutine listmg *** fatal error ***',
     &' linked list too long ***',/,' choose a better hash-function')
      stop

c-----------------------------------------------------------------------
 
c     entry point for checking if a wavevector is in the garbage list
 
c     input:    gr
c     output:   iplace  ..... i > 0 is the place in the garbage list
c                             0 if gr not among the garbage
 
      entry garbag (iut,gr,iplace)
 
c     search for gr in linked list
 
c     point to the garbage list
      ipoint = igarb0
      do 320 i = 1 , nlist/nwords
 
c       list exhausted
        if (ipoint .eq. nil) then
 
c         gr was not found in the mesh:
 
          iplace = 0
          return
        endif
 
c       compare gr with this element
 
        do j = 1,3
          if (abs(gr(j) - rlist(ipoint+j)) .gt. tol) goto 315
        enddo
 
c       gr was located in the garbage list
 
        iplace = i
        return
 
c       next element of list
  315   continue
        ihash = ipoint
        ipoint = list(ihash)
  320 continue
 
c     list too long
 
      write (*,*) ' ***subroutine garbag - garbage list too long***'
      stop

      end
