      subroutine spchek(a1,a2,a3,mdim,rsq,rvec,nkpt,kpt,wtk,
     &  nsym,symmet,irmag,rmag,smag)

c This subroutine checks special point quadrature grids
c by computing Sum(ikpt) [wtk(ikpt)*A_m(kpt(ikpt))]
c where A_m(kpt) = (1/numsh_m)*Sum(i) [cos(2*Pi*kpt.R(i)] for |R(i)|=|R|_m,
c the shells m are ordered by size, |R|_m, and numsh_m is the number of
c real space primitive translations R(i) which have |R(i)|=|R|_m.
c At the first shell, |R|=0, the sum is 1.
c A good grid will cause the sum to vanish for as many shells as possible.
c This routine finds the first nonvanishing shell and returns its |R|.
c See Monkhorst and Pack, Phys. Rev. B13, 5188 (1976).

c Doug Allan   March 1995.

c Input:
c  a1(3),a2(3),a3(3)=dimensional primitive translations in real space,
c   in cartesian coordinates
c  mdim=calling routine dimension for rsq, rvec
c  rsq(mdim)=array for holding |R|^2 values
c  rvec(3,mdim)=integer array for holding real space translations
c  nkpt=number of k points in special point grid
c  kpt(3,nkpt)=coordinates of k points in basis of reciprocal lattice
c   translations
c  wtk(nkpt)=weights associated with each k point.
c  nsym=number of symmetries of the LATTICE (not crystal)
c  symmet(3,3,nsym)=transposes of symmetry matrices for the lattice
c Output:
c  irmag=index of shell (0 is R=0) at which first failure occurs
c  rmag=|R| of the first shell for which first failure occurs
c  smag=value of sum at first failure

      implicit none
      integer mdim,nkpt,irmag,nsym
      integer rvec(3,mdim),symmet(3,3,nsym)
      double precision rmag,smag
      double precision a1(3),a2(3),a3(3),rsq(mdim),kpt(3,nkpt),wtk(nkpt)

      double precision rmet(3,3)
      integer rv(3)
      integer count,ir1,ir2,ir3,nr,j,j1,ikpt,numsh,shell,ish,
     &  nvec,nshc,ii,jj,nref,nsub,nmatch,isubsh,nsubsh,nshell,number
      double precision rsquar,twopi,fac,dot

c     nsh is a temp storage array for counting the number of vectors
c     in each shell.  need about 4*number of shells expected.
      integer msh
      parameter (msh=2000)
      integer nsh(msh,2)

c     ind is used to keep track of different stars of real space
c     vectors which have the same length
      integer mind
      parameter (mind=768)
      integer ind(mind)

c     mr sets an upper limit on a loop which studies the ranges
c     -mr to mr in each dimension; probably will run out of storage
c     before hitting end of this loop
      integer mr
      parameter (mr=1000)

c     Tol sets the minimum separation between the |R|^2 of shells
      double precision tol
      parameter (tol=1.d-8)

      do j=1,msh
       nsh(j,1)=0
       nsh(j,2)=0
      enddo
      nsh(1,2)=1

      twopi = 8.d0 * atan(1.d0)

c     Define metric tensor in real space
      rmet(1,1)=a1(1)*a1(1)+a1(2)*a1(2)+a1(3)*a1(3)
      rmet(2,1)=a2(1)*a1(1)+a2(2)*a1(2)+a2(3)*a1(3)
      rmet(3,1)=a3(1)*a1(1)+a3(2)*a1(2)+a3(3)*a1(3)
      rmet(2,2)=a2(1)*a2(1)+a2(2)*a2(2)+a2(3)*a2(3)
      rmet(3,2)=a3(1)*a2(1)+a3(2)*a2(2)+a3(3)*a2(3)
      rmet(3,3)=a3(1)*a3(1)+a3(2)*a3(2)+a3(3)*a3(3)
      rmet(1,2)=rmet(2,1)
      rmet(1,3)=rmet(3,1)
      rmet(2,3)=rmet(3,2)

c     count keeps track of the total number of vectors under consideration
c     at any time; shells already computed get removed from the count
      count=0
c     shell counts up the number |R|'s for which fac has been computed
      shell=0
c     ish is used to alternate the parts of the nsh array
      ish=2
c     nshell counts up actual symmetry-related shells
c     (counts subshells separately)
      nshell=0

c     Run a loop over larger and larger shells

      do 500 nr=1,mr

c      alternate between the two parts of nsh
       ish=3-ish
 
       do 100 ir3 = -nr,nr
       do 100 ir2 = -nr,nr
       do 100 ir1 = -nr,nr

c       Skip over shells which were done previously
        if (abs(ir1).eq.nr .or. abs(ir2).eq.nr .or. abs(ir3).eq.nr
     $      .or. nr.eq.1 ) then

         count=count+1

c        Compute |R|^2
         rsquar=rmet(1,1)*dble(ir1*ir1)+rmet(2,2)*dble(ir2*ir2)+
     $          rmet(3,3)*dble(ir3*ir3)+
     $    2.d0*(rmet(2,1)*dble(ir2*ir1)+rmet(3,2)*dble(ir3*ir2)+
     $          rmet(1,3)*dble(ir1*ir3))

         if (count.gt.mdim) then
          write(6,1010) count,mdim
 1010     format(/,'spchek error: count=',i12,' exceeds mdim=',i12,/,
     &    '==> array dimension in calling routine is too small for',
     &    ' number of',/,'shell vectors needed.  Raise mdim in calling',
     &    ' routine.',/)
          stop
         endif

         rsq(count)=rsquar
         rvec(1,count)=ir1
         rvec(2,count)=ir2
         rvec(3,count)=ir3
        endif
  100  continue

c      Sort the existing set of stored vectors by |R|^2
c      (sort3 also makes the same rearrangement of rvec)

       call sort3(count,rsq,rvec)

c      Arrange the existing vectors into shells, including a grouping
c      into subshells related by symmetry

       rsquar=rsq(1)
       numsh=1
       nsh(numsh,ish)=0
       do j=1,count
        if (rsq(j).lt.rsquar+tol) then
         nsh(numsh,ish)=nsh(numsh,ish)+1
        else
c        Next shell
         rsquar=rsq(j)
         numsh=numsh+1
c        prevent array overrun
         if (numsh.gt.msh) then
          write(6,1030) numsh,msh
 1030     format(/,' spchek error: numsh=',i10,' exceeds dimensioned',
     &    ' msh=',i10,/,' Raise msh in spchek and recompile.',/)
          stop
         endif
         nsh(numsh,ish)=1
        endif
       enddo

c      Evaluate the test sum for all shells which are "complete", i.e.
c      for which all vectors seem to be accounted for (as tested by
c      number of vectors in shell not changed since last iteration)

c      nvec keeps track of the actual number of vectors in the complete shells
c      nshc keeps track of the corresponding number of shells
       nvec=0
       nshc=0
       do j=1,numsh
c       Do the numbers found in successive iterations agree?
        if (nsh(j,ish).eq.nsh(j,3-ish)) then

         if (nsh(j,ish).gt.1) then

c         Break into subshells, if any
          nmatch=1
          nsub=1
          nref=nvec+1

          if (nsh(j,ish) .gt. mind) then
           write(6,1040) mind
 1040      format(/,'spchek error: out of space in array ind',
     &     ' with mind=',i8,
     &     /,'Raise dimension mind in spchek and recompile.',/)
           stop
          endif

          do j1=1,nsh(j,ish)
           ind(j1)=0
          enddo
          ind(1)=1
          nsubsh=nsh(j,ish)
          do isubsh=1,nsubsh

           do j1=nvec+2,nvec+nsh(j,ish)
c           Check whether additional vectors in shell are related by symm

            if (ind(j1-nvec).eq.0) then
             do ii=1,nsym
              do jj=1,3
               rv(jj)=symmet(1,jj,ii)*rvec(1,j1)+
     &                symmet(2,jj,ii)*rvec(2,j1)+
     &                symmet(3,jj,ii)*rvec(3,j1)
              enddo
              if (rv(1).eq.rvec(1,nref) .and.
     &            rv(2).eq.rvec(2,nref) .and.
     &            rv(3).eq.rvec(3,nref)) then
c              matches
               ind(j1-nvec)=nsub
               nmatch=nmatch+1
               if (nmatch.eq.nsh(j,ish)) then
c               if all vectors of this magnitude are accounted for, skip out
                goto 102
               else
                goto 103
               endif
c              does not match
              endif
c             complete loop on symmetries
             enddo
c            complete ind(j1-nvec)=0
            endif
  103       continue
c           complete j1 loop on real space vectors if given length
           enddo

c          Find next subshell
           do jj=nvec+2,nvec+nsh(j,ish)
             if (ind(jj-nvec).eq.0) then
              nref=jj
              nsub=nsub+1
              goto 101
             endif
           enddo
           write(6,1050) rsq(nvec+1)
 1050      format(/,'spchek error: did not account for all vectors',
     &     /,'of length |R|^2 =',f12.5,/,
     &     'could not find next subshell.',/)
           stop
  101      continue
c          complete isubsh loop over subshells
          enddo

          write(6,1060) rsq(nvec+1)
 1060     format(/,'spchek error: did not account for all vectors',
     &    /,'of length |R|^2 =',f12.5,/)
          stop

  102     continue
         endif
  

c        Compute sum on k and R
         do ii=1,nsub
          fac=0.d0
          number=0
          do j1=nvec+1,nvec+nsh(j,ish)
           if (ind(j1-nvec).eq.ii) then
            do ikpt=1,nkpt
             dot=kpt(1,ikpt)*dble(rvec(1,j1))+
     &           kpt(2,ikpt)*dble(rvec(2,j1))+
     &           kpt(3,ikpt)*dble(rvec(3,j1))
             fac=fac+wtk(ikpt)*cos(twopi*dot)
             number=number+1
            enddo
           endif
          enddo

          fac=fac/dble(2*number)

c         Look for nonzero value of sum and skip at first beyond |R|=0.
          if (abs(fac).gt.1.d-8) then
            goto 501
          endif

c         debugging
c         write(6,2010) nshell+ii,fac,rsq(nvec+1)
c2010     format(i10,f15.12,f10.5)

         enddo

         nshell=nshell+nsub

        else

c        Skip out when the nsh first disagree
         goto 105

        endif

        shell=shell+1
        nvec=nvec+nsh(j,ish)
        nshc=nshc+1

c       complete loop on all possible shells
       enddo

  105  continue

c      Remove numsh entries from the lists, collapsing down to the
c      lowest indices (this is to save storage)

       do j=1,count-nvec
        rvec(1,j)=rvec(1,j+nvec)
        rvec(2,j)=rvec(2,j+nvec)
        rvec(3,j)=rvec(3,j+nvec)
        rsq(j)=rsq(j+nvec)
       enddo
       count = count - nvec
       j1 = j1 - nvec

       do j=1,numsh-nshc
        nsh(j,1)=nsh(j+nshc,1)
        nsh(j,2)=nsh(j+nshc,2)
       enddo
       numsh=numsh-nshc

  500 continue
  501 continue

c     Get final summary of shell index (0 is |R|=0), |R|, and fac
      irmag=nshell+1
      rmag=sqrt(rsq(nvec+1))
      smag=fac

      return
      end
