//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "NEB/NEBPotential.h"
#include "ParticleBase/ParticleUtility.h"
#include "Particle/NNEngine.h"
#include "Message/Communicate.h"
#include "Message/TagMaker.h"
#ifdef HAVE_MPI
#include <mpi.h>
#endif
namespace OHMMS{

  enum {pot_energy=0,
	chain_dL,
	chain_dR,
	chain_theta,
	f_norm2,
	fc_norm2,
	neb_end};

  NEBPotential::NEBPotential():SpringConst(0.5), neb_mode(1) {
    Ename.resize(neb_end);
    Ename[pot_energy] = " Epot ";
    Ename[chain_dL] = " Left ";
    Ename[chain_dR] = " Right ";
    Ename[chain_theta] = " Angle ";
    Ename[f_norm2] = " |F| ";
    Ename[fc_norm2] = " |Fc|";
  }

  NEBPotential::~NEBPotential() {
    delete myPtcl;
    DEBUGMSG("Calling NEBPotential::~NEBPotential()");
  }

  void NEBPotential::setParticleSet(Particle_t* ptclset) {
     // do nothing
  }

  /** initialize NEBPotential
   * @param pot physical potential that evaluates forces
   * @param Images neb chains
   */
  void NEBPotential::init(Interaction_t* pot, vector<Particle_t*>& Images) {

    realPotential=pot;

    if(myPtcl == 0) {
      myPtcl = new Particle_t;
      myPtcl->addDynamicAttributes(true);
      myPtcl->Lattice.copyGrid(Images[0]->Lattice);
      myPtcl->getSpeciesSet()=Images[0]->getSpeciesSet();
      myPtcl->assign(*Images[0]); 
    }
    //set the particle to the physical potential
    realPotential->setParticleSet(myPtcl);

    size_type nat = Images[0]->getTotalNum();
    Ft.setUnit(PosUnit::CartesianUnit);
    Lpos.setUnit(PosUnit::CartesianUnit);
    Rpos.setUnit(PosUnit::CartesianUnit);
    
    Ft.resize(nat);
    Fc.resize(nat);
    Lpos.resize(nat);
    Rpos.resize(nat);

    E.resize(neb_end,Images.size());

    for(int ic=0; ic<Images.size(); ic++) {
      Images[ic]->F.resize(nat);
    }

    myPtcl->R.setUnit(Images[0]->R.getUnit());
    int LastChain=Images.size();
    for(int ic=0; ic<LastChain; ic++) {
      myPtcl->R=Images[ic]->R;
      realPotential->setCurrentPos(myPtcl->R);
      E(pot_energy,ic) = realPotential->evalEnergy();
    }
  }

  ////THIS is for optimizaton
  //void NEBPotential::init(Interaction_t* pot, vector<Particle_t*>& Images) {
  //  realPotential=pot;
  //  myPtcl=Images[0];
  //  myPtcl->addDynamicAttributes(true);
  //  realPotential->setParticleSet(myPtcl);
  //  size_type nat = Images[0]->getTotalNum();
  //  Ft.setUnit(PosUnit::CartesianUnit);
  //  Lpos.setUnit(PosUnit::CartesianUnit);
  //  Rpos.setUnit(PosUnit::CartesianUnit);
  //  
  //  Ft.resize(nat);
  //  Fc.resize(nat);
  //  Lpos.resize(nat);
  //  Rpos.resize(nat);
  //  E.resize(neb_end,Images.size());
  //  for(int ic=0; ic<Images.size(); ic++) {
  //    Images[ic]->F.resize(nat);
  //  }
  //  myPtcl->R.setUnit(Images[0]->R.getUnit());
  //  myPtcl->R=Images[0]->R;
  //  realPotential->setCurrentPos(myPtcl->R);
  //  E(pot_energy,0) = realPotential->evalEnergy();
  //  NNEngine* myLayout=myPtcl->getNNEngine();
  //  GridRadius = Rmax = myLayout->Rscale*realPotential->getRmax();
  //  CollectVisitingForces=realPotential->collectVisitingForces();
  //  int LastChain=Images.size()-1;
  //  for(int ic=1; ic<LastChain; ic++) {
  //    nnListEngine = Images[ic]->getNNEngine();
  //    if(nnListEngine == 0) {
  //      nnListEngine = NNEngineBuilder::create(realPotential->getLayoutXmlNode());
  //    } else {
  //      ERRORMSG("\tUsing an existing layout ")
  //    }
  //    Images[ic]->setNNEngine(nnListEngine);
  //    Images[ic]->updateLayout(Rmax,GridRadius);
  //    nnListEngine->initialize(*Images[ic],CollectVisitingForces);
  //    myPtcl->setNNEngine(nnListEngine);
  //    realPotential->setCurrentPos(Images[ic]->R);
  //    E(pot_energy,ic) = realPotential->evalEnergy();
  //  }
  //  myPtcl->setNNEngine(myLayout);
  //  realPotential->setCurrentPos(Images[LastChain]->R);
  //  E(pot_energy,LastChain) = realPotential->evalEnergy();
  //  
  //}

  NEBPotential::Scalar_t 
  NEBPotential::evalForce(ParticlePos_t& f0, unsigned stressmode) {
    return E(pot_energy,WorkingPtcl++);
  }


  void NEBPotential::getNEBForce(vector<Particle_t*>& Images) {

    size_type nat = Images[0]->getTotalNum();

    Tensor_t stress;

    int LastChain=Images.size()-1;
    int ic=1;
    int inn=0;
    while(ic<LastChain) {
      //convert(Images[ic]->Lattice, Images[ic]->R, R);
      //ParticlePos_t& R = *(Images[ic]->getVectorAttrib(TRAJECTORY_TAG));
      Images[ic]->convert(Images[ic]->R, myPtcl->R);
      Images[ic-1]->convert(Images[ic-1]->R, Lpos);
      Images[ic+1]->convert(Images[ic+1]->R, Rpos);

      Particle_t::ParticleLayout_t& Lattice = Images[ic]->Lattice;

      Lpos-=myPtcl->R;
      Rpos-=myPtcl->R;

      ParticlePos_t& F0(Images[ic]->F);
      realPotential->setCurrentPos(myPtcl->R);
      E(pot_energy,ic) = realPotential->evalForce(F0,0);

      // k is a const
      // calculate the force on the chain
      //Fc = SpringConst*(Lpos+Rpos) - SpringConst2*Ft;
      Fc = SpringConst*(Lpos+Rpos);

      // calculate the tangent for NEB
      if(neb_mode == 0) {
	F0 += Fc;
	E(f_norm2,ic) = Dot(F0,F0);
	E(fc_norm2,ic) = Dot(Fc,Fc);
      } else {

	//Lpos -= Ft;
	//Rpos -= Ft;
	Scalar_t d1 = Dot(Lpos,Lpos);
	Scalar_t d2 = Dot(Rpos,Rpos);
	Scalar_t d12 = Dot(Lpos,Rpos);

	Scalar_t x = sqrt(d1);
	Scalar_t y = sqrt(d2);
	Scalar_t norm = 1.0e0/(d1*y+d2*x+1e-12);
	
	d1 *= norm;
	d2 *= norm;

	Ft = d1*Rpos - d2*Lpos;
	Scalar_t tnorm = Dot(Ft,Ft);

	if(tnorm > 0) {
	  norm = 1.0e0/tnorm; 
	  Ft *= norm;
	  double dfproj = Dot(Ft,Fc)-Dot(Ft,F0);
	  F0 += dfproj*Ft;
	} else {
	  Ft = 0.0e0;
	}
    
	E(chain_dL,ic) = x;
	E(chain_dR,ic) = y;
	E(chain_theta,ic) = acos(d12/x/y+1e-12)*57.29577951308232;
	E(f_norm2,ic) = Dot(F0,F0);
	E(fc_norm2,ic) = Dot(Fc,Fc);
      }

      ic++;
    }

    //reset to zero
    WorkingPtcl = 0;
  }

  bool NEBPotential::put(xmlNodePtr cur){
    
#if defined(HAVE_LIBXML2)
    if(xmlHasProp(cur, (const xmlChar *) "mode")) {
      neb_mode = atoi((const char*)(xmlGetProp(cur, (const xmlChar *) "mode")));
    }
    cur = cur->xmlChildrenNode;
    while(cur != NULL) {
      if ((!xmlStrcmp(cur->name, (const xmlChar *)"parameter"))) {
	string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
	if(aname == "spring") {
	  putContent(SpringConst,cur);
	}
      }
      cur = cur->next;
    }
#endif
    SpringConst2 = SpringConst*2.0;
    return true;
  }
  void NEBPotential::print(ostream& os, int outputlevel) const {
  
  }

  NEBPotential::Scalar_t 
  NEBPotential::evalEnergy() {
    return 0.0;
  }

  bool
  NEBPotential::evalDynamicMatrix(DynMatrix_t&) {
    return false;
  }

  bool NEBPotential::get(ostream& os) const { 
    return true;
  }
  bool NEBPotential::put(istream&) {return true;}
  void NEBPotential::reset() {
    
  }


  RecordNEB::~RecordNEB() {
  }

  void RecordNEB::reset(const char* fileroot) {
  
    FileRoot = fileroot;
    LogFile = FileRoot;
    LogFile.append(".neb.dat");
  
    ofstream fout(LogFile.c_str());
    fout.setf(ios::left,ios::adjustfield);
    fout << "#Progress dat of NEB chains " << endl << "#     ";
    for(int i=0; i<pot_ref.Ename.size(); i++)
      fout <<setw(15) << pot_ref.Ename[i];
    fout << endl;
  }

  void RecordNEB::report(int iter) {
  
    if(stride && iter%stride == 0) {
      // two fixed ends are reported by the first and last mpi node
      int offset = 0;
      int nproc = Controller->ncontexts();
      int myid =  Controller->mycontext();
      int firstchain = 0;
      int lastchain = ptcl_ref.size();
      if(nproc > 1) {
  #if defined(HAVE_MPI)
        TagMaker neb_report_tag;
        if(myid == 0) {
  	firstchain = 0;
  	lastchain--;
  
  	///send the last chain index 
  	offset = lastchain-1;
  	reportNEB(iter,firstchain,lastchain);
   
  	MPI_Send(&offset, 1, MPI_INT, 1, neb_report_tag(),MPI_COMM_WORLD);
        } else {
  	MPI_Status status;
  	MPI_Recv(&offset, 1, MPI_INT, myid-1, neb_report_tag(),MPI_COMM_WORLD, 
  		 &status);
  	firstchain = 1;
  	if(myid < nproc-1) lastchain--;     
  
  	reportNEB(iter,firstchain,lastchain,myid,offset);
  
  
  	if(myid < nproc-1) {
  	  offset += lastchain-1;
  	  MPI_Send(&offset, 1, MPI_INT, myid+1, neb_report_tag(),MPI_COMM_WORLD);
  	}
        }
  #endif
      } else {
        reportNEB(iter,firstchain,lastchain);
      }
    }
  }

 void RecordNEB::reportNEB(int iter, int firstchain, int lastchain, int inode, int offset) {
 
   ofstream fout(LogFile.c_str(),ios::app);
   fout.setf(ios::scientific, ios::floatfield);
   fout.setf(ios::right,ios::adjustfield);
   if(!inode) fout << "#NEB iteration = " << iter << endl;
 
   for(int ic=firstchain; ic<lastchain; ic++) {
     fout << setw(3) << ic+offset;
     for(int i=0; i<pot_ref.Ename.size(); i++) {
       fout <<setw(15) << pot_ref.E(i,ic);
     }
     fout << endl;
   }
   
   char filename[128];
   sprintf(filename,"%s.iter=%d.xyz",FileRoot.c_str(),iter);
   
   ofstream *nebreport;
   if(inode) {
     nebreport = new ofstream(filename, ios::app);
   } else {
     nebreport = new ofstream(filename);
   }
 
   nebreport->setf(ios::scientific, ios::floatfield);
   nebreport->setf(ios::right,ios::adjustfield);
   
   SpeciesSet& tspecies = ptcl_ref[0]->getSpeciesSet();
   const Particle_t::ParticleIndex_t& ionid = ptcl_ref[0]->GroupID;
   for(int ic=firstchain; ic<lastchain; ic++) {
     //const ParticlePos_t& r = *(ptcl_ref[ic]->getVectorAttrib(TRAJECTORY_TAG));
     ptcl_ref[ic]->convert(ptcl_ref[ic]->R, pot_ref.Ft);    
     (*nebreport) << pot_ref.Ft.size() << endl << endl;
     for(int i=0; i<pot_ref.Ft.size(); i++)
       (*nebreport) << tspecies.Name[ionid[i]]
 		   << pot_ref.Ft[i] << endl;
   }
   delete nebreport;
 }

  bool RecordNEB::put(xmlNodePtr cur){
  
    xmlChar* t=xmlGetProp(cur, (const xmlChar *) "stride");
    if(t) {
      stride = atoi((const char*)t);
    }
    return true;
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 384 $   $Date: 2005-09-05 16:06:48 -0500 (Mon, 05 Sep 2005) $
 * POOMA_VERSION_ID: $Id: NEBPotential.cpp 384 2005-09-05 21:06:48Z jnkim $ 
 ***************************************************************************/
