//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "NanoApp/ArpackSKTB.h"
#include "Message/Communicate.h"
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseMatrix2MatrixCopy.h"
#include "SparseMatrix/Matrix2SparseMatrixProduct.h"
#include "Particle/NNEngine.h"
#include "Utilities/Clock.h"
namespace OHMMS {

  ArpackSKTB::ArpackSKTB(): 
    eig_solver(0), NeedAtomicProperties(false), WriteForDx(false)
 { 
    myRecorder = new ESOutput(*this);
    CollectVisitingForces = false;
  }

  ArpackSKTB::~ArpackSKTB() { 
    if(eig_solver) delete eig_solver;
  }

  bool ArpackSKTB::get(ostream&) const {
    return true;
  }

  bool ArpackSKTB::put(istream&){
    return true;
  }


  void ArpackSKTB::reset() {
  }
  
  bool ArpackSKTB::put(xmlNodePtr cur){
    m_cur = cur;
    return true;
  }

  void ArpackSKTB::tbInit() {

    //Before initializing physical data, process the xml elements
    int mynode = Controller->mycontext();
    xmlNodePtr cur = m_cur->children;
    while(cur != 0) {
      string cname((const char*)(cur->name));
      if(cname == "solver") {
	//eig_solver->put(cur);
	int pid = mynode;
	xmlChar *p = xmlGetProp(cur,(const xmlChar*)"node");
	if(p) {
	  pid = atoi((const char*)p);
	}
	if(pid == mynode) {
	  int item = Spectrum.size();
	  Spectrum.push_back(new ArpackParam());
	  Spectrum[item]->put(cur);
	}
      } else if(cname == "evaluate") {
        NeedAtomicProperties = true;
	xmlChar *p = xmlGetProp(cur,(const xmlChar*)"dx");
	if(p) {
	  WriteForDx = atoi((const char*)p);
	}
      }
      cur = cur->next;
    }

    //if spectrum is not specificed. Peform SM with zero shift
    if(Spectrum.empty()) {
      Spectrum.push_back(new ArpackParam());
    }

    int ncv = 0;
    for(int i=0; i<Spectrum.size(); i++) {
      ncv = max(ncv,Spectrum[i]->ncv);
    }
    //Overwrite the number of states by the maximum of the request eige states per solve
    NumStates = ncv;

    //Do the normal initialization
    setESParameters(myPtcl);

    States.N = BasisDim;
    States.M = NumStates;
    Gamma = true; 
    if(!KSet) {
      KSet = new KPointBase;
      KSet->addKpt(SingleParticlePos_t(0,0,0),1);
    }
    NumKpts = KSet->size();
    Occ.resize(NumKpts,NumStates);
    E.resize(NumKpts,NumStates);
    Occ = 0.0e0;
    E = 0.0e0;
  }

  ArpackSKTB::Scalar_t 
  ArpackSKTB::tbEnergy(){

    LOGMSG("ArpackSKTB::tbEnergy(): calculating eigen states and values. Rmax = " << Rmax)

    myPtcl->getPairlist(Rmax);

    int setmode = 0;
    setHamiltonian(setmode);

    if(eig_solver==0)  eig_solver = new ArpackSparseSolver(States);

    string fname(getName());
    fname.append(".h5");
    //hid_t hdfFile;
    hdfFile =  H5Fopen(fname.c_str(),H5F_ACC_RDWR,H5P_DEFAULT);
    if(hdfFile<0)
      hdfFile = H5Fcreate(fname.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);

    char gname[128];

    hid_t hgrp = H5Gcreate(hdfFile, "/BasisSet", 0);
    HDFAttribIO<vector<int> > blockout(Hmat.index()->ColBlock);
    blockout.write(hgrp,"blockoffset");
    H5Gclose(hgrp);

    Pooma::Clock timer;
    HDFAttribIO<ArpackSolver_t> out(*eig_solver);
    double cpu_io=0.0;
    for(int i=0; i<Spectrum.size(); i++) {

      Spectrum[i]->assign(*eig_solver);

      timer.start();
      if(Identity) 
	eig_solver->getEigvectors(Hmat, E);
      else
	eig_solver->getEigvectors(Hmat, Smat, E);
      timer.stop();

      LOGMSG("Arpack solver " << timer.cpu_time())
      eig_solver->checkEigvectors(Hmat,log());

      timer.start();
      sprintf(gname,"/Spectrum%d",i);
      hdfGroup = H5Gcreate(hdfFile, gname, 0);
      out.write(hdfGroup,gname);
      if(NeedAtomicProperties) getAtomicProperties(i);
      H5Gclose(hdfGroup);
      timer.stop();
      cpu_io += timer.cpu_time();
    }

    H5Fclose(hdfFile);
    LOGMSG("IO timing " << cpu_io)
    return 0.0;
  }

  void ArpackSKTB::getAtomicProperties(int i) {

    const int nstates = Spectrum[i]->ncv;
    RealMatrix_t aden(nstates,myPtcl->getTotalNum());
    
#pragma omp parallel
    {
      const ArpackSparseSolver::WorkArray_t &psi = eig_solver->eigVec;
      int myID = omp_get_thread_num();
      int first = myPtcl->getNNEngine()->first_id(myID);
      int last =  myPtcl->getNNEngine()->last_id(myID);
      const SparseIndex_t& hIndex = *(Hmat.index());
      for(int il=0; first<last; il++, first++) {
	int ib_first = hIndex.col_loc(first);
	int ib_last = hIndex.col_loc(first+1);
	for(int is=0; is<nstates; is++) {
	  real_type sum=0;
	  for(int ib= ib_first; ib<ib_last; ib++) 
	    sum+=pow(psi[is*BasisDim+ib],2); 
	  aden(is,first)=sum;
	}
      }
    }

    //split columns
    HDFAttribIO<RealMatrix_t> out(aden,true);
    out.write(hdfGroup,"AtomicOrbitalDensity");
    
    if(WriteForDx) { // don't need it anymore 
      char tfile[64];
      sprintf(tfile,"%s.set=%i.dx",getName().c_str(),i);
      ofstream fout(tfile);
      fout.setf(ios::scientific);
      fout.precision(5);
      fout << "object 1 class array type float rank 1 shape 3 items " 
	   << myPtcl->getTotalNum()
	   << " data follows " << endl;
      for(int iat=0; iat<myPtcl->getTotalNum(); iat++) {
	fout << myPtcl->R[iat] << endl;
      }
      const int ncol = 8;
      int ntot = myPtcl->getTotalNum();
      int nrow = ntot/ncol;
      int nfit = nrow*ncol;      

      for(int is=0; is<nstates; is++) {
	fout << "object " << is+2
	     << " class array type float rank 1 shape 1 items " 
	     << myPtcl->getTotalNum() << " data follows" << endl;
	for(int iat=0; iat<nfit;) {
	  for(int ic=0; ic<ncol; ic++,iat++)
	    fout << setw(12) << aden(is,iat);
	  fout << endl;
	}
	for(int iat=nfit; iat<ntot; iat++) fout << setw(12) << aden(is,iat);
	fout << endl;
      }
      for(int is=0; is<nstates; is++) {
        fout << "#list of the states " << endl;
        fout << "object \"state" << is << "\" class field" << endl;
        fout << "component \"positions\" value 1" << endl;
        fout << "component \"data\" value " << is+2<< endl;
      }
    }
  }
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 255 $   $Date: 2005-01-14 11:41:16 -0600 (Fri, 14 Jan 2005) $
 * $Id: ArpackSKTB.cpp 255 2005-01-14 17:41:16Z jnkim $ 
 ***************************************************************************/
