//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002, 2003 - by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <set>
#include "OhmmsUnits.h"
#include "ParticleBase/ParticleUtility.h"
#include "Utilities/SimpleParser.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "SlaterKosterTB/SlaterKosterTB.h"
#include "ParticleSets.h"
#include "ProjectData.h"
#include "IonicPotential.h"
#include "ElectronPhononCoupling.h"
using namespace OHMMS;

ElectronPhononCoupling::ElectronPhononCoupling(OhmmsElementSet& nodelist, const char* aname): 
  OhmmsElementBase(aname), NodeList(nodelist), 
  Width(-1.0) {
}

ElectronPhononCoupling::~ElectronPhononCoupling(){
}

bool 
ElectronPhononCoupling::action(bool fullmatrix){

  RecordList.reset();

  Particle_t* workingptcl
    = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle)))->get();

  SlaterKosterTB* pot  
    = dynamic_cast<SlaterKosterTB*>(dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential))->get()); 

  if(!pot) {
    ERRORMSG("Only quantum Hamiltonian can be used for electron-phonon couplings")
    return false;
  }

  if(Width>0) pot->Sigma = Width;
  pot->set(workingptcl);
  int nat = workingptcl->getTotalNum();
  ParticlePos_t F(nat);
  ParticleScalar_t Eloc(nat);
  workingptcl->advance();
  F = 0;
  Eloc = 0;
  Tensor_t stress;
  pot->evalForce(F, Eloc, stress, 0);
  RecordList.report(0);

  ProjectData* project 
    = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));

  RealMatrix_t Vep;

  string elphout(project->CurrentRoot());
  elphout.append(".elph");

  ofstream fout(elphout.c_str());
  fout.setf(ios::scientific, ios::floatfield);
  Scalar_t eavg;
  vector<int> eband_t;

  fout << "#phonon    frequency    |<f|dH|i>|^2 |<f|dH|i>|^2/E " << endl;
  if(fullmatrix) {//using density matrix

    for(int i=0; i<eband.size(); i++) {

      Scalar_t lambda = 0.0;

      for(int iph = 0; iph<Phonon.size(); iph++) {
	cout << "deltae " << DeltaE[iph] << endl;
	Scalar_t gp = pot->el_ph_coupling(*eband[i], Phonon[iph], eband_t, DeltaE[iph]);
	int deg = eband[i]->size()*eband_t.size();

	fout << setw(6) << Phonon[iph].ID
	     << setw(14) << setprecision(3) << Phonon[iph].Omega2;
	if(deg) {
 	  fout << setw(14) << setprecision(3) << gp
	       << setw(14) << setprecision(3) << gp/Phonon[iph].Omega2
	       << setw(14) << setprecision(3) << gp/Phonon[iph].Omega2/static_cast<Scalar_t>(deg)
	       << endl;
	} else {
 	  fout << setw(14) << setprecision(3) << 0.0
	       << setw(14) << setprecision(3) << 0.0
	       << setw(14) << setprecision(3) << 0.0
	       << endl;
	}
	//lambda += gp/omega2[thisphonon];
      }
      //fout  << setw(18) << setprecision(5) << lambda << endl;
    }
  } else {//eigen values due to the distortion is calculated

    // save original configration to R0
    ParticlePos_t R0(POSSTR,"r0");
    R0.resize(nat);
    workingptcl->Lattice.convert2Cart(workingptcl->R, R0);
    workingptcl->R.setUnit(PosUnit::CartesianUnit);

    RealMatrix_t E(pot->E);

    for(int i=0; i<eband.size(); i++) {

      for(int iph = 0; iph<Phonon.size(); iph++) {
	//int thisphonon = pband[iph];
	workingptcl->R = R0 + 0.01*Phonon[iph].Z;
	workingptcl->advance();
	pot->evalEnergy();

	Scalar_t deps2=0.0,deps;
	vector<int>::iterator it = eband[i]->begin();
	while(it != eband[i]->end()) {
	  int esub = (*it);
	  deps = pot->E(0,esub)-E(0,esub);
	  deps2 += deps*deps;
	}
      }
    }
  }
  return true; 
}

bool ElectronPhononCoupling::get(ostream& os) const {
  return true;
}
bool ElectronPhononCoupling::put(istream& is){
  return true;
}
void ElectronPhononCoupling::reset(){
}

#if defined(HAVE_LIBXML2)
bool 
ElectronPhononCoupling:: put(xmlNodePtr cur){

  string modefile, freqfile;
  vector<string> slist;
  bool fullmatrix = true;

  vector<int> pband;
  xmlNodePtr curref = cur;
  cur = cur->xmlChildrenNode;
  while (cur != NULL) {
    string cname((const char*)(cur->name));
    if (cname == "particlesets") {
      NodeList.getNode(Key_particle)->put(cur);
    } else if (cname == "potential") {
      NodeList.getNode(Key_potential)->put(cur);
    } else if(cname == "electronband") {
      int n = atoi((const char*)(xmlGetProp(cur, (const xmlChar *) "size")));
      eband.push_back(new vector<int>(n));
      putContent(*(eband.back()),cur);
    } else if(cname == "phononband") {
      pband.resize(atoi((const char*)(xmlGetProp(cur, (const xmlChar *) "size"))));
      putContent(pband,cur);
    } else if(cname == "record") {
      RecordList.put(cur);
    }
    cur = cur->next;
  }


  Particle_t* workingptcl
    = (dynamic_cast<ParticleSets*>(NodeList.getNode(Key_particle))->get();
  workingptcl->put(cur);

  int nat = workingptcl->getTotalNum();
  int totalmode = nat*OHMMS_DIM;
  vector<Scalar_t> omega2(nat*OHMMS_DIM);
  vector<Scalar_t> homega(nat*OHMMS_DIM);

  cur = curref->xmlChildrenNode;
  while (cur != NULL) {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"PhononEnergy"))) {
      if(xmlHasProp(cur, (const xmlChar *) "file")) {
	ifstream 
	  efile((const char*)(xmlGetProp(cur, (const xmlChar *) "file")));
	if(!efile) {
	  ERRORMSG("phonon energy is not found. Missing " << modefile)
          return false;
	}
	int n = getwords(slist,efile);
	if(n>0) {
	  int pmode_in = atoi(slist[0].c_str());
	  pband.resize(pmode_in);
	  for(int iph=0; iph<pmode_in; iph++) {
	    getwords(slist,efile);
	    pband[iph] = atoi(slist[0].c_str());
	    omega2[pband[iph]] = atof(slist[1].c_str());
	    homega[pband[iph]] = atof(slist[2].c_str())*thz_to_eV;
	  }
	} else {
	  for(int iph=0; iph<totalmode; iph++) {
	    getwords(slist,efile);
	    omega2[iph] = atof(slist[1].c_str());
	    homega[iph] = atof(slist[2].c_str())*thz_to_eV;
	  }
	  if(!pband.size()) {
	    pband.resize(totalmode);
	    for(int iph=0; iph<totalmode; iph++) pband[iph] = iph;
	  }
	}
      }
    }
    cur = cur->next;
  }

  int counter = 0;
  SingleParticlePos_t gamma;
  Phonon.reserve(pband.size());
  DeltaE.resize(pband.size(),0.0);

  cur = curref->xmlChildrenNode;
  while (cur != NULL) {
    if((!xmlStrcmp(cur->name, (const xmlChar *)"EnergyWidth"))) {
      putContent(Width,cur);
      DeltaE.resize(pband.size(),-1.0);
    }
    if((!xmlStrcmp(cur->name, (const xmlChar *)"PhononMode"))) {
      if(xmlHasProp(cur, (const xmlChar *) "file")) {
	ifstream pfile((const char*)(xmlGetProp(cur, (const xmlChar *) "file")));
	if(!pfile) {
	  ERRORMSG("phonon mode is not found. Missing " << modefile)
	    return false;
	}
	ParticlePos_t tp(nat);
	for(int iph=0; iph<totalmode; iph++) {
	  for(int j=0; j<nat; j++) {
	    pfile >> tp[j];
	  }
	  vector<int>::iterator it = find(pband.begin(), pband.end(), iph);
	  if(it != pband.end()) {
	    int thismode = (*it);
	    Phonon.push_back(PhononMode());
	    Phonon[counter].ID = thismode;
	    Phonon[counter].Omega2 = omega2[thismode];
	    Phonon[counter].E = homega[thismode];
	    Phonon[counter].Q = gamma;
	    Phonon[counter].Z.resize(nat); Phonon[counter].Z = tp;
	    counter++;
	  }
	}
      }
    }
    cur = cur->next;
  }
  return action(fullmatrix);
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ElectronPhononCoupling.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
