//////////////////////////////////////////////////////////////////
// (c) Copyright 2003  by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "OhmmsApp/NEBRun.h"
#include "OhmmsApp/RecordPropertyList.h"
#include "AppBase/ProjectData.h"
#include "OhmmsApp/PropagatorBuilder.h"
#include "OhmmsApp/IonicPotential.h"
#include "ParticleBase/ParticleUtility.h"
#include "ParticleIO/ParticleParser.h"
#include "ParticleIO/RecordParticle.h"
#include "NEB/NEBPotential.h"
#include "NEB/NEBPropagator.h"
#include "Message/Communicate.h"
#include "Message/TagMaker.h"
#include "Propagators/PropagatorBase.h"

//namespace OHMMS {
using namespace OHMMS;

  enum {chain_tag = 0,
	firstchain_tag,
	lastchain_tag,
	step_tag, 
	nproc_tag, 
	node_tag, 
	left_tag, 
	right_tag, 
	send_to_right,
	send_to_left,
	fix_ends, 
	with_ends,
	num_tags};


  NEBRun::NEBRun(OhmmsElementSet& nodelist, const char* aname): 
    OhmmsElementBase(aname), NodeList(nodelist), 
    nebPot(0), nebMover(0), nebRecorder(0),ptcl_ptr(NULL)
  { 

    I.resize(num_tags);
    I.Name[chain_tag] = "images";  I[chain_tag] = 8;
    I.Name[step_tag] = "steps";    I[step_tag]  = 10;
    I.Name[nproc_tag] = "mpinode"; I[nproc_tag] = 1;
    I.Name[node_tag] = "thisnode"; I[node_tag]  = 0;
    I.Name[left_tag] = "left";     I[left_tag]  = 0;
    I.Name[right_tag] = "right";   I[right_tag] = 0;
    I.Name[fix_ends] = "condition";  I[fix_ends]  = 1;
    I.Name[with_ends] = "interpolate";  I[with_ends]  = 1;
  }

  // WDR: handler implementations for NEBRun
  bool NEBRun::get(ostream& os) const {

    return true;
  }

  bool NEBRun::put(istream& is) {
    return true;
  }

  void NEBRun::reset() {  
    if(ptcl_ptr) {

      xmlNodePtr last=ptcl_ptr->last;
      xmlNodePtr cur = ptcl_ptr->children->next;
      while(cur != last) {
	xmlNodePtr t = cur;
	cur=cur->next;
	xmlUnlinkNode(t);
	xmlFreeNode(t);
      }

      for(int ic=1; ic<=I[chain_tag]; ic++) {
	xmlNodePtr anode =xmlNewNode(ptcl_ptr->ns,(const xmlChar*)"particleset");
	char newfile[128];
	sprintf(newfile,"%s.c%03d.restart.xml",RestartConfig.c_str(),ic);
	xmlNewProp(anode,(const xmlChar*)"src",(const xmlChar*)newfile);
        xmlAddPrevSibling(last,anode);
      }        
    }
  }

  ///Process the input data and DoNEB
  bool NEBRun:: put(xmlNodePtr cur){

    //create NEBPotential and NEBPropagator
    if(!nebPot) nebPot = new NEBPotential;
    if(!nebMover) nebMover = new NEBPropagator;

    int dump_stride = 1;
    xmlNodePtr neb_ptr=NULL;
    xmlNodePtr ensemble_ptr = NULL;
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
      string cname((const char*)(cur->name));
      if(cname == "neb") {
	neb_ptr = cur;
      } else if(cname == "particlesets") {
	ptcl_ptr = cur;
      } else if(cname == "particlelayout") {
	NodeList.getNode(Key_playout)->put(cur);
      } else if(cname == "loop") {
	if(xmlHasProp(cur, (const xmlChar *) I.Name[step_tag].c_str())) {
	  I[step_tag] =
	    atoi((const char*)(xmlGetProp(cur, (const xmlChar *) I.Name[step_tag].c_str())));
	}
	xmlChar *t= xmlGetProp(cur, (const xmlChar *)"stride");
	if(t) {
          dump_stride = atoi((const char*)t);
	}
      } else if(cname == "ensemble") { 
        nebMover->put(cur);
      } else  if(cname == "potential") {
	NodeList.getNode(Key_potential)->put(cur);
      }
      cur = cur->next;
    }

    if(!ptcl_ptr) {
      ERRORMSG("\tNo particle set exist do anything. Do nothing")
      return false;
    }

    //first count the number of sets to read : if less than two, get out
    cur = ptcl_ptr->children;
    int nchilds = 0;
    while(cur != NULL) {
      string cname((const char*)cur->name);
      if (cname == "particleset") {
	nchilds++;
      } else if(cname == "particlelayout") {
	NodeList.getNode(Key_playout)->put(cur);
      }
      cur= cur->next;
    }
    if(nchilds < 2) {
      ERRORMSG("\tNo valid particleset seems to exist.")
      return false;
    }

    if(neb_ptr) {
      cur = neb_ptr;
      //process the attributes
      xmlAttrPtr att = cur->properties;
      while(att != NULL) {
	string aname((const char*)(att->name));
	string vname((const char*)(att->children->content));
	if(aname == I.Name[chain_tag]) {
	  I[chain_tag]  = atoi(vname.c_str());
	} else if(aname == I.Name[fix_ends]) {
          if(vname == "fixed") I[fix_ends] = 1;
          else                 I[fix_ends] = 0;
	} else if(aname == I.Name[with_ends]) {
	  I[with_ends] = atoi(vname.c_str());
	  xmlSetProp(cur,att->name,(const xmlChar*)"0");
	}
	att = att->next;
      }
      nebPot->put(cur);
    }

    if(nchilds == 2) {//overwrite the interpolation property
      LOGMSG("\tNEBRun:: will interpolate the chains between the two ends")
      I.Name[with_ends] = 1;
    }

    cur = ptcl_ptr->children;
    if(I[with_ends]) { //read two configurations and interpolate
      while(cur != NULL) {
	if ((!xmlStrcmp(cur->name, (const xmlChar *)"particleset"))) {
	  Particle_t* aptcl = new Particle_t;
	  ParticleParser aHandle(*aptcl);
	  aHandle.put(cur);
	  Images.push_back(aptcl);
	}
	cur= cur->next;
      }
      DistributeChains();
    } else {
      I[chain_tag] = nchilds - 2;
      //distribute the chains among the processors
      DistributeChains();
      int i=-1; 
      while(cur != NULL) {
	if ((!xmlStrcmp(cur->name, (const xmlChar *)"particleset"))) {
	  if(i>=I[firstchain_tag] && i<I[lastchain_tag]) {
	    Particle_t* aptcl = new Particle_t;
	    ParticleParser aHandle(*aptcl);
	    aHandle.put(cur);
	    Images.push_back(aptcl);
            aptcl->convert2Cart(aptcl->R);
	  }
	  i++;
	}
	cur = cur->next;
      }
    }

    //time to make chains
    MakeChains();
    
    //number of inner chains
    int nc = I[lastchain_tag]-I[firstchain_tag] - 2;
    if(ptcl_save.empty()) {
      for(int ic=0; ic<nc; ic++) 
	ptcl_save.push_back(new SaveParticle(*(Images[ic+1])));
    }

    //prepare output
    ProjectData* project
      = dynamic_cast<ProjectData*>(NodeList.getNode(Key_project));

    //remove the process id
    RestartConfig = project->CurrentRoot();
    int ic = RestartConfig.find(".p");
    if(ic >= 0) {
      RestartConfig.erase(RestartConfig.begin()+ic,RestartConfig.end());
    }

    if(!nebRecorder) nebRecorder = new RecordNEB(Images,*nebPot);
    nebRecorder->stride = dump_stride;
    nebRecorder->reset(RestartConfig.c_str());
    char froot[128];
    int icg = ChainIndex[I[node_tag]]+1;
    for(int ic=0; ic<nc; ic++,icg++) {
      sprintf(froot,"%s.c%03d",RestartConfig.c_str(),icg);
      ptcl_save[ic]->stride = dump_stride;
      ptcl_save[ic]->reset(froot);
    }

    //initialize Potential
    nebPot->init((dynamic_cast<IonicPotential*>(NodeList.getNode(Key_potential)))->get(),Images);
    nebMover->setImages(Images);
    nebMover->initialize(0,nebPot,false);
    LOGMSG("End initializing images")
    return DoNEB();

  }

  bool NEBRun::DoNEB() {
    ////////////////////////////////////
    ///Do NEB!!!!
    ////////////////////////////////////
    bool moving = true;
    int iter = 0;
    int nc = ptcl_save.size();
    while(iter < I[step_tag] && moving) { // time-stepping
      nebRecorder->report(iter);
      nebPot->getNEBForce(Images);
      for(int ic=0; ic<nc; ic++) {
	ptcl_save[ic]->report(iter);
      }
      moving = nebMover->propagate(nebPot);
      iter++;
    }

    nebPot->getNEBForce(Images);
    nebRecorder->report(iter);
    for(int ic=0; ic<nc; ic++) {
      ptcl_save[ic]->report(0);
    }

    return true;
  }

  /** Pre-calculate the chain distribution
   */
  void NEBRun::DistributeChains(){

    //====================================================
    // Global Controller is called to determine how images 
    // are initialized and exchanged.
    // @todo This is a place to use grouped communicator.
    //====================================================
    I[nproc_tag] = Controller->ncontexts();
    I[node_tag] = Controller->mycontext();
    I[right_tag] = I[node_tag]+1;
    I[left_tag] = I[node_tag]-1;
    if(I[right_tag] == I[nproc_tag]) I[right_tag] = -1;

    ChainIndex.resize(I[nproc_tag]+1);

    //////////////////////////
    // create message tags
    //////////////////////////
    TagMaker move_to_right;
    TagMaker move_to_left;
    I[send_to_right] = move_to_right();
    I[send_to_left] = move_to_left();

    if(I[with_ends]) { // interpolation
      I[firstchain_tag] = -1;
      I[lastchain_tag] = 1;
    } else {

      int n = I[chain_tag]/I[nproc_tag];
      int dn = I[chain_tag]%I[nproc_tag];
      //vector<int> nc(I[nproc_tag]+1);
      ChainIndex[0] = 0;
      for(int i=0; i<I[nproc_tag]; i++) {
	if(i<dn) { 
	  ChainIndex[i+1] = ChainIndex[i] + n + 1;
	}
	else {
	  ChainIndex[i+1] = ChainIndex[i] + n;
	}
      }
      I[firstchain_tag] = ChainIndex[I[node_tag]]-1;
      I[lastchain_tag]  = ChainIndex[I[node_tag]+1]+1;
    }
  }

  void NEBRun::MakeChains(){

    int SpatialGrid = Particle_t::ParticleLayout_t::SPATIAL_GRID;
    if(I[with_ends]) {

      /// overwrite total number of chains to have the same number of images per node
      int localchain =I[chain_tag]/I[nproc_tag];

      for(int ip=0; ip<=I[nproc_tag]; ip++) ChainIndex[ip] = ip*localchain;

      I[chain_tag] = localchain*I[nproc_tag];
      I[firstchain_tag] = 0;
      I[lastchain_tag]  = localchain+2;

      ParticlePos_t Ri, Rf, delta;
      int natom = Images[0]->getTotalNum();
      delta.create(natom);
      Ri.create(natom);
      Rf.create(natom);

      delta.setUnit(PosUnit::LatticeUnit);
      Ri.setUnit(PosUnit::LatticeUnit);
      Rf.setUnit(PosUnit::LatticeUnit);

      double dn = 1.0e0/(static_cast<double>(I[chain_tag])+1.0e0);

      Particle_t::ParticleLayout_t lat;
      //lat = Images[0]->Lattice;
      lat.copy(Images[0]->Lattice);
      lat.R = 0.5*(Images[0]->Lattice.R + Images[1]->Lattice.R);
      lat.reset();
      lat.copyGrid(Images[0]->Lattice);

      Images[0]->applyBC(Images[0]->R, Ri);
      Images[1]->convert(Images[1]->R, Rf);

      delta=Rf-Ri;
      Images[0]->getMinimumDisplacement(delta);
      delta *= dn;

      Particle_t::Tensor_t 
	dlat = (Images[1]->Lattice.R - Images[0]->Lattice.R)*dn;
      Particle_t::Tensor_t 
	lat0 = Images[0]->Lattice.R;
      double ictot = static_cast<double>(I[node_tag]*localchain);

      for(int ic=0; ic<localchain; ic++) {
	Particle_t* aptcl = new Particle_t;
	aptcl->assign(*Images[0]);
	Images.push_back(aptcl);
	char newname[64];
	sprintf(newname,"%s_clone%d",Images[0]->getName().c_str(),ic);
	aptcl->setName(newname);
      }


      Ri += ictot*delta; // left configuration  
      Images[0]->convert2Cart(Ri, Images[0]->R);
      Images[0]->Lattice.R = lat0 + ictot*dlat;
      Images[0]->Lattice.reset();

      for(int ic=1; ic<I[lastchain_tag]; ic++) {
	//copy the lattice and re-evaluate the cell
        Images[ic]->Lattice.copy(lat);
	Images[ic]->Lattice.R = Images[ic-1]->Lattice.R + dlat;
	Images[ic]->Lattice.reset();
	Ri += delta;
	Images[ic]->convert2Cart(Ri, Images[ic]->R);
        //Probably not necessary
	Images[ic]->Lattice.copyGrid(Images[0]->Lattice);
      }

    } else {
      ///reset the local chain indices
      I[lastchain_tag] -= I[firstchain_tag];
      I[firstchain_tag] = 0;
    }
  }


//}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 385 $   $Date: 2005-09-05 17:02:25 -0500 (Mon, 05 Sep 2005) $
 * $Id: NEBRun.cpp 385 2005-09-05 22:02:25Z jnkim $ 
 ***************************************************************************/

