//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <iostream>
#include <iomanip>
#include "Message/OpenMP.h"
#include "OhmmsNumeric/MatrixOperators.h"
#include "OhmmsNumeric/Blasf.h"
#include "OhmmsNumeric/OhmmsBlasf.h"

#ifdef ADD_
#define abcdq    abcdq_
#define twoims   twoims_
#define atbsym   atbsym_
#define dotrace  dotrace_
#define xasymbc  xasymbc_
#endif

extern "C" {
  void abcdq(const double*, const double*, const double*, const double*,
	     double*, const int& n, const int& m);
  void atbsym(const double*, const double*, double*,
	      const int& , const int& );
  void twoims(const double*, double*, const int& , const int& );

  double dotrace(const double*, const double*, const int&);

  void xasymbc(const double*, const double*, const double&, double*,
	       const int&, const int&);
}


void OHMMS::debug_print(const RealMatrix_t& a) {

#ifdef OHMMS_UBLAS
  for(int i=0; i<a.size1(); i++) {
    for(int j=0; j<a.size2(); j++)
      cout << setw(10) << setprecision(3) << a(i,j);
    cout << endl;
  }
#elif defined(OHMMS_BLITZ)
  cout << a << endl;
#else
  for(int i=0; i<a.rows(); i++) {
    for(int j=0; j<a.cols(); j++)
      std::cout << setw(10) << setprecision(3) << a(i,j);
    std::cout << endl;
  }
#endif
}

void 
OHMMS::addDiagonal(RealMatrix_t& q, RealMatrix_t::Type_t x){

#ifdef OHMMS_UBLAS
  if(q.size1() != q.size2()) {
    cerr << "Illegal operation." << endl;
    return;
  }
  for(int i=0; i<q.size1(); i++) q(i,i) += x;
#elif  defined(OHMMS_BLITZ)
  if(q.length(0) != q.length(1)) {
    cerr << "Illegal operation." << endl;
    return;
  }
#else
  if(q.rows() != q.cols()) {
    cerr << "Illegal operation." << endl;
    return;
  }
  for(int i=0; i<q.rows(); i++) q(i,i) += x;
#endif
}

OHMMS::RealMatrix_t::Type_t
OHMMS::Dot(const RealMatrix_t& a, const RealMatrix_t& b) {
#ifdef OHMMS_UBLAS
  return OhmmsBlas::dot(a.size1()*a.size2(),a.data().begin(),b.data().begin());
#elif defined(OHMMS_BLITZ)
  return OhmmsBlas::dot(a.size(),a.data(),b.data());
#else
  return OhmmsBlas::dot(a.rows()*a.cols(),a.data(),b.data());
#endif
}

void 
OHMMS::axpy(RealMatrix_t::Type_t a, 
            const RealMatrix_t& x, RealMatrix_t& y) {
#ifdef OHMMS_UBLAS
  OhmmsBlas::axpy(x.size1()*x.size2(), a, x.data().begin(), y.data().begin());
#elif defined(OHMMS_BLITZ)
  OhmmsBlas::axpy(x.size(), a, x.data(), y.data());
#else
  OhmmsBlas::axpy(x.rows()*x.cols(), a, x.data(), y.data());
#endif
}


void 
OHMMS::copy(const RealMatrix_t& A, RealMatrix_t& B) {
  //copy function
  //\todo Maybe need to check the column dimension but for now assume that
  // two matrices have (n,m) and (n',m) where n == n' or  n != n'
#ifdef OHMMS_UBLAS
  int n = A.size2()*min(A.size1(),B.size1());
  OhmmsBlas::copy(n, A.data().begin(), B.data().begin());
#elif defined(OHMMS_BLITZ)
  int n = A.length(0)*min(A.length(1),B.length(1));
  OhmmsBlas::copy(n, A.data(), B.data());
#else
  int n = A.rows()*min(A.cols(),B.cols());
  OhmmsBlas::copy(n, A.data(), B.data());
#endif
}

// transpose(A)*A -> symmetric matrix
void 
OHMMS::xAAt_plus_yC(RealMatrix_t::Type_t x, const RealMatrix_t& A, RealMatrix_t::Type_t y, RealMatrix_t& tAA) {

  const char Uplo = 'L';
  const char Trans = 'T';
#ifdef OHMMS_UBLAS
  dsyrk(Uplo, Trans, tAA.size2(), A.size2(), x, A.data().begin(), A.size2(),
        y, tAA.data().begin(), tAA.size2());
#elif defined(OHMMS_BLITZ)
  dsyrk(Uplo, Trans, tAA.length(1), A.length(1), x, A.data(), A.length(1),  
	y, tAA.data(), tAA.length(1));
#else
  dsyrk(Uplo, Trans, tAA.cols(), A.cols(), x, A.data(), A.cols(),  
	y, tAA.data(), tAA.cols());
#endif
  fill_lower(tAA);
}

void 
OHMMS::xABt_plus_yC(RealMatrix_t::Type_t x, const RealMatrix_t& A, const RealMatrix_t& B,
		    RealMatrix_t::Type_t y, RealMatrix_t& C) {

  const char transa = 'T';
  const char transb = 'N';
#ifdef OHMMS_UBLAS
  dgemm(transa, transb,	B.size1(), C.size2(), A.size2(),
	x, B.data().begin(),B.size2(), A.data().begin(), A.size2(), y, 
	C.data().begin(), C.size2());

#elif defined(OHMMS_BLITZ)
  dgemm(transa,transb, B.length(0), C.length(1), A.length(1),
	x,B.data(), B.length(1), A.data(), A.length(1), y, C.data(), C.length(1));
#else
  dgemm(transa, transb,	B.rows(), C.cols(), A.cols(),
	x,B.data(),B.cols(), A.data(), A.cols(), y, C.data(), C.cols());
#endif

}


void 
OHMMS::xAB_plus_yC(RealMatrix_t::Type_t x, const RealMatrix_t& A, const RealMatrix_t& B,
		   RealMatrix_t::Type_t y,  RealMatrix_t& C) {

  // A(n,m) B(m,k) = C(n,k) (C/C++) <-> b(k,m) a(m,n) = c(k,n) (fortran)
  const char transa = 'N';
  const char transb = 'N';

#ifdef OHMMS_UBLAS
  dgemm(transa, transb, B.size2(), A.size1(), B.size1(),
	x, B.data().begin(), B.size2(), A.data().begin(), A.size2(),
	y, C.data().begin(), C.size2());

#elif defined(OHMMS_BLITZ)
  dgemm(transa, transb, B.length(1), A.length(0), B.length(0),
	x, B.data(), B.length(1), A.data(), A.length(1),
	y, C.data(), C.length(1));
#else
  dgemm(transa, transb, B.cols(), A.rows(), B.rows(),
	x, B.data(), B.cols(), A.data(), A.cols(),
	y, C.data(), C.cols());
#endif
}

void 
OHMMS::xABt_xBAt_plus_yC(RealMatrix_t::Type_t x, 
			 const RealMatrix_t& bra, const RealMatrix_t& ket,
			 RealMatrix_t::Type_t y, RealMatrix_t& C) {
  
  const char Uplo = 'L';
  const char Trans = 'T';
#ifdef OHMMS_UBLAS
  dsyr2k(Uplo,Trans, C.size2(),bra.size2(),x, 
	 bra.data().begin(), bra.size2(),
    	 ket.data().begin(), ket.size2(), y, 
	 C.data().begin(), C.size2());
#elif defined(OHMMS_BLITZ)
  dsyr2k(Uplo,Trans, C.length(1), bra.length(1),x, 
	 bra.data(), bra.length(1), ket.data(), ket.length(1), y,  
	 C.data(), C.length(1));
#else
  dsyr2k(Uplo,Trans, C.cols(),bra.cols(),x, 
	 bra.data(), bra.cols(), ket.data(), ket.cols(), y,  
	 C.data(), C.cols());
#endif
  fill_lower(C);
}

OHMMS::RealMatrix_t::Type_t
OHMMS::Trace(const RealMatrix_t& X, const RealMatrix_t& Y) {
  return Dot(X,Y);
  //return OhmmsBlas::dot(X.size1()*X.size2(),X.data().begin(),Y.data().begin());
  //return dotrace(X.data().begin(), Y.data().begin(), X.size1());
}


void OHMMS::fill_lower(RealMatrix_t& aa) {

#ifdef OHMMS_UBLAS
  for(int i=0; i<aa.size1(); i++)
    for(int j=0;j<i; j++) aa(i,j) = aa(j,i);
#elif defined(OHMMS_BLITZ)
  for(int i=0; i<aa.length(0); i++)
    for(int j=0;j<i; j++) aa(i,j) = aa(j,i);
#else
  for(int i=0; i<aa.rows(); i++)
    for(int j=0;j<i; j++) aa(i,j) = aa(j,i);
#endif
}

/*
void 
OHMMS::TwoIMinusS(const RealMatrix_t& psi, RealMatrix_t& q){
  twoims(psi.data().begin(),q.data().begin(),psi.size1(),psi.size2());
}


void OHMMS::AtB(const RealMatrix_t& bra, const RealMatrix_t& ket, 
		RealMatrix_t& H) {
  atbsym(bra.data().begin(), ket.data().begin(), H.data().begin(),
	 bra.size1(), bra.size2());
}

void OHMMS::ABminusCD(const RealMatrix_t& A, const RealMatrix_t& B,
		      const RealMatrix_t& C, const RealMatrix_t& D,
		      RealMatrix_t& G) {


  const char transa = 'N';
  const char transb = 'N';
  const double one = 1.0;
  const double minusone = -1.0;
  const double zero=0.0; 

  dgemm(transa,transb,B.size2(), A.size2(), B.size1(),
	one,B.data().begin(),B.size2(),
	A.data().begin(), A.size2(), zero, G.data().begin(), G.size2());

  dgemm(transa,transb,D.size2(), C.size2(), D.size1(),
	minusone,D.data().begin(),D.size2(),
	C.data().begin(), C.size2(), one, G.data().begin(), G.size2());

}


void  OHMMS::QHPsiMinusHPsi(const RealMatrix_t& Q, const RealMatrix_t& Psi_1, 
			    const RealMatrix_t& H, const RealMatrix_t& Psi, 
			    RealMatrix_t& G) {

  abcdq(Q.data().begin(), Psi_1.data().begin(),
	H.data().begin(), Psi.data().begin(),
	G.data().begin(), Psi_1.size1(), Psi_1.size2());
}


//c = alph*a*b : a = symmetric
void 
OHMMS::product(const RealMatrix_t& a, 
	       const RealMatrix_t& b, double alpha,
	       RealMatrix_t& c) {

  xasymbc(a.data().begin(), b.data().begin(), alpha,c.data().begin(),
       c.size1(), c.size2());
}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: MatrixOperators.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

