//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "TransformParticle.h"
#include "ParticleBase/ParticleFunctions.h"
#include "Utilities/OhmmsSpecies.h"
#include "Utilities/SpeciesCollection.h"
#include "Message/OpenMP.h"

namespace OHMMS {

  /*! \class PtclGeometryBase
   *  \author Jeongnim Kim
   *  \brief An abstract class to define a shape of a particle system
   */
  struct PtclGeometryBase: public OhmmsElementBase {

    typedef Particle_t::ParticlePos_t         ParticlePos_t;
    typedef Particle_t::ParticleIndex_t       ParticleIndex_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;

    PtclGeometryBase() { }
    ~PtclGeometryBase() { }

    //virtual void set(const vector<string>&) {}
    virtual void set(Particle_t&) { }

    virtual bool keepit(SingleParticlePos_t& p, int& id) {
      return true;
    }

  };

  /*!\class SphericalQD
   *\author JeongnimKim
   *\brief SphericalQD 
   */
  struct SphericalQD:public PtclGeometryBase {

    SphericalQD():CenterInUnit(true) { }
    ~SphericalQD() { }

    SingleParticlePos_t Center;
    double Radius2;
    bool CenterInUnit;
    bool AutoCenter;

    bool get(xmlDocPtr doc) const { return true;}

    bool put(xmlNodePtr cur){
      SpeciesBase* species = SpeciesCollection::getSpecies();
      double a0 = 1;
      AutoCenter = true;
      while(cur != NULL) {
	if ((!xmlStrcmp(cur->name, (const xmlChar *)"parameter"))) {
	  string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "role"));
	  if(aname == "scale") {
	    putContent(a0,cur);
	  } else if(aname == "radius") {
	    putContent(Radius2,cur);
	  } else if(aname == "center") {
	    putContent(Center,cur);
            AutoCenter = false;
	  } 
	}
	cur = cur->next;
      }
      ///scale radiums by the scaling factor
      Radius2 *= a0;
      if(!AutoCenter) { Center *= a0; } 
      Radius2 = Radius2*Radius2;
      return true;
    }

    void set(Particle_t& p) { 
      //Center = lattice.toCart(Center);
      //Radius2  = Radius2*Radius2;
      p.Lattice.BoxBConds[0]=0;
      p.Lattice.BoxBConds[1]=0;
      p.Lattice.BoxBConds[2]=0;
      if(AutoCenter) {
        Center = 0.5*(p.Lattice.a(0) + p.Lattice.a(1)+ p.Lattice.a(2));
      }
    }
    bool keepit(SingleParticlePos_t& p, int& id) {
      SingleParticlePos_t pp = p - Center;
      if(dot(pp,pp) < Radius2) return true;
      else              return false;
    }

    bool get(ostream& os) const {
      os << "center = " << Center << endl;
      os << "Radius = " << Radius2 << endl;
      return true;
    }

    bool put(istream& is) {
      return true;
    }

    void reset() {

    }

  };

  bool ShapeParticle::put(xmlNodePtr cur) {

    typedef Particle_t::ParticlePos_t         ParticlePos_t;
    typedef Particle_t::ParticleIndex_t       ParticleIndex_t;
    typedef Particle_t::SingleParticlePos_t   SingleParticlePos_t;
    typedef Particle_t::SingleParticleIndex_t SingleParticleIndex_t;

    SingleParticleIndex_t uc_grid(1);
    PtclGeometryBase* geom = NULL;

    const int mpigrid  = ParticleLayout_t::MPI_GRID;
    const int ompgrid  = ParticleLayout_t::OMP_GRID;
    const int finegrid = ParticleLayout_t::SPATIAL_GRID;
    const int gl       = ParticleLayout_t::GridLevel;
    vector<SingleParticleIndex_t> grid(gl,SingleParticleIndex_t(1));

    bool resetlattice = true;
    string selectedset = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
    cur = cur->xmlChildrenNode;
    while (cur != NULL) {
      if(!xmlStrcmp(cur->name, (const xmlChar *)"parameter")) {
        string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "role"));
	if(aname == "unitcell_grid") {
	  putContent(uc_grid,cur);
	} else if(aname == "grid") {
	  putContent(grid[finegrid],cur);
	} else if(aname == "ompgrid") {
	  putContent(grid[ompgrid],cur);
	} else if(aname == "mppgrid") {
	  putContent(grid[mpigrid],cur);
	} else if(aname == "lattice") {
	  putContent(out_.Lattice.R,cur);
	  resetlattice = false;
	}
      }  
      if (!xmlStrcmp(cur->name, (const xmlChar *)"geometry")){
	string shape =  (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
	if(shape == "sphere") {
	  geom = new SphericalQD;
	}
	if(geom) geom->put(cur->xmlChildrenNode);
	else {
	  ERRORMSG(shape << " is not implemented.")
	  return false;
	}
      }
      cur=cur->next;
    }

    ///\todogoing to use selectedset
    int natom = in_.getLocalNum();
    int ntot = uc_grid[0]*uc_grid[1]*uc_grid[2]*natom;

#ifdef ENABLE_OPENMP
    int  NumThreads = omp_get_max_threads();
    if(NumThreads != grid[1][0]*grid[1][1]*grid[1][2]) {
      grid[1][0] = 1; grid[1][1] = 1; grid[1][2] = NumThreads;
    }
#else
    int  NumThreads = 1;
#endif


    ///Prepare lattice
    if(resetlattice) {
      out_.Lattice.set(in_.Lattice,&uc_grid[0]);
    } else {
      out_.Lattice.reset();
    }

    //PartitionGrid(out_.Lattice,grid);
    ///Use only the grid partition
    out_.Lattice.makeGrid(grid);

    ParticleLayout_t& lattice = out_.Lattice;
    geom->set(out_);
  
    ///first convert the input position to unit, if not in unit
    convert2Unit(in_.Lattice,in_.R);
    vector<int> GroupID;
    vector<SingleParticlePos_t> Pos;
    int newid;
    SingleParticlePos_t del,pos;
    const ParticlePos_t& R = in_.R;
    for(int ic=0; ic<uc_grid[0]; ic++) {
      del[0] = static_cast<Scalar_t>(ic);
      for(int jc=0; jc<uc_grid[1]; jc++) {
	del[1] = static_cast<Scalar_t>(jc);
	for(int kc=0; kc<uc_grid[2]; kc++) {
	  del[2] = static_cast<Scalar_t>(kc);
	  for(int nb=0; nb<in_.getLocalNum(); nb++) {
	    pos = in_.Lattice.toCart(R[nb]+del);
	    newid = in_.GroupID[nb];	
	    if(geom->keepit(pos,newid)) {
	      Pos.push_back(pos);
	      GroupID.push_back(newid);
	    }
	  }
	}
      }
    }
    out_.create(Pos.size());
    out_.R.InUnit = false; /// using cartesian
    for(int iat=0; iat<Pos.size(); iat++) out_.R[iat] = Pos[iat];
    for(int iat=0; iat<GroupID.size(); iat++) out_.GroupID[iat] = GroupID[iat];
    for(int iat=0; iat<GroupID.size(); iat++) out_.ID[iat] = iat;
    //out_.NeedPartition = true;
    //out_.partition(false);
    return true;
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ShapeParticle.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
