//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
/***************************************************************************
 * C++ implementation of F90 module of the TBMD program written by
 * Dr. F. Kirchhoff and Prof. F. S. Khan Dept. of Electrical Engineering, 
 * Ohio State University  
 ***************************************************************************/
//#include "PolyMinFunc.h"

//Two points, Quadratic functional
double line_minimization(double x1, double x2, 
                         double f1, double f2, double df1,
                         double& fmin, vector<double>& fit) {
  double a, b, c, det;
  double x1sq = x1*x1;
  double x2sq = x2*x2;
  double x12 = x1*x2;

  det = (x1sq - 2.0e0*x12 + x2sq);

  a = (-f1 + f2 + df1*x1 - df1*x2)/det;

  b = (2.0e0*f1*x1 - 2.0e0*f2*x1 - df1*x1sq + df1*x2sq)/det;

  c = (f2*x1sq - 2.0e0*f1*x12 + df1*x1sq*x2 + 
       f1*x2sq - df1*x1*x2sq)/det;

  double x = -0.5e0*b/a;

  fmin = x*(a*x + b) + c;

  fit[2] = a;
  fit[1] = b;
  fit[0] = c;
  return x;
}

//Three points, Quadratic functional
double line_minimization(double x1, double x2, double x3,
                          double f1, double f2, double f3,
                          double& fmin, vector<double>& fit) {
  double a,b,c,det;
  
  double x1sq = x1*x1;
  double x2sq = x2*x2;
  double x3sq = x3*x3;


  det = (x1sq*x2 - x1*x2sq - x1sq*x3 + 
         x2sq*x3 + x1*x3sq - x2*x3sq);
  det = 1.e0/det;
  a = (-f2*x1 + f3*x1 + f1*x2 - f3*x2 - f1*x3 + f2*x3)*det;
        
  b = (f2*x1sq - f3*x1sq - f1*x2sq + 
       f3*x2sq + f1*x3sq - f2*x3sq)*det;

  c = (f3*x1sq*x2 - f3*x1*x2sq - f2*x1sq*x3 + 
       f1*x2sq*x3 + f2*x1*x3sq - f1*x2*x3sq)*det;

  double x = -0.5e0*b/a;

  fmin = x*(a*x + b) + c;

  fit[2] = a;
  fit[1] = b;
  fit[0] = c;
  return x;
}

//Four points, Cubic functional
double line_minimization(double x1, double x2, double x3, double x4,
                          double f1, double f2, double f3, double f4,
                          double& fmin, vector<double>& fit) {

  double a, b, c, d, det, delta;
  double x1_min, x2_min, ddf1, ddf2;

  a = f1/((-x1 + x2)*(-x1 + x3)*( x1 - x4)) + 
      f2/((-x1 + x2)*(-x2 + x3)*(-x2 + x4)) + 
      f3/(( x2 - x3)*(-x1 + x3)*(-x3 + x4)) + 
      f4/(( x1 - x4)*( x2 - x4)*(-x3 + x4));

  b =  f4*(x1 + x2 + x3)/(( x3 - x4)*(-x1 + x4)*(-x2 + x4)) + 
       f3*(x1 + x2 + x4)/(( x2 - x3)*(-x1 + x3)*( x3 - x4)) + 
       f2*(x1 + x3 + x4)/((-x1 + x2)*(-x2 + x3)*( x2 - x4)) + 
       f1*(x2 + x3 + x4)/((-x1 + x2)*(-x1 + x3)*(-x1 + x4));

  c =  f4*(x1*x2 + x1*x3 + x2*x3)/((-x1 + x4)*(-x2 + x4)*(-x3 + x4)) + 
       f3*(x1*x2 + x1*x4 + x2*x4)/((-x1 + x3)*(-x2 + x3)*( x3 - x4)) + 
       f2*(x1*x3 + x1*x4 + x3*x4)/((-x1 + x2)*(-x2 + x3)*(-x2 + x4)) + 
       f1*(x2*x3 + x2*x4 + x3*x4)/((-x1 + x2)*(-x1 + x3)*(x1 - x4));

  d =   f1*x2*x3*x4/(( x1 - x2)*(-x1 + x3)*( x1 - x4)) + 
        f2*x1*x3*x4/((-x1 + x2)*(-x2 + x3)*( x2 - x4)) + 
        f4*x1*x2*x3/(( x3 - x4)*(-x1 + x4)*(-x2 + x4)) + 
        f3*x1*x2*x4/(( x1 - x3)*( x2 - x3)*(-x3 + x4));


  fit[3] = a;
  fit[2] = b;
  fit[1] = c;
  fit[0] = d;

  delta = 4.0e0*( b*b - 3.0e0*a*c );

  if(delta<0.0e0) {
    return -1.0;
  } else {
    x1_min = ( -2.0e0*b - sqrt(delta) )/(6.0e0*a);
    x2_min = ( -2.0e0*b + sqrt(delta) )/(6.0e0*a);

    ddf1 = 6.0e0*a*x1_min + 2.0e0*b;
    ddf2 = 6.0e0*a*x2_min + 2.0e0*b;

    double x;
    if(ddf1 > 0.0e0) {
      x = x1_min;
    } else if (ddf2 > 0.0e0) {
      x = x2_min;
    } else {
      ERRORMSG("minum_cubic : 2nd derivative < 0 at both extrema !" << endl);
    }

    fmin = x*(x*(a*x + b) + c) + d;
    return x;
  }
}

//Three points, Cubic functional
double line_minimization(double x1, double x2, double x3,
                         double f1, double f2, double f3, double df1,
                         double& fmin, vector<double>& fit) {
  double a, b, c, d, det, delta;
  double x1_min, x2_min, ddf1, ddf2;

//  double x12sq = pow(-x1 + x2, 2);
//  double x13sq = pow(-x1 + x3, 2);
  double x12sq = (-x1 + x2)*(-x1+x2);
  double x13sq = (-x1+x3)*(-x1 + x3);
  a =  f2/(x12sq*(x2 - x3)) + 
      df1/((-x1 + x2)*(-x1 + x3))   + 
       f3/(x13sq*(-x2 + x3))+ 
       f1*(-2.0e0*x1 + x2 + x3)/(x12sq*x13sq);

  b =  f3*(2.0e0*x1 + x2)/(( x2 - x3)*x13sq) + 
       f2*(2.0e0*x1 + x3)/(x12sq*(-x2 + x3)) + 
      df1*(x1 + x2 + x3)/((-x1 + x2)*(x1 - x3))  + 
       f1*(3.0e0*x1*x1 - x2*x2 - x2*x3 - x3*x3)/ 
      (x12sq*x13sq);

  c =  f3*x1*(x1 + 2.0e0*x2)/(x13sq*(-x2 + x3)) + 
       f2*x1*(x1 + 2.0e0*x3)/(x12sq*(x2 - x3))  + 
      df1*(x1*x2 + x1*x3 + x2*x3)/((-x1 + x2)*(-x1 + x3)) + 
       f1*x1*(-3.0e0*x1*x2 + 2.0e0*x2*x2 - 
       3.0e0*x1*x3 + 2.0e0*x2*x3 + 2.0e0*x3*x3)/ 
       (x12sq*x13sq);

  d = df1*x1*x2*x3/((-x1 + x2)*(x1 - x3)) + 
       f3*x1*x1*x2/((x2 - x3)*x13sq) + 
       f2*x1*x1*x3/(x12sq*(-x2 + x3)) + 
       f1*x2*x3*(3.0e0*x1*x1 - 2.0e0*x1*x2 - 
       2.0e0*x1*x3 + x2*x3)/ 
       (x12sq*x13sq);

  fit[3] = a;
  fit[2] = b;
  fit[1] = c;
  fit[0] = d;

  delta = 4.0e0*( b*b - 3.0e0*a*c );
  if(delta<0.0e0) {
    return -1.0;
  } else {

    x1_min = ( -2.0e0*b - sqrt(delta) )/(6.0e0*a);
    x2_min = ( -2.0e0*b + sqrt(delta) )/(6.0e0*a);
    ddf1 = 6.0e0*a*x1_min + 2.0e0*b;
    ddf2 = 6.0e0*a*x2_min + 2.0e0*b;

    double x;
    if(ddf1 > 0.0e0) {
      x = x1_min;
    } else if (ddf2 > 0.0e0) {
      x = x2_min;
    } else {
      ERRORMSG("minum_cubic : 2nd derivative < 0 at both extrema !" << endl);
    }

    fmin = x*(x*(a*x + b) + c) + d;
    return x;
  }
}

double line_minimization(double x1, double x2,
                         double f1, double f2, double df1, double df2,
                         double& fmin, vector<double>& fit, unsigned& info) {
  double a, b, c, d, det, delta;
  double x1_min, x2_min, ddf1, ddf2;

  info = 0;
  double x1s = x1*x1, x1c = x1*x1s, x1q = x1s*x1s;
  double x2s = x2*x2, x2c = x2*x2s, x2q = x2s*x2s;


  det = (x1q - 4.0e0*x1c*x2 + 6.0e0*x1s*x2s - 
       4.0e0*x1*x2c + x2q);

  a = (- 2.0e0*f1*x1 +  2.0e0*f2*x1 + df1*x1s + 
       df2*x1s +  2.0e0*f1*x2 -  2.0e0*f2*x2 - 
       2.0e0*df1*x1*x2 -  2.0e0*df2*x1*x2 + df1*x2s + 
       df2*x2s)/det;      

  b = (3.0e0*f1*x1s - 3.0e0*f2*x1s - 
       df1*x1c - 2.0e0*df2*x1c + 3.0e0*df2*x1c*x2 - 
       3.0e0*f1*x2s + 3.0e0*f2*x2s + 3.0e0*df1*x1*x2s - 
       2.0e0*df1*x2c - df2*x2c)/det;

  c = (df2*x1q - 6.0e0*f1*x1s*x2 + 6.e0*f2*x1s*x2 + 
       2.0e0*df1*x1c*x2 + 6.0e0*f1*x1*x2s -           
       6.0e0*f2*x1*x2s - 3.0e0*df1*x1s*x2s -        
       3.0e0*df2*x1s*x2s + 2.0e0*df2*x1*x2c + df1*x2q)/det;

  d = (f2*x1q - 4.0e0*f2*x1c*x2 - df2*x1q*x2 +       
       3.0e0*f1*x1s*x2s+ 3.0e0*f2*x1s*x2s- 
       df1*x1c*x2s + 2.0e0*df2*x1c*x2s - 
       4.0e0*f1*x1*x2c + 2.0e0*df1*x1s*x2c - 
       df2*x1s*x2c + f1*x2q - df1*x1*x2q)/det;

  fit[3] = a;
  fit[2] = b;
  fit[1] = c;
  fit[0] = d;

  delta = 4.0e0*( b*b - 3.0e0*a*c );

  if(delta<0.0e0) {
    return -1.0;
  } else {

    x1_min = ( -2.0e0*b - sqrt(delta) )/(6.0e0*a);
    x2_min = ( -2.0e0*b + sqrt(delta) )/(6.0e0*a);
    ddf1 = 6.0e0*a*x1_min + 2.0e0*b;
    ddf2 = 6.0e0*a*x2_min + 2.0e0*b;

    double x;
    if(ddf1 > 0.0e0) {
      x = x1_min;
    } else if (ddf2 > 0.0e0) {
      x = x2_min;
    } else {
      ERRORMSG("minum_cubic : 2nd derivative < 0 at both extrema !" << endl);
    }

    fmin = x*(x*(a*x + b) + c) + d;
    return x;
  }
}

/*
void find_poly_extremas(vector<double>& e, vector<double>& z, 
                        unsigned& number_of_zeros){
// e: polynomial coefficients, e[0] = zeroth order 
//                             e[n] = nth order
  unsigned n = e.size(); 
  vector<dcomplex> a(n), root(n);
  vector<double> radius(n), apoly(n), apolyr(n);
  vector<bool> err(n);
  unsigned nitermax = 30;

  n--; // order of the polynomial

  double norm = 1.0e0/static_cast<double>(n)/e[n];

  for(int i=1; i<= n; i++) 
    a[i] = complex(norm*static_cast<double>*e[i], 0.0e0);      

  polzeros(a, root, radius, err, apoly, apolyr);
n, a.begin(), eps, big, small, nitermax,
           root, radius, err, iter, apoly, apolyr);

  eps = 0.000001
  number_of_zeros = 0
    for(i=1;i< n; i++) {
      if(abs(root(i)) <= eps) {
      } else {
        z(number_of_zeros) = real(root(i));
        number_of_zeros = number_of_zeros + 1
	  }
    }

}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: line_min_ftns.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
