//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#include "Particle/EventDetectorBase.h"
#include "Utilities/OhmmsSpecies.h"
#include "Utilities/SpeciesCollection.h"

#include <stdio.h>
#include <iostream>
#include <fstream>
using namespace std;

//MOD 11112000
EventDetectorBase::EventDetectorBase():
  RefPos(POSSTR,"edbrefpos"), DeltaR(POSSTR,"edbdisp"),
  RefEne(SCALARSTR,"edbrefene"), Ene(POSSTR, "edbene"),
  PtclRef(NULL), LocalNum(0)
{
  AdjustCM = false; Rcut = 0.5; Counter = -1;
  TraceBC = false; // trance boundary crossing
}

EventDetectorBase::EventDetectorBase(int argc, char **argv):
  RefPos(POSSTR,"edbrefpos"), DeltaR(POSSTR,"edbdisp"),
  RefEne(SCALARSTR,"edbrefene"), Ene(POSSTR, "edbene"),
  PtclRef(NULL), LocalNum(0)
{
  AdjustCM = false; Rcut = 0.5; Counter = -1;
  int i = 0;
  while(i<argc) {
   if(!strcmp(argv[i],"--rcut")) {
     Rcut = atof(argv[++i]);
   } else if(!strcmp(argv[i],"--adjustcm")) { 
     AdjustCM = true;
   } else if(!strcmp(argv[i],"--tracebc")) {
     TraceBC = true;
   } else {
     options.push_back(argv[i]);
   }
    i++;
  }
}

EventDetectorBase::EventDetectorBase(vector<string>& argv):
  RefPos(POSSTR,"edbrefpos"), DeltaR(POSSTR,"edbdisp"),
  RefEne(SCALARSTR,"edbrefene"), Ene(POSSTR, "edbene"),
  PtclRef(NULL), LocalNum(0)
{
  AdjustCM = false; Rcut = 0.5; Counter = -1;
  getoptions(argv);
}

void EventDetectorBase::getoptions(vector<string>& argv) {
  int i = 0;
  while(i<argv.size()) {
   if(argv[i] == "--rcut") {
     Rcut = atof(argv[++i].c_str());
   } else if(argv[i] == "--adjustcm") { 
     AdjustCM = true;
   } else if(argv[i] == "--tracebc") {
     TraceBC = true;
   } else {
     options.push_back(argv[i]);
   }
    i++;
  }
}

EventDetectorBase::~EventDetectorBase(){

}

void 
EventDetectorBase::initialize(Particle_t* p, const char* name){
  
  if(LocalNum > 0) {
    // already inintialize
    // clean up to prepare a new detector
    Counter = -1;
    LocalNum = 0;
  }

  PtclRef = p;
  FileRoot = name;
  Species.erase(Species.begin(), Species.end());
  SpeciesBase* species = SpeciesCollection::getSpecies();
  Species.reserve(species->getTotalNum());
  for(int i=0; i<species->getTotalNum(); i++) {
    Species.push_back(species->Name[i]);
  }
}

bool 
EventDetectorBase::detect(ParticlePos_t& newPos, Scalar_t& drmax) {

  drmax = 0.0e0;

  Scalar_t tmp = 0.0e0;
  Scalar_t RcutSq = Rcut*Rcut;
 
  SingleParticlePos_t newcm = 0.0e0,dr, dr2,accumbc = 0.0e0;
  int i=0;

  // assuming all the positions are in cartesian coordinates
  // but should be generalized.
  if(AdjustCM) {

    for(i=0; i<LocalNum; i++) {
      newcm += newPos[i];
      DeltaR[i] = newPos[i]-RefPos[i];
      dr = PtclRef->Lattice.BConds.apply(PtclRef->Lattice.toUnit(DeltaR[i]));
      dr2 = PtclRef->Lattice.toCart(dr);
      // return the change, all in the unit of supercell
      accumbc += (dr2 - DeltaR[i]);
    }

    Scalar_t dninv = 1.0e0/static_cast<Scalar_t>(LocalNum);
    newcm = dninv*(newcm+accumbc) - RefCM;

    i = 0;
    do {

      dr = PtclRef->Lattice.BConds.apply(
           PtclRef->Lattice.toUnit(DeltaR[i]-newcm));
      dr2 = PtclRef->Lattice.toCart(dr);
      tmp = dot(dr2,dr2);
      drmax = max(tmp,drmax);
      if(drmax > RcutSq) {
        return true; 
      }
      i++;
    } while(i < LocalNum);

  } else {
      i=0;
      do {
        dr = PtclRef->Lattice.BConds.apply(
           PtclRef->Lattice.toUnit(newPos[i]-RefPos[i]));
        dr2 = PtclRef->Lattice.toCart(dr);
        tmp = dot(dr2,dr2);
        drmax = max(tmp,drmax);
        if(drmax > RcutSq) return true;
        i++;
      } while(i < LocalNum);
  }
  return false;
}

//////////////////////////////////////////////////////////////////////////////
void 
EventDetectorBase::setRefConfig(ParticlePos_t& newPos, ParticleScalar_t& newE) {

  // first time to set the reference config
  if(Counter < 0) {
    LocalNum = PtclRef->getLocalNum();
    RefPos.create(LocalNum);
    DeltaR.create(LocalNum);
    RefEne.create(LocalNum);
    if(TraceBC) {
      AccumBC.create(LocalNum);
      AccumBC = 0.0;
    }
  }
   
  RefPos = newPos;
  RefEne = newE;
///////////////////////////////////////////////
// Add special functions to trace boundary crossings if TraceBC == true
//      if(TraceBC) {
//        for(int i=0; i<LocalNum; i++) {
//          RefPos[i] = 
//            PtclRef->Lattice.BConds.apply(PtclRef->Lattice.toUnit(newPos[i]));
//        }
//        AccumBC += (newPos - refPos);
//      }
///////////////////////////////////////////////
  if(AdjustCM) {
    RefCM = 0.0e0;
    for(int i=0; i<LocalNum; i++) RefCM += RefPos[i];
    double dninv = 1.0e0/static_cast<double>(LocalNum);
    RefCM *= dninv;
  }

  Counter++;
}

void 
EventDetectorBase::saveRefConfig(int frame){

  if(frame == 0) {
    ofstream fxmol(FileRoot.c_str());
    fxmol.close();
  }

  double etot = 0.0e0;
  for(int i=0; i<LocalNum; i++) etot += RefEne[i];

  ofstream fxmol(FileRoot.c_str(),ios::app);
  fxmol.setf(ios::fixed, ios::floatfield);
  fxmol << LocalNum << endl;
  fxmol << "TRANS = " << frame << " Epot = " << etot << endl;

  if(TraceBC) {
    for(int i=0; i<LocalNum; i++) {
      fxmol << Species[PtclRef->GroupID[i]]
            << RefPos[i]
	    << setw(12) << setprecision(4) << RefEne[i]
	    << AccumBC[i] << endl;
    }
  } else {
    for(int i=0; i<LocalNum; i++) {
      fxmol << Species[PtclRef->GroupID[i]]
	    << RefPos[i]
	    << setw(12) << setprecision(4) << RefEne[i] << endl;
    }
  }
  fxmol.close();
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: EventDetectorBase.cpp 151 2004-11-02 23:51:14Z jnkim $ 
***************************************************************************/
