//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Message/OpenMP.h"
#include "Utilities/OhmmsInfo.h"
#include "Particle/NNEngine.h"
#include <algorithm>
#include "Utilities/UtilityFunctions.h"
namespace OHMMS {


/** Allocate pair-list data
 * @param ptcl the particleset whose list data will be stored in PairAttrib
 * @param needvisitingforce true, if the interaction uses visiting force
 *
 * Default data management is handled by this function once the data distribution
 * is determined. The default distribution of the particles over the threads is based
 * on the sequence of the particles, see makeFairPartition. 
 * The size of PairAttrib is normally the number of OpenMP threads but is not limited
 * to the thread number. With OpenMP, it is not necessary to re-evaluate OmpAttrib, the
 * thread index for each particle. However, a derived NNEngine manages the update
 * if it uses spatial-dependent, dynamic partition.
 */
void NNEngine::allocateAttributes(Particle_t& ptcl, bool needvisitingforce) {

  needvisitingforce = needvisitingforce && PairAttrib.size()>1;

  if(needvisitingforce) {
    OmpAttrib = ptcl.getIndexAttrib(openmp_tag);
  }

#pragma omp parallel 
  {
    omp_int_t ip = omp_get_thread_num();
    if(!PairAttrib[ip]) {
      PairAttrib[ip]= new PtclPairList_t;
    }
    //PairAttrib[ip]->MultiSet = (PairAttrib.size()>1);
    Index_t first=PtclDist[ip], last = PtclDist[ip+1];
    if(MaxPairsPerPtcl>0) {
      PairAttrib[ip]->set(ip,first,last);
      PairAttrib[ip]->resizePairs(last-first,MaxPairsPerPtcl);
    }
    if(OmpAttrib) for(Index_t j=first; j<last; j++) (*OmpAttrib)[j]=ip;
  }

}

void NNEngine::collectForce(ParticlePos_t& f, omp_int_t ip) { 

  //do nothing
}


void 
NNEngine::makeNewPairs(Particle_t& ptcl, const ParticlePos_t& pos) {

#pragma omp parallel for schedule(static,1)
  for(Index_t ip=0; ip<NumThreads; ip++) {
    PtclPairList_t& nnlist = *PairAttrib[ip];
    evalPairsPerThread(ptcl.Lattice,pos, ptcl.GroupID, nnlist);

    //map<string,vector<MetisConnection*>*>::iterator it(Connections.begin());
    //map<string,vector<MetisConnection*>*>::iterator it_end(Connections.end());
    //while(it != it_end) {
    //  vector<MetisConnection*>& v(*((*it).second));
    //  v[ip]->refresh(nnlist); 
    //  //v[ip]->refresh(nnlist.M, nnlist); 
    //  ++it;
    //}
  }

  //sync has the identical function
  TotalNumPairs = 0; 
  PtclDist[0]=0;
  for(Index_t ip=0; ip<NumThreads; ip++) {
    TotalNumPairs += PairAttrib[ip]->getTotNadj();
    PtclDist[ip+1]=PtclDist[ip]+PairAttrib[ip]->getLocalNum();
    PairAttrib[ip]->set(ip,PtclDist[ip],PtclDist[ip+1]);
  }
}

void NNEngine::updatePairs(Particle_t& ptcl, const ParticlePos_t& del) {

#pragma omp for schedule(static,1) nowait
  for(Index_t ip=0; ip<NumThreads; ip++) {
    updatePairsPerThread(del,*PairAttrib[ip]);
  }
}

void 
NNEngine::updatePairsPerThread(const ParticlePos_t& del, PtclPairList_t& myppl) {

  Index_t n(myppl.getLocalNum());
  for(Index_t iL=0; iL<n; iL++) {
    PtclPairList_t::iterator nit(myppl.begin(iL)),nit_end(myppl.end(iL));
    const SingleParticlePos_t& d(del[myppl.I[iL]]);
    while(nit != nit_end) {
      (*nit).addDelta(del[(*nit).Next]-d);
      ++nit;
      //SingleParticlePos_t dr((*nit).displ + del[(*nit).Next]-d); 
      //(*nit++).set(sqrt(dot(dr,dr)),dr);
    }
    //for(int nnloc=myppl.M[iL]; nnloc<myppl.M[iL+1]; nnloc++) {
    //  PairData_t& thispair= myppl[nnloc];
    //  SingleParticlePos_t dr = thispair.displ + del[thispair.J]-del[iat]; 
    //  thispair.set(sqrt(dot(dr,dr)),dr);
    //}
  }
}

void NNEngine::rescale(Scalar_t ascale) {
//#pragma omp parallel 
//  {
//    PairAttrib[ip]->rescale(ascale);
//  }
}


//TESTING with ForceConstant: OpenMP is not allowed
void NNEngine::makeMove(Index_t iatMove, const SingleParticlePos_t& displ) {

  PtclPairList_t& myppl(*PairAttrib[0]);
  SingleParticlePos_t mdispl(-1.0*displ);
  PtclPairList_t::iterator pij(myppl.begin(iatMove)),pij_end(myppl.end(iatMove));
  while(pij != pij_end) {
    int jat=(*pij).Next;
    (*pij).addDelta(mdispl);

    PtclPairList_t::iterator pji(myppl.begin(jat)),pji_end(myppl.end(jat));
    bool search_pji=true;
    while(search_pji && pji != pji_end) {
      if((*pji).Next == iatMove) {
        (*pji).addDelta(displ); //(*jit).sep=pij.sep; (*jit).displ=-pij.displ;
        search_pji=false;
      }
      ++pji;
    }
    ++pij;
  }
}

void NNEngine::restoreMove(Index_t i) {
}

/**update the neighbor-list of the iat particle
 *
 * The connectivity is unchanged but the distance and directional
 * relation has changed.
 *\todo the function is not fully implemented and not yet used by other functions
 */
void NNEngine::update(const Particle_t& p, Index_t iat) {
  sync();
}
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 377 $   $Date: 2005-08-28 15:32:58 -0500 (Sun, 28 Aug 2005) $
 * $Id: NNEngine.1.cpp 377 2005-08-28 20:32:58Z jnkim $ 
 ***************************************************************************/

