//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <list>
#include <iostream>
#include <fstream>
#include "Message/OpenMP.h"
#include "Message/Communicate.h"
#include "Particle/NNRegGridEngine.h"
#include "ParticleBase/ParticleUtility.h"
#include "Particle/PLayoutFunctions.h"
using namespace OHMMS;

NNRegGridEngine::NNRegGridEngine(){
  LOGMSG("Creating NNRegGridEngine")
  setID(ThisMethod);
  UseLayout = true;
}

NNRegGridEngine::~NNRegGridEngine() { 
}

/** prepare the spatial grid to perform linked-cell calculations
 *@param p the particle whose neighbor lists are evaluated by this engine.
 *@param rmax the maximum interaction radius
 *@return true NNRegiGridEngine uses the cell index of the current configuration
 * 
 *- make a linked-cell list with the spatial grid
 *- apply boundary conditions to the current position p.R
 *- create the data sets for each grid
 *- update the layout information, such as cell indices with p.R
 *- partition the local particles over the number of threads
 */ 
bool
NNRegGridEngine::initialize(Particle_t& p, bool needvisitingforce) {

  DEBUGMSG("NNRegGrid::initialize")
  Rmax = p.Lattice.InteractionRadius;
  RmaxSq = Rmax*Rmax;


  //p.applyBC(p.R,p.curR);
  if(Cell==0) {
    Cell = new ParticleIndex_t(p.getLocalNum());
  }

  ParticleLayout_t::Grid_t* sg=p.Lattice.getGrid(SPATIAL_GRID);
  if(sg->empty()) sg->createData();
  p.curR.setUnit(PosUnit::LatticeUnit);
  Index_t nloc_max = updateGridData(p.R,p.curR, p.Lattice);


  if(MaxConnectedCells != p.Lattice.MaxConnections) { 
    makeFairPartition(p);
    MaxConnectedCells = p.Lattice.MaxConnections;
    MaxPtclsPerCell = nloc_max;
    MaxPairsPerPtcl= (MaxConnectedCells+1)*(MaxPtclsPerCell+1);
    if(PresetMaxNN>0) {
      MaxPairsPerPtcl = std::min(PresetMaxNN,MaxPairsPerPtcl);
    }
    allocateAttributes(p,UseVisitingForce);
    LOGMSG("\tLinked cell data")
    LOGMSG("\t  Max connected cells " << MaxConnectedCells)
    LOGMSG("\t  Max numer of particles per cell " << MaxPtclsPerCell)
    LOGMSG("\t  Max numer of pairs per particle " << MaxPairsPerPtcl)
  }

  LOGMSG("NNRegGridEngine::UseVisitingForce = false")
  UseVisitingForce=false;

  return true;
}

NNRegGridEngine::Index_t 
NNRegGridEngine::updateGridData(const ParticlePos_t& pos, 
  ParticlePos_t& pu, ParticleLayout_t& lattice) {

  PLayoutFunc::applyBC(lattice,pos,pu);

  if(lattice.MaxConnections > MaxConnectedCells) {
    WARNMSG("\tPotential problems with the linked cells")
  }

  //Work with spacial grid
  ParticleLayout_t::Grid_t* restrict sg=lattice.getGrid(SPATIAL_GRID);
  Index_t nat(pos.size());
  sg->clearData();
  for(int iat=0; iat<nat; iat++) {
    Index_t iloc = sg->loc(pu[iat]);
    (*Cell)[iat] = iloc;
    sg->addData(iloc,iat);
  }
  return sg->getMaxDataPerGrid();
}

void 
NNRegGridEngine::evalPairsPerThread(const ParticleLayout_t& lattice,
    const ParticlePos_t& pos, const ParticleIndex_t& grID,
    PtclPairList_t& myppl) {

  myppl.reset(); // reset the pair lists

  const ParticleIndex_t&  cell = *Cell;
  Index_t LocalNum(myppl.getLocalNum());
  Index_t iat(myppl.First);

  const ParticleLayout_t::Grid_t* restrict sg=lattice.getGrid(SPATIAL_GRID);

  for(Index_t iL=0; iL<LocalNum; iL++,iat++) {
    Index_t ic = cell[iat]; //get the cell index
    myppl.I[iL] = iat;    //save the global index
    ParticleLayout_t::const_iterator it = sg->beginData(ic);
    ParticleLayout_t::const_iterator it_end = sg->endData(ic);
    while(it != it_end) { ///check the particles in the same grid
      Index_t jat = *it++;
      if(iat == jat) continue;
      SingleParticlePos_t dr(lattice.toCart(pos[jat]-pos[iat]));
      Scalar_t rsq(dot(dr,dr));
      if(rsq< RmaxSq && rsq > RminSq) {
        myppl.addPair(iL,jat,jat,grID[jat],sqrt(rsq),dr);
      }          
    }//it != it_end
      
    //==================================================
    // iterate over the neighboring cells
    //==================================================
    for(Index_t nc=lattice.first_connected(ic); nc<lattice.last_connected(ic);nc++) {
      Index_t jc = lattice.id_connected(nc);//the cell index of the neighbor cell
      SingleParticlePos_t del = lattice.bc(nc);
      ParticleLayout_t::const_iterator jt(sg->beginData(jc));
      ParticleLayout_t::const_iterator jt_end(sg->endData(jc));
      while(jt != jt_end) {//check the particles of the current neighbor cell
	Index_t jat = *jt++;
	SingleParticlePos_t dr(lattice.toCart(pos[jat]-pos[iat]) + del);
	Scalar_t rsq(dot(dr,dr));
	if(rsq< RmaxSq && rsq > RminSq) {
          myppl.addPair(iL,jat,jat,grID[jat],sqrt(rsq),dr);
	}
      }//jt!=jt_end
    }//for(Index_t nc;;)
  }//go thru all the particles for this thread
    
  myppl.checkPairs(UseVisitingForce);
}

bool NNRegGridEngine::putExtra(xmlNodePtr cur){
  return true;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: NNRegGridEngine.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 *******************************************************************/
