//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <set>
#include <list>
#include <numeric>
#include <algorithm>
#include <iterator>
#include <fstream>
#include "Message/OpenMP.h"
#include "Message/Communicate.h"
#include "OhmmsPETE/OhmmsMatrix.h"
#include "Particle/NNSpatialEngine.h"
#include "Particle/PLayoutFunctions.h"
#include "Utilities/OhmmsInfo.h"
using namespace OHMMS;

int 
NNSpatialEngine::updateGridData(const ParticlePos_t& pos, ParticlePos_t& pu, 
    ParticleLayout_t& lattice) {

  //This is where particles will be swapped
  PLayoutFunc::applyBC(lattice,pos,pu);

  ParticleLayout_t::Grid_t* sg(lattice.getGrid(SPATIAL_GRID));
  sg->clearData();

  for(int iat=0; iat<NumLocalPtcls; iat++) {
    int iloc = sg->loc(pu[iat]);
    (*Cell)[iat] = iloc;
    sg->addData(iloc,iat);
  }

  setMessagePatterns(sg);
  return sg->getMaxDataPerGrid();
}

void 
NNSpatialEngine::evalPairsPerThread(const ParticleLayout_t& lattice, 
                                    const ParticlePos_t& pos, 
                                    const ParticleIndex_t& grID,
                                    PtclPairList_t& myppl) {

  NumLocalPtcls=myPtcl->getLocalNum();
  GhostList.resize(NumGhosts);

  sendrecvGhostParticles(pos);

  myppl.reset(); // reset the pair lists

  const ParticleIndex_t&  cell(*Cell);
  const ParticleLayout_t::Grid_t& sg(*lattice.getGrid(SPATIAL_GRID));

  SingleParticlePos_t dr;
  int iat(myppl.First);//iat/iL are identical but keep them separate for openmp
  int grouptype; //species type
  for(int iL=0;iL<NumLocalPtcls; iL++, iat++) {
    int ic(cell[iat]); //get the cell index
    myppl.I[iL] = iat;    //save the local index
    ParticleLayout_t::const_iterator it(sg.beginData(ic));
    ParticleLayout_t::const_iterator it_end(sg.endData(ic));
    while(it != it_end) { ///check the particles in the same grid
      int jat = *it++;
      if(iat == jat) continue;
      if(jat<NumLocalPtcls) {
        grouptype=grID[jat];
        dr=lattice.toCart(pos[jat]-pos[iat]);
      } else {
        const GhostDataType& g(GhostList[jat-NumLocalPtcls]);
        grouptype= g.GroupID;
        dr=lattice.toCart(g.R-pos[iat]);
      }
      Scalar_t rsq(dot(dr,dr));
      if(rsq< RmaxSq && rsq > RminSq) {
        myppl.addPair(iL,jat,jat,grouptype,sqrt(rsq),dr);
      }          
    }//it != it_end
      
    //==================================================
    // iterate over the neighboring cells
    //==================================================
    for(int nc=lattice.first_connected(ic); nc<lattice.last_connected(ic);nc++) {
      int jc = lattice.id_connected(nc);//the cell index of the neighbor cell
      SingleParticlePos_t del = lattice.bc(nc);
      ParticleLayout_t::const_iterator jt(sg.beginData(jc));
      ParticleLayout_t::const_iterator jt_end(sg.endData(jc));
      while(jt != jt_end) {//check the particles of the current neighbor cell
	int jat = *jt++;
        if(jat<NumLocalPtcls) {
          grouptype=grID[jat];
          dr=lattice.toCart(pos[jat]-pos[iat]) + del;
        } else {
          const GhostDataType& g(GhostList[jat-NumLocalPtcls]);
          grouptype= g.GroupID;
          dr=lattice.toCart(g.R-pos[iat]) + del;
        }
	Scalar_t rsq = dot(dr,dr);
	if(rsq< RmaxSq && rsq > RminSq) {
	  myppl.addPair(iL,jat,jat,grouptype,sqrt(rsq),dr);
	}
      }//jt!=jt_end
    }//for(int nc;;)

  }//go thru all the particles for this thread
    
  myppl.checkPairs(UseVisitingForce);

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: NNSpatialEngine.1.cpp,v 1.1 2005/05/28 22:30:01 jnkim Exp 
 ****************************************************************************/
