//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <list>
#include "Message/OpenMP.h"
#include "Message/Communicate.h"
#include "Particle/NNRegGridMethod.h"
using namespace OHMMS;
#include "ParticleBase/ParticleUtility.h"

NNRegGridMethod::NNRegGridMethod():NNMethodBase(ThisMethod) { }

NNRegGridMethod::~NNRegGridMethod() { }

void 
NNRegGridMethod::updateR(Particle_t& p, const ParticlePos_t& pos) {
  if(pos.InUnit) {
    for(int iG=0; iG<p.getLocalNum(); iG++){
      p.R[iG] = p.Lattice.BConds.wrap(pos[iG]);
    }
  } else {
    for(int iG=0; iG<p.getLocalNum(); iG++){
      p.R[iG] = p.Lattice.BConds.wrap(p.Lattice.toUnit(pos[iG]));
    }
  }
}

void 
NNRegGridMethod::partition(Particle_t& p) {
  ///p.partition(false);
}

void 
NNRegGridMethod::get(const Particle_t& p, PtclPairList_t& myppl, int myID) {

  myppl.reset(); // reset the pair lists

  int ntot  = p.getTotalNum();
  if(CellID.size()<ntot) CellID.resize(ntot);
  const ParticleLayout_t& lat = p.Lattice;
  const ParticlePos_t& R = p.R;
  vector<list<int> > nsub(lat.dGrid[2]->getTotalNum());
  for(int iG=0; iG<ntot; iG++){
    int iloc = p.Lattice.dGrid[2]->loc(R[iG]);
    CellID[iG] = iloc;
    nsub[iloc].push_back(iG);
  }

  int first = p.Lattice.dGrid[1]->PtclDist[myID];
  int last = p.Lattice.dGrid[1]->PtclDist[myID+1];
  int maxnn = 0;
  int iL = 0;
  SingleParticlePos_t dr;
  Scalar_t rsq;    
  for(int ig=first; ig<last; ig++, iL++) {
    int iG = p.ID[ig];
    int ic = CellID[iG]; //get the cell index
    myppl.I[iL] = iG;  //save the global index
    int nn = 0;
    ///get the neighbors within a cell
    list<int>::iterator it = nsub[ic].begin();
    while(it != nsub[ic].end()) {
      int jG = (*it);
      if(iG != jG) { 
	dr = lat.toCart(R[jG]-R[iG]);
	rsq = dot(dr,dr);
	if(rsq< RmaxSq && rsq > RminSq) {
	  rsq = sqrt(rsq);
	  Scalar_t rinv = 1.0e0/rsq;
	  myppl.add(iL,jG,rsq,rinv*dr);
	  nn++;
	}          
      }
      it++;
    }///

    ////////////////////////////////
    // neighboring cells
    ////////////////////////////////
    for(int nc=NextCell.first(ic); nc<NextCell.last(ic);nc++) {
      int jc = NextCell.id(nc);
      list<int>::iterator jt = nsub[jc].begin();
      SingleParticlePos_t del = NextCell.bc(nc);
      while(jt != nsub[jc].end()) {
	int jG = *jt;
	dr = lat.toCart(R[jG]-R[iG]) + del;
	rsq = dot(dr,dr);
	if(rsq< RmaxSq && rsq > RminSq) {
	  rsq = sqrt(rsq);
	  Scalar_t rinv = 1.0e0/rsq;
	  myppl.add(iL,jG,rsq,rinv*dr); //  logical index jg is used
	  nn++;
	}
	jt++;
      }
    }
    maxnn = (maxnn>nn)? maxnn:nn;
  }//go thru all the atoms

  myppl.add(iL,-1,rsq,dr);
  myppl.setMaxNadj(maxnn);

}


int 
NNRegGridMethod::init(Particle_t& p, Scalar_t rmax) {

  RmaxSq = rmax*rmax;
  int ntot = p.getTotalNum();
  ParticleLayout_t& Lattice = p.Lattice;

/*
  ///going to move it to OhmmsParticle
  bitset<OHMMS_DIM> checkbound;
  for(int idim=0; idim<OHMMS_DIM; idim++) {
    if(!p.Lattice.BoxBConds[idim]) checkbound.set(idim);
  }
    
  if(checkbound.any()) {
    SingleParticlePos_t lower(1e12,1e12,1e12);
    SingleParticlePos_t upper(-1e12,-1e12,-1e12);
    for(int i=0; i<ntot; i++) {
      for(int idim=0; idim<OHMMS_DIM; idim++) {
        lower[idim] = min(lower[idim],p.R[i][idim]);
        upper[idim] = max(upper[idim],p.R[i][idim]);
      }
    }
    if(!p.R.InUnit) {
      lower = Lattice.toUnit(lower);
      upper = Lattice.toUnit(upper);
    }
   
    bool outofbound = false;
    for(int idim=0; idim<OHMMS_DIM; idim++) {
      if(checkbound[idim]) {
        if(lower[idim]<0.0 || upper[idim] >= 1.0) outofbound = true;
      }
    }
    if(outofbound) {
      ERRORMSG("lower bound [0,1) " << lower)
      ERRORMSG("upper bound (0,1) " << upper)
      Shift = 0.5-0.5*(lower+upper); 
      ERRORMSG("Going to shift all positions by " << Shift)
    } 
  }
*/

  vector<list<int> > nsub(Lattice.dGrid[2]->getTotalNum());
  if(p.R.InUnit) {
    for(int iG=0; iG<ntot; iG++){
      p.R[iG] = Lattice.BConds.wrap(p.R[iG]);
    }
  } else {
    for(int iG=0; iG<ntot; iG++){
      p.R[iG]=Lattice.BConds.wrap(Lattice.toUnit(p.R[iG]));
    }
  }
  p.R.InUnit = true;
  int iloc;
  for(int iG=0; iG<ntot; iG++) {
    iloc = Lattice.dGrid[2]->loc(p.R[iG]);
    nsub[iloc].push_back(iG);
  }

  size_t maxnat = 0;
  for(size_t ig=0; ig<nsub.size(); ig++) 
    maxnat = max(maxnat,nsub[ig].size());

  ///make a linked-cell list with Grid[2]
  int maxnc = NextCell.connect(p.Lattice,rmax,2);

  simplePartition(p);

  //if(Controller->master()){
  //  p.Lattice.print(cout);
  //}
  LOGMSG("\tNumber of maximum linked cells " << maxnc)
  return (maxnc+1)*(maxnat+1);
}

bool NNRegGridMethod::putExtra(xmlNodePtr cur){
  return true;
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: NNRegGridMethod.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

