//////////////////////////////////////////////////////////////////
// (c) Copyright 2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_PARTICLEBASE_H
#define OHMMS_PARTICLEBASE_H

#include <string>
#include <vector>
#include <map>
#include "ParticleTags.h"


/**@file ParticleBase.h
 *@brief Declaration of ParticleBase<PT>
 */

/** Container class for particle attributes. 
 *
 * The template parameter PT is a trait class which provides typedefs. 
 * See, e.g., ParticleTraits.h
 */
template<class PT>
class ParticleBase: public ParticleTags {

public:

  typedef typename PT::ParticleLayout_t      ParticleLayout_t;
  typedef typename PT::size_type             size_type;
  typedef typename PT::difference_type       difference_type;
  typedef typename PT::Index_t               Index_t;
  typedef typename PT::Scalar_t              Scalar_t;
  typedef typename PT::SingleParticleIndex_t SingleParticleIndex_t;
  typedef typename PT::SingleParticlePos_t   SingleParticlePos_t;
  typedef typename PT::Tensor_t              Tensor_t;

  ///Integer quantity associated with each atom, e.g., ID, GroupID
  typedef typename PT::ParticleIndex_t       ParticleIndex_t;

  /// Scalar quantity associated with each atom, e.g., Local Energy
  typedef typename PT::ParticleScalar_t      ParticleScalar_t;

  /// D-dimensional vector quantity associated with each atom, e.g., R
  typedef typename PT::ParticlePos_t         ParticlePos_t;

  /// D-dimensional tensor quantity associated with each atom, e.g., R
  typedef typename PT::ParticleTensor_t      ParticleTensor_t;

  /// iterator for the map between an attribute name and particle attribute object
  typedef map<string,OhmmsObject*>::iterator  PAListIterator;
  typedef map<string,OhmmsObject*>::const_iterator  PAListConstIterator;

  /// Internal counter  
  size_type Counter;

  /// The number of local particles
  size_type LocalNum;

  /** The number of global particles. 
   *
   * GlobalNum is normally identical to LocalNum. For distributed particle system over MPI nodes,
   * GlobalNum is the sum of LocalNums on each node.
   */
  size_type GlobalNum;

  /// ParticleLayout
  ParticleLayout_t  Lattice; 

  /**@defgroup BasicAttribs Common particle attributes
   *@{*/
  /// unique, persistent ID for each particle 
  ParticleIndex_t   ID;  

  /// Group(Species) ID
  ParticleIndex_t   GroupID;  

  /// Position
  ParticlePos_t    R;    
 
  /// Instantaneous Position in unit used by layout/nnlist engines, not added to myAttribList
  ParticlePos_t    curR;    
  /*@}*/

  /// Default constructor
  ParticleBase();

  /// Copy constructor
  ParticleBase(const ParticleBase<PT>& P) { assign(P); }

  /// Virtual destructor
  virtual ~ParticleBase();

  /** return a type id of the named type
   *@param tname type nape
   *@return matching enum value
   */
  inline difference_type getAttribType(const string& tname) const {
    typename map<string,size_type>::const_iterator it = AttribTypeMap.find(tname);
    if(it == AttribTypeMap.end()) 
      return -1;
    else
      return (*it).second;
  }

  /// return the number of attributes
  inline size_type getNumAttrib() const { return AttribList.size();}

  /// return the const_iterator pointing to the first attribute
  inline PAListConstIterator first_attrib() const { return AttribList.begin(); } 

  /// return the const_iterator pointing to the last attribute
  inline PAListConstIterator last_attrib() const { return AttribList.end(); } 

  /**@defgroup AccessAttribs Functions to access particle attributes
   *
   * In order to enable fast access using index to the attributes of the basic datatype, 
   * a vector storage for each data type, INDEX (ParticleIndex_t), VAL (ParticleScalar_t), 
   * POS (ParticlePos_t) and TENZOR(ParticleTensor_t), is separately kept.
   * An external user should save the return value of addAttribute functions
   * to utilize the access functions get(Index|Scalar|Vector|Tensor)Attrib.
   * @{*/
  /** check if a named attribute exists 
   * @param attrib_name name of the attribute
   * @return true, if the attribute is found.
   */
  bool hasAttrib(const std::string& attrib_name) const;

  /** add a named attribute of a type
   * @param tname datatype of a new attribute
   * @param oname name of a new attribute
   * @return index to access to the attribute
   *
   * A new attribute is created by this and added to myAttribList. 
   */
  size_type addAttribute(const std::string& tname, const std::string& oname);
  /** add a new attribute of index type
   * @param pa reference to ParticleIndex_t object
   * @return index to access to the attribute
   *
   * The object is created externally and this object only keeps the pointerer
   * to the physical data. 
   */
  size_type addAttribute(ParticleIndex_t& pa);
  size_type addAttribute(ParticleScalar_t& pa);
  size_type addAttribute(ParticlePos_t& pa);
  size_type addAttribute(ParticleTensor_t& pa);

  /** return a pointer to the name attribute of index type*/
  ParticleIndex_t*  getIndexAttrib(const std::string& aname); 
  /** return a pointer to the name attribute of scalar type*/
  ParticleScalar_t* getScalarAttrib(const std::string& aname);
  /** return a pointer to the name attribute of vector type*/
  ParticlePos_t*    getVectorAttrib(const std::string& aname);
  /** return a pointer to the name attribute of tensor type*/
  ParticleTensor_t* getTensorAttrib(const std::string& aname);

  /** return a const pointer to the name attribute of index type
   * @param i the index return by addAttribute
   */
  inline const ParticleIndex_t*  getIndexAttrib(size_type i) const { return INDEX[i];}
  /** return a const pointer to the name attribute of scalar type*/
  inline const ParticleScalar_t* getScalarAttrib(size_type i) const { return VAL[i];}
  /** return a const pointer to the name attribute of vector type*/
  inline const ParticlePos_t*    getVectorAttrib(size_type i) const { return POS[i];}
  /** return a const pointer to the name attribute of tensor type*/
  inline const ParticleTensor_t* getTensorAttrib(size_type i) const { return TENZOR[i];}
  inline ParticleIndex_t*  getIndexAttrib(size_type i)  { return INDEX[i];}
  inline ParticleScalar_t* getScalarAttrib(size_type i) { return VAL[i];}
  inline ParticlePos_t*    getVectorAttrib(size_type i) { return POS[i];}
  inline ParticleTensor_t* getTensorAttrib(size_type i) { return TENZOR[i];}
  /* @{*/

  void create(size_type m);
  void create(const vector<size_type>& agroup);

  void resize(size_type m);
  void clear();
  void erase(size_type first, size_type last);

  virtual void assign(const ParticleBase<PT>& ptclin) {
    create(ptclin.getLocalNum());
    Lattice.copy(ptclin.Lattice);
    R.InUnit = ptclin.R.InUnit;
    R = ptclin.R;
    ID = ptclin.ID;
    GroupID = ptclin.GroupID;
  }

  ///return the number of local particles
  inline size_type getLocalNum() const { return LocalNum;}

  ///return the number of global particles
  inline size_type getTotalNum() const { return GlobalNum;}

  ///return the number of groups
  inline size_type groups() const { return SubPtcl.size()-1;}
  
  ///return the first index of a group i
  inline size_type first(size_type igroup) const { return SubPtcl[igroup];}
  
  ///return the last index of a group i
  inline size_type last(size_type igroup) const { return SubPtcl[igroup+1];}
  
  /// Returns the current counter
  inline size_type  current() const { return Counter;}

  /// Increments the counter
  inline void advance() { Counter++;}

  /// Sets the counter
  inline void setCounter(size_type i = 0) { Counter = i;}

protected:

  ///array to handle a group of distinct particles per species
  std::vector<size_type>             SubPtcl;

  ///AttribTypeMap[name] maps the attribute type to the index
  map<string,size_type>             AttribTypeMap;
  ///Name2Index[name] maps the named attribute to the index
  map<string,size_type>             Name2Index;
  ///AttribList[name] maps the named attribute to the pointer to the attribute
  map<string,OhmmsObject*>    AttribList;
  ///Storage of the attributes of index type
  vector<ParticleIndex_t*>    INDEX;
  ///Storage of the attributes of scalar type
  vector<ParticleScalar_t*>   VAL;
  ///Storage of the attributes of vector type
  vector<ParticlePos_t*>      POS;
  ///Storage of the attributes of tensor type
  vector<ParticleTensor_t*>   TENZOR;
  ///Storage of the attributes created by this
  vector<OhmmsObject*>        myAttribList;
};

#include "ParticleBase/ParticleBase.cpp"
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 369 $   $Date: 2005-07-22 20:45:59 -0500 (Fri, 22 Jul 2005) $
 * $Id: ParticleBase.h 369 2005-07-23 01:45:59Z jnkim $ 
 ***************************************************************************/
