//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_PARTICLEUTILITY_H
#define OHMMS_PARTICLEUTILITY_H
//////////////////////////////////////////////////////////
// functors to generate vectors with each element [-0.5,0.5)
//////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////
// specialized of RandomVector in Utilities/RandomGenerator.h for TinyVector<T,D>
//////////////////////////////////////////////////////////

#include "Utilities/RandomGenerator.h"

template<class T, std::size_t D, class RNG>
struct 
RandomVector<TinyVector<T,D>, RNG> {

  typedef TinyVector<T,D> Return_t;
  typedef std::size_t     size_type;

  RandomVector(RNG& gen): RG(gen){ }

  inline Return_t operator()() {
    TinyVector<T,D> res;  
    for(size_type i=0; i<D; i++) res[i] = 2*RG()-1;
    return res;
  }

  inline Return_t norm() {
    Return_t res;
    for(size_type i=0; i<D; i++) res[i] = RG();
    T norm = 1.0/sqrt(dot(res,res));
    return norm*res;
  }

  inline void get(Return_t& a, Return_t& b) {
    for(size_type i=0; i<D; i++) RG.bivariate(a[i], b[i]);
  }

  RNG& RG;
};


// specialized for TinyVector<T,D>
template<class RNG>
struct RandomVector<TinyVector<double,3>, RNG> {
  typedef TinyVector<double,3> Return_t;

  RandomVector(RNG& gen): RG(gen){ }

  inline Return_t operator()() {
    return Return_t(2*RG()-1, 2*RG()-1, 2*RG()-1);
  }

  inline Return_t norm() {
    double x,y,s;
    s = 2.0e0;
    while(s > 1.0e0) {
      x = 2.0e0*RG() - 1.0e0;  // [-0.5,0.5)
      y = 2.0e0*RG() - 1.0e0;  // [-0.5,0.5)
     s = x*x + y*y;
    }
    double z = 2.0e0*sqrt(1.0e0-s);
    return Return_t(z*x, z*y, 1.0e0-2.0e0*s);
  }

  inline void get(Return_t& a, Return_t& b) {
    RG.bivariate(a[0], b[0]);
    RG.bivariate(a[1], b[1]);
    RG.bivariate(a[2], b[2]);
  }
  RNG& RG;
};

//////////////////////////////////////////////////////////
// Templated function to generate random Vector<T>
//////////////////////////////////////////////////////////
// forward declaration
//template<class T, unsigned D> class CrystalLattice;
template<class T> class ParticleAttrib;

///////////////////////////////////////////////////////////
// specialization of RandomSqence with ParticleAttrib<T>
///////////////////////////////////////////////////////////
template<class T, class RNG>
struct RandomSequence<ParticleAttrib<T>, RNG> { 

  static void apply(ParticleAttrib<T>& v, RandomVector<T,RNG>& rnd) {
    for(int i=0; i<v.size(); i++) v[i] = rnd(); 
  }
};


///////////////////////////////////////////////////////////
// Functor to generate a sequence of Normalized vectors
///////////////////////////////////////////////////////////

template<class RV, class RNG>
struct RandomNormSequence { };

template<class T, class RNG>
struct RandomNormSequence<ParticleAttrib<T>, RNG> { 

  static 
  inline void apply(ParticleAttrib<T>& v, RandomVector<T,RNG>& rnd) {
    for(std::size_t i=0; i<v.size(); i++) v[i] = rnd.norm(); 
  }

  static 
  inline void apply(ParticleAttrib<T>& v, RandomVector<T,RNG>& rnd,
      std::size_t first, std::size_t last) {
    for(std::size_t i=first; i<last; i++) v[i] = rnd.norm(); 
  }
};

/////////////////////////////////////////////////////////////////
/*\fn template<class T, unsigned D>
 *    T Dot(const ParticleAttrib<TinyVector<T, D> >& pa,
 *      const ParticleAttrib<TinyVector<T, D> >& pb) 
 * \return a dot product of an array
 */
/////////////////////////////////////////////////////////////////
template<class T, std::size_t D>
T Dot(const ParticleAttrib<TinyVector<T, D> >& pa,
      const ParticleAttrib<TinyVector<T, D> >& pb) {
  T sum = 0;
  for(int i=0; i<pa.size(); i++) {
    sum += dot(pa[i],pb[i]);
  }
  return sum;
}

template<class T, std::size_t D>
T Dot(const ParticleAttrib<TinyVector<T, D> >& pa,
      const ParticleAttrib<TinyVector<T, D> >& pb, T& pmax) {
  T sum = 0;
  pmax=0.0;
  for(int i=0; i<pa.size(); i++) {
    T x = dot(pa[i],pb[i]);
    pmax = max(pmax,x);
    sum += x;
  }
  return sum;
}

template<class T, std::size_t D>
void  normalize(ParticleAttrib<TinyVector<T, D> >& pa)
{
  T factor = Dot(pa,pa);
  factor = 1.0/sqrt(factor);
  pa *= factor;
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: ParticleUtility.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
