/////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
/////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file PtclPairListBase.h
 * @brief Declaration of PairDistData<T,D> and  PtclPairListBase.
 */
#ifndef OHMMS_PTCLPAIRLISTBASE_H
#define OHMMS_PTCLPAIRLISTBASE_H

#include <vector>
#include <string>
#include <algorithm>

#ifndef OHMMS_TINYVECTOR_H
#include "OhmmsPETE/TinyVector.h"
#endif
#ifndef	OHMMS_TENSOR_H
#include "OhmmsPETE/Tensor.h"
#endif
#include "ParticleBase/ParticleAttrib.h"
 
/** data to store distance relation
 *
 * A prototype for the third template parameter of  PtclPairListBase<T,D,PDT>
 */
template<class T, std::size_t D, class IDType=int>
struct PairDistData {

  typedef IDType      index_type;
  typedef T           value_type;
  typedef std::size_t size_type;
  ///unique index of this neigbor point
  size_type Next;
  ///Location where the visiting force will go to
  size_type Where;
  ///the group id of the connected point
  index_type Group;
  ///Cartesian Distance
  T sep;
  ///Carteisan displacement
  TinyVector<T,D> displ;

  ///default construtor
  inline PairDistData(){}

  ///construtor
  inline PairDistData(size_type next,size_type loc, index_type gtype,T r, 
      const TinyVector<T,D>& dr):
    Next(next),Where(loc),Group(gtype),sep(r),displ(dr) {}

  /** update this pair list with delta
   * @param delta carteisan difference
   */
  inline void addDelta(const TinyVector<T,D>& delta) {
    displ += delta; 
    sep = sqrt(dot(displ,displ)); 
  }

  inline void 
  set (index_type next, size_type loc, index_type gtype, T r, const TinyVector<T,D>& dr) {
    Next=next, Where=loc; Group=gtype; sep = r; displ = dr; 
  }
};

/** Container for nearest-neighbor data.
 *
 *  <ul>
 *  \li M[i] = locator for the first nearest-neighbor ptcl of the i-th ptcl
 *  \li M[i+1] - M[i] = number of nn ptcls of the i-thc ptcl
 *  \li Attributes of integer, scalar, vector and tenor types for a particle
 *      pair are accessed via index \f$l\f$ where \f$ M[i]\le l < M[i+1]\f$.
 *  \li A new named attribute can be added.
 *  \li Estimating the mixmum bound for the number of nearest neighbors is important.
 *  </ul>
 *  Need to fix size_type to avoid any problem with 32/64.
 */
template<class PDT, class T=double, std::size_t D=3>
struct PtclPairListBase {

  typedef PDT                             PairData_t;
  typedef typename PairData_t::index_type index_type;
  typedef std::size_t                     size_type;
  typedef std::ptrdiff_t                  difference_type;
  typedef TinyVector<T,D>                 SingleParticlePos_t;
  typedef Tensor<T,D>                     Tensor_t;
  typedef ParticleAttrib<size_type>       IndexList_t;

//#if defined(HAVE_LIBBOOST)
//  typedef std::vector<PDT,boost::object_pool<PDT> > PairContainer_t;
//#else
//  typedef std::vector<PDT> PairContainer_t;
//#endif
  typedef std::vector<PDT> PairContainer_t;
  typedef typename PairContainer_t::iterator           iterator;
  typedef typename PairContainer_t::const_iterator     const_iterator;

  //bool MultiSet;
  size_type myID;
  size_type First;
  size_type Last;
  size_type Visitors;
  size_type LocalNum, TotalNumAdj, MaxNN;

  ///The energy of this pair-list
  T Energy;

  /// The internal stress
  Tensor_t Stress;

  /// ID
  IndexList_t  I;

//  /** Locator: M[i+1]-M[i] is the number of neighbor for the ith particle
//   *
//   *The index i [0,M.size()-1)
//   */
//  IndexList_t  M;
//
  vector<PairContainer_t*> Pairs;

  ParticleAttrib<SingleParticlePos_t> Fv;

  /// Constructor
  PtclPairListBase(): myID(0),First(0),Last(0), Visitors(0),LocalNum(0), TotalNumAdj(0), MaxNN(0)
  { }

  /// Destructor
  ~PtclPairListBase() {
    for(size_type i=0; i<Pairs.size(); i++) delete Pairs[i];
  }

  inline void set(size_type pid, size_type first, size_type last) {
    myID=pid; First=first; Last=last;
  }

  /// Number of particles
  inline size_type getLocalNum() const { return LocalNum; } 

  /// get the total Number of pairs
  inline size_type getTotNadj() const { return TotalNumAdj;}

  /// get the Maximum number of neighbors per atom for memory management
  inline size_type getMaxNadj() const { return MaxNN;}

  /** get the number of neighbors of the i-th ptcl.
   *@param i the local particle index
   *@return the number of neighbors i=[0,M.size()-1)
   */
  inline size_type nadj(size_type i) const { return Pairs[i]->size();}

  inline const PairContainer_t* getPairs(size_type i) const { return Pairs[i];}
  inline const_iterator begin(size_type i) const { return Pairs[i]->begin();}
  inline const_iterator end(size_type i) const { return Pairs[i]->end();}

  inline iterator begin(size_type i) { return Pairs[i]->begin();}
  inline iterator end(size_type i) { return Pairs[i]->end();}

  /** Sets the maximum number of neighbors
   @param nnmax the maximum number of neighbors per particle
   */
  inline void setMaxNadj(size_type maxnn) { 
    MaxNN = (maxnn>0)?maxnn:MaxNN;
  }

  /** Copy the pair list
   * @param rhs pairlist to be copied
   */
  void copy(const PtclPairListBase<PDT,T,D>& rhs) {
    myID=rhs.myID;
    First=rhs.First;
    Last=rhs.Last;
    Visitors=rhs.Visitors;
    resizePairs(rhs.LocalNum,rhs.MaxNN);
    I = rhs.I;
    for(size_type i=0; i<Pairs.size(); i++) {
      *Pairs[i]=*(rhs.Pairs[i]);
      //Pairs[i]->insert(rhs.Pairs[i]->begin(),rhs.Pairs[i]->end());
    }
    this->checkPairs(false);
  }

  /** Resizes neighbor lists. Removes everything.
   *@param m the number of particles for this list
   *@param maxnn maximum (best guess) number of neighbos per particle
   */
  inline void resizePairs(size_type m, size_type maxnn) 
  {
    I.resize(m);
    //M.resize(m+1);
    LocalNum = m;
    MaxNN = (maxnn>0)? maxnn: MaxNN;
    difference_type dm(m-Pairs.size());
    if(dm>0){//need more
      while(dm) {
        Pairs.push_back(new PairContainer_t); --dm;
      }
    } else if(dm<0) { //remove the extra
      typename vector<PairContainer_t*>::iterator pit(Pairs.begin()+m),pit_end(Pairs.end());
      while(pit != pit_end) {
        delete *pit; ++pit;
      }
      Pairs.erase(Pairs.begin()+m, Pairs.end());
    }
    for(size_type i=0; i<Pairs.size(); i++) {Pairs[i]->reserve(MaxNN);}
    reset();
  }

  /** Print info of this list */
  void print(ostream& os) const 
  {
    os << "Total Number of Pairs" << Pairs.size() << endl;
  }

  /** Reset the internal counter to accumulate nnlist
   */
  inline void reset() 
  {
    for(size_type i=0; i<Pairs.size(); i++) Pairs[i]->clear();
  }

  /** Add a neighbor data
   *@param iL local index of the paritcle
   *@param next ID of the neighbor
   *@param group particle type
   *@param where local locator of the neighbor
   *@param r distance of this pair
   *@param dr displacement vector \f$R_{next}-R_{iL}\f$
   *
   *Three vectors are used to store 
   *- J the ID of the neighbor
   *- R the distance of the pair
   *- dR the directional cosine of the pair
   */
  inline void 
    addPair(size_type iL, size_type next, size_type where, index_type gtype,
        T r, const SingleParticlePos_t& dr) 
  {
    Pairs[iL]->push_back(PairData_t(next,where,gtype,r,dr));
  }

  /** Evaluate counters
   * @param maxnn maximum number of neighbors
   * @param needfv if true, allocate Fv
   *
   * This replaces previous calls by NNEngine objects:
   * - addPair
   * - setMaxNadj
   * - Fv.resize
   */
  inline void checkPairs(bool needfv) {
    //M[0]=0;
    //int maxnn=MaxNN;
    //for(int i=0; i<LocalNum; i++) {
    //  maxnn = std::max(maxnn,Pairs[i]->size());
    //  M[i+1]=Pairs[i]->size()+M[i];
    //}
    //TotalNumAdj = M[LocalNum];
    size_type maxnn(0);
    TotalNumAdj=0;
    typename vector<PairContainer_t*>::const_iterator pit(Pairs.begin());
    typename vector<PairContainer_t*>::const_iterator pit_end(Pairs.end());
    while(pit != pit_end) {
      size_type nn((*pit)->size());
      maxnn = std::max(maxnn,nn);
      TotalNumAdj +=nn;
      ++pit;
    }
    MaxNN=maxnn;
    if(needfv) Fv.resize(TotalNumAdj);
  }
};

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: PtclPairListBase.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
