//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 David A. Richie and Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   David A. Richie
//   Department of Physics, Ohio State University
//   174 W 18th Avenue
//   Columbus, OH 43210
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <bitset>
using std::bitset;

#ifdef USE_OPENMP
#include <omp.h>
#endif
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <vector>
using namespace std;

#include "ParticleIO/ParticleInputFactory.h"
#include "ParticleIO/XYZParticleIO.h"
#include "ParticleIO/TbmdParticleIO.old.h"
#include "Utilities/OhmmsInfo.h"
#include "Utilities/SimpleParser.h"
#include "ParticleIO/PtclGeometryBase.h"
#include "ParticleBase/ParticleUtility.h"

void 
buildOMPParticle(Particle_t& p, int *uc_grid, vector<string>&);

void 
ParticleInputFactory::createParticle(Particle_t& p, istream& fpos) {

  ///////////////////////////////////////////////////////////////
  // first read lattice information
  ///////////////////////////////////////////////////////////////
  int uc_grid[OHMMS_DIM];
  for(int idim=0; idim<OHMMS_DIM; idim++) uc_grid[idim] = 1;

  //apply to be expanded to partition the grid
  //makeLattice<ParticleLayout_t>::apply(p.Lattice, uc_grid, fpos);

  //ParticleInputFactory::read_lattice(fpos, p.Lattice, uc_grid);
  vector<string> slist, vlist, shapes;

  bool tofindpos = true;
  p.Lattice.R.diagonal(1e6);
  p.Lattice.reset();

  do {
    int nw = getXwords(slist,fpos);//remove xml related stuff
    if(nw == 0) continue;
    if(slist[0] == "unitcell") {
      ParticleInputFactory::read_lattice(fpos, p.Lattice, uc_grid);
    } else if(slist[0] == "ptclset"){
      tofindpos = false;
      int isub=1;
      while(isub<slist.size()) {
	if(slist[isub] == "oname") {
	  p.setName(slist[++isub].c_str());
	}
	isub++;
      }
    }
  } while(tofindpos);

  string fileformat, filename;
  getwords(slist,fpos);  fileformat = slist[1]; // expect format
  getwords(slist,fpos);  filename   = slist[1]; // expect filename


  //////////////////////////////////////////////////
  if(fileformat == "tbmd") {
    LOGMSG("\t reading tbmd file from " << filename) 
    TbmdParticleIO xyzin(p, 0, filename.c_str());
    if(filename == "none") {
      xyzin.read(fpos);
    } else {
      ifstream fxmol(filename.c_str());
      xyzin.read(fxmol);
    }
  }
  else if(fileformat == "xyz") {
    LOGMSG("\t reading xyz file from " << filename) 
    XYZParticleIO xyzin(p, 0, filename.c_str());
    if(filename == "none") {
      xyzin.read(fpos);
    } else {
      ifstream fxmol(filename.c_str());
      xyzin.read(fxmol);
    }
  }

  //////////////////////////////////////
  // check if shape is given (need to generalize)
  //////////////////////////////////////
  //fpos.rewind();
  fpos.seekg(ios::beg);
  bool tofindshape = true;
  do {
    getwords(slist,fpos);
    if(fpos.eof()) tofindshape = false;
    else if(slist[0] == "<shape>") {
      LOGMSG("\t Shaping of the system is required")
	shapes.push_back("<shape>");
      getwords(shapes, fpos, "</shape>");
      tofindshape = false;
    }
  } while(tofindshape);

  int ngtot = uc_grid[0]*uc_grid[1]*uc_grid[2];

  if(ngtot > 1 || !shapes.empty()) {
   
    LOGMSG("\t using buildOMPParticle to expand or shape a supercell.")
    LOGMSG("\t the ordering of particles is based on grid information")
    buildOMPParticle(p, uc_grid, shapes);
    p.NeedPartition = false; // input configuration has been already partitioned
  }//
}

void 
ParticleInputFactory::createParticle(Particle_t& p, int argc, char **argv) {
  vector<string> arglist;
  for(int i=0;i<argc; i++) arglist.push_back(argv[i]);
  ParticleInputFactory::createParticle(p,arglist);
}


void 
ParticleInputFactory::createParticle(Particle_t& p, vector<string>& arglist) {

  //assign a default file-root based on MDParameter members
  //string sysname, latfile;
  char sysname[256];
  bool read_old = true;
  int it = 0;

  while(it < arglist.size()) {
    if(arglist[it] == "ptcl_in") {
      //sysname = arglist[++it];
      sprintf(sysname,"%s", arglist[++it].c_str());
      read_old = false;
    }
    it++;
  }

  ifstream fpos(sysname);
  if(!fpos) {
    ERRORMSG("Missing " << sysname << " file for the particle information ")
    return;
  }
  ParticleInputFactory::createParticle(p,fpos);
}

/*\fn void  read_lattice(istream& fpos, ParticleLayout_t& lat, int *ng)
 *\param fpos input stream 
 *\param lat  lattice object to initialize
 *\param ng   3D integers to expand the input system
 */
void 
ParticleInputFactory::read_lattice(istream& fpos, ParticleLayout_t& lat, int *ng) {

  enum {mppgrid=ParticleLayout_t::MPI_GRID, 
	ompgrid, 
	finegrid};

  typedef ParticleLayout_t::SingleParticleIndex_t SingleParticleIndex_t;
  vector<SingleParticleIndex_t> mgrid(3,SingleParticleIndex_t(1));

  ng[0] = 1;  ng[1] = 1;  ng[2] = 1;
  vector<string> vlist,slist;
  int ic=0;

  Scalar_t a0tmp=1.0;
  Scalar_t vtmp[] = {1.0e0, 0.0e0, 0.0e0, 
		     0.0e0, 1.0e0, 0.0e0, 
		     0.0e0, 0.0e0, 1.0e0};

  bool tofindlattice = true;
  do {
    int argc = getXwords(slist,fpos);
    if(argc == 0) continue;
    vlist = slist;
    if(slist[0] == "a0") {
      getXwords(vlist, fpos, "/a0");
      a0tmp = atof(vlist[1].c_str());
    } else if(slist[0] == "lattice") {
      getXwords(vlist, fpos, "/lattice");
      for(int ij=0; ij<9; ij++) vtmp[ij] = atof(vlist[ij+1].c_str());
    }else if(slist[0] == "mppgrid") {
      getXwords(vlist, fpos, "/mppgrid");
      mgrid[mppgrid][0] = atoi(vlist[1].c_str());
      mgrid[mppgrid][1] = atoi(vlist[2].c_str());
      mgrid[mppgrid][2] = atoi(vlist[3].c_str());
    } else if(slist[0] == "ompgrid") {
      getXwords(vlist, fpos, "/ompgrid");
      mgrid[ompgrid][0] = atoi(vlist[1].c_str());
      mgrid[ompgrid][1] = atoi(vlist[2].c_str());
      mgrid[ompgrid][2] = atoi(vlist[3].c_str());
    } else if(slist[0] == "grid") {
      getXwords(vlist, fpos, "/grid");
      mgrid[finegrid][0] = atoi(vlist[1].c_str());
      mgrid[finegrid][1] = atoi(vlist[2].c_str());
      mgrid[finegrid][2] = atoi(vlist[3].c_str());
    } else if(slist[0] == "unit_cell_grid") {
      getXwords(vlist, fpos, "/unit_cell_grid");
      ng[0] = atoi(vlist[1].c_str());
      ng[1] = atoi(vlist[2].c_str());
      ng[2] = atoi(vlist[3].c_str());
    } else if(slist[0] == "bconds") { // boundary conditions
      getXwords(vlist, fpos, "/bconds");
        if(vlist[1][0] == 'n' || vlist[1][0] == 'N') {
          lat.BoxBConds[0] = false;
        }
        if(vlist[2][0] == 'n' || vlist[2][0] == 'N') {
          lat.BoxBConds[1] = false;
        }
        if(vlist[3][0] == 'n' || vlist[3][0] == 'N') {
          lat.BoxBConds[2] = false;
        }
    } else if(slist[0] == "/unitcell") {
      tofindlattice = false;
    }
  } while(tofindlattice);

#ifdef USE_OPENMP
    int  NumThreads = omp_get_max_threads();
    if(NumThreads != ompgrid[0]*ompgrid[1]*ompgrid[2]) {
      ompgrid[0] = 1; ompgrid[1] = 1; ompgrid[2] = NumThreads;
    }
#else
    int  NumThreads = 1;
#endif
    lat.makeGrid(mgrid);
//     ////////////////////////////////////////////////////////////////
//     // grid partitions are created
//     ////////////////////////////////////////////////////////////////
//     int mppgrid_id = lat.makeGrid(mppgrid,-1);
//     int ompgrid_id = lat.makeGrid(ompgrid,mppgrid_id);
//     for(int idim=0; idim<OHMMS_DIM; idim++) {
//       if(ompgrid[idim] > finegrid[idim]) ompgrid[idim] = finegrid[idim];
//       finegrid[idim] /= ompgrid[idim];
//     }
//     lat.makeGrid(finegrid,ompgrid_id);
//     lat.set(a0tmp, vtmp);
//     for(int idim=0; idim<OHMMS_DIM; idim++) {
//       lat.Grid[idim] = finegrid[idim]*ompgrid[idim];
//     }
}


/* \fn void buildOMPParticle(Particle_t& p, int *uc_grid, vector<string>& options
 * \param p an input configuration to be expanded
 * \param uc_grid expansion dimension
 * \param extra options to shape a system
 */
void
buildOMPParticle(Particle_t& p, int *uc_grid, vector<string>& options) {

  int natom = p.getTotalNum();

  // temporary positions and ionid
  Particle_t::ParticlePos_t   pin(natom);
  Particle_t::ParticleIndex_t ionid(natom);
  pin.InUnit = true;
  convert(p.Lattice,p.R,pin);

  ionid = p.IonID;
  int ntot = uc_grid[0]*uc_grid[1]*uc_grid[2]*natom;

  p.R.InUnit = false;
  convert(p.Lattice,pin,p.R);

  ParticleLayout_t lat = p.Lattice;
  p.Lattice.set(lat,uc_grid); // expand the lattice

  p.create(ntot-natom);
  int nacc = natom;
  Particle_t::SingleParticlePos_t del;
  for(int ic=0; ic<uc_grid[0]; ic++) {
    del[0] = static_cast<Scalar_t>(ic);
    for(int jc=0; jc<uc_grid[1]; jc++) {
      del[1] = static_cast<Scalar_t>(jc);
      for(int kc=0; kc<uc_grid[2]; kc++) {
	del[2] = static_cast<Scalar_t>(kc);
	if(ic == 0 && jc == 0 && kc == 0) continue;
	for(int nb=0; nb<natom; nb++) {
	  p.R[nacc] = lat.toCart(pin[nb]+del);
	  p.ID[nacc] = nacc;
	  p.IonID[nacc] = ionid[nb];
	  nacc++;
	}
      }
    }
  }
  p.NeedPartition = true;
  p.partition(false);

//    int NumThreads = p.Lattice.dGrid[1]->getTotalNum();

//    //number of particles per omp node
//    vector<int> nat(NumThreads,0);

//    ParticleLayout_t lat = p.Lattice;
//    p.Lattice.set(lat,uc_grid); // expand the lattice
  
//    bool posInUnit = p.R.InUnit;
//    // initial distribution
//    p.Lattice.dGrid[0]->distribute(natom);
//    p.Lattice.dGrid[1]->distribute(natom);
//    p.Lattice.dGrid[2]->PtclDist.resize(p.Lattice.dGrid[2]->getTotalNum()+1);

//    ///////////////////////////////////////////////////////////
//    // pre-evaluate the number of particles per omp node
//    // omp parallelization of this routine is not so critical
//    // Do not create any big temporary object
//    ///////////////////////////////////////////////////////////
//    if(NumThreads == 1 && options.empty()) {
//      // on a single processor without shaping, 
//      //the problem size can be calculated
//      p.create(natom*(uc_grid[0]*uc_grid[1]*uc_grid[2]-1));
//      p.Lattice.dGrid[0]->PtclDist[1] = p.getLocalNum();
//      p.Lattice.dGrid[1]->PtclDist[1] = p.getLocalNum();
//      nat[0] = p.getLocalNum();

//    } else {

//      // first step to estimate the number of particles per OMP grid
//      // nat is evaluated
//      typedef ParticleLayout_t::PtclGrid_t PtclGrid_t;
//      typedef Particle_t::SingleParticlePos_t SingleParticlePos_t;

//      PtclGrid_t& ompgrid = *p.Lattice.dGrid[1];//omp level

//      PtclGeometryBase* sysgeom = createPtclGeometry(options,lat);

//      SingleParticlePos_t dx(1.0/static_cast<Scalar_t>(uc_grid[0]),
//  			   1.0/static_cast<Scalar_t>(uc_grid[1]),
//  			   1.0/static_cast<Scalar_t>(uc_grid[2]));

//      SingleParticlePos_t del,pos,upos;
//      bool simplecount = (NumThreads == 1);

//      int newid = 0; // not used
//      for(int ic=0; ic<uc_grid[0]; ic++) {
//        del[0] = static_cast<Scalar_t>(ic);
//        for(int jc=0; jc<uc_grid[1]; jc++) {
//  	del[1] = static_cast<Scalar_t>(jc);
//  	for(int kc=0; kc<uc_grid[2]; kc++) {
//  	  del[2] = static_cast<Scalar_t>(kc);

//  	  for(int nb=0; nb<pin.size(); nb++) {
//  	    pos = lat.toCart(pin[nb]+del); 
//  	    if(sysgeom->keepit(pos,newid)) {
//  	      if(simplecount) 
//  		nat[0]++;
//  	      else
//  		nat[ompgrid.loc(p.Lattice.toUnit(pos))]++;
//  	    }
//  	  }
//  	}
//        }
//      }

//      if(sysgeom) delete sysgeom;// clean up

//      int nattot = 0;
//      ompgrid.PtclDist[0] = 0;
//      for(int inode=0; inode<NumThreads; inode++) {
//        nattot += nat[inode];
//        ompgrid.PtclDist[inode+1] = nattot;
//      }

//      ///////////////////////////////////
//      // create particle sets
//      ///////////////////////////////////
//      p.create(nattot-pin.size());
//    }

//    //SGI OPENMP compiler bug: parallel for is not working with C++
//    //  int myID = 0;
//    //#pragma omp parallel for 
//    for(int myID =0; myID < NumThreads; myID++) {
//  //  #pragma omp parallel
//  //    {
//  //  #ifdef USE_OPENMP
//  //      int myID = omp_get_thread_num();
//  //  #else
//  //      int myID = 0;
//  //  #endif
//      typedef ParticleLayout_t::PtclGrid_t PtclGrid_t;
//      typedef Particle_t::SingleParticlePos_t SingleParticlePos_t;

//      const PtclGrid_t& ompgrid = *p.Lattice.dGrid[1];//omp level
//      PtclGrid_t& bgrid = *p.Lattice.dGrid[2];//base level

//      // starting particle index on this OMP node
//      int iG = ompgrid.PtclDist[myID]; 
//      int inode = bgrid.NodeDist[myID];
//      int fnode = bgrid.NodeDist[myID+1];
//      vector<int> nat_loc(fnode-inode, 0);
//      int ntot_loc0 = ompgrid.PtclDist[myID+1]-iG;
//      // temporary particles before ordering according to the finegrid 
//      //(lowest level)
//      vector<int> iloc(ntot_loc0);
//      vector<int> cellid(ntot_loc0);
//      vector<SingleParticlePos_t> ploc(ntot_loc0);

//      PtclGeometryBase* sysgeom = createPtclGeometry(options,lat);
//      SingleParticlePos_t dx(1.0/static_cast<Scalar_t>(uc_grid[0]),
//  			   1.0/static_cast<Scalar_t>(uc_grid[1]),
//  			   1.0/static_cast<Scalar_t>(uc_grid[2]));
//      SingleParticlePos_t del,pos,upos;

//      int newid; // not used
//      int ntot_loc = 0; // local number of particles

//      // it is possible that uc_grid is not commensurate with ompgrid
//      for(int ic=0; ic<uc_grid[0]; ic++) {
//        del[0] = static_cast<Scalar_t>(ic);
//        for(int jc=0; jc<uc_grid[1]; jc++) {
//  	del[1] = static_cast<Scalar_t>(jc);
//  	for(int kc=0; kc<uc_grid[2]; kc++) {
//  	  del[2] = static_cast<Scalar_t>(kc);

//  	  for(int nb=0; nb<pin.size(); nb++) {

//  	    pos = lat.toCart(pin[nb]+del); 
//              upos = p.Lattice.toUnit(pos); 
//  	    // does not belong to this node
//  	    if(ompgrid.loc(upos) != myID) continue;
//              newid = ionid[nb];	
//              if(sysgeom->keepit(pos,newid)) {
//                int thisnode =  bgrid.loc(upos);
//                iloc[ntot_loc] = newid;
//                ploc[ntot_loc] = upos;
//                cellid[ntot_loc] = thisnode;
//                nat_loc[thisnode-inode] ++;
//  	      ntot_loc++;
//  	    }
//  	  }
//  	}
//        }
//      }

//      bgrid.PtclDist[inode] = iG;
//      for(int thisnode=inode; thisnode<fnode; thisnode++) {
//        bgrid.PtclDist[thisnode+1] 
//  	= bgrid.PtclDist[thisnode] + nat_loc[thisnode-inode];
//      }

//      for(int thisnode=inode; thisnode<fnode; thisnode++) {
//        int iG2 = bgrid.PtclDist[thisnode];
//        for(int iL = 0; iL<ntot_loc; iL++) {
//          if(cellid[iL] != thisnode) continue;
//          p.R[iG2] = ploc[iL];
//          p.IonID[iG2] = iloc[iL];
//          p.ID[iG2] = iG2;
//  	iG2++;
//        }
//      }
//      if(sysgeom) delete sysgeom;// clean up

//    } // run over the ompgrid


}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ParticleInputFactory.cpp 151 2004-11-02 23:51:14Z jnkim $
 ***************************************************************************/
