//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_PYRAMID_GEOMETRY_H
#define OHMMS_PYRAMID_GEOMETRY_H

/*! \class PyramidalQD a inherited class from PtclGeometryBase
 *  \author Jeongnim Kim
 *  \brief 
 <shape>
 pyramid
 a0    value               
 center value value value    # 3D coordinate of the base center
 base   value value theta         # 2D base
 height  value               # distance from the base to the tip
 wetting value               # thickness
 exchange A B
 exchange C C
 </shape>

 \note All the valuables are in the unit of a0.
 
 exchange is used to specify how to swap species in the BARRIER.
 IMPORTANT: GaAs/InAs where As in the inner and outer layers are identical,
 add "exchange As As". I decided to force this to reduce additional
 boolean evalulation.

 <shape>
 pyramid
 a0 5.6
 center 2 2 2
 base   2 2
 height 1
 exchange In Ga
 exchange As As
 </shape>
 */

struct PyramidalQD:public PtclGeometryBase {

  PyramidalQD():InUnit(true){}
  ~PyramidalQD() { }

  SingleParticlePos_t Center;
  double Base[2];
  double Height, taninv;
  double WL[3];
  bool InUnit;
  map<int,int> exchange;

  void set(const vector<string>& argv) {

    SpeciesBase* species = SpeciesCollection::getSpecies();
    vector<string>::const_iterator it = argv.begin();
    double a0 = 1;
    while(it != argv.end()) {
      if((*it) == "a0") {
	a0 = atof((*(++it)).c_str());
      } else if((*it) == "height") {
	Height = atof((*(++it)).c_str());
      } else if((*it) == "base") {
	Base[0] = atof((*(++it)).c_str());
	Base[1] = atof((*(++it)).c_str());
      } else if((*it) == "wetting") {
	WL[0] = atof((*(++it)).c_str());
      } else if((*it) == "center") {
	Center[0] = atof((*(++it)).c_str());
	Center[1] = atof((*(++it)).c_str());
	Center[2] = atof((*(++it)).c_str());
      } else if((*it) == "exchange") {
        int a = species->getSpeciesID(*(++it));
        int b = species->getSpeciesID(*(++it));
        exchange[a] = b;
      }
      it++;
    }

    taninv = Base[0]/Height*0.5;
    Height *= a0;
    Base[0] *= a0; Base[1] *= a0;
    WL[0] *= a0; WL[1] = WL[0];
  }

  void set(const ParticleLayout_t& lattice) { 
    if(InUnit) {
      Center = lattice.toCart(Center);
    }
    WL[0] = Center[2]-WL[1] ;// bottom of the wetting layer
    WL[1] = Center[2];       // top of the wetting layer == base z
    WL[2] = Center[2]+Height;// tip of the pyramid
  }

  bool keepit(SingleParticlePos_t& p, int& id) {

    if(p[2]>= WL[0] && p[2] < WL[1]) // in the wetting layer, do nothing
      return true;

    if(p[2]<WL[0] || p[2] > WL[2]) { // buffer layer -> exchange species
      id = exchange[id];
    } else {
      double bmax = taninv*(WL[2]-p[2]);//for a given z, find pyramid
      if((fabs(p[0]-Center[0]) > bmax)|| (fabs(p[1]-Center[1]) > bmax)) {
	id = exchange[id];
      }
    }
    return true;

  }

};

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: PyramidalQD.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
