//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <bitset>
using std::bitset;
#include <stdio.h>
#include "ParticleIO/XYZParticleIO.h"
#include "ParticleIO/ParticleInputTemplate.h"
#include "ParticleBase/ParticleUtility.h"

void 
XYZParticleIO::read(istream& fxmol) {

  if(ioMode == INPUT) {

    DEBUGMSG("Performing XYZParticleIO::read()")

    SpeciesBase* species = SpeciesCollection::getSpecies();

    // will be converted into unit-cell uni vectors
    MyPtcl.R.InUnit = true; 

    vector<string> slist;
    int argc = getwords(slist,fxmol);
    if(!isdigit(slist[0][0])) {
      ERRORMSG("Input file does not confirm xmol format. No Action is taken.")
      return;
    }

    unsigned natom = atoi(slist[0].c_str());
    getwords(slist,fxmol);//line-skip
    
    if(natom) {
      int loc = MyPtcl.getLocalNum();
      MyPtcl.create(natom);
      int i=0;
      SingleParticlePos_t pos;
      while(i<natom) {
        argc =  getwords(slist, fxmol);
        MyPtcl.ID[loc] = i;
        MyPtcl.IonID[loc] = species->getSpeciesID(slist[0]);
        getValue<SingleParticlePos_t>::get(pos,slist.begin()+1);
	MyPtcl.R[loc] = 
	  MyPtcl.Lattice.BConds.wrap(MyPtcl.Lattice.toUnit(pos));
        loc++; i++;
      }

      //expandLattice<ThisParticleLayout_t>::apply(MyPtcl,uc_grid);
      //MyPtcl.update(Particle_t::LOCupdate);//local grid update
      //MyPtcl.update(Particle_t::IDupdate); //ID is reassigned
    }
  } else {
    ERRORMSG("read for OUPUT/PROGRESS mode is not allowed.\n")
  }
}

void XYZParticleIO::record(int iter) {
  if(ioMode == PROGRESS) {
    DEBUGMSG("Performing XYZParticleIO::record()")
    if(!initialized) { setup(); initialized = true;}

    Rcart.InUnit = false;
    Rcart.resize(MyPtcl.getLocalNum());

    recordXYZ(iter); 
  } else {
    ERRORMSG("record for INPUT/OUPUT mode is not allowed.\n")
  }
}

void XYZParticleIO::write() {

  if(ioMode == OUTPUT) {

    WARNMSG("Performing XYZParticleIO::write(), not suitable for restart")
    WARNMSG("USE hdf5 or tbmd to dump particle positions")
    if(!initialized) {
      if(Species.empty()) {
	SpeciesBase* species = SpeciesCollection::getSpecies();
	int nspecies = species->getTotalNum();
	Species.reserve(nspecies);
	for(int i=0; i<nspecies; i++) {
	  Species.push_back(species->Name[i]);
	}
      }
      sprintf(xmolfile,"%s.data",FileRoot.c_str());
      sprintf(metafile,"%s.meta",FileRoot.c_str());
      initialized = true;
    }


    ofstream fxmol(xmolfile); // always overwrite
    fxmol << MyPtcl.getLocalNum()<< endl << endl;
    if(MyPtcl.R.InUnit) {
      for(int i=0; i<MyPtcl.getLocalNum(); i++) {
	fxmol << Species[MyPtcl.IonID[i]] 
	      << MyPtcl.Lattice.toCart(MyPtcl.R[i]) 
	      << endl;
      }
    } else {
      for(int i=0; i<MyPtcl.getLocalNum(); i++) {
	fxmol << Species[MyPtcl.IonID[i]] <<  MyPtcl.R[i]  << endl;
      }
    }
    fxmol.close();
    fxmol.open(metafile);
    fxmol << "<unitcell>" << endl;
    fxmol << MyPtcl.Lattice.R;
    //MyPtcl.Lattice.print(fxmol,false);
    fxmol << "</unitcell>" << endl;
    fxmol << "<ptclset oname=\"" << MyPtcl.getName() << "\">" << endl;
    fxmol << "\t<format> xyz </format> " << endl;
    fxmol << "\t<name> " << xmolfile << " </name>" <<  endl;
    for(int i=3; i<MyPtcl.getNumAttrib(); i++) {
      OhmmsObject& ooref = MyPtcl.getAttrib(i);  
      fxmol << ooref.typeName() << " " << ooref.objName() << endl;
      //ooref.write_data(fxmol);
      ooref.get(fxmol);
    }
  } else {
    ERRORMSG("write for INPUT/PROGRESS  mode is not allowed.\n")
  }
}

void XYZParticleIO::setup() {

  enum XMOLATTRIB {E, Q, F, V};

  bitset<4> attribMode;
  attribMode.reset();

  int i=0; 
  while(i<options.size()) {
    if(options[i] == "--E") attribMode.set(E);
    else if(options[i] == "--Q") attribMode.set(Q);
    else if(options[i] == "--F") attribMode.set(F);
    else if(options[i] == "--V") attribMode.set(V);
    i++;
  }

  sprintf(xmolfile,"%s.xyz",FileRoot.c_str());
  ofstream fxmol(xmolfile);
  fxmol.close();

  if(attribMode.any()) {
    if(attribMode[E]) {
      if(MyPtcl.hasAttrib(ENERGY_TAG)) 
	e_ref = MyPtcl.getScalarAttrib(ENERGY_TAG);
      else {
	ERRORMSG("Energy is not found among the Particle Attributes.\n");
	attribMode.flip(E);
      }
    }

    if(attribMode[Q]) {
      if(MyPtcl.hasAttrib(VAL_CHARGE_TAG)) 
	e_ref = MyPtcl.getScalarAttrib(VAL_CHARGE_TAG);
      else 
      {
        ERRORMSG("Charge is not found among the Particle Attributes.\n");
        attribMode.flip(Q);
      }
    }

    if(attribMode[F]) {
      if(MyPtcl.hasAttrib(FORCE_TAG)) 
	f_ref = MyPtcl.getVectorAttrib(FORCE_TAG);
      else 
	{
	  ERRORMSG("Force is not found among the Particle Attributes.\n");
	  attribMode.flip(F);// cannot perform
	}
    }

    if(attribMode[V]) {
      if(MyPtcl.hasAttrib(VELOCITY_TAG)) 
	f_ref = MyPtcl.getVectorAttrib(VELOCITY_TAG);
      else
	{
	  ERRORMSG("Velocity is not found among the Particle Attributes.\n");
	  attribMode.flip(V);// cannot perform
	}
    }
  }

  //calculate the integer value
  xmolMode = (attribMode[E] || attribMode[Q])
    + (attribMode[F] || attribMode[V])*2;

  if(Species.empty()) {
    SpeciesBase* species = SpeciesCollection::getSpecies();
    int nspecies = species->getTotalNum();
    Species.reserve(nspecies);
    for(int i=0; i<nspecies; i++) {
      Species.push_back(species->Name[i]);
    }
  }
}

void XYZParticleIO::recordXYZ(int iter) {

  DEBUGMSG("XYZParticleOut::init() is completed.\n");
  ofstream fxmol(xmolfile, ios::app);
  fxmol.setf(ios::fixed, ios::floatfield);  
  int nat = MyPtcl.getLocalNum();
  bool toCart = MyPtcl.R.InUnit;

  if(MyPtcl.hasAttrib(TRAJECTORY_TAG)) {
    convert(MyPtcl.Lattice, *(MyPtcl.getVectorAttrib(TRAJECTORY_TAG)), Rcart);
  } else {
    convert(MyPtcl.Lattice, MyPtcl.R, Rcart);
  }

  if(MyPtcl.hasAttrib("sumbc")) {
    Rcart +=  *(MyPtcl.getVectorAttrib("sumbc"));
  }

  fxmol << nat << endl <<"#step = " << iter <<  endl;

  switch(xmolMode) {
  case(1):// position + scalar
    for(int i=0; i<nat; i++) {
      fxmol << Species[MyPtcl.IonID[i]]    << Rcart[i]
	    << setw(15) << setprecision(5)  << (*e_ref)[i] << endl;
    }
    break;

  case(2):// position + vector

    for(int i=0; i<nat; i++) 
      fxmol << Species[MyPtcl.IonID[i]]  << Rcart[i] << (*f_ref)[i] << endl;
    break;

  case(3):// position + scalar + vector

    for(int i=0; i<nat; i++) 
      fxmol << Species[MyPtcl.IonID[i]]  << Rcart[i]
	    << setw(15) << setprecision(5)  << (*e_ref)[i] 
	    << (*f_ref)[i] << endl;
    break;

  default:
    for(int i=0; i<nat; i++) 
      fxmol << Species[MyPtcl.IonID[i]]  << Rcart[i] << endl;
    break;
  }

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: XYZParticleIO.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

