// -*- C++ -*-
/***************************************************************************
 *
 * The POOMA Framework
 * 
 * This program was prepared by the Regents of the University of
 * California at Los Alamos National Laboratory (the University) under
 * Contract No.  W-7405-ENG-36 with the U.S. Department of Energy (DOE).
 * The University has certain rights in the program pursuant to the
 * contract and the program should not be copied or distributed outside
 * your organization.  All rights in the program are reserved by the DOE
 * and the University.  Neither the U.S.  Government nor the University
 * makes any warranty, express or implied, or assumes any liability or
 * responsibility for the use of this software
 *
 * Visit http://www.acl.lanl.gov/POOMA for more details
 *
 ***************************************************************************/

#include "Pooma/CommCreator.h"
#include "Pooma/Communicate.h"
#include "Pooma/Profiler.h"
#ifdef USE_MPI
#include "Pooma/CommMPI.h"
#endif

#include <string.h>

list<Communicate*> CommCreator::commlist;

// static data for this file
static char *CommLibraryNames[CommCreator::COMMLIBRARIES] =
  { "aclmpl", "pm", "mpi", "shmempi", "pvm", "serial" };
static char *CommLibraryList = "aclmpl, pm, mpi, shmempi, pvm, or serial";


/////////////////////////////////////////////////////////////////////////
// return the name of the Nth library
const char *CommCreator::getLibraryName(int n) {
  TAU_PROFILE("CommCreator::getLibraryName()", "char * (int n)", TAU_MESSAGE);
  if (n >= 0 && n < COMMLIBRARIES)
    return CommLibraryNames[n];
  else
    return 0;
}


/////////////////////////////////////////////////////////////////////////
// return a list of all the libraries, as a single string
const char *CommCreator::getAllLibraryNames() {
  TAU_PROFILE("CommCreator::getAllLibraryNames()", "char *()", TAU_MESSAGE);
  return CommLibraryList;
}


/////////////////////////////////////////////////////////////////////////
bool CommCreator::supported(int cm) {
  TAU_PROFILE("CommCreator::supported()", "bool (int)", TAU_MESSAGE);
  if(cm == MPI) 
    return true;
  else
    return false;
}


/////////////////////////////////////////////////////////////////////////
// return the index of the given named library, or (-1) if not found
int CommCreator::libindex(const char *nm) {
  TAU_PROFILE("CommCreator::libindex()", "int (char *)", TAU_MESSAGE);
  for (int i=0; i < COMMLIBRARIES; ++i) {
    if (strcasecmp(nm, getLibraryName(i)) == 0)
      return i;
  }

  // if here, it was not found
  return (-1);
}


/////////////////////////////////////////////////////////////////////////
// create a new Communicate object.  Arguments = type, argc, argv, num nodes,
// whether to do initialization or not (ignored by some libs).
// If the type code is illegal, or the comm library is not supported,
// return 0.
Communicate *CommCreator::create(int cm, int& argc, char**& argv, int nodes,
                                 bool doinit) {
  TAU_PROFILE("CommCreator::create()",
	      "Communicate * (int, int, char **, int)", TAU_MESSAGE);

  if(commlist.empty()) {
    Communicate *comm = 0;

    ///when the list is empty, needs to initialize the global engine
    doinit = true; //if (!commlist.empty()) doinit = false;

    if (cm == MPI) {
#ifdef USE_MPI
      comm = new CommMPI(argc, argv, nodes, doinit);
#endif
    } else {
      comm = new Communicate(argc, argv, nodes);
    }
    
    if(comm) commlist.push_back(comm);
    return comm;
  } else {
    return commlist.front();
  }
}

/*!\fn Communicate* CommCreator::get(int i)
 *\param i id (reserved)
 *\return Communicate to handle communication
 */
Communicate* CommCreator::get(int i) {
  int argc = 0;
  char **argv = 0;
#if defined(USE_MPI)
  int cm = MPI;
#else
  int cm = SERIAL;
#endif
  return create(cm,argc,argv);
//    Communicate* comm = 0;
//    if(commlist.empty()) {
//      int argc = 0;
//      char **argv = 0;
//  #if defined(USE_MPI)
//      int cm = MPI;
//  #else
//      int cm = SERIAL;
//  #endif
//      comm = create(cm,argc,argv);
//    } 
//    return commlist.front();
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * POOMA_VERSION_ID: $Id: CommCreator.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
