// -*- C++ -*-
/***************************************************************************
 *
 * The POOMA Framework
 * 
 * This program was prepared by the Regents of the University of
 * California at Los Alamos National Laboratory (the University) under
 * Contract No.  W-7405-ENG-36 with the U.S. Department of Energy (DOE).
 * The University has certain rights in the program pursuant to the
 * contract and the program should not be copied or distributed outside
 * your organization.  All rights in the program are reserved by the DOE
 * and the University.  Neither the U.S.  Government nor the University
 * makes any warranty, express or implied, or assumes any liability or
 * responsibility for the use of this software
 *
 * Visit http://www.acl.lanl.gov/POOMA for more details
 *
 ***************************************************************************/

#ifndef COMM_CREATOR_H
#define COMM_CREATOR_H

#include <list>
using std::list;

/***********************************************************************
 * 
 * CommCreator is a factory class which is used to create specific
 * subclasses of the base class Communicate.  To create a new library-
 * specific Communicate, the user can either instantiate such a class
 * directly, or they can call the static method
 *   CommCreator::create(CommCreator::CreateMethod, int argc, char *argv[],
 *                       int nodes)
 *
 ***********************************************************************/


// forward declarations
class Communicate;


// class definition
class CommCreator {

public:
  // enumeration of communication libraries
  enum { ACLMPL, PM, MPI, SHMEMPI, PVM, SERIAL, COMMLIBRARIES };

public:
  // constructor and destructor
  CommCreator() { } 
  ~CommCreator() { }

  //
  // informative methods
  //

  // return the number of libraries available
  static int getNumLibraries() { return COMMLIBRARIES; }

  // return the name of the Nth library
  static const char *getLibraryName(int);

  // return a list of all the libraries, as a single string
  static const char *getAllLibraryNames();

  // is the given comm library available here?
  static bool supported(int);
  static bool supported(const char *nm) { return supported(libindex(nm)); }

  // is the given comm library name one we recognize?
  static bool known(const char *nm) { return (libindex(nm) >= 0); }

  //
  // Communicate create methods
  //

  // create a new Communicate object.  Arguments = type, argc, argv, num nodes,
  // do init
  static Communicate *create(int, int&, char** &, int = (-1), bool = true);

  // same as above, but specifying the type as a string instead of an int
  static Communicate *create(const char *nm, int& argc, char **& argv,
			     int nodes= (-1), bool doinit = true) {
    return create(libindex(nm), argc, argv, nodes, doinit);
  }

  static Communicate* get(int i = 0);

private:
  // return the index of the given named library, or (-1) if not found
  static int libindex(const char *);
  static list<Communicate*> commlist;
};

#endif // COMM_CREATOR_H

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * POOMA_VERSION_ID: $Id: CommCreator.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
