//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Richard Hennig and Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/*!\class AlloyEAMPotential
 * \author Richard Hennig
 * \author Jeongnim Kim
 * \brief Derived class from PotentialBase.  Implements AlloyEAM potential for single
 *  species.
 * \note  
 * Implemented by Richard Hennig 6/2001
 *
 * Based on MAlloyEAMPotential originally written by Thomas Lenosky, LLNL, 12/8/99
 *
 * Modified to confirm new interfaces of PotentialBase Jeongnim Kim 9/27/2001
 ************************************************************************/
#ifndef OHMMS_EAMPOTENTIAL_PARAMETER_H
#define OHMMS_EAMPOTENTIAL_PARAMETER_H

#include "Utilities/OhmmsInfo.h"
#include "Utilities/SimpleParser.h"
#include "Potentials/CubicSplineType.h"

/** EAM parameter set
 *
 * Contain spline parameters for EAM potentials and various splines.
 * Using SplineType defined in CubicSplineType.h
 */
template<class T>
struct EAMParam {
  //!<Range of pair potential
  T clsphircut; 
  //!<Range for rho(r)
  T clsrhorcut;   
  //!<Range for f(rho)
  T clsfrhocut;   
  //Number of species
  int      nspecies;  
  // SplineTypeEAM phi, f, rho;
  vector<SplineType> phi, f, rho; 
  
  EAMParam() { }
  
  EAMParam(const char *fname, int num);

  void check(int num);

  void readEAMParam(const char *fname, int num);

  T rcutmax() { 
    return (clsphircut > clsrhorcut)? clsphircut : clsrhorcut;
  }
};

template<class T>
EAMParam<T>::EAMParam(const char *fname, int num) {
  readEAMParam(fname, num);
  check(num);
}

template<class T>
void EAMParam<T>::check(int num) {
  nspecies = num;
  
  //symmeterize the pair interactions
  for(int i=0; i<num; i++) {
    for(int j=i+1; j<num; j++) {
      phi[j*num+i] = phi[i*num+j];
    }
  }
  
  clsphircut = phi[0].X[phi[0].Npt];
  clsrhorcut = rho[0].X[rho[0].Npt];
  clsfrhocut = f[0].X[1];
  for (int i=1; i<num; i++) {
    clsphircut = max(clsphircut, phi[i].X[phi[i].Npt]);
    clsrhorcut = max(clsrhorcut, rho[i].X[rho[i].Npt]);
    clsfrhocut = min(clsfrhocut, f[1].X[1]);
  }
}

template<class T>
void 
EAMParam<T>::readEAMParam(const char *fname, int nspec) {
  T econv = 1.0;
  nspecies = nspec;
  
  std::ifstream fin(fname);
  if(!fin) { 
    ERRORMSG(fname << " is not found in this directory")
    ERRORMSG("Exiting very ungracefully: EAMPotential::initialize()")
    exit(1);
  }
  DEBUGMSG("Read potential parameter from file "<< fname)
    
  phi = vector<SplineType>(nspecies*nspecies);
  rho = vector<SplineType>(nspecies);
  f   = vector<SplineType>(nspecies);
  
  vector<string> slist;
  do {
    while ((getwords(slist, fin))==0);
    
    if (slist[0] == "<eunit>") {
      if (slist[1] == "K")   econv = 8.617385692E-5;
      if (slist[1] == "eV")  econv = 1.0;
      if (slist[1] == "erg") econv = 6.241506363E11;
      DEBUGMSG("Energy units: " << slist[1].c_str() << " (factor: " << econv << ")")
    }
    
    if (slist[0] == "<potential>") {
      if (slist[1] == "pair") {
	int i = atoi(slist[2].c_str())-1;
	int j = atoi(slist[3].c_str())-1;
	DEBUGMSG("Read pairpotential " << i << " " << j)
	  if (i>j) {int k=i; i=j; j=k;}
	phi[i*nspecies+j].readSpline(fin, econv);
	phi[i*nspecies+j].getSpline();
	//if (i!=j) phi[j*nspecies+i] = phi[i*nspecies+j];
      }
      if (slist[1] == "dens") {
	int i = atoi(slist[2].c_str())-1;
	DEBUGMSG("Read density function " << i)
	  rho[i].readSpline(fin, 1.0);
	rho[i].getSpline();
      }
      if (slist[1] == "embed") {
	int i = atoi(slist[2].c_str())-1;
	DEBUGMSG("Read embedding function " << i)
          f[i].readSpline(fin, econv);
	f[i].getSpline();
      }
    }
  } while(slist[0] != "end");
}

#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: EAMParam.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
