//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/** @file PotentialCollection.1.cpp
 * @brief Declare member functions with large computations.
 *
 * OpenMP/MPI directives are used by these member functions.
 * Some compilers have trouble compiling large cpp files and
 * it is not a bad idea to compile multiple sizable cpp files.
 */
#include <numeric>
#include "Potentials/PotentialCollection.h"
#include "Potentials/ShortRangePotential.h"
#include "Particle/NNEngine.h"
#include "Message/OpenMP.h"
#include "Utilities/OhmmsInfo.h"

namespace OHMMS {

  bool PotentialCollection::evalDynamicMatrix(DynMatrix_t&) {
    return false;
  }
  
  PotentialCollection::Scalar_t 
  PotentialCollection::evalEnergy() {

    Scalar_t esum = 0.0e0;
    for(size_type l=0; l<H_long.size(); l++) 
      esum += H_long[l]->evalEnergy();

#pragma omp parallel 
    {
      size_type np = nnListEngine->size();
#pragma omp for nowait 
      for(size_type myID=0; myID<np; myID++) {
        Scalar_t e=0.0;
        for(size_type s=0; s<H_short.size(); s++) {
  	  e += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID),1.0);
        }
        nnListEngine->getPairAttrib(myID).Energy=e;
      }
    }//end-of-omp
   
    for(size_type myID=0; myID<nnListEngine->size(); myID++) {
      esum += nnListEngine->getPairAttrib(myID).Energy;
    }

    return nnListEngine->collect(esum);
  }

  PotentialCollection::Scalar_t 
  PotentialCollection::evalForce(ParticlePos_t& f0, unsigned fmode) {
    //update the pair relations
    //setCurrentPos(myPtcl->R, myPtcl->dR);
    unsigned stressmode=(fmode&1);
    Scalar_t esum=0.0;
    Stress=0.0;
    f0=0.0;
    for(size_type l=0; l<H_long.size(); l++) {
      esum += H_long[l]->evalForce(f0,stressmode);
    }
    
    size_type myID = 0;
    for(size_type s=0; s<H_short.size(); s++) {
      esum += H_short[s]->evalForceNN(nnListEngine->getPairAttrib(myID),f0,stressmode);
      Stress += nnListEngine->getPairAttrib(myID).Stress;
    }

    if(CollectVisitingForces) nnListEngine->endCollect(f0);
    return nnListEngine->collect(esum);
  }

  PotentialCollection::Scalar_t 
  PotentialCollection::evalForceAndPressure(ParticlePos_t& f, Scalar_t& dedv) {


    Scalar_t esum = evalForce(f,1), eplus=0.0,eminus=0.0;

    const Scalar_t dilation =    1.001000000000000;
    const Scalar_t contraction = 0.999000000000000;

#pragma omp parallel reduction(+:eplus,eminus)
    {
      omp_int_t myID = omp_get_thread_num();
      for(size_type s=0; s<H_short.size(); s++) {
	eplus += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID), dilation);
      }
      for(size_type s=0; s<H_short.size(); s++) {
	eminus += H_short[s]->evalEnergyNN(nnListEngine->getPairAttrib(myID), contraction);
      }
    }//end-of-omp

    //FIX IT FOR MPI
    //eplus = myPtcl->getNNEngine()->collect(eplus);
    //eminus = myPtcl->getNNEngine()->collect(minus);
    Scalar_t dV = (pow(dilation,3)-pow(contraction,3))*myPtcl->Lattice.Volume;
    dedv = (eplus-eminus)/2.0/dV;
    return esum;
  }  

  void 
  PotentialCollection::setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell) {
    //First update ParticleSet::curR to be inside the unit cell
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
    myPtcl->applyBConds(true,myPtcl->curR.getUnit());
  }

  void 
  PotentialCollection::setCurrentPos(ParticlePos_t& pos) {
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(*myPtcl,myPtcl->curR);
  }

  void 
  PotentialCollection::setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& del) {
    nnListEngine->updatePairs(*myPtcl,del);
  }

 /*
  void 
  PotentialCollection::setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell) {

    Counter = 0;
    myPtcl->advance();

    if(!UseNNList) return;

    //First update ParticleSet::curR to be inside the unit cell
    myPtcl->curR.setUnit(PosUnit::LatticeUnit);
    //myPtcl->applyBC(pos,myPtcl->curR);
    //nnListEngine->updateGridData(myPtcl->curR,myPtcl->Lattice);
    nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
    nnListEngine->makeNewPairs(myPtcl->Lattice,myPtcl->curR, myPtcl->ID);
    if(wrap) myPtcl->applyBConds(true,PosUnit::LatticeUnit);

  }

  void 
  PotentialCollection::setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& dr) {
//    nnListEngine->updatePairs(dr);
    Counter++;
    myPtcl->advance();
      if(UseNNList) {
        if(nnListEngine->needCleanUp(Counter)) {
          myPtcl->curR.setUnit(PosUnit::LatticeUnit);
          nnListEngine->updateGridData(pos,myPtcl->curR,myPtcl->Lattice);
          nnListEngine->makeNewPairs(myPtcl->Lattice,myPtcl->curR, myPtcl->ID);
        } else {
          nnListEngine->updatePairs(dr);
        }
      }
  }
  */
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: PotentialCollection.1.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/

