//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

/**@file PotentialCollection.h
 * @brief Define PotentialCollection class.
 */
#ifndef OHMMS_POTENTIAL_COLLECTION_H
#define OHMMS_POTENTIAL_COLLECTION_H

#include "OhmmsData/OhmmsElementBase.h"
#include "Potentials/PotentialBase.h"

namespace OHMMS {

  class ShortRangePotential;

  /** Inherited class from PotentialBase 
   *
   * Using Composite pattern. Other application-level classes, such as 
   * PropagatorBase's, use PotentialCollection. The functions ensure that
   * the computationally expensive tasks are done only once by this
   * class and the constituting PotentialBase* and ShortRangePotential*
   * use the updated information to evaluate the physical interactions.
   */
  class PotentialCollection :  public PotentialBase
    {

    public:

      ///constructor
      PotentialCollection();

      ///destructor
      ~PotentialCollection();

      //@{Implement virtual functions of OhmmsElementBase
      virtual bool get(ostream&) const;
      virtual bool put(istream&);
      virtual bool put(xmlNodePtr cur);
      virtual void reset();
      //@}

      //@{Implement virtual functions of PotentialBase
      virtual void setParticleSet(Particle_t* ptclset);
      virtual Scalar_t evalEnergy();
      /** reimplement
       * @param fv new forces are assigned
       * @param fmode if(fmode&1) evaluate stress; if(fmode&2) update pair list
       *
       * Default fmode=0.
       */
      virtual Scalar_t evalForce(ParticlePos_t& fv, unsigned fmode = 0);
      virtual bool evalDynamicMatrix(DynMatrix_t&);
      //@}      

      //@{New member functions for PotentialCollection
      Scalar_t evalForceAndPressure(ParticlePos_t& f, Scalar_t& dedv);
      void addPotential(PotentialBase* aterm, const string& aname, 
          xmlNodePtr cur);
      void addShortRangePotential(ShortRangePotential* aterm, 
          const string& aname, xmlNodePtr cur);
      bool remove(const string& aname);      
      bool addRecord(RecordPropertyList& reclist, xmlNodePtr cur);

      /** assign xmlNodePtr to Particle Layout object
       * @param pl_ptr an xml node
       */
      void setLayout(xmlNodePtr pl_ptr);

      /** Set the current position to be pos
       *@param pos position array
       */
      virtual void setCurrentPos(ParticlePos_t& pos);

      /** Set the current position to be pos
       *@param pos position array
       *@param wrap bool, if true, call myPtcl->applyBConds()
       *@param update bool, if true (default), update the cell information.
       */
      virtual void setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell=true);

      /** Set the current position to be pos and set the displacement vector dr
       *@param pos position array
       *@param dr  displacement array
       */
      void setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& dr);

      Scalar_t getEnergy();

    private:

      ///Flag to check if everything is initialized.
      bool Ready;

      ///Flag to ensure that the unit of the ParticlePos_t used by NNEngine
      bool UseLatticeUnit;

      ///Flag to create and update the data for each grid point
      bool UseGridData;

      /// Index for the type of nearest-neighbo lists
      int UseNNList;

      /// Index for the type of nearest-neighbo lists
      int LastAddedTerm;

      /// xml node to define layout
      xmlNodePtr LayoutXmlNode;

      /// Total energy
      Scalar_t PotEnergy;

      typedef pair<size_type,xmlNodePtr> CheckType;
      std::map<string,CheckType> H_short_list;
      std::map<string,CheckType> H_long_list;
      std::vector<ShortRangePotential*> H_short;
      std::vector<PotentialBase*> H_long;

      ParticleIndex_t Cell;
      void updatePairlist(bool update_list);
      void initialize();
    };
  
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: PotentialCollection.h 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
