//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Richard Hennig and Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <string>
#include <vector>
using namespace std;
#include "Potentials/AlloyEAMPotential.h"
#include "Potentials/CubicSplineType.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"
#include "Message/OpenMP.h"
#include "ParticleBase/ParticleMessage.h"

/*! \class AlloyEAMParam
 *  \brief Contain spline parameters for AlloyEAM potentials.
 *         Contains various splines for a AlloyEAM potential
 *  \note Using SplineType defined in CubicSplineType.h
 */
struct OHMMS::AlloyEAMParam {
  typedef double Scalar_t;
  Scalar_t clsphircut;    //!<Range of pair potential
  Scalar_t clsrhorcut;    //!<Range for rho(r)
  Scalar_t clsfrhocut;    //!<Range for f(rho)
  int      nspecies;          //Number of species
  vector<SplineType> phi, f, rho;  // SplineTypeAlloyEAM phi, f, rho;

  AlloyEAMParam(const char *fname, int num) {
    readAlloyEAMParam(fname, num);
    clsphircut = phi[0].X[phi[0].Npt];
    clsrhorcut = rho[0].X[rho[0].Npt];
    clsfrhocut = f[0].X[1];
    for (int i=1; i<num; i++) {
      clsphircut = max(clsphircut, phi[i].X[phi[i].Npt]);
      clsrhorcut = max(clsrhorcut, rho[i].X[rho[i].Npt]);
      clsfrhocut = min(clsfrhocut, f[1].X[1]);
    }
  }
  void readAlloyEAMParam(const char *fname, int num);
  Scalar_t rcutmax() { 
    return (clsphircut > clsrhorcut)? clsphircut : clsrhorcut;
  }
};
using namespace OHMMS;

/*! \fn AlloyEAMPotential::AlloyEAMPotential()
 *  \brief Default constructor of AlloyEAMPotential
 */
AlloyEAMPotential::AlloyEAMPotential(): mySpecies(NULL){
  Name = "AlloyEAM";
}

/*! \fn AlloyEAMPotential::AlloyEAMPotential(int argc, char** argv) 
 *  \param argc number of arguments
 *  \param argv list of arguments
 */
AlloyEAMPotential::AlloyEAMPotential(int argc, char** argv): 
mySpecies(NULL){
  Name = "AlloyEAM";
}

AlloyEAMPotential::~AlloyEAMPotential() {
  for(int i=0; i<AlloyEAMSet.size(); i++) delete AlloyEAMSet[i];
}

void AlloyEAMPotential::initialize() {

  if(AlloyEAMSet.empty()) {

    string potfile("AlloyEAMPOT");

    Rmax = 0.0;
    mySpecies = SpeciesCollection::getSpecies();
    int imass = SpeciesCollection::addAttrib("mass");
    int irmax = SpeciesCollection::addAttrib("rmax");

    cerr << "Data file: " << potfile.c_str() << "\n";

    ifstream fin(potfile.c_str());
    if(!fin) { 
      cerr << potfile.c_str() << " is not found in this directory.\n";
      cerr << "Exiting very ungracefully.\n";
      exit(1);
    }
    cerr << "Read species parameters from file "<< potfile.c_str() << "\n";

    vector<string> slist;
    int nspecies = 0;
    do {
      while ((getwords(slist, fin))==0);
      if (slist[0] == "species") {
	int newid =SpeciesCollection::addSpecies(slist[1].c_str());
	(*mySpecies)(imass,newid)=atof(slist[2].c_str());
	nspecies ++;
	cerr << "New Species: " << slist[1].c_str() << " (" << newid
	     << ") " << "m= " << (*mySpecies)(imass,newid) <<"\n";
      }
    } while(slist[0] != "<potential>");

    cerr << "Number of species: " << nspecies << "\n";
    AlloyEAMSet.push_back(new AlloyEAMParam(potfile.c_str(), nspecies));

    Rmax = max(Rmax, AlloyEAMSet[0]->rcutmax());
    cerr << "Set Rmax: " << Rmax <<"\n";
    for (int i=0; i<nspecies; i++) (*mySpecies)(irmax,0) = Rmax*0.5;
  }
}

Scalar_t 
AlloyEAMPotential::e(const ParticlePos_t& R, int i, int myid) 
{
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myid);
  int nn = nnlist.nadj(i);
  Scalar_t  r;
  Scalar_t  rho = 0.0e0, elocal =0.0e0;
  SingleParticlePos_t dr;

  for(int jL=0; jL<nn; ++jL) 
    {
      int j = nnlist.iadj(i,jL);  //atomic index of jL-th nn of i

      if(R.InUnit) 
	dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
      else
	dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

      r = sqrt(myPtcl->Lattice.Dot(dr,dr));

      // check all the nearest atoms within a cutoff of phi
      if (r<AlloyEAMSet[0]->clsphircut) {
	elocal += 0.5*AlloyEAMSet[0]->phi[myPtcl->IonID[i]*AlloyEAMSet[0]->nspecies+myPtcl->IonID[j]].splint(r,0);
	//phi(r),double counting
      }

      if (r<AlloyEAMSet[0]->clsrhorcut) {
	rho += AlloyEAMSet[0]->rho[myPtcl->IonID[j]].splint(r,0);
      }
    }

  elocal += AlloyEAMSet[0]->f[myPtcl->IonID[i]].splint(rho,0);
  return elocal; // return the energy
}

Scalar_t 
AlloyEAMPotential::evalEnergy() {
  //NOTE PotentialBase::evalEnergy interface is modified 09-27-2001
  //PotentialBase::Scalar_t 
  //AlloyEAMPotential::evalEnergy(SimpleParticle& inSys) {

  getPairList(); // using a pair-list
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(0);
  int LocalNum = myPtcl->getLocalNum();

  //resize(nnlist.getMaxNadj()); // resize aux vectors

  Scalar_t  r, elocal=0, rho;
  int i,nn;
  for(i=0; i<LocalNum; i++) {

    //if(inSys.IonID[i] != AlloyEAMSpecies) continue; 
    nn = nnlist.nadj(i);
    //emaX = 0.0e0;
    rho = 0.0e0;
    for(int jL=0; jL<nn; ++jL) {
      int j = nnlist.iadj(i,jL);  //atomic index of jL-th nn of i
      int loc = nnlist.loc(i,jL); //location of R, dR of the pair

      //if(inSys.IonID[j] == AlloyEAMSpecies) continue;

      r = nnlist.R[loc];

      // check all the nearest atoms within a cutoff of phi
      if (r<AlloyEAMSet[0]->clsphircut) {
	elocal += 
	  0.5*AlloyEAMSet[0]->phi[myPtcl->IonID[i]*AlloyEAMSet[0]->nspecies+myPtcl->IonID[j]].splint(r,0);
	//phi_ij(r), double counting
      }

      if (r<AlloyEAMSet[0]->clsrhorcut) {
	rho += AlloyEAMSet[0]->rho[myPtcl->IonID[j]].splint(r,0);
      }
    }
    elocal += AlloyEAMSet[0]->f[myPtcl->IonID[i]].splint(rho,0); 
  }
  return elocal; // return the energy
}


Scalar_t 
AlloyEAMPotential::evalForce(ParticlePos_t& f0, ParticleScalar_t& q,
			     Tensor_t& stress, unsigned stressmode) {

  DEBUGMSG("Entering AlloyEAMPotential::evalForce.\n")
  getPairList(); // using a pair-list

  Scalar_t esum = 0.0;
  if(nnListEngine->size() >1) {
#pragma omp parallel
    {
      int myID = omp_get_thread_num();
      ParticlePos_t ft(f0.size());
      Tensor_t stressloc;
      Scalar_t esumloc = evalForceNN(nnListEngine->getPairAttrib(myID),
				     myID,ft,q,stressloc,stressmode);
#pragma omp critical
      {    
	esum += esumloc;
	stress += stressloc;
	f0 += ft;
      }
    }
  } else 
    esum = evalForceNN(nnListEngine->getPairAttrib(0),0,
		       f0,q,stress,stressmode);
  if(nnListEngine->distributed()) {
    gsum(f0,0);
    gsum(q,0);
    gsum(esum,0);
  }
  return esum;
}


Scalar_t 
AlloyEAMPotential::evalForceNN(const PtclPairList_t& nnlist, int myID,
			       ParticlePos_t& f0, ParticleScalar_t& q, 
			       Tensor_t& stress, unsigned stressmode) {

  // storing three-body function
  //int max_est = (4*nnlist.getTotNadj())/nnlist.getLocalNum();
  int max_est = nnlist.getMaxNadj();

  //temporary data for multi-body functions
  vector<SaveRho> srho(max_est);
  vector<SaveF>   sf(max_est);
  vector<SaveG>   sg(max_est*max_est);

  int nnloc = 0;
  Scalar_t esumloc = 0.0;
  int first =  nnListEngine->first_id(myID);
  int last =    nnListEngine->last_id(myID);
  for(int iL = 0; iL < nnlist.getLocalNum(); iL++) {
    int iG = nnlist.I[iL]; // get the index, physical index
    int num_nn = nnlist.nadj(iL);
    int itype = myPtcl->IonID[iG];
    int n3 = 0;
    int n2 = 0;
    Scalar_t elocal = 0.0;
    Scalar_t val;

    for(int nni = 0; nni< num_nn; nni++, nnloc++) {
      int jG = myPtcl->ID[nnlist.iadj(iL,nni)];
      Scalar_t r = nnlist.R[nnloc];

      if(r < AlloyEAMSet[0]->clsphircut) {
	int jtype = myPtcl->IonID[jG];
	Scalar_t deriv = 
	  AlloyEAMSet[0]->phi[itype*AlloyEAMSet[0]->nspecies+jtype].splint(r, val);
	SingleParticlePos_t dr = nnlist.dR[nnloc];
	elocal += val*0.5e0;                     // double counting
	SingleParticlePos_t df = 0.5e0*deriv*dr; // double counting
	
	f0[iG] += df;
	f0[jG] -= df;
	stress -= r*outerProduct(df, dr);
	  
	if(r < AlloyEAMSet[0]->clsrhorcut) { 
	  // save data associated with rho(r) 
	  deriv = AlloyEAMSet[0]->rho[jtype].splint(r, val, 0);
	  srho[n2].id = jG;      // ID
	  srho[n2].r = r;       // distance
	  srho[n2].rho = val;   // rho(r)
	  srho[n2].drho = deriv;// d rho(r)/ dr
	  srho[n2].dr = dr;     // (r[j]-r[i])/r
	  n2++;
	}
      }
    } // for(; plisti != plistf; ++plisti)
    
    // calculating x for U(x) for the embedded atom term
    Scalar_t emaX = 0.e0;
    for(int nj=0; nj < n2; nj++) {
      emaX += srho[nj].rho;
    }
    
    // Done with emaX, calculate the EMA term
    //Scalar_t dux;               // du/dx, used for a force prefactor
    Scalar_t dux = AlloyEAMSet[0]->f[itype].splint(emaX, val);
    elocal +=val; 
      
    for(int nj=0; nj < n2; nj++) { // forces from drho
      int jG = srho[nj].id;
      SingleParticlePos_t df = dux * srho[nj].drho * srho[nj].dr;
      f0[iG] += df;
      f0[jG] -= df;
      stress -= srho[nj].r*outerProduct(df, srho[nj].dr);
    }
    
    q[iG] = elocal;
    esumloc += elocal;
  }

  return esumloc;
}

bool 
AlloyEAMPotential::evalDynamicMatrix(DynMatrix_t&) {

  return false;

}

//void AlloyEAMPotential::resize(int num_nn) {
//      //Check if the auxilary vectors are big enough
//      if(srho.size()< num_nn) {
//        srho.erase(srho.begin(), srho.end());
//        srho.reserve(num_nn*2);
//        for(int is=0; is< 2*num_nn; is++) srho.push_back(SaveRho());
//      }
//      if(sf.size()< num_nn) {
//        sf.erase(sf.begin(), sf.end());
//        sf.reserve(num_nn*2);
//        for(int is=0; is< 2*num_nn; is++) sf.push_back(SaveF());
//      }
//      if(sg.size()< num_nn*num_nn) {
//        sg.erase(sg.begin(), sg.end());
//        sg.reserve(num_nn*num_nn);
//        for(int is=0; is< num_nn*num_nn; is++) sg.push_back(SaveG());
//      }
//}

void AlloyEAMParam::readAlloyEAMParam(const char *fname, int nspec) {
  Scalar_t econv = 1.0;
  nspecies = nspec;

  ifstream fin(fname);
  if(!fin) { 
    cerr << fname << " is not found in this directory.\n";
    cerr << "Exiting very ungracefully.\n";
    exit(1);
  }
  cerr << "Read potential parameter from file "<< fname << "\n";

  phi = vector<SplineType>(nspecies*nspecies);
  rho = vector<SplineType>(nspecies);
  f   = vector<SplineType>(nspecies);

  vector<string> slist;

  do {
    while ((getwords(slist, fin))==0);

    if (slist[0] == "<eunit>") {
      if (slist[1] == "K")   econv = 8.617385692E-5;
      if (slist[1] == "eV")  econv = 1.0;
      if (slist[1] == "erg") econv = 6.241506363E11;
      cerr << "Energy units: " << slist[1].c_str() << " (factor: " << econv << ")\n";
    }

    if (slist[0] == "<potential>") {
      if (slist[1] == "pair") {
	int i = atoi(slist[2].c_str())-1;
	int j = atoi(slist[3].c_str())-1;
	cerr << "Read pairpotential " << i << " " << j << "\n";
	if (i>j) {int k=i; i=j; j=k;}
	phi[i*nspecies+j].readSpline(fin, econv);
	phi[i*nspecies+j].getSpline();
	if (i!=j) phi[j*nspecies+i] = phi[i*nspecies+j];
      }
      if (slist[1] == "dens") {
	int i = atoi(slist[2].c_str())-1;
	cerr << "Read density function " << i << "\n";
	rho[i].readSpline(fin, 1.0);
	rho[i].getSpline();
      }
      if (slist[1] == "embed") {
	int i = atoi(slist[2].c_str())-1;
	cerr << "Read embedding function " << i << "\n";
	f[i].readSpline(fin, econv);
	f[i].getSpline();
      }
    }
  } while(slist[0] != "end");
}

bool AlloyEAMPotential::get(ostream& os) const { 
  return true;
}
bool AlloyEAMPotential::put(istream&) {return true;}
void AlloyEAMPotential::reset() {

}
#if defined(ENABLE_LIBXML2)
bool AlloyEAMPotential::put(xmlNodePtr cur){

  string potfile = (const char*)(xmlGetProp(cur, (const xmlChar *) "file"));

  Rmax = 0.0;
  mySpecies = SpeciesCollection::getSpecies();
  int imass = SpeciesCollection::addAttrib("mass");
  int irmax = SpeciesCollection::addAttrib("rmax");

  cerr << "Data file: " << potfile.c_str() << "\n";

  ifstream fin(potfile.c_str());
  if(!fin) { 
    cerr << potfile.c_str() << " is not found in this directory.\n";
    cerr << "Exiting very ungracefully.\n";
    exit(1);
  }
  cerr << "Read species parameters from file "<< potfile.c_str() << "\n";

  vector<string> slist;
  int nspecies = 0;
  do {
    while ((getwords(slist, fin))==0);
    if (slist[0] == "species") {
      int newid =SpeciesCollection::addSpecies(slist[1].c_str());
      (*mySpecies)(imass,newid)=atof(slist[2].c_str());
      nspecies ++;
      cerr << "New Species: " << slist[1].c_str() << " (" << newid
	   << ") " << "m= " << (*mySpecies)(imass,newid) <<"\n";
    }
  } while(slist[0] != "<potential>");

  cerr << "Number of species: " << nspecies << "\n";
  AlloyEAMSet.push_back(new AlloyEAMParam(potfile.c_str(), nspecies));

  Rmax = max(Rmax, AlloyEAMSet[0]->rcutmax());
  cerr << "Set Rmax: " << Rmax <<"\n";
  for (int i=0; i<nspecies; i++) (*mySpecies)(irmax,0) = Rmax*0.5;

  return true;
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: AlloyEAMPotential.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

