// -*- C++ -*-
/***************************************************************************
 *
 * The POOMA Framework
 * 
 * This program was prepared by the Regents of the University of
 * California at Los Alamos National Laboratory (the University) under
 * Contract No.  W-7405-ENG-36 with the U.S. Department of Energy (DOE).
 * The University has certain rights in the program pursuant to the
 * contract and the program should not be copied or distributed outside
 * your organization.  All rights in the program are reserved by the DOE
 * and the University.  Neither the U.S.  Government nor the University
 * makes any warranty, express or implied, or assumes any liability or
 * responsibility for the use of this software
 *
 * Visit http://www.acl.lanl.gov/POOMS for more details
 *
 ***************************************************************************/


#include "Potentials/KeatingParam.h"
#include "Potentials/KeatingPotential.h"

template<class playout_t>
KeatingPotential<playout_t>::KeatingPotential(){  
   // default is to read from KeatingParam
  TAU_PROFILE("KetaingPotential::KeatingPotential()", " ", TAU_USER);
  KP.readKeatingParam("KeatingParam");
//  KP.print(cout);
}

template<class playout_t>
KeatingPotential<playout_t>::KeatingPotential(const char *fname) {
  TAU_PROFILE("KetaingPotential::KeatingPotential()", " ", TAU_USER);
  KP.readKeatingParam(fname);
}

template<class playout_t>
void KeatingPotential<playout_t>::reportHeader(ostream& o) const{
  o << "  BStretch    BBending ";
}

template<class playout_t>
void KeatingPotential<playout_t>::print(ostream& o) const {
  o << setw(12) << setprecision(5) << BondStretchingE 
    << setw(12) << setprecision(5) << BondBendingE; 
}


template<class playout_t>
PotentialBase<playout_t>::Element_t
KeatingPotential<playout_t>::evalForce(SimpleParticle<playout_t>& p,
                                  ParticlePos_t& f,
                                  bool assignEnergy, bool virial) {

  TAU_PROFILE("KetaingPotential::evalForce()", "void () ", TAU_USER);

  int i,j,k, itype, jtype, ktype, i2,i3;
  int nspecies   = KP.nSpecies;
  Element_t rcutsq = 1.20e0*KP.rsqmax(); 

  int LocalNum = p.getLocalNum();
  int TotalNum = p.getTotalNum();

  //localQ[0] -> local sum of twobody energies
  //localQ[1] -> local sum of threebody energies
  //localQ[2] -> local sum of pressure
  Element_t  localQ[3] = {0.0e0, 0.0e0, 0.0e0};
  Element_t globalQ[3] = {0.0e0, 0.0e0, 0.0e0};

  for(i=0; i< LocalNum; i++) { // going thru atomic indices

    itype = p.IonID[i]; // checking the species

    Element_t e2=0.0e0, e3=0.0e0, z = 0.0e0;
    ParticleLayout_t::pair_iterator pji,pki;
    ParticleLayout_t::pair_iterator pjf,pkf;

    p.getLayout().getPairlist(i, pji, pjf, p); // neighbor lists
    
    Element_t r2ij,r2ik;

    for(; pji != pjf; ++pji) {

      j = pji->first;
      jtype = p.IonID[j];

      r2ij = pji->second;
      if(r2ij < rcutsq) {
        z += 1.0e0;
        i2 = nspecies*itype+jtype; // getting the index for twobody potential

        SingleParticlePos_t dij = p.R[j] - p.R[i];
	SingleParticlePos_t rvec = KP.df2(i2,r2ij)*dij;

        f[i] += rvec;
        f[j] -= rvec;

        e2 += KP.f2(i2,r2ij);
        if(virial) localQ[2] += dot(rvec,dij);

        p.getLayout().getPairlist(i, pki, pkf, p); // neighbor lists
	// checking three body term
        for(; pki < pjf;++pki) { // three body
          k = pki->first;
          if(k > j) {
          ktype = p.IonID[k];
          r2ik = pki->second;
          if(r2ik < rcutsq) {
            // Always only count j-i-k type triple bonds
            i3 = nspecies*(jtype*nspecies+itype)+ktype;
            SingleParticlePos_t dik = p.R[k]-p.R[i];
            Element_t thex = dot(dij,dik);
            Element_t xx = KP.df3(i3,thex);
            f[i] += xx*(dij+dik);
            f[j] -= xx*dik;
            f[k] -= xx*dij;
            e3 += KP.f3(i3,thex);
            if(virial) {
  	      localQ[2] -= 2.0e0*dot(dij,xx*dik); // check this
	    }
	  }
          }
	}
      }
    }
    p.Coord[i] = z;
    // after proper normalization
    if(assignEnergy) p.ELocal[i] = e2+e3;

    localQ[0] += e2;
    localQ[1] += e3;
  }

  if(assignEnergy || virial) 
    reduce(localQ, localQ+3, globalQ, OpAddAssign());

  if(assignEnergy)  {
     p.PEnergy = globalQ[0]+globalQ[1];
     BondStretchingE = globalQ[0];
     BondBendingE = globalQ[1]; 
  }

  if(virial)        p.Pressure =  globalQ[2]/3.0;

  p.collect_forces(f);

  return globalQ[0]+globalQ[1]; // return the energy
}     

template<class playout_t>
PotentialBase<playout_t>::Element_t
KeatingPotential<playout_t>::evalForce(SimpleParticle<playout_t>& p,
                       ParticlePos_t& f,
                       ParticlePos_t& fb,
                       ParticleIndex_t& gb,
                       bool assignEnergy, bool virial) {
  TAU_PROFILE("KetaingPotential::evalForce()", "void () ", TAU_USER);

  int j, k, itype, jtype, ktype, i2,i3;
  int nspecies = KP.nSpecies;
  Element_t rcutsq = 1.20e0*KP.rsqmax(); 

  //localQ[0] -> local sum of twobody energies
  //localQ[1] -> local sum of threebody energies
  //localQ[2] -> local sum of pressure
  Element_t  localQ[3] = {0.0e0, 0.0e0, 0.0e0};
  Element_t globalQ[3] = {0.0e0, 0.0e0, 0.0e0};
  for(int i=0; i< p.getLocalNum(); i++) { // going thru atomic indices
    itype = p.IonID[i]; // checking the species
  
    Element_t e2=0.0e0, e3=0.0e0, z = 0.0e0;
    ParticleLayout_t::pair_iterator plisti, pki0, pki;
    ParticleLayout_t::pair_iterator plistf;
  
    p.getLayout().getPairlist(i, plisti, plistf, p); // neighbor lists
    pki0 = plisti; // for three body term
  
    Element_t r2ij,r2ik;
  
    for(; plisti != plistf; ++plisti) {
  
      j = plisti->first;
      jtype = p.IonID[j];
  
      r2ij = plisti->second;
    
      if(r2ij < rcutsq) {
        z += 1.0e0;
        i2 = nspecies*itype+jtype; // getting the index for twobody potential
  
        SingleParticlePos_t dij = p.R[j] - p.R[i];
        SingleParticlePos_t rvec = KP.df2(i2,r2ij)*dij;
  
        if(gb[i]) {
          fb[i] += rvec;
          fb[j] -= rvec;
        } else {
          f[i] += rvec;
          f[j] -= rvec;
        }

        e2 += KP.f2(i2,r2ij);
        if(virial) localQ[2] += dot(rvec,dij);
  
        // checking three body term
        for(pki = pki0; pki<plisti;++pki) { // three body
          k = pki->first;
          ktype = p.IonID[k];
          r2ik = pki->second;
          if(r2ik < rcutsq) {
            i3 = nspecies*(itype+jtype*nspecies)+ktype;
            SingleParticlePos_t dik = p.R[k]-p.R[i];
            Element_t xx = KP.df3(i3,dot(dij,dik));
            if(gb[i]) {
              fb[i] += xx*(dij+dik);
              fb[j] -= xx*dik;
              fb[k] -= xx*dij;
	    } else {
              f[i] += xx*(dij+dik);
              f[j] -= xx*dik;
              f[k] -= xx*dij;
	    }
            e3 += KP.f3(i3,dot(dij,dik));
            if(virial) {
              localQ[2] -= 2.0e0*dot(dij,xx*dik); // check this
            }
          }
        }
      }
    }
    if(assignEnergy) p.ELocal[i] = e2+e3;
    localQ[0] += e2;
    localQ[1] += e3;
  }

  if(assignEnergy || virial) 
    reduce(localQ, localQ+2, globalQ, OpAddAssign());

  if(assignEnergy)  p.PEnergy = globalQ[0]+globalQ[1];
  if(virial)        p.Pressure =  globalQ[2]/3.0;

  return globalQ[0]+globalQ[1]; // return the energy
}     

template<class playout_t>
PotentialBase<playout_t>::Element_t
KeatingPotential<playout_t>::evalEnergy(SimpleParticle<playout_t>& p,
                                        bool assignEnergy){

  TAU_PROFILE("KetaingPotential::evalForce()", "void () ", TAU_USER);

  int i,j,k, itype, jtype, ktype, i2,i3;
  int nspecies   = KP.nSpecies;
  Element_t rcutsq = 1.20e0*KP.rsqmax(); 

  int LocalNum = p.getLocalNum();
  int TotalNum = p.getTotalNum();

  //localQ[0] -> local sum of twobody energies
  //localQ[1] -> local sum of threebody energies
  //localQ[2] -> local sum of pressure
  Element_t  localQ[2] = {0.0e0, 0.0e0};
  Element_t globalQ[2] = {0.0e0, 0.0e0};

  for(i=0; i< LocalNum; i++) { // going thru atomic indices

    itype = p.IonID[i]; // checking the species

    Element_t e2=0.0e0, e3=0.0e0, z = 0.0e0;
    ParticleLayout_t::pair_iterator pji,pki;
    ParticleLayout_t::pair_iterator pjf,pkf;

    p.getLayout().getPairlist(i, pji, pjf, p); // neighbor lists
    
    Element_t r2ij,r2ik;

    for(; pji != pjf; ++pji) {

      j = pji->first;
      jtype = p.IonID[j];

      r2ij = pji->second;
      if(r2ij < rcutsq) {
        z += 1.0e0;
        i2 = nspecies*itype+jtype; // getting the index for twobody potential

        SingleParticlePos_t dij = p.R[j] - p.R[i];
        e2 += KP.f2(i2,r2ij);

        p.getLayout().getPairlist(i, pki, pkf, p); // neighbor lists
	// checking three body term
        for(; pki < pjf;++pki) { // three body
          k = pki->first;
          if(k > j) {
            ktype = p.IonID[k];
            r2ik = pki->second;
            if(r2ik < rcutsq) {
              // Always only count j-i-k type triple bonds
              i3 = nspecies*(jtype*nspecies+itype)+ktype;
              SingleParticlePos_t dik = p.R[k]-p.R[i];
	      // Element_t thex = dot(dij,dik);
              e3 += KP.f3(i3,dot(dij,dik));
	    }
          }
	}
      }
    }
    p.Coord[i] = z;
    // after proper normalization
    if(assignEnergy) p.ELocal[i] = e2+e3;
    localQ[0] += e2;
    localQ[1] += e3;
  }

  if(assignEnergy)
    reduce(localQ, localQ+2, globalQ, OpAddAssign());

  if(assignEnergy){
     p.PEnergy = globalQ[0]+globalQ[1];
     BondStretchingE = globalQ[0];
     BondBendingE = globalQ[1]; 
  }

  return globalQ[0]+globalQ[1]; // return the energy
}     

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * POOMA_VERSION_ID: $Id: KeatingPotential.org.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

