//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
/////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifdef USE_OPENMP
#include <omp.h>
#endif
#include <vector>
using namespace std;
#include "Particle/NearestNeighborBase.h"
#include "Utilities/SpeciesCollection.h"

template<class ScaleFunc>
TwoBodyPotential<ScaleFunc>::TwoBodyPotential(int argc, char **argv)
:Phi(NULL)
{
  for(int i=0; i<argc; i++) {
    options.push_back(argv[i]);
  }
}

template<class ScaleFunc>
void TwoBodyPotential<ScaleFunc>::initialize() {

  if(!Phi) Phi = new ScaleFunc;
  Name = "TwoBody"; // name is LJ
  SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
  int newid, irmax, imass;
  newid =SpeciesCollection::addSpecies("Ar");
  irmax = SpeciesCollection::addAttrib("rmax");
  imass = SpeciesCollection::addAttrib("mass");
  (*mySpecies)(imass,newid)= 39.948;
  Rmax = Phi->rmax();
}
  
template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e(const ParticlePos_t& R, int i, int myID) {

  // use pre-calculated neighbor lists
  ////////////////////////////////////////
  // only works if the logical and physical orders are identical
  ////////////////////////////////////////
  int iL = i - nnListEngine->first_id(myID);
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

  Scalar_t esum = 0;
  SingleParticlePos_t dr;
  Scalar_t r;
  int num_nn = nnlist.nadj(iL);

  for(int nni=0; nni < num_nn; ++nni) {
    int j = myPtcl->ID(nnlist.iadj(iL,nni));  //atomic index of jL-th nn of i
    if(R.InUnit) 
      dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
    else
      dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

    r = sqrt(myPtcl->Lattice.Dot(dr,dr));
    if(r < Rmax) {
      esum += Phi->f(r);
    }
  }  // for jL < nn

  return esum;
}
  
template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e1(const ParticlePos_t& R, int i, 
				Scalar_t h, Scalar_t k, Scalar_t l, Scalar_t c, int myID) {

  // use pre-calculated neighbor lists
  ////////////////////////////////////////
  // only works if the logical and physical orders are identical
  ////////////////////////////////////////
  int iL = i - nnListEngine->first_id(myID);
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

  Scalar_t esum = 0;
  SingleParticlePos_t dr;
  Scalar_t r;
  int num_nn = nnlist.nadj(iL);

  for(int nni=0; nni < num_nn; ++nni) {
    int j = myPtcl->ID(nnlist.iadj(iL,nni));  //atomic index of jL-th nn of i
    if(R.InUnit) 
      dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
    else
      dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

    r = sqrt(myPtcl->Lattice.Dot(dr,dr));
    
    double LHSPlane;
    LHSPlane = dr[0]*h*12.0 + dr[1]*k*12.0 + dr[2]*l*12.0;

    if (r < Rmax) 
      if(LHSPlane > c )
        esum += Phi->f(r);
      else if (LHSPlane==c)
        esum += 0.5*Phi->f(r);
    
  }
   
  return esum;
}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e1a(const ParticlePos_t& R, int i, 
				 Scalar_t h, Scalar_t k, Scalar_t l, Scalar_t c, int myID) {

  // use pre-calculated neighbor lists
  ////////////////////////////////////////
  // only works if the logical and physical orders are identical
  ////////////////////////////////////////
  int iL = i - nnListEngine->first_id(myID);
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

  Scalar_t esum = 0;
  SingleParticlePos_t dr;
  Scalar_t r;
  int num_nn = nnlist.nadj(iL);
  double LHSPlane;
  double phi, junk;
  double myPi=3.141592653589;

  for(int nni=0; nni < num_nn; ++nni) {
    int j = myPtcl->ID(nnlist.iadj(iL,nni));  //atomic index of jL-th nn of i
    if(R.InUnit) 
      dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
    else
      dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

    r = sqrt(myPtcl->Lattice.Dot(dr,dr));
    
    LHSPlane = dr[0]*h*12.0 + dr[1]*k*12.0 + dr[2]*l*12.0;

    if (r < Rmax) 
      if(LHSPlane > c+0.01 )
        esum += Phi->f(r);
      else if (LHSPlane<=c+0.01 && LHSPlane>=c-0.01 ){
        phi=atan(Phi->df(r,junk));
        cout << phi/myPi << endl;
        esum += (0.5 - phi/myPi )*Phi->f(r);
      }
    }
   
  return esum;
}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e1w(const ParticlePos_t& R, int i, 
				 Scalar_t h, Scalar_t k, Scalar_t l, Scalar_t c, 
				 Scalar_t w, int myID) {

  // use pre-calculated neighbor lists
  ////////////////////////////////////////
  // only works if the logical and physical orders are identical
  ////////////////////////////////////////
  int iL = i - nnListEngine->first_id(myID);
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

  Scalar_t esum = 0, w2; 
  SingleParticlePos_t dr;
  Scalar_t r;
  int num_nn = nnlist.nadj(iL);
  double LHSPlane;
  double phi, junk;
  double myPi=3.141592653589;
  
  w2=w*0.5;

  for(int nni=0; nni < num_nn; ++nni) {
    int j = myPtcl->ID(nnlist.iadj(iL,nni));  //atomic index of jL-th nn of i
    if(R.InUnit) 
      dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
    else
      dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

    r = sqrt(myPtcl->Lattice.Dot(dr,dr));
    
    LHSPlane = dr[0]*h*12.0 + dr[1]*k*12.0 + dr[2]*l*12.0;

    if (r < Rmax) 
      if(LHSPlane < c+w2 && LHSPlane >= c-w2)
        esum += Phi->f(r);
  }
   
  return esum;
}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e2(const ParticlePos_t& R, int i, 
				Scalar_t h, Scalar_t k, Scalar_t l, Scalar_t c, int myID) {

  // use pre-calculated neighbor lists
  ////////////////////////////////////////
  // only works if the logical and physical orders are identical
  ////////////////////////////////////////
  int iL = i - nnListEngine->first_id(myID);
  const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);

  Scalar_t esum = 0;
  SingleParticlePos_t dr;
  Scalar_t r;
  int num_nn = nnlist.nadj(iL);

  for(int nni=0; nni < num_nn; ++nni) {
    int j = myPtcl->ID(nnlist.iadj(iL,nni));  //atomic index of jL-th nn of i
    if(R.InUnit) 
      dr = myPtcl->Lattice.BConds.apply(R[j]-R[i]);
    else
      dr = myPtcl->Lattice.BConds.apply(myPtcl->Lattice.toUnit(R[j]-R[i]));

    r = sqrt(myPtcl->Lattice.Dot(dr,dr));
    
    double LHSPlane;
    LHSPlane = dr[0]*h*12.0 + dr[1]*k*12.0 + dr[2]*l*12.0;

    if (r < Rmax) 
      if(LHSPlane < c )
        esum += Phi->f(r);
      else if (LHSPlane==c)
        esum += 0.5*Phi->f(r);
    
    }

  return esum;
}


template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::evalEnergy() {

  DEBUGMSG("Entering TwoBodyPotential::evalEnergy.\n")
  getPairList(); // using a pair-list

  Scalar_t  esum =0.0e0;
  //For total energy calculations, the physical and logical indices do not matter
#if defined(SGI) && defined(USE_OPENMP)
#pragma omp parallel reduction(+:esum)
 {
    int myID = omp_get_thread_num();
#else
#pragma omp parallel for reduction(+:esum)
  for(int myID=0; myID<nnListEngine->size(); myID++)  {
#endif

    const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);
    Scalar_t r;
    int nnloc = 0;
    Scalar_t  elocal = 0.0e0;
    for(int iL = 0; iL < nnlist.getLocalNum(); iL++) {
      int nn0 = nnloc; // save the starting M[iL]
      int num_nn = nnlist.nadj(iL);
      for(int nni = 0; nni< num_nn; nni++, nnloc++) {
        r = nnlist.R[nnloc];
        if(r < Rmax) {
          esum += Phi->f(r);
        }
      }
    }

  }//end-of-omp

  return esum;
}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::evalForce(ParticlePos_t& f0, ParticleScalar_t& q,
				       Tensor_t& stress, unsigned stressmode) {

  getPairList(); // using a pair-list
  Scalar_t esum = 0.0e0;
  Scalar_t p00 = 0.0e0, p11=0.0e0, p22 = 0.e0, p01 = 0.0, p02=0.0, p12 = 0.0;

#if defined(SGI) && defined(USE_OPENMP)
#pragma omp parallel reduction(+:esum,p00,p11,p22,p01,p02,p12)
 {
    int myID = omp_get_thread_num();
#else
#pragma omp parallel for reduction(+:esum,p00,p11,p22,p01,p02,p12)
  for(int myID=0; myID<nnListEngine->size(); myID++)  {
#endif
    const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);
    //int iG = nnListEngine->myAtoms[myID][0];
    int nnloc = 0;
    Scalar_t  esumloc = 0.0, phi;
    Tensor_t  stressloc;

    for(int iL = 0; iL < nnlist.getLocalNum(); iL++) {

      int iG = nnlist.I[iL]; // get the index, physical index
      int num_nn = nnlist.nadj(iL);
      Scalar_t elocal = 0.0; // local energy to be accumulated

      for(int nni = 0; nni< num_nn; nni++, nnloc++) {
        //int jG = nnlist.iadj(iL,nni);
        Scalar_t r = nnlist.R[nnloc];

        if(r < Rmax) {
          Scalar_t dphi = Phi->df(r,phi);
          elocal += phi;
          SingleParticlePos_t df = dphi*nnlist.dR[nnloc];
          f0[iG] += 2*df; // openmp optimal, does not matter anyway
          if(stressmode) 
	    stressloc -= r*outerProduct(df, nnlist.dR[nnloc]);
	}
      }
      q[iG] = elocal;
      esumloc += elocal;
    }

    esum += esumloc;
    if(stressmode) {
      p00 += stressloc(0,0);
      p11 += stressloc(1,1);
      p22 += stressloc(2,2);
      p01 += stressloc(0,1);
      p02 += stressloc(0,2);
      p12 += stressloc(1,2);
    }
  }//end-of-omp

  if(stressmode)
    stress += Tensor_t(p00,p01,p02,-p01,p11,p12,-p02,-p12,p22);
  return esum;
}

template<class ScaleFunc>
bool 
TwoBodyPotential<ScaleFunc>::evalDynamicMatrix(DynMatrix_t&) {

  return false;

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: TwoBodyPotential.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
