//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_GRIDBASE_TWOBODYPOTENTIAL_H
#define OHMMS_GRIDBASE_TWOBODYPOTENTIAL_H

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <map>
using namespace std;
#include "Potentials/PotentialBase.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"
#include "Lattice/UniformGridConnection.h"
#include "ParticleBase/ParticleUtility.h"

namespace OHMMS {
template<class ScaleFunc>
class TwoBodyPotential:public PotentialBase {

public:

  TwoBodyPotential(){ }
  TwoBodyPotential(int argc, char **argv); 
  TwoBodyPotential(vector<string>& arglist);

  ~TwoBodyPotential() { }
  void initialize();

  Scalar_t evalEnergy();
  Scalar_t evalForce(ParticlePos_t& f, ParticleScalar_t& e, Tensor_t&, 
                     unsigned stressmode = 0);
  bool evalDynamicMatrix(DynMatrix_t&);

  Scalar_t e(const ParticlePos_t&, int i, int myid = 0);

protected:
  Matrix<int> Bond;
  Scalar_t RmaxSq, RminSq;
  vector<string> options;
  vector<ScaleFunc> Phi;
  UniformGridConnection<Scalar_t,OHMMS_DIM> NextCell;

};

template<class ScaleFunc>
TwoBodyPotential<ScaleFunc>::TwoBodyPotential(int argc, char **argv)
{
  for(int i=0; i<argc; i++) {
    options.push_back(argv[i]);
  }
}

template<class ScaleFunc>
TwoBodyPotential<ScaleFunc>::TwoBodyPotential(vector<string>& arglist)
{
  options.insert(options.end(), arglist.begin(), arglist.end());
}

template<class ScaleFunc>
void TwoBodyPotential<ScaleFunc>::initialize() {

  string potfile("POT");
  int i=0;
  while(i<options.size()) {
    if(options[i] == "--potfile") { potfile = options[++i];}
    i++;
  }
  Name = "TwoBody"; // tobody potential

  SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
  int newid, imass;
  imass = SpeciesCollection::addAttrib("mass");

  Rmax = 0.0;
  ifstream fin(potfile.c_str());
  if(fin) {
    Scalar_t rc_min = 100;
    vector<string> slist,vlist;
    bool tofindpot=true;
    do{
      int nw=getXwords(slist,fin);
      if(nw == 0) continue;
      if(slist[0] == "Potential") {tofindpot = false;}
    } while(tofindpot && !fin.eof());
    
    // temporary mapping before setting up Bond(itype,jtype) 
    map<pair<int,int>, int> bondmap;
    
    do {
      int nw=getXwords(slist,fin);
      if(nw == 0) continue;
      if(slist[0] == "species") {
	getXwords(vlist, fin, "/species");
	int iargc=0, newid;
	while(iargc<vlist.size()) {
	  if(vlist[iargc] == "name") {
	    newid =SpeciesCollection::addSpecies(vlist[++iargc].c_str());
	    iargc++;
	  } else if(vlist[iargc] == "mass") {
	    (*mySpecies)(imass,newid)=atof(vlist[++iargc].c_str()); // -m
	    iargc++;
	  } 
	  iargc++;
	}
      } else if(slist[0] == "pairpot") {
	int iat =SpeciesCollection::addSpecies(slist[2].c_str());
	int jat =SpeciesCollection::addSpecies(slist[4].c_str());
	int nbond = Phi.size();
	bondmap[pair<int,int>(iat,jat)] = nbond;
	getXwords(vlist, fin, "/pairpot");
	Phi.push_back(ScaleFunc());
	Phi[nbond].set(vlist.begin(),vlist.end());
	Rmax = max(Phi[nbond].rmax(),Rmax);
	rc_min = min(Phi[nbond].rmin(),rc_min);
      } else if(slist[0] == "/Potential") {
	tofindpot = true;
      }
    } while(!fin.eof() && !tofindpot);
    
    // resize the bond map
    Bond.resize(mySpecies->getTotalNum(),mySpecies->getTotalNum());
    Bond = -1;
      map<pair<int,int>, int>::iterator bmit = bondmap.begin();
      while(bmit != bondmap.end()) {
	Bond((*bmit).first.first, (*bmit).first.second) = (*bmit).second;
	bmit++;
      }
    RmaxSq = Rmax*Rmax;
    RminSq = rc_min*rc_min;
  } else {
    if(Phi.empty())  {
      newid =SpeciesCollection::addSpecies("Ar");
      Phi.push_back(ScaleFunc());
      (*mySpecies)(imass,newid)= 39.948;
      Rmax = Phi[0].rmax();
      RmaxSq = Rmax*Rmax;
      RminSq = 1e-6;
      Bond.resize(1,1);
      Bond(0,0) = 0;
    }
  }
}
  
template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::e(const ParticlePos_t& R, int i, int myID) {

  return 0;

}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::evalEnergy() {

  DEBUGMSG("Entering TwoBodyPotential::evalEnergy.\n")
  //if(Counter == -1)  MakeGridConnection(*myPtcl,Rmax,2);
  if(Counter == -1)  {
    NextCell.connect(myPtcl->Lattice,Rmax,2);
    convert2Cart(myPtcl->Lattice,myPtcl->R);
  }
  Counter++;

  myPtcl->partition(false);

  Scalar_t  esum =0.0e0;

#pragma omp parallel reduction(+:esum)
 {
#ifdef SE_OPENMP
    int myID = omp_get_thread_num();
#else
    int myID = 0;
#endif
    const ParticleLayout_t::PtclGrid_t& bg = *(myPtcl->Lattice.dGrid[2]);
    const ParticlePos_t& Rref = myPtcl->R;
    bool inunit = Rref.InUnit;

    SingleParticlePos_t dr;
    Scalar_t rsq;

    for(int ic=bg.NodeDist[myID]; ic<bg.NodeDist[myID+1]; ic++) {

      int ni=bg.PtclDist[ic];
      int nf=bg.PtclDist[ic+1];

      for(int ig=ni; ig<nf; ig++) {
	int iG = myPtcl->ID[ig];// physical index
	int itype = myPtcl->IonID[iG];

	////////////////////////////////
	// in-cell 
	////////////////////////////////
	for(int jg=ni; jg<nf; jg++) {

	  int jG = myPtcl->ID[jg]; // physical index
	  if(iG == jG) continue; //do not interact with itself

	  //////////////////////
	  //dr = lat.toCart(p.R[jG]-p.R[iG]);
	  //////////////////////
#ifdef INLINE_TINYVECTOR
	  dr[0] = Rref[jG][0]-Rref[iG][0];
	  dr[1] = Rref[jG][1]-Rref[iG][1];
	  dr[2] = Rref[jG][2]-Rref[iG][2];
	  rsq = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2];
#else
	  dr = Rref[jG]-Rref[iG];
	  rsq = dot(dr,dr);
#endif
	  if(rsq > RmaxSq || rsq < RminSq) continue;
	  int ibond = Bond(itype,myPtcl->IonID[jG]);
	  if(ibond>=0) esum += Phi[ibond].f(sqrt(rsq));
	}

	//for(int nc=0; nc<bg.ConnectedNodes[ic].size(); nc++) {
	  //int jc = bg.ConnectedNodes[ic][nc];
	  //SingleParticlePos_t del = bg.Shifts[ic][nc];
	for(int nc = NextCell.first(ic); nc< NextCell.last(ic); nc++) {
	  int jc = NextCell.id(nc);
	  // corrections due to the boundary conditions
	  SingleParticlePos_t del =NextCell.bc(nc);

	  for(int jg=bg.PtclDist[jc]; jg<bg.PtclDist[jc+1]; jg++) {
	    int jG = myPtcl->ID[jg];
#ifdef INLINE_TINYVECTOR
	    dr[0] = Rref[jG][0]-Rref[iG][0]+del[0];
	    dr[1] = Rref[jG][1]-Rref[iG][1]+del[1];
	    dr[2] = Rref[jG][2]-Rref[iG][2]+del[2];
	    rsq = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2];
#else
	    dr = Rref[jG]-Rref[iG]+del;
	    rsq = dot(dr,dr);
#endif
	    if(rsq > RmaxSq || rsq < RminSq) continue;
	    int ibond = Bond(itype,myPtcl->IonID[jG]);
	    if(ibond>=0) esum += Phi[ibond].f(sqrt(rsq));
	  }
	}
      }
    }
  }//end-of-omp

  return esum;
}

template<class ScaleFunc>
Scalar_t 
TwoBodyPotential<ScaleFunc>::evalForce(ParticlePos_t& f0, ParticleScalar_t& q,
				       Tensor_t& stress, unsigned stressmode) {

  if(Counter == -1)  {
    NextCell.connect(myPtcl->Lattice,Rmax,2);
    convert2Cart(myPtcl->Lattice,myPtcl->R);
  }

  //if(Counter == -1)  MakeGridConnection(*myPtcl,Rmax,2);
  Counter++;

  myPtcl->partition(false);

  Scalar_t esum = 0.0e0;
  Scalar_t p00 = 0.0e0, p11=0.0e0, p22 = 0.e0, p01 = 0.0, p02=0.0, p12 = 0.0;

#pragma omp parallel reduction(+:esum,p00,p11,p22,p01,p02,p12)
 {
#ifdef SE_OPENMP
    int myID = omp_get_thread_num();
#else
    int myID = 0;
#endif
    const ParticleLayout_t::PtclGrid_t& bg = *(myPtcl->Lattice.dGrid[2]);
    const ParticlePos_t& Rref = myPtcl->R;

    SingleParticlePos_t dr;
    Scalar_t rsq;
    Scalar_t  esumloc = 0.0, phi;
    Tensor_t stressloc;

    for(int ic=bg.NodeDist[myID]; ic<bg.NodeDist[myID+1]; ic++) {

      int ni=bg.PtclDist[ic];
      int nf=bg.PtclDist[ic+1];

      for(int ig=ni; ig<nf; ig++) {
	int iG = myPtcl->ID[ig];// physical index
	int itype = myPtcl->IonID[iG];
	Scalar_t eat = 0.0;

	for(int jg=ni; jg<nf; jg++) {//in-cell

	  int jG = myPtcl->ID[jg]; // physical index
	  if(iG == jG) continue; //do not interact with itself

#ifdef INLINE_TINYVECTOR
	  dr[0] = Rref[jG][0]-Rref[iG][0];
	  dr[1] = Rref[jG][1]-Rref[iG][1];
	  dr[2] = Rref[jG][2]-Rref[iG][2];
	  rsq = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2];
#else
	  dr = Rref[jG]-Rref[iG];
	  rsq = dot(dr,dr);
#endif
	  if(rsq > RmaxSq || rsq < RminSq) continue;
	  int ibond = Bond(itype,myPtcl->IonID[jG]);
	  if(ibond<0) continue;

	  Scalar_t r = sqrt(rsq);
          Scalar_t dphi = Phi[ibond].df(r,phi)/r;
          eat += phi;
          SingleParticlePos_t df = dphi*dr;
          f0[iG] += 2*df; // openmp optimal, does not matter anyway
          if(stressmode) 
	    stressloc -= outerProduct(df, dr);
	}// jg, incell

	for(int nc = NextCell.first(ic); nc< NextCell.last(ic); nc++) {
	  int jc = NextCell.id(nc);
	  // corrections due to the boundary conditions
	  SingleParticlePos_t del =NextCell.bc(nc);

	  //for(int nc=0; nc<bg.ConnectedNodes[ic].size(); nc++) {
	  //int jc = bg.ConnectedNodes[ic][nc];
	  //// corrections due to the boundary conditions
	  //SingleParticlePos_t del = bg.Shifts[ic][nc];

	  for(int jg=bg.PtclDist[jc]; jg<bg.PtclDist[jc+1]; jg++) {
	    int jG = myPtcl->ID[jg];
#ifdef INLINE_TINYVECTOR
	    dr[0] = Rref[jG][0]-Rref[iG][0]+del[0];
	    dr[1] = Rref[jG][1]-Rref[iG][1]+del[1];
	    dr[2] = Rref[jG][2]-Rref[iG][2]+del[2];
	    rsq = dr[0]*dr[0] + dr[1]*dr[1] + dr[2]*dr[2];
#else
	    dr = Rref[jG]-Rref[iG]+del;
	    rsq = dot(dr,dr);
#endif
	    if(rsq > RmaxSq || rsq < RminSq) continue;
	    int ibond = Bond(itype,myPtcl->IonID[jG]);
	    if(ibond<0) continue;
	    Scalar_t r = sqrt(rsq);
	    Scalar_t dphi = Phi[ibond].df(r,phi)/r;
	    eat += phi;
	    SingleParticlePos_t df = dphi*dr;
	    f0[iG] += 2*df; // openmp optimal, does not matter anyway
	    if(stressmode) 
	      stressloc -= outerProduct(df, dr);

	  }// jg of neighboring cells
	}//neighboring cells
	q[iG] = eat;
	esumloc += eat;
      }//ig
    }//basegrids of a thread

    esum += esumloc;
    if(stressmode) {
      p00 += stressloc(0,0);
      p11 += stressloc(1,1);
      p22 += stressloc(2,2);
      p01 += stressloc(0,1);
      p02 += stressloc(0,2);
      p12 += stressloc(1,2);
    }
  }//end-of-omp

  if(stressmode)
    stress += Tensor_t(p00,p01,p02,-p01,p11,p12,-p02,-p12,p22);
  return esum;
}

template<class ScaleFunc>
bool 
TwoBodyPotential<ScaleFunc>::evalDynamicMatrix(DynMatrix_t&) {

  return false;

}
}
#endif


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: TwoBodyPotentialG.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
