//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/EPNHydroVerlet.h"
//#include "Propagators/EVNVerlet.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

EPNHydroVerlet::EPNHydroVerlet()
  : T_ref(0.0), P_ref(0.0), Dt(1.0), Friction(0.0), VolMass(1.0)
{
  pressure_ind = E.add("Pressure");
  pv_ind = E.add("P*dV");

  m_param.add(T_ref, "temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
  m_param.add(P_ref_kB,"pressure","kB");
  m_param.add(VolMass,"volmass","au");
}

void 
EPNHydroVerlet::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;

  Scalar_t tmp = Friction*Dt*0.5e0;
  c0 = 2.0e0/(1.0e0+tmp);
  c1 = 1.0e0 - c0;
  c2 = Dt*Dt/(1.0e0+tmp);
  oneover2dt = 0.5e0/Dt;
  P_ref = P_ref_kB/pressure_to_kbar;

  if(!Initialized) {

    LOGMSG("External Pressure " << P_ref_kB << " kB " << P_ref << " eV??? ")
    initializeBase(ptclin,potin,makecopy);

    // Calculate the volumes
    Vol0 = myPtcl->Lattice.Volume;
    Vol1 = Vol0;
    Vol2 = Vol0;

    ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
    ParticlePos_t& R1 = *myPtcl->getVectorAttrib("r1");

    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addPositionAttrib(R1);

    R1.setUnit(PosUnit::CartesianUnit);

    Initialized = true;
  }

  /*! set the velocity to a reference temperature only when the clock is zero    
   * When the clock E[NOW] > 0, it assumes that the simulation restarts from
   * the previous configuration.
   */
  if(ParameterModified)
    setTemperature(RescaleTemperature,T_ref);
  else
    applyBConds(true);
}

void 
EPNHydroVerlet::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticlePos_t& R1 = *myPtcl->getVectorAttrib("r1");

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
    R1 = myPtcl->R;
    myPtcl->R += Dt*V;
  } else {
    E[KINETIC_ENERGY] = 0.0;
    V = 0.0;
    R1 = myPtcl->R;
  }

  ScaledV = 1.0;
  applyBConds(true);
}


bool 
EPNHydroVerlet::propagate(Interaction_t* pot) {

  if(!Initialized) initialize(myPtcl,pot,false);

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  ParticlePos_t& dR = *myPtcl->getVectorAttrib(displacement_tag);
  ParticlePos_t& R1 = *myPtcl->getVectorAttrib("r1");

  ParticlePos_t& R0 = myPtcl->R;

  Tensor_t stress;
  myPot->setCurrentPos(myPtcl->R,false);
  E[POT_ENERGY] = pot->evalForce(F,1);
  evalAccel(F);
  const Scalar_t onethird = 1.0/3.0;
  E[pressure_ind] = (trace(stress) + 2.0e0*E[KINETIC_ENERGY])*onethird/Vol0;
  VolF = E[pressure_ind] -  P_ref;

  //free energy
  E[FREE_ENERGY] = E[POT_ENERGY] - Vol0*P_ref;
  Vol2 = Vol1;
  Vol1 = Vol0;
  Vol0 = c0*Vol1 + c1*Vol2 + c2*VolF/VolMass;
  E[pv_ind] = (E[pressure_ind]-P_ref)*Vol0;
  DEBUGMSG("stress = " << stress << " (P-Pext)*V " << E[pv_ind] << "VolF =" << VolF)

  Scalar_t volinv = 1.0e0/Vol1;
  Scalar_t dvol = (Vol0 - Vol2)*volinv;
  Scalar_t aux1 = c0 + onethird*(Dt*Dt*VolF*volinv - 0.5*onethird*dvol*dvol)/VolMass;

  //dR store R1 := R(t-dt)
  dR = R1;
  R1 = R0;
  R0 = aux1*R1+c1*dR + c2*F;

  Scalar_t aux2 = -onethird*dvol*oneover2dt;
  V = oneover2dt*(R0 - dR) + aux2*R1;

  //myPtcl->Lattice *= pow(Vol0/Vol1, onethird); // scale the lattice
  Scalar_t vscale =pow(Vol0/Vol1, onethird);
  Tensor_t newlat = vscale*myPtcl->Lattice.R;
  myPtcl->updateLayout(newlat);

  evalKineticEnergy();

  //Need to add the kinetic energy of Volume
  Scalar_t dvdt = (Vol0-Vol2)*oneover2dt;
  Scalar_t ke_vol = 0.5*VolMass*dvdt*dvdt;
  E[FREE_ENERGY] += E[KINETIC_ENERGY]+ke_vol; // add the kinetic energy to the free energy

  myPot->setCurrentPos(R0,false);

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  ///pressure criterion
  //return(E[TEMPERATURE] > zeroKelvin || fabs(E[pv_ind])>zeroKelvin);
  return ((E[TEMPERATURE]+ke_vol*eV_to_Kelvin)>KE_min_K);
}


void 
EPNHydroVerlet::applyBConds(bool wrapping) {
  myPot->setCurrentPos(myPtcl->R,wrapping);
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: EPNHydroVerlet.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
