//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/EVNVerlet.h"
#include "ParticleBase/ParticleUtility.h"
using namespace OHMMS;

EVNVerlet::EVNVerlet(): 
Rid(0),T_ref(0.0), Dt(1.0), Friction(0.0)
{ 
  /*! add EVNVerlet specific <parameter/>  */
  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
}


void 
EVNVerlet::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  //friction set to zero for EVN simulation. sdmin should be used instead.
  if(myName == "evn") Friction = 0.0;

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  Scalar_t tmp = Friction*Dt*0.5e0;
  c0 = 2.0e0/(1.0e0+tmp);
  c1 = 1.0e0 - c0;
  c2 = Dt*Dt/(1.0e0+tmp);
  oneover2dt = 0.5e0/Dt;

  initializeBase(ptclin,potin,makecopy);

  if(!Initialized) { 

    ParticlePos_t& V = myPtcl->V;

    //add an attribute for previous position and save the index to access the attribute
    Rid =myPtcl->addAttribute(postype_tag, "r1");
    ParticlePos_t& R1 = *myPtcl->getVectorAttrib(Rid);

    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addPositionAttrib(R1);

    ///////////////////////////////////
    // these values are always in cartesian coordinates
    ///////////////////////////////////
    myPtcl->convert2Cart(myPtcl->R);
    R1.setUnit(PosUnit::CartesianUnit);

    Initialized = true;
  }

  if(ParameterModified) 
    setTemperature(RescaleTemperature,T_ref);
  else 
    applyBConds(true);

  RescaleTemperature = false;
}

void 
EVNVerlet::setTemperature(bool rescale, Scalar_t tref) {

  ParticlePos_t& V = myPtcl->V;
  ParticlePos_t& R1 = *myPtcl->getVectorAttrib(Rid);

  //if tref>0, overwrites the target temperature
  if(tref>zeroKelvin) { T_ref = tref;}

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
    R1 = myPtcl->R;
    myPtcl->R = R1 + Dt*V;
  }else {
    V = 0.0;
    R1 = myPtcl->R;
  }

  ScaledV = 1.0;

  applyBConds(true);

} 

bool 
EVNVerlet::propagate(Interaction_t* pot) {

  if(Counter == UpdateStride) {
    Counter = 0;
    myPot->setCurrentPos(myPtcl->R);
    myPtcl->dR=0.0;
  }

  size_type first=0;
  size_type last=myPtcl->getLocalNum();
  const ParticleIndex_t& gid = myPtcl->GroupID;

  //myPot->setCurrentPos(myPtcl->R, myPtcl->dR);
  
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F, 2);

  ParticlePos_t& F = myPtcl->F;
  ParticlePos_t& V = myPtcl->V;
  ParticlePos_t& dR = myPtcl->dR;
  ParticlePos_t& R1 = *(myPtcl->getVectorAttrib(Rid));

  Scalar_t ke=0.0;
  for(size_type iat=0; iat<last; iat++) {
    Index_t itype(gid[iat]);
    SingleParticlePos_t r2(R1[iat]);
    R1[iat]=myPtcl->R[iat];
    myPtcl->R[iat]=c0*R1[iat]+c1*r2+c2*minv[itype]*F[iat];
    dR[iat]=myPtcl->R[iat]-R1[iat];
    SingleParticlePos_t v(oneover2dt*(myPtcl->R[iat]-r2));
    V[iat]=v;
    ke += mass[itype]*dot(v,v);
  }
  //dR = R1;
  //R1 = myPtcl->R;
  //myPtcl->R = c0*R1 + c1*dR + c2*F;
  //V = oneover2dt*(myPtcl->R - dR);
  //dR.setUnit(PosUnit::CartesianUnit);
  //dR = myPtcl->R-R1;

  //update the current position for the interaction with R and dR
  //myPot->setCurrentPos(myPtcl->R,dR);

  //evalKineticEnergy();
  E[KINETIC_ENERGY]= 0.5*myPtcl->collect(ke);
  E[TEMPERATURE]   = E[KINETIC_ENERGY]*TwoOverDOF*eV_to_Kelvin;

  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY];

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  return (E[TEMPERATURE]>KE_min_K);
}

void EVNVerlet::applyBConds(bool wrapping) 
{
  myPot->setCurrentPos(myPtcl->R, wrapping);
  Counter = 0;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 375 $   $Date: 2005-07-31 14:19:41 -0500 (Sun, 31 Jul 2005) $
 * $Id: EVNVerlet.cpp 375 2005-07-31 19:19:41Z jnkim $ 
 ***************************************************************************/
