//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim and Nichols A. Romero
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// Implemented by Jeongnim Kim, May 2002
// Modifided by Jeongnim Kim and Nichols A. Romero, May/30 2002
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <complex>
using namespace std;

#include "Propagators/ForceConstant.h"
#include "OhmmsNumeric/Matrix.h"
#if defined(HAVE_LIBHDF5)
#include "ParticleIO/HDFParticleAttrib.h"
#endif
#include "ParticleIO/ParticleParser.h"
#include "ParticleIO/ParticleLayoutIO.h"
using namespace OHMMS;
#include "CrystalSymmetry/KPointBase.h"
#include "Utilities/GaussianDOS.h"
#include "Utilities/OhmmsInfo.h"

//extern KPointBase* createKPointMesh(Particle_t&, istream&);

void ForceConstant::printFreq(const vector<Scalar_t>& Lambda, int writenode) {
  if(!writenode && PrintOut[PhononFrequency]) {

    //string fcname=myPtcl->getName();
    string fcname(myPtcl->getName());
    fcname.append(".pfreq");
    ofstream fcout(fcname.c_str());
    fcout.setf(ios::scientific, ios::floatfield);
    fcout.setf(ios::right,      ios::adjustfield);
    int nat = Lambda.size()/OHMMS_DIM;  

    /////////////////////////////////////////////////////
    // IMPORTANT: internal mass is in eV, A, fsec unit
    // Phonon frequencies will be converted into 1/cm
    // amu = amu_old/1e2, where amu_old = 1.036427207361627e-28
    // because of the mass conversion taken earlier.
    // old implementation used:
    // Scalar_t amu   = 1.036427207361627e-28;// [eV/(Ang/s)^2]
    // Scalar_t masst =12; // carbon
    ////////////////////////////////////////////////////
    const Scalar_t c_cm  = 2.99792458e10; // [cm/s]
    const Scalar_t amu   = 1.0e-30;// [eV/(Ang/s)^2]
    const Scalar_t pi    = 3.14159265358979; // pi
    complex<Scalar_t> comega;
    vector<Scalar_t> freqs(Lambda.size());
    //////////////////////////////////////////////////////////
    //IMPORTANT: units are converted here.
    // Format = mode Lambda(eV/A^2)  omega (1/cm)
    //////////////////////////////////////////////////////////
    int ii=0;// eigen mode index
    fcout << "#    mode        eV/A^2       1/cm (real)   thz" << endl;
    for(int i=0; i<nat; i++) {
      Scalar_t massfac = 1.0/sqrt(mass[ myPtcl->GroupID[i] ]*amu)/(2*pi)/c_cm; 
      Scalar_t thz = massfac*0.02998;
      for(int ik=0; ik<OHMMS_DIM; ik++) {
	//comega = sqrt(complex<Scalar_t>(DMSolver.Lambda[ii]-eig0,0.0));          
	comega = sqrt(complex<Scalar_t>(Lambda[ii]));          
	fcout << setw(10) << ii 
	      << setw(15) << setprecision(5) << Lambda[ii]
	      << setw(15) << setprecision(5) 
	      << comega.real()*massfac
	      << setw(15) << setprecision(5) 
	      << comega.real()*thz << endl;
	freqs[ii] = comega.real()*thz;
	ii++;
      }
    }

    fcout.close();
    if(GetDOS) {
      Scalar_t w = 1.0/static_cast<Scalar_t>(nat);
      Scalar_t max_freq = freqs.back();    
      fcname= myPtcl->getName(); 
      fcname.append(".vdos");
      fcout.open(fcname.c_str());
      fcout << "#Vibrational DOS: THz   DOS " << endl;
      Scalar_t gwidth = 0.1;
      max_freq += 5*gwidth;
      int nbin = static_cast<int>(max_freq/gwidth)*4;
      GaussianDOS<Scalar_t> vdos(0.0,max_freq,nbin);
      for(int i=0; i<freqs.size(); i++) {
	vdos.accumulate(freqs[i],gwidth,w);
      }
      vdos.print(fcout);
    }
  }
}
#if defined(HAVE_LIBHDF5)
/*!\void ForceConstant::printForceConstant
 * \brief print out forces for each displacement for other phonon codes
 */
void 
ForceConstant::printForceConstant(const ParticlePos_t& F, 
				  int iat, int j, bool plus) {
  if(PrintOut[FC]) {
    static int i=-1;//used to open and close
    static hid_t hdfFile;
    static string fieldname;
    string fcname=myPtcl->getName();
    fcname.append(".fc.h5");
    if(i<0) {
      hdfFile = H5Fcreate(fcname.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
      i=iat;
    }
    else {
      hdfFile = H5Fopen(fcname.c_str(), H5F_ACC_RDWR, H5P_DEFAULT);
//       hid_t dataset = H5Dopen(hdfFile,fieldname.c_str());
//       cout << "found a dataset " << dataset << endl;
//       H5Dclose(dataset);
    }

    char modename[128];
    if(plus)  
      sprintf(modename,"/p%dd%d_add",iat,j);
    else
      sprintf(modename,"/p%dd%d_sub",iat,j);
    fieldname = modename;
    HDFAttribIO<ParticlePos_t> hdf_p(const_cast<ParticlePos_t&>(F));
    hdf_p.write(hdfFile,modename);
    H5Fclose(hdfFile);
  }
}

/*!\void ForceConstant::printMode
 * \brief print out the phonon mode and frequencies in 1/cm
 */
void 
ForceConstant::printMode(const RealMatrix_t& FC, int first, int last, int writenode) {

  if(!writenode == 0 && PrintOut[PhononMode]) {
    string fcname=myPtcl->getName();
    fcname.append(".pmode.h5");
    int nat = FC.rows()/OHMMS_DIM;  
    string pname("EigVector");
    hid_t   hdfFile = H5Fcreate(fcname.c_str(), H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);
    hsize_t dims[3];
    dims[0] = FC.rows();
    dims[1] = FC.rows()/OHMMS_DIM;
    dims[2] = OHMMS_DIM;
    hid_t dataspace  = H5Screate_simple(3, dims, NULL);
    hid_t dataset =  
      H5Dcreate(hdfFile,pname.c_str(),H5T_NATIVE_DOUBLE, dataspace, H5P_DEFAULT);
    hid_t ret = 
      H5Dwrite(dataset, H5T_NATIVE_DOUBLE, H5S_ALL, H5S_ALL, H5P_DEFAULT, 
	       FC.data());
    H5Sclose(dataspace);
    H5Dclose(dataset);
    H5Fclose(hdfFile);
  }
}
#else
/*!\void ForceConstant::printForceConstant
 * \brief print out forces for each displacement for other phonon codes
 */
void 
ForceConstant::printForceConstant(const ParticlePos_t& F, int iat, int j, bool plus) {
  if(PrintOut[FC]) {
    static int i=-1;//used to open and close
    char fcname[128];
    sprintf(fcname,"%s.fc",myPtcl->getName());
    ofstream fcout;
    if(i<0) {
      fcout.open(fcname);
      i=0;
    } else {
      fcout.open(fcname,ios::app);
    }
    fcout.setf(ios::scientific, ios::floatfield);
    fcout.setf(ios::right,      ios::adjustfield);
    fcout << "Displacement " << iat << " " << j << " ";
    if(plus) fcout << Displacement << endl;
    else     fcout << -Displacement << endl;
    fcout << F;
  }
}

/*!\void ForceConstant::printMode
 * \brief print out the phonon mode and frequencies in 1/cm
 */
void 
ForceConstant::printMode(const RealMatrix_t& FC, int first, int last, int writenode) {

  if(PrintOut[PhononMode]) {
    char fcname[128];
    sprintf(fcname,"%s.pmode",myPtcl->getName());
    ofstream fcout(fcname);
    fcout.setf(ios::scientific, ios::floatfield);
    fcout.setf(ios::right,      ios::adjustfield);
    int nat = FC.rows()/OHMMS_DIM;  
    for(int i=first; i<last; i++) { //for(int i=0; i<FC.rows(); i++) {
      int jj=0;
      for(int j=0; j<nat; j++) {
	for(int jk=0; jk<OHMMS_DIM; jk++) {
	  fcout << setw(15) << setprecision(5) << FC(i,jj++);
	}   
	fcout << endl;
      }
    }
  }
}
#endif

#if defined(HAVE_LIBXML2)

bool ForceConstant::put(xmlNodePtr cur){

  xmlDocPtr doc = cur->doc;
  xmlNsPtr ns = cur->ns; 

  //WriteForceConstant = 0;
  myName = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
  cur = cur->xmlChildrenNode;
  PrintOut.set(PhononFrequency);

  xmlXPathContextPtr context;
  xmlXPathObjectPtr result = NULL;

  while(cur != NULL) {
    string cname((const char*)(cur->name));
    if (cname == "parameter" || cname == "Parameter") {
      string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
      if(aname == "displacement") {
        putContent(Displacement,cur);
      }
    }  else if(cname == "record") { 
      string aname = (const char*)(xmlGetProp(cur, (const xmlChar *) "name"));
      if(aname == "forceconstant") PrintOut.set(FC);
      else if(aname == "eigenmode") PrintOut.set(PhononMode);
      else if(aname == "!frequency") PrintOut.flip(PhononFrequency);
    } else if(cname == "primarycell") {
      xmlNodePtr cur_pc;
      if(xmlHasProp(cur,(const xmlChar*)"ref")) {
	const xmlChar* pname = xmlGetProp(cur,(const xmlChar*)"ref");
	context = xmlXPathNewContext(doc);
	result = xmlXPathEvalExpression(pname,context);
	if(!xmlXPathNodeSetIsEmpty(result->nodesetval)) {
	  cur_pc = result->nodesetval->nodeTab[0]->xmlChildrenNode;
	}
      } else {
	cur_pc = cur->xmlChildrenNode;
      }
      while(cur_pc != NULL) {
        string cname1((const char*)(cur_pc->name));
	if (cname1 == "unitcell") {
	  LatticeParser lat(PrimaryCell.Lattice);
	  lat.put(cur_pc);
	} else if(cname1 == "particleset") {
	  ParticleParser aHandle(PrimaryCell);
	  aHandle.put(cur_pc);
	}
	cur_pc = cur_pc->next;
      }
    } else if(cname == "kpoint") {
      KPointBase* k=PrimaryCell.getKSet();
      if(k == 0) {
        k = new KPointBase;
        k->ref_ = &PrimaryCell;
        PrimaryCell.setKSet(k);
        k->put(cur);
      }
    }
    cur = cur->next;
  }

  if(result) xmlXPathFreeObject(result);

  if(PrimaryCell.getTotalNum()) {
    KPointBase* k=PrimaryCell.getKSet();
    if(k == 0) {
      WARNMSG("Primary cell is set but no q-point is provided. Adding Gamma point")
      k = new KPointBase;
      k->addKpt(SingleParticlePos_t(0,0,0),1);
      k->ref_ = &PrimaryCell;
      PrimaryCell.setKSet(k);
    } 
    //k->print(cout);
  }


  return true;
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 376 $   $Date: 2005-08-28 15:32:27 -0500 (Sun, 28 Aug 2005) $
 * $Id: ForceConstant.IO.cpp 376 2005-08-28 20:32:27Z jnkim $ 
 ***************************************************************************/
