//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include <stdio.h>
#include "Propagators/PropagatorBase.h"
#include "ParticleBase/ParticleUtility.h"
#include "Propagators/ConstraintBase.h"
#include "Message/OpenMP.h"
#include <limits>
using namespace OHMMS;


//!< Constructor.
PropagatorBase::PropagatorBase(): 
  Counter(-1), 
  UpdateStride(20),
  myPtcl(0), 
  myPot(0),
  myConstraint(0),
  ParameterModified(true), 
  RescaleTemperature(false)
{
  KE_min_K = zeroKelvin;

  TwoOverDOF = -1.0e0;

  ///initialize E array
  E.resize(5);
  E.Name[NOW] = "Clock";
  E.Name[FREE_ENERGY] = "Etot"; 
  E.Name[POT_ENERGY] = "Epot"; 
  E.Name[KINETIC_ENERGY] = "Ekin"; 
  E.Name[TEMPERATURE] = "T"; 
  E[NOW] = 0.0;
  E[FREE_ENERGY] = 0.0;
  E[POT_ENERGY] = 0.0;
  E[KINETIC_ENERGY] = 0.0;
  E[TEMPERATURE] = 0.0;

  MultiSpecies = false;
  Initialized = false;
  OwnParticle = false;
  myConstraint = new ConstraintBase;

  ///add parameters
  m_param.add(E[NOW],"mdclock","fsec");
  m_param.add(KE_min_K,"minT","K");
  m_param.add(RescaleTemperature,"rescale-T","bool");
  m_param.add(UpdateStride,"update","int");

#ifdef HAVE_LIBBOOST
  int np = omp_get_max_threads();
  reng.resize(np);
  unsigned int baseseed = static_cast<unsigned int>(std::time(0));
#pragma omp parallel for 
  for(int ip=0; ip<np; ip++) {
    reng[ip]= new BoostRandom<Scalar_t>();
    reng[ip]->init(baseseed+ip);
  }
#endif
}

/// destructor: remove a particle set created by this
PropagatorBase::~PropagatorBase() {
  DEBUGMSG("Calling PropagatorBase::~PropagatorBase()");
  if(OwnParticle && myPtcl)  delete myPtcl;
  if(myConstraint) delete myConstraint;
}

void PropagatorBase::resetBase() {

  SpeciesSet& species = myPtcl->getSpeciesSet();
  size_type imass = species.addAttrib("mass");
  size_type nspecies = species.getTotalNum();
  if(nspecies>mass.size()) {
    mass.resize(nspecies);
    minv.resize(nspecies);
    for(size_type i=0; i < nspecies; i++) {
      mass[i] = species(imass,i)*amu_to_mass_unit;
      if(mass[i] > numeric_limits<Scalar_t>::epsilon()) minv[i] = 1.0e0/mass[i];
    }
// #if defined(__ICC) || defined(__ECC)
//     if(nspecies == 1) {
//       mass[0] = species(imass,0)*amu_to_mass_unit;
//       if(mass[0] > small) minv[0] = 1.0e0/mass[0];
//     } else {
// #endif
//       for(int i=0; i < nspecies; i++) {
// 	mass[i] = species(imass,i)*amu_to_mass_unit;
// 	if(mass[i] > small) minv[i] = 1.0e0/mass[i];
//       }
// #if defined(__ICC) || defined(__ECC)
//     }
// #endif
    if(nspecies>1) MultiSpecies = true;
    else           MultiSpecies = false;
  }
}

/** Initialize myPtcl and myPot for a propagator
 *@param ptclin assigned to myPtcl
 *@param potin assigned to myPot
 *@param makecopy if true, make a hard copy of the particleset.
 *@param getforce if true, request the particleset to create force attribute
 *
 *Derived classes from PropagatorBase calls this function while initialization.
 *When ptclin is already assigned to myPtcl, nothing needs to be done.
 */
void PropagatorBase::initializeBase(Particle_t* ptclin, Interaction_t*  potin, 
				    bool makecopy, bool getforce) {
  myPot = potin;

  if(myPtcl != ptclin) { 
    if(makecopy) {
      WARNMSG("\tInitialize the ParticleSet by the input configuration by Copying")
      OwnParticle = true;
      myPtcl = new Particle_t;
      myPtcl->assign(*ptclin); 
    } else {
      if(OwnParticle) {
        ERRORMSG("\tCannot assign the input configuration because the object is owned by the Propagator")
      } else {
        //Probably need to check if I have own anything.
        OwnParticle = false;
        LOGMSG("\tInitialize the ParticleSet by the input configuration")
        myPtcl = ptclin;
      }
    }

    //add Particle Attribute for the force
    if(getforce) {
      myPtcl->addDynamicAttributes(TDEnsemble[STATIC_BIT]);
    } else {
      LOGMSG("\t"<<myName << " propagator does not need force")
    }

    // step 2: initialization of constraint: this is questionable
    myConstraint->set(myPtcl);

    if(TwoOverDOF < 0.0e0) {
      DOF = myConstraint->dof();
      TwoOverDOF = 2.0/DOF;
      LOGMSG("\tThe number of particles " << myPtcl->getTotalNum() 
          << " Degrees of Freedom " << DOF)
    }

    resetBase();
  }

  LocIndex0.resize(omp_get_max_threads()+1);
  FairDivide(myPtcl->getLocalNum(),omp_get_max_threads(),LocIndex0);
}

PropagatorBase::Scalar_t 
PropagatorBase::evalPotEnergy(Interaction_t* pot) 
{
  if(pot) {
    myPtcl->advance();
    E[FREE_ENERGY]= E[POT_ENERGY] = pot->evalEnergy();
  } 
  return E[POT_ENERGY]; 
}

void PropagatorBase::scaleTemperature(bool rescale, Scalar_t tref){

  ScaledV = 1.0;
  RescaleTemperature = false;

  resetBase();
  if(!(myPtcl->hasAttrib(velocity_tag))) {
    WARNMSG("void Propagator::setTemperature() cannot proceed because V is not created.")
    return; // Velocity is not created. Do nothing
  }

  ParticlePos_t& Vref = *myPtcl->getVectorAttrib(velocity_tag);

  if(tref < zeroKelvin) 
    {
      //WARNMSG(tref << " is smaller than " << zeroKelvin << ". Assign zero velocities")
      Vref = 0; // set to Zero temperature
      return;
    }

  if(!Initialized) { // Called before propagate
    DEBUGMSG("V has never been assigned. Assign random vectors for " << tref << " K.")
    setVelocity(tref);
    ScaledV = -1.0;
    return;
  }

  // The simulation has been going on. 
  if(rescale) {
    DEBUGMSG("V has been assigned. Rescale the velocities to " << tref << " K.")
    evalKineticEnergy();
    ScaledV = sqrt(tref/E[TEMPERATURE]);
    Vref *= ScaledV;
  } else{
    DEBUGMSG("new random vectors are assigned to V at " << tref << " K.")
    setVelocity(tref);
    ScaledV = -1.0;
  }
}

bool PropagatorBase::get(ostream& os) const {
  os << "<ensemble name=\"" << myName << "/>"<< endl;
  m_param.get(os);
  os << "</ensemble>" << endl;
  return true;
}

/** Process the <ensemble name="aname" engine="class-name" method="defunct"/>
 *@param cur the xml node to work on whose name is ensemble
 *@return true, if any parameter is given.
 *
 * This is the main method to define the parameters for a propagator.
 */
bool PropagatorBase::put(xmlNodePtr cur) {

  const xmlChar* aname=xmlGetProp(cur, (const xmlChar *) "name");
  if(aname)  myName = (const char*)aname;

  xmlNodePtr parent = cur;
  bool something = setParameters(cur);

  ParameterSet::iterator jt(m_param.begin());
  ParameterSet::iterator jt_end(m_param.end());
  while(jt != jt_end) {
    (*jt).second->add(parent);++jt;
  }

  if(E[NOW]>1e-6 && !RescaleTemperature) {
    ParameterModified = false;
  }

  return ParameterModified;
}

/** Process external parameters to control a Propagator
 * @param cur the xml node to work on
 * @return true, if the node has any valid parameter registered by a Propagator
 *
 * The return value can be used to signal that any external parameter has been
 * modified, so that it call for a proper initialization.
 */
bool PropagatorBase::setParameters(xmlNodePtr cur) {

  bool something = m_param.put(cur);
  ParameterModified = something;
  return ParameterModified;
}

void PropagatorBase::reset() {
  map<string,OhmmsElementBase*>::iterator it(m_param.begin());
  map<string,OhmmsElementBase*>::iterator it_end(m_param.end());
  while(it != it_end) {
    (*it).second->reset();++it;
  }
}

bool PropagatorBase::put(std::istream& ) {
  return true;
}

/*@param old an existing propagator
 *
 *Experimental. Set the clock with the old value and clear attribute lists 
 *of the particle set.
 */
void PropagatorBase::copy(PropagatorBase* old) {
  E[NOW] = old->E[NOW];
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 364 $   $Date: 2005-07-13 13:33:56 -0500 (Wed, 13 Jul 2005) $
 * $Id: PropagatorBase.cpp 364 2005-07-13 18:33:56Z jnkim $ 
 ***************************************************************************/
