//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifndef OHMMS_PARTICLEPROPAGATORBASE_H
#define OHMMS_PARTICLEPROPAGATORBASE_H

#include "OhmmsUnits.h"
#include "Particle/ParticleSet.h"
//#include "Potentials/PotentialCollection.h"
#include "Potentials/PotentialBase.h"
#include "OhmmsData/RecordProperty.h"
#include "OhmmsData/ParameterSet.h"
#include "Propagators/RecordPropagator.h"
#ifdef HAVE_LIBBOOST
#include "Utilities/BoostRandom.h"
#endif
#include <bitset>

namespace OHMMS {

  // forward declaration
  class ConstraintBase;

  /*! \class PropagatorBase
   *  \author Jeongnim Kim
   *  \brief Abstract class from which propagators are inherited.
   */
  class PropagatorBase: 
    public ParticleTags,
    public OhmmsElementBase { 

    ///RecordPropagator
    friend class RecordPropagator;

    ///SavePropagator
    friend class SavePropagator;

  public: 

    /** enumeration to define a thermodyanmical ensemble */
    enum {STATIC_BIT, ET_BIT, VP_BIT, NU_BIT};
    enum {NOW,FREE_ENERGY,POT_ENERGY,KINETIC_ENERGY,TEMPERATURE};

    typedef ParticleSet                     Particle_t;
    typedef Particle_t::size_type           size_type;
    typedef Particle_t::Index_t             Index_t;
    typedef Particle_t::Scalar_t            Scalar_t;
    typedef Particle_t::SingleParticlePos_t SingleParticlePos_t;
    typedef Particle_t::Tensor_t            Tensor_t;
    typedef Particle_t::ParticleIndex_t     ParticleIndex_t;
    typedef Particle_t::ParticleScalar_t    ParticleScalar_t;
    typedef Particle_t::ParticlePos_t       ParticlePos_t;

    typedef Particle_t::ParticleLayout_t    ParticleLayout_t;
    typedef PotentialBase             Interaction_t;

    /// constructor
    PropagatorBase(); 

    /// destructor
    virtual ~PropagatorBase();  

    ///implement virtual functions of OhmmsElementBase
    bool get(std::ostream& ) const;
    bool put(std::istream& );
    virtual bool put(xmlNodePtr cur);

    /** Prepare for a new simulation
     *
     * Calling Propagator::reset indicates that propagations following reset
     * belong to a new simulation. Output engine will open new files. 
     * If the SeqID has not been incremented, 
     * the existing output files will be overwritten.
     * Default is to evaluate parameters used for time-stepping and  set Counter = -1
     */
    void reset();

    /** get the value of a property
     *@param i the index of the property
     *
     *Each propagator has a number of properties, such as energy, clock.
     */
    inline Scalar_t getProperty(size_type i) const {
      return E[i];
    }

    template<class T1>
    inline void setParameter(const std::string& aname, T1 x) {
      m_param.setValue(aname,x);
    }

    bool setParameters(xmlNodePtr cur);

    /** initialize the object with a Particle_t*
     *
     *@param ptclin the ParticleSet whose trajectories will be updated by the propagator.
     *@param potin the PotentialCollection* for force/energy evaluations.
     *@param makecopy if true, make a hard copy of the ParticleSet
     *
     *ptclin is assigned to myPtcl. Normally, there will be only one instance of
     *ParticleSet during a simulation but sometimes, it is better to keep copies
     *to mix different propagators.
     */
    virtual void initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy = false) =0;

    /** Set the temperature and coordinates 
     * \param rescale if true, the temperature is to be rescaled
     * \param tref target temperature to which the velocity will be assigned
     *
     * Each derived class has to reimplemet setTemperature. PropagatorBase class scaleTemperature
     * provides a generic method to assign random velocities with Boltzman factor.
     */
    virtual void setTemperature(bool rescale, Scalar_t tref) = 0; 

    /** Adjust trajectory-related attributes for a given boundary condition.
     *@param wrapping  bool, if true, apply boundary condition
     *
     *When wrapping is ture, the layout class re-distribute particles based on their positions
     *All the ParticleAttrib's affected by boundary conditions should be modified.
     */
    virtual void applyBConds(bool wrapping) = 0;

    /** Update particle trajectories according to the integration algorithm.
     *@param pot A pointer to a Potential.
     *@return true if the Kinetic energy is smaller than a critical value.
     *
     *A derived class implements a numerical integration scheme to move particles.
     *The stopping criteria depends on the propagator.
     */
    virtual bool propagate(Interaction_t*) = 0; 


    /// return the particle owned by this
    const Particle_t* getPtcl() const { return myPtcl;}  

    /// return the particle owned by this object
    Particle_t* getPtcl() { return myPtcl;}  

    /** Evaluate an energy using an external potential 
     *@param apot A PotentialCollection*
     *@return the potential energy
     */
    Scalar_t evalPotEnergy(Interaction_t*);

    /// calculate the kinetic energy: using V
    void evalKineticEnergy(); 

    /// copy an old propagator to a new one
    void copy(PropagatorBase* old); 

  protected:

    ///bits to define thermodynamical ensemble
    std::bitset<4> TDEnsemble;

    /// True, if a simulation cell contains more than one species.
    bool MultiSpecies;

    /// False, until propagate is called
    bool Initialized;

    /** True, if any external parameter m_param is modified
     *
     * put(xmlNodePtr cur) is the only valid mean to modify the external
     * parameter, such as reference temperature and pressure. When
     * the parameter is reset, than the flag will be true and subclasses
     * will take actions according to their algorithms.
     */
    bool ParameterModified;

    /// True, if the particle is created by this Propagator
    bool OwnParticle;

    /// Internal counter. To be synchronized with SimpleParticle::Counter
    int Counter;

    /// Stride to call expensive update of the Particle Layout
    int UpdateStride;

    /// Boolean to force rescaling
    bool RescaleTemperature;

    /// degrees of freedom
    Scalar_t TwoOverDOF, DOF; 

    /// kinetic energy tolerance, default is zeroKelvin
    Scalar_t KE_min_K;

    /** scaling factor of the velocity
     *
     *ScaledV is reset by scaleTemperature. 
     * - ScaledV < 0, new random velocities are assigned.
     * - ScaledV > 0, the velocities are rescaled.
     * - ScaledV == 1, the velocities are unchanged.
     */
    Scalar_t ScaledV;

    /// array of extensive energy
    RecordNamedProperty<Scalar_t> E;

    /// internal copy of mass, species-dependent
    vector<Scalar_t> mass;    

    /// internal copy of massinv
    vector<Scalar_t> minv;   

    /// collection of particles
    Particle_t* myPtcl;  

    /// potential/interaction energy
    Interaction_t* myPot;

    /// constraint object
    ConstraintBase* myConstraint;

    /// a set of external parameters for a propagator
    ParameterSet m_param;

    std::vector<size_type> LocIndex0;

    std::vector<RecordProperty*> MCFRecorder;

#ifdef HAVE_LIBBOOST
    std::vector<BoostRandom<Scalar_t>*> reng;
#endif

    /** initialize the PropagatorBase
     * @param ptclin the ParticleSet to propagate
     * @param potin  the Interaction
     * @param makecopy bool, if true, make a copy of a ParticleSet
     * @param getforce bool, if true, the propagator will need force.
     *
     */
    void initializeBase(Particle_t* ptclin, Interaction_t* potin, 
        bool makecopy = false, bool getforce = true);

    /** scale the random velocities to a target temperature
     *@param rescale, true if the velocity is to be rescaled
     *@param tref, the target temperature
     *
     *Set the scaling factor ScaledV
     */
    void scaleTemperature(bool rescale, Scalar_t tref); 

    /*! \fn void setVelocity(Scalar_t temperature)
     *  \param temperature A reference temperature in Kelvin to which the vies are set.
     *  \return true if the velocity was reset
     *  \brief New random velocies are assigned if temperature > zero.
     */
    bool setVelocity(Scalar_t temperature);

    /*! \fn void evalAccel(ParticlePos_t& f)
     *  \param f A force type in eV/A to be converted to the acceleration unit.
     *  \brief Potentials return forces in eV/A. If necessary, convert the forces 
     *   to accelerations.
     */
    void evalAccel(ParticlePos_t& );

    void resetBase();

    Tensor_t evalKineticPressure(const ParticlePos_t& vref);
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 376 $   $Date: 2005-08-28 15:32:27 -0500 (Sun, 28 Aug 2005) $
 * $Id: PropagatorBase.h 376 2005-08-28 20:32:27Z jnkim $ 
 ***************************************************************************/
