//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TPNHydroLangevin.h"
#include "ParticleBase/RandomSeqGenerator.h"
using namespace OHMMS;

TPNHydroLangevin::TPNHydroLangevin()
{
  initParameters(E,m_param);
  m_param.add(Volume,"volume","AA^3");
  m_param.add(pVolume,"pvolume","eV-fsec");
}


void 
TPNHydroLangevin::initialize(Particle_t* ptclin, 
			     Interaction_t* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = -1;
  initializeBase(ptclin,potin,makecopy);
  Volume=myPtcl->Lattice.Volume;
  VolMass=OHMMS_DIM*myPtcl->getTotalNum()*T_ref*Kelvin_to_eV/VolFreq/VolFreq;
  checkParameters(myName);

  if(!Initialized) {
    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);

    Vr.resize(myPtcl->getLocalNum());
    Initialized = true;
  }

  /*! set the velocity to a reference temperature whenever
   * any of parameters is modified.
   */
  if(ParameterModified) 
    setTemperature(RescaleTemperature,T_ref);
  else {
    evalKineticEnergy();
    myPtcl->updateLayout(myPtcl->Lattice.R);
    applyBConds(true);
    if(AddNoise) setSigma(T_ref,minv);
  }
  ParameterModified=false;
}

void 
TPNHydroLangevin::setTemperature(bool rescale, Scalar_t tref) {


  if(tref>zeroKelvin) T_ref = tref;

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  if(T_ref>zeroKelvin) {
    scaleTemperature(rescale,T_ref);
  } else {
    ERRORMSG("TPN simulation at zero temperature?")
  }
  ScaledV = 1.0;

  pVolume = 0.0;
  applyBConds(true);

  if(AddNoise) setSigma(T_ref,minv);
}


bool 
TPNHydroLangevin::propagate(Interaction_t* pot) {

  if(!Initialized) initialize(myPtcl,pot,false);

  DoverDOF = OHMMS_DIM/DOF;

  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);

  //update Volume(t+dt/2)
  //Scalar_t Volume_h = Volume+dtover2*DV*pVolume*VolMassInv;
  Scalar_t Volume_h = Volume+dtover2W*DV*pVolume;

  //update p_Volume(t+dt/2), momentum of the volume
  //Scalar_t pVolume_h = (1.0-chi_vol)*pVolume
  //  +dtover2*(OHMMS_DIM*Volume_h*(E[pressure_ind]-P_ref)+DoverDOF*P_k);

  Scalar_t pVolume_h = cf_vol*pVolume
    +dtover2*(OHMMS_DIM*Volume_h*(E[pressure_ind]-P_ref)+DoverDOF*P_k);

  if(AddNoise) pVolume_h += SigmaVol*randomSource();

  //update the particle velocities V(t+dt/2)
  //Scalar_t vfac = (1.0-chi_vel)-dtover2*(1.0+DoverDOF)*pVolume_h*VolMassInv;
  Scalar_t vfac = cf_vel-dtover2W*(1.0+DoverDOF)*pVolume_h;
  V = vfac*V+dtover2*F;

  if(AddNoise) {
    makeGaussRandom(Vr);
    for(int i=0; i<myPtcl->getLocalNum(); i++) {
      V[i] += SigmaV[myPtcl->GroupID[i]]*Vr[i];
    }
  }

  //update the particle position R(t+dt)
  //Scalar_t rfac = 1.0+Dt*pVolume_h*VolMassInv;
  Scalar_t rfac = 1.0+dtoverW*pVolume_h;

  myPtcl->R = rfac*myPtcl->R+Dt*V;

  //update the unit cell: Volume_h -> myPtcl->Lattice.Volume
  const double oneoverdim = 1.0/OHMMS_DIM;
  Scalar_t volscale =pow(Volume_h/myPtcl->Lattice.Volume,oneoverdim);
  Tensor_t newlat = volscale*myPtcl->Lattice.R;
  myPtcl->updateLayout(newlat);
  myPot->setCurrentPos(myPtcl->R,false);

  //Evaluate gradients, pressure and dE/dV
  Tensor_t stress;
  //E[POT_ENERGY] = myPot->evalForceAndPressure(F,Eloc,stress,dEdVol);
  E[POT_ENERGY] = myPot->evalForce(F,Eloc,stress,1);
  evalAccel(F);
  P_p = trace(stress);
  DV = OHMMS_DIM*myPtcl->Lattice.Volume;

  //update the velocity V(t+dt) 
  V = vfac*V+dtover2*F;
  if(AddNoise) {
    makeGaussRandom(Vr);
    for(int i=0; i<myPtcl->getLocalNum(); i++) {
      V[i] += SigmaV[myPtcl->GroupID[i]]*Vr[i];
    }
  }

  evalKineticEnergy();
  P_k = 2.0*E[KINETIC_ENERGY];

  //Evalaute the internal pressure inclduing the pressure tensor, dE/dV
  //Using the relations:
  //dEdVol = -0.5*P_p/DV;
  //E[pressure_ind] = (P_p+P_k)/DV-dEdVol;
  E[pressure_ind] = (1.5*P_p+P_k)/DV;

  //update the volume momentum p_Volume(t+dt)
  //pVolume = (1.0-chi_vol)*pVolume_h
  //  +dtover2*(OHMMS_DIM*Volume_h*(E[pressure_ind]-P_ref)+DoverDOF*P_k);
  pVolume = cf_vol*pVolume_h+dtover2*(OHMMS_DIM*Volume_h*(E[pressure_ind]-P_ref)+DoverDOF*P_k);
  if(AddNoise) pVolume += SigmaVol*randomSource();

  //update the volume Volume(t+dt)
  //Volume = Volume_h+dtover2*DV*pVolume*VolMassInv;
  Volume = Volume_h+dtover2W*DV*pVolume;

  volscale =pow(Volume/myPtcl->Lattice.Volume,oneoverdim);
  newlat = volscale*myPtcl->Lattice.R;
  myPtcl->updateLayout(newlat);

  E[pv_ind] = E[pressure_ind]*Volume;
  E[pk_ind] = pVolume*pVolume*VolMassInv*0.5;    
  E[FREE_ENERGY] = E[POT_ENERGY]+E[KINETIC_ENERGY]+E[pk_ind]+P_ref*Volume;
  E[volume_ind] = Volume;

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  ///pressure criterion
  return  ((E[TEMPERATURE]+E[pk_ind]*eV_to_Kelvin) > KE_min_K);
}


void 
TPNHydroLangevin::applyBConds(bool wrapping) {

  myPot->setCurrentPos(myPtcl->R,wrapping);
  ParticlePos_t& F = *myPtcl->getVectorAttrib(force_tag);
  ParticleScalar_t& Eloc = *myPtcl->getScalarAttrib(energy_tag);
  Tensor_t stress;
  //E[POT_ENERGY] = myPot->evalForceAndPressure(F,Eloc,stress,dEdVol);
  E[POT_ENERGY] = myPot->evalForce(F,Eloc,stress,1);
  evalAccel(F);

  P_p = trace(stress);
  P_k = 2.0*E[KINETIC_ENERGY];
  DV = OHMMS_DIM*myPtcl->Lattice.Volume;

  //Evaluate the instantaneous  pressure
  //E[pressure_ind] = (P_p+P_k)/DV-dEdVol;
  E[pressure_ind] = (1.5*P_p+P_k)/DV;
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 273 $   $Date: 2005-01-17 16:13:53 -0600 (Mon, 17 Jan 2005) $
 * $Id: TPNHydroLangevin.cpp 273 2005-01-17 22:13:53Z jnkim $ 
 ***************************************************************************/
