//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TPNLangevin.h"
#include "ParticleBase/ParticleBoostUtility.h"
#include "Message/OpenMP.h"
using namespace OHMMS;

TPNLangevin::TPNLangevin()
{
  CellMask=1.0;
  initParameters(E,m_param);
  m_param.add(Lattice0,"lattice","AA");
  m_param.add(pLattice0,"plattice","eV-fsec");
  m_param.add(CellMask,"conditions","none");
}


void 
TPNLangevin::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = 1;

  initializeBase(ptclin,potin,makecopy);

  // Calculate the volumes
  Volume = myPtcl->Lattice.Volume;
  Lattice0 = myPtcl->Lattice.R;

  if(!Initialized) {//allocation-related issue is done once
    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addRestartAttrib(velocity_tag);
    myPtcl->addRestartAttrib(force_tag);
    Initialized = true;
  }

  if(myName == "sdmin") {
    VolMass=(DOF+OHMMS_DIM)*20*Kelvin_to_eV/OHMMS_DIM/VolFreq/VolFreq;
  } else {
    VolMass=(DOF+OHMMS_DIM)*T_ref*Kelvin_to_eV/OHMMS_DIM/VolFreq/VolFreq;
  }

  checkParameters(myName);
  Stress_ext = 0.0;
  Stress_ext.diagonal(P_ref);

  if(AddNoise && !Vr.size()) Vr.resize(myPtcl->getLocalNum());

  if(ParameterModified) {
    setTemperature(RescaleTemperature,T_ref);
  }  else {
    evalKineticEnergy();
    setSigma(T_ref,minv);
  }

  DoverDOF = 1.0/DOF;
  Volume = myPtcl->Lattice.Volume;
  VolumeInv = 1.0/Volume;

  //NN list is evaluated while checking the BConds
  applyBConds(true);
  //1 to evaluate stress and use the NN list 
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,1);
  Stress_ke = evalKineticPressure(myPtcl->V);

  Tensor_t stress=myPot->getStress();
  Stress_ind = VolumeInv*(1.5*stress + Stress_ke);
  P_p = trace(stress);
  P_k = 2.0*E[KINETIC_ENERGY];

  dtover2m=minv;
  for(int i=0; i<dtover2m.size(); i++) dtover2m[i]*=dtover2;
  ParameterModified = false;
}

void 
TPNLangevin::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  setSigma(T_ref,minv);

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  scaleTemperature(rescale,T_ref);

  ScaledV = 1.0;
  pLattice0 = 0.0;
}


bool 
TPNLangevin::propagate(Interaction_t* pot) {

  //Set the current position
  if(Counter == UpdateStride) {
    Counter=0;
    myPot->setCurrentPos(myPtcl->R);
  }

  Scalar_t c1 = dtover2*myPtcl->Lattice.Volume;
  //update Lattice(t+dt/2)
  Lattice_h = Lattice0 + dtover2W*dot(pLattice0,Lattice0);
  //update pLattice(t+dt/2)
  Scalar_t pDiag = dtover2*DoverDOF*P_k;
  if(AddNoise) pDiag += SigmaVol*randomSource();
  pLattice_h = 0.0;
  pLattice_h.diagonal(pDiag);

  //pLattice_h += (1.0-chi_vol)*pLattice0+c1*(Stress_ind-Stress_ext);
  for(int i=0;i<Tensor_t::Size;i++) 
    pLattice_h(i) += cf_vol*pLattice0(i)+c1*(Stress_ind(i)*CellMask(i)-Stress_ext(i));
  pLattice_h.symmetrize();

  Scalar_t pg = trace(pLattice_h);
  Scalar_t vfac = cf_vel-dtover2W*DoverDOF*pg;

  //update V(t,dt/2) and R(t+dt). Note that c2/c0 = 2
  Tensor_t pgt_v = dtover2W*pLattice_h;
  Tensor_t pgt_r = dtoverW*pLattice_h;

#pragma omp parallel
  {
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];

    const ParticlePos_t& F = myPtcl->F;
    const ParticleIndex_t& gid = myPtcl->GroupID;
    ParticlePos_t& V = myPtcl->V;
    ParticlePos_t& dR = myPtcl->dR;
    Tensor_t hinv = inverse(transpose(Lattice0));
    Tensor_t dht = transpose(Lattice_h-Lattice0);
    Tensor_t dh=dot(dht,hinv);

    RandomGaussianVector<SingleParticlePos_t,BoostRandom<Scalar_t> > gg(*reng[ip]);
    for(int i=first; i<last; i++) {
      int itype=gid[i];
      V[i] = vfac*V[i]+dtover2m[itype]*F[i]+dot(pgt_v,V[i]) +SigmaV[itype]*gg();
      SingleParticlePos_t dr(Dt*V[i]+dot(pgt_r,myPtcl->R[i]));
      dR[i] = dr + dot(dh,myPtcl->R[i]+dr);
      //myPtcl->R[i] = myPtcl->R[i]+dt_t*V[i]+dot(pgt_r,myPtcl->R[i]);
    }
  }

  //update the unit cell: Volume_h -> myPtcl->Lattice.Volume
  myPtcl->updateLayout(Lattice_h);

  //3 to evaluate both NN list with dR and stress
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,3);

  Volume = myPtcl->Lattice.Volume;
  VolumeInv = 1.0/Volume;
  Tensor_t stress=myPot->getStress();
  P_p = trace(stress);

  Stress_ke=0.0;
#pragma omp parallel
  {
    int ip = omp_get_thread_num();
    int first = LocIndex0[ip];
    int last = LocIndex0[ip+1];
    for(int i=first; i<last; i++) myPtcl->R[i]+=myPtcl->dR[i];

    const ParticleIndex_t& gid = myPtcl->GroupID;
    ParticlePos_t& F = myPtcl->F;
    ParticlePos_t& V = myPtcl->V;
    std::vector<Scalar_t> sigv(SigmaV);
    std::vector<Scalar_t> m_t(mass);
    std::vector<Scalar_t> cf(dtover2m);
    Scalar_t vfac_t=vfac;
    Tensor_t pgt_v_t = pgt_v;
    Tensor_t ke_loc;

    //update the particle velocities V(t+dt/2) and R(t+dt)
    RandomGaussianVector<SingleParticlePos_t,BoostRandom<Scalar_t> > gg(*reng[ip]);
    for(int i=first; i<last; i++) {
      int itype = gid[i];
      V[i]=vfac*V[i]+dtover2m[itype]*F[i]+dot(pgt_v,V[i])+SigmaV[itype]*gg();
      ke_loc += mass[itype]*outerProduct(V[i],V[i]);
      //SingleParticlePos_t v(vfac_t*V[i]+cf[itype]*F[i]+dot(pgt_v_t,V[i])+sigv[itype]*gg());
      //ke_loc += m_t[itype]*outerProduct(v,v);
    }

#pragma omp critical
    {
      Stress_ke+=ke_loc;
    }
  }

  E[KINETIC_ENERGY] = 0.5*trace(Stress_ke);

  //Stress_ke = evalKineticPressure(V);
  Stress_ind = VolumeInv*(1.5*stress+Stress_ke);
  P_k = 2.0*E[KINETIC_ENERGY];
  E[pressure_ind] = trace(Stress_ind)*VolumeInv;

  c1 = dtover2*Volume;
  pDiag = dtover2*DoverDOF*P_k;
  if(AddNoise) pDiag += SigmaVol*randomSource();
  pLattice0 = 0.0;
  pLattice0.diagonal(pDiag);

  //for(int i=0;i<Tensor_t::Size;i++) stress(i) = Stress_ind(i)*CellMask(i);
  //pLattice0 += (1.0-chi_vol)*pLattice_h+c1*(stress-Stress_ext);
  for(int i=0;i<Tensor_t::Size;i++) 
    pLattice0(i) += cf_vol*pLattice_h(i)+c1*(Stress_ind(i)*CellMask(i)-Stress_ext(i));
  pLattice0.symmetrize();

  //update Lattice(t+dt/2)
  Lattice0 = Lattice_h + dtover2W*dot(pLattice_h,Lattice_h);
  myPtcl->updateLayout(Lattice0);
  Volume = myPtcl->Lattice.Volume;

  E[pv_ind] = E[pressure_ind]*Volume;
  E[pk_ind] = pLattice0.squared()*VolMassInv*0.5;    

  E[FREE_ENERGY] = E[POT_ENERGY]+E[KINETIC_ENERGY]+E[pk_ind]+P_ref*Volume;
  E[TEMPERATURE] = E[KINETIC_ENERGY]*TwoOverDOF*eV_to_Kelvin;
  E[volume_ind] = Volume;

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  ///pressure criterion
  return  ((E[TEMPERATURE]+E[pk_ind]*eV_to_Kelvin) > KE_min_K);
}


void 
TPNLangevin::applyBConds(bool wrapping) {

  myPot->setCurrentPos(myPtcl->R,wrapping);
  Counter = 0;
}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 314 $   $Date: 2005-03-01 17:08:22 -0600 (Tue, 01 Mar 2005) $
 * $Id: TPNLangevin.OMP.cpp 314 2005-03-01 23:08:22Z jnkim $ 
 ***************************************************************************/
