//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2004- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TPNLangevin.h"
#include "ParticleBase/RandomSeqGenerator.h"
using namespace OHMMS;

TPNLangevin::TPNLangevin()
{
  CellMask=1.0;
  initParameters(E,m_param);
  m_param.add(Lattice0,"lattice","AA");
  m_param.add(pLattice0,"plattice","eV-fsec");
  m_param.add(CellMask,"conditions","none");
}


void 
TPNLangevin::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy){

  // reset, time coefficients with  new or old MD parameters
  Counter = 1;

  initializeBase(ptclin,potin,makecopy);

  // Calculate the volumes
  Volume = myPtcl->Lattice.Volume;
  Lattice0 = myPtcl->Lattice.R;

  if(!Initialized) {//allocation-related issue is done once
    //add position-related data to OhmmsParticle::ConfigList
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addRestartAttrib(velocity_tag);
    myPtcl->addRestartAttrib(force_tag);
    Initialized = true;
  }

  if(myName == "sdmin") {
    VolMass=(DOF+OHMMS_DIM)*20*Kelvin_to_eV/OHMMS_DIM/VolFreq/VolFreq;
  } else {
    VolMass=(DOF+OHMMS_DIM)*T_ref*Kelvin_to_eV/OHMMS_DIM/VolFreq/VolFreq;
  }

  checkParameters(myName);
  Stress_ext = 0.0;
  Stress_ext.diagonal(P_ref);

  if(Vr.size()==0) Vr.resize(myPtcl->getLocalNum());

  if(ParameterModified) {
    setTemperature(RescaleTemperature,T_ref);
  }  else {
    evalKineticEnergy();
    setSigma(T_ref,minv);
    applyBConds(true);
  }

  DoverDOF = 1.0/DOF;
  Volume = myPtcl->Lattice.Volume;
  VolumeInv = 1.0/Volume;
  myPot->setCurrentPos(myPtcl->R,myPtcl->dR);
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,0);
  Stress_ke = evalKineticPressure(myPtcl->V);

  Tensor_t stress=myPot->getStress();
  Stress_ind = VolumeInv*(1.5*stress + Stress_ke);
  P_p = trace(stress);
  P_k = 2.0*E[KINETIC_ENERGY];

  dtover2m=minv;
  for(int i=0; i<dtover2m.size(); i++) dtover2m[i]*=dtover2;
  ParameterModified = false;
}

void 
TPNLangevin::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  setSigma(T_ref,minv);

  ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);

  scaleTemperature(rescale,T_ref);

  ScaledV = 1.0;
  pLattice0 = 0.0;
  applyBConds(true);
}


bool 
TPNLangevin::propagate(Interaction_t* pot) {

  //Set the current position
  if(Counter == UpdateStride) {
    Counter=0;
    myPot->setCurrentPos(myPtcl->R);
  }

  Scalar_t c1 = dtover2*myPtcl->Lattice.Volume;
  //update Lattice(t+dt/2)
  Lattice_h = Lattice0 + dtover2W*dot(pLattice0,Lattice0);
  //update pLattice(t+dt/2)
  Scalar_t pDiag = dtover2*DoverDOF*P_k;
  if(AddNoise) pDiag += SigmaVol*randomSource();
  pLattice_h = 0.0;
  pLattice_h.diagonal(pDiag);

  //pLattice_h += (1.0-chi_vol)*pLattice0+c1*(Stress_ind-Stress_ext);
  for(int i=0;i<Tensor_t::Size;i++) 
    pLattice_h(i) += cf_vol*pLattice0(i)+c1*CellMask(i)*(Stress_ind(i)-Stress_ext(i));
  pLattice_h.symmetrize();

  Scalar_t pg = trace(pLattice_h);
  Scalar_t vfac = cf_vel-dtover2W*DoverDOF*pg;

  //update V(t,dt/2) and R(t+dt). Note that c2/c0 = 2
  Tensor_t pgt_v(dtover2W*pLattice_h);
  Tensor_t pgt_r(dtoverW*pLattice_h);
  Tensor_t hinv(inverse(transpose(Lattice0)));
  Tensor_t dht(transpose(Lattice_h-Lattice0));
  Tensor_t dh(dot(dht,hinv));
  //vfac to the diagonal elements
  pgt_v += vfac;

  const ParticleIndex_t& gid=myPtcl->GroupID;
  ParticlePos_t::const_iterator F(myPtcl->F.begin());
  ParticlePos_t::iterator V(myPtcl->V.begin());
  ParticlePos_t::iterator dR(myPtcl->dR.begin());
  ParticlePos_t::iterator R(myPtcl->R.begin());
  size_type first=0, last=myPtcl->getLocalNum();

  if(AddNoise) makeGaussRandom(Vr);

  for(size_type iat=first; iat<last; iat++) {
    Index_t itype=gid[iat];
    SingleParticlePos_t v(dtover2m[itype]*(*F)+dot(pgt_v,*V) +SigmaV[itype]*Vr[iat]);
    SingleParticlePos_t dr(Dt*v+dot(pgt_r,*R));
    *V = v;
    *dR = dr + dot(dh,dr+*R);
    ++F;++V;++dR;
    //V[iat] = vfac*V[iat]+dtover2m[itype]*F[iat]+dot(pgt_v,V[iat]) +SigmaV[itype]*Vr[iat];
    //SingleParticlePos_t dr(Dt*V[iat]+dot(pgt_r,myPtcl->R[iat]));
    //dR[iat] = dr + dot(dh,myPtcl->R[iat]+dr);
    //Original formulation
    //myPtcl->R[i] = myPtcl->R[i]+Dt*V[i]+dot(pgt_r,myPtcl->R[i]);
  }

  //update the unit cell: Volume_h -> myPtcl->Lattice.Volume
  myPtcl->updateLayout(Lattice_h);
  Volume = myPtcl->Lattice.Volume;
  VolumeInv = 1.0/Volume;

  //myPot->setCurrentPos(myPtcl->R,false);
  //Evaluate gradients, pressure and dE/dV
  //stress=0.0;
  //E[POT_ENERGY] = myPot->evalForce(F,Eloc,stress,1);
  //evalAccel(F);

  //myPot->setCurrentPos(myPtcl->R,myPtcl->dR);
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,3);

  Tensor_t stress=myPot->getStress();
  P_p = trace(stress);

  //Add the displacement vectors
  myPtcl->R += myPtcl->dR;

  //update the particle velocities V(t+dt/2) and R(t+dt)
  if(AddNoise) makeGaussRandom(Vr);

  F=myPtcl->F.begin();
  V=myPtcl->V.begin();

  for(size_type iat=first; iat<last; iat++) {
    Index_t itype=gid[iat];
    *V = dtover2m[itype]*(*F)+dot(pgt_v,*V)+SigmaV[itype]*Vr[iat];
    //V[iat] = vfac*V[iat]+dtover2m[itype]*F[iat]+dot(pgt_v,V[iat])+SigmaV[itype]*Vr[iat];
    ++F;++V;
  }

  Stress_ke = evalKineticPressure(myPtcl->V);
  Stress_ind = VolumeInv*(1.5*stress+Stress_ke);

  //for(int i=0; i<3; i++) 
  //  Stress_ind(i,i) = VolumeInv*(1.5*stress(i,i)+Stress_ke(i,i));
  P_k = 2.0*E[KINETIC_ENERGY];
  E[pressure_ind] = trace(Stress_ind)/3.0;

  c1 = dtover2*Volume;
  pDiag = dtover2*DoverDOF*P_k;
  if(AddNoise) pDiag += SigmaVol*randomSource();
  pLattice0 = 0.0;
  pLattice0.diagonal(pDiag);

  //for(int i=0;i<Tensor_t::Size;i++) stress(i) = Stress_ind(i)*CellMask(i);
  //pLattice0 += (1.0-chi_vol)*pLattice_h+c1*(stress-Stress_ext);
  for(size_type i=0;i<Tensor_t::Size;i++) 
    pLattice0(i) += cf_vol*pLattice_h(i)+c1*CellMask(i)*(Stress_ind(i)-Stress_ext(i));
  pLattice0.symmetrize();

  //update Lattice(t+dt/2)
  Lattice0 = Lattice_h + dtover2W*dot(pLattice_h,Lattice_h);
  myPtcl->updateLayout(Lattice0);
  Volume = myPtcl->Lattice.Volume;

  E[pv_ind] = E[pressure_ind]*Volume;
  E[pk_ind] = pLattice0.squared()*VolMassInv*0.5;    

  E[FREE_ENERGY] = E[POT_ENERGY]+E[KINETIC_ENERGY]+E[pk_ind]-P_ref*Volume;
  E[volume_ind] = Volume;

  Counter++; // increment the internal counter
  E[NOW] += Dt;

  ///pressure criterion
  return  ((E[TEMPERATURE]+E[pk_ind]*eV_to_Kelvin) > KE_min_K);
}


void 
TPNLangevin::applyBConds(bool wrapping) {

  myPot->setCurrentPos(myPtcl->R,wrapping);
  Counter=0;
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 376 $   $Date: 2005-08-28 15:32:27 -0500 (Sun, 28 Aug 2005) $
 * $Id: TPNLangevin.cpp 376 2005-08-28 20:32:27Z jnkim $ 
 ***************************************************************************/
