//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Propagators/TVNLangevin.h"
#include "ParticleBase/ParticleUtility.h"
#include "Message/OpenMP.h"
#include <limits>
using namespace OHMMS;

TVNLangevin::TVNLangevin(): T_ref(0.0), Dt(1.0), Friction(0.05)
{ 
  m_param.add(T_ref,"temperature","K");
  m_param.add(Dt,"timestep","fsec");
  m_param.add(Friction,"friction","fsec-1");
}

void 
TVNLangevin::initialize(Particle_t* ptclin, Interaction_t* potin, bool makecopy) {

  Counter = 0;

  if(!Initialized) {
    initializeBase(ptclin,potin,makecopy);
    setSigma(T_ref, Dt, Friction);
    myPtcl->convert2Cart(myPtcl->R);
    myPtcl->addPositionAttrib(myPtcl->R);
    myPtcl->addRestartAttrib(velocity_tag);
    myPtcl->addRestartAttrib(force_tag);
    Initialized = true;
  }
  
  if(ParameterModified) 
    setTemperature(RescaleTemperature,T_ref);
  else
    applyBConds(true);

  if(E[NOW]<numeric_limits<Scalar_t>::epsilon()) {
    E[POT_ENERGY] = myPot->evalForce(myPtcl->F,0); 
  }


  ParameterModified = false;
}

void 
TVNLangevin::setTemperature(bool rescale, Scalar_t tref) {

  if(tref>zeroKelvin) T_ref = tref;

  if(T_ref>zeroKelvin) {
    setSigma(T_ref, Dt, Friction);
    ParticlePos_t& V = *myPtcl->getVectorAttrib(velocity_tag);
    scaleTemperature(rescale,T_ref);
    ScaledV = 1.0;
    applyBConds(true);
  }
}


bool  TVNLangevin::propagate(Interaction_t* pot) {

  if(Counter == UpdateStride) {
    Counter = 0; 
    myPot->setCurrentPos(myPtcl->R); 
  }

  omp_int_t ip=0;
  size_type first =0;
  size_type last = myPtcl->getLocalNum();
  const ParticleIndex_t& gid = myPtcl->GroupID;

  //Random Number Generators usually are not thread-safe
  //engine to generate a set of correlated vectors, CorrelatedVectorGenerator
  RandomVector<SingleParticlePos_t, RandomGenerator_t> CVG(Random);
  register SingleParticlePos_t r1, r2;

  ParticlePos_t::const_iterator F=myPtcl->F.begin();
  ParticlePos_t::iterator V = myPtcl->V.begin();
  ParticlePos_t::iterator dR = myPtcl->dR.begin();
  for(size_type iat=first; iat<last; iat++) {
    CVG.get(r1,r2); // correlated random vectors 
    Index_t itype = gid[iat];
    r1*=OneOverSqrtM[itype];
    r2*=OneOverSqrtM[itype];
    SingleParticlePos_t a(minv[itype]*(*F));
    SingleParticlePos_t v(*V);
    *dR = SigmaR*r1+c1dt*v+c2dt2*a;
    *V = CorrRVx*r1+CorrRVy*r2+c0*v+c12dt*a;
    ++F;++dR;++V;
  }

  //myPot->setCurrentPos(myPtcl->R,myPtcl->dR);
  E[POT_ENERGY] = myPot->evalForce(myPtcl->F,2); 

  myPtcl->R += myPtcl->dR;

  F = myPtcl->F.begin();
  V = myPtcl->V.begin();
  Scalar_t ke=0.0;
#pragma ivdep
  for(size_type iat=first; iat<last; ++iat) {
    Index_t itype = gid[iat];
    SingleParticlePos_t v(*V+c2dtoverm[itype]*(*F));
    ke += mass[itype]*dot(v,v);
    *V=v;
    ++F;++V;
  }

  E[KINETIC_ENERGY] = 0.5*myPtcl->collect(ke);
  E[TEMPERATURE] = TwoOverDOF*eV_to_Kelvin*E[KINETIC_ENERGY];
  E[FREE_ENERGY] = E[POT_ENERGY] + E[KINETIC_ENERGY];

  ++Counter;
  E[NOW] += Dt;
  return (E[TEMPERATURE] > KE_min_K);
}


void TVNLangevin::applyBConds(bool wrapping) 
{
  myPot->setCurrentPos(myPtcl->R, wrapping);
  Counter = 0;
}

void
TVNLangevin::setSigma(Scalar_t Temp, Scalar_t dt, Scalar_t friction) {
     
  // if the friction is not set, use 0.01 
  if(friction < numeric_limits<Scalar_t>::epsilon()) 
    Friction = 0.05/dt;
  else 
    Friction=friction;
  Dt = dt;
  Scalar_t gamma = Friction*Dt;
  Scalar_t bi = 1/gamma;
  c0 = exp(-gamma);
  Scalar_t x = Temp*Kelvin_to_eV;
  SigmaR = Dt*sqrt(x*bi*(2.0e0-bi*(3.0e0-4.0e0*c0+c0*c0)));
  SigmaV = sqrt(x*(1.0e0-c0*c0));
  //Scalar_t sigrv = Dt*x*bi*(1.0e0-c0)*(1.0e0-c0)/SigmaR/SigmaV;
  Scalar_t sigrv = Dt*x*bi*(1.0e0-c0)*(1.0e0-c0)/SigmaR/SigmaV;
  CorrRVx = SigmaV*sigrv;
  CorrRVy = SigmaV*sqrt(1.0e0-sigrv*sigrv);
  c1 = (1.0e0-c0)*bi; c1dt = c1*Dt; c2 = (1.0e0-c1)*bi; 

  OneOverSqrtM.resize(minv.size());
  for(int i=0; i<minv.size(); ++i) OneOverSqrtM[i]=sqrt(minv[i]);

  Scalar_t c2dt = c2*Dt;  
  c2dt2 = c2dt*Dt; 
  c12dt = (c1-c2)*Dt;
  c2dtoverm.resize(minv.size());
  for(int i=0; i<minv.size(); ++i) c2dtoverm[i]=c2dt*minv[i];
  /*
  // variables independent of species
  Scalar_t sigr = Dt*sqrt(x*bi*(2.0e0-bi*(3.0e0-4.0e0*c0+c0*c0)));
  Scalar_t sigv = sqrt(x*(1.0e0-c0*c0));
  Scalar_t sigrv = Dt*x*bi*(1.0e0-c0)*(1.0e0-c0);

  int nspecies = mass.size();
  if(SigmaR.size() < nspecies) {
    SigmaR.resize(2*nspecies);
    SigmaV.resize(2*nspecies);
    CorrRVx.resize(2*nspecies);
    CorrRVy.resize(2*nspecies);
    c2dtoverm.resize(2*nspecies);
    c2dt2overm.resize(2*nspecies);
    c12dtoverm.resize(2*nspecies);
  }
  for(int i=0; i<nspecies; i++) {
    Scalar_t sqrtminv = sqrt(minv[i]);
    SigmaR[i] = sigr*sqrtminv;
    SigmaV[i] = sigv*sqrtminv;
    //CorrRVx[i] = sigrv*minv[i]/SigmaR[i]/SigmaV[i];
    CorrRVx[i] = sigrv/sigr/sigv;
    CorrRVy[i] = sqrt(1.0e0-CorrRVx[i]*CorrRVx[i]);
    CorrRVx[i] *= SigmaV[i];
    CorrRVy[i] *= SigmaV[i];
    c2dtoverm[i] = c2dt*minv[i];
    c2dt2overm[i] = c2dt2*minv[i];
    c12dtoverm[i] = c12dt*minv[i];
    //WARNMSG("\tIntel Bug on IA64 for nspecies = 1 case")
  }
  */
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: TVNLangevin.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
