//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "QuantumBase/ElectronStructure.h"

namespace OHMMS {

  //initialize static data
  const int ElectronStructure::BiIter;
  const ElectronStructure::real_type ElectronStructure::BiTol = 1.0e-8;

  ElectronStructure::real_type
  ElectronStructure::getBandEnergy() {

    SumOcc = 0.0e0;
    dSumOcc = 0.0e0;
    dSumEnergy = 0.0;
    if(ZeroTemperature) {
      for(int i=0; i<NumStates; i++) {
	for(int ik=0; ik<NumKpts; ik++) {
	  if(SumOcc < NumElectrons) Occ(ik,i) = 1.0;
	  else                      Occ(ik,i) = 0.0;
	  SumOcc += SpinDeg*Occ(ik,i)*KSet->W[ik];
	}
      }
    }else {
      if(!FixedFermiEnergy) bisection(); 
      if(!dOcc.size())  dOcc.resize(NumKpts,NumStates);
      for(int ik=0; ik<NumKpts; ik++) {
	real_type wk = KSet->W[ik];
	for(int i=0; i<NumStates; i++) {
	  Occ(ik,i) = fermi(E(ik,i), Mu);
	  dOcc(ik,i) = Occ(ik,i)*(Occ(ik,i)-1.0);
	  SumOcc += SpinDeg*Occ(ik,i)*wk;
	  dSumOcc += dOcc(ik,i)*wk;
	  dSumEnergy += dOcc(ik,i)*E(ik,i)*wk;
	}
      }
      dSumEnergy /= Sigma;
    } 

    SumEnergy = 0.0e0;
    for(int ik=0; ik<NumKpts; ik++) {
      real_type wk = SpinDeg*KSet->W[ik];
      for(int i=0; i<NumStates; i++) {
	SumEnergy += Occ(ik,i)*E(ik,i)*wk;
      }
    }

    findHomoLevel();
    if(ZeroTemperature) {
      double emin = E(0,Homo);
      int lumo = Homo+1;
      double emax = E(0,lumo);
      for(int ik=1; ik<NumKpts; ik++) {
	emin = max(E(ik,Homo),emin);
	emax = min(E(ik,lumo),emax);
      }
      Mu = (emin+emax)*0.5;
    }
    return SumEnergy;
  }

  void ElectronStructure::resize(int ndim, real_type nels, int nstates) {
    BasisDim = ndim;
    NumElectrons = nels;
    NumStates = nstates;
    setElectronStates();
  }

  void ElectronStructure::resize() {
    setElectronStates();
  }

  void ElectronStructure::setElectronStates() { 

    // determine based on the number of states requted
    if(Sigma > 1.0e-5)  ZeroTemperature = false;
    else                ZeroTemperature = true;

    if(NumStates < 0) {
      if(ZeroTemperature) {
	NumStates = static_cast<int>(NumElectrons/SpinDeg);
      } else {
	int num = static_cast<int>(NumElectrons/SpinDeg + NumElectrons/10);
	NumStates = (num < BasisDim) ? num: BasisDim;
      }
    }

    //set the parameters in EigenStates
    States.set(BasisDim,NumStates);
  }

  void ElectronStructure::bisection() {

    real_type nel0 = (NumElectrons - NumCoreElectrons)/SpinDeg;
    int l = static_cast<int>(nel0);

    real_type emax = -1e5, emin = 1e5; 
    int nend = NumStates-1;
    for(int ik=0; ik<NumKpts; ik++) emax = max(E(ik,nend),emax);
    for(int ik=0; ik<NumKpts; ik++) emin = min(E(ik,0),emin);

    real_type ef = 0.0e0;
    for(int ik=0; ik<NumKpts; ik++) {
      ef += (E(ik,l) + E(ik,l+1))*0.5e0*KSet->W[ik];
    }

    real_type ef0 = ef;
    real_type elow = emin;
    real_type ehi = emax;
    for(int i= 0; i< BiIter; i++) {
      real_type nel = nelec(ef); 
      real_type err = nel - nel0;
      if(fabs(err) < BiTol) { Mu = ef; return;}
      if(err < 0.0e0) elow = ef;
      else            ehi = ef;
      ef = (elow + ehi)*0.5e0;
      if(fabs(ef - ef0) < BiTol) { Mu = ef; return;}
      ef0 = ef;
      if(ef > emax || ef < emin) { return;}
    }
    Mu = ef;
  }

  ElectronStructure::real_type
  ElectronStructure::nelec(real_type ef) {

    real_type nel = 0.0e0;
    for(int ik=0; ik<NumKpts; ik++) {
      for(int i=0; i<NumStates; i++){
	nel += fermi(E(ik,i),ef)*KSet->W[ik];
      }
    }
    return nel;
  }


  int ElectronStructure::findHomoLevel() {
    //Homo = NumStates-1;
    const real_type small = 1e-10;
    for(int ik=0; ik<NumKpts; ik++) {
      int i= Homo-10; // start 10 band below old Homo level
      if(i<0) i = 0;  // reset the starting index to 0
      //while(E(ik,i) >= Mu && i > 0) {
      while(Occ(ik,i) > small && i < NumStates) {
	i++; 
      } 
      Homo = i;
    }
    return Homo;
  }

}
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: ElectronStructure.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
