//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
// (c) Copyright 1998-2002 by Jeongnim Kim
//   All rights reserved.
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//                    CONTACT INFORMATION
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////
// \author Jeongnim Kim
//
//   Polak-Ribiere factor (not used)
//      double x1 = dot(g,g)-dot(gprev,g);
//      double x2 = dot(gprev,gprev);
//      gamma = -x1/(x2+1e-40);
//   
//      gamma = dot(g,g)/(dot(gprev,gprev)+1e-40); // <g|g>/<gp|gp>
//
//////////////////////////////////////////////////////////////////////
#ifndef OHMMS_BANDBYBANDCONJUGATEGRADIENT_H
#define OHMMS_BANDBYBANDCONJUGATEGRADIENT_H
#include "Utilities/OhmmsInfo.h"
#include "EigSolver/OhmmsBlasf.h"
#include "OhmmsPETE/OhmmsVector.h"

template<class Ham_t, class Wfs_t>
struct BandByBandConjGradient {

  typedef Wfs_t::Type_t  T;
  int BasisDim;
  Vector<T> hpsi, g, gprev, cg, hcg, psi;
  BandByBandConjGradient():BasisDim(0) { }
  void solve(const Ham_t& ham, Wfs_t& states, T* eigval);
};

template<class T>
inline 
T dot(const Vector<T>& a, const Vector<T>& b) {
  return OhmmsBlas::dot(a.size(),a.begin(), b.begin());
}

inline 
void normalize(Vector<double>& a) {
  double tmp = dot(a,a);
  tmp = 1.0/sqrt(tmp);
  OhmmsBlas::scal(a.size(),tmp, a.begin());
}

template<class Ham_t, class Wfs_t>
void 
BandByBandConjGradient<Ham_t,Wfs_t>::solve(const Ham_t& ham, Wfs_t& states, 
                                           T* eigval) {

  // allocate the data
  int nband = states.nrow();
  if(BasisDim < states.ncol()) {
    BasisDim = states.ncol();
    psi.resize(BasisDim);
    hpsi.resize(BasisDim);
    g.resize(BasisDim);
    cg.resize(BasisDim);
    hcg.resize(BasisDim);
    gprev.resize(BasisDim);
  }

  T gamma, residue;
  const T etol = 1e-12;
  const T restol = 1e-12;
  T lambda, e1, e2, theta, costheta, sintheta;
  int maxiter = 100;
  int iband = 0;
  do {

    states.orthonormalize(iband);
    states.copy(iband, psi.begin());// copy iband to psi    

    ham.apply(psi.begin(), hpsi.begin());

    lambda = dot(psi,hpsi); 
    g = lambda*psi - hpsi;

    /////////////////////////////////////////////
    //if no precondition is used, Ham_t::precondition should include
    // states.orthogonalize(g.begin(), 0, iband);
    //
    //For precondition,
    //  states.orthogonalize(g.begin(), 0, iband-1);
    //    hamiltonian specific precondition 
    //  states.orthogonalize(g.begin(), 0, iband);
    ////////////////////////////////////////////
    ham.precondition(g.begin(),states,iband);

    residue = dot(g,g);
    gprev = g;
    cg = g;
    normalize(cg);

    int iter = 0;
    do {
      ///////////////////////////////////////////////////////
      // Lineminimization using cosine and sine expansion
      // |psi> = cos(theta)*|psi> + sin(theta)*|cg>
      ///////////////////////////////////////////////////////
      e1 = dot(hpsi,cg);

      ham.apply(cg.begin(), hcg.begin());

      e2 = dot(hcg,cg);
      theta = 0.5*atan(2.0*e1/(lambda - e2));
      costheta = cos(theta);
      sintheta = sin(theta);
      T eproj = costheta*(costheta*lambda+2.0*sintheta*e1)
	+sintheta*sintheta*e2;

      psi = costheta*psi + sintheta*cg;
      hpsi = costheta*hpsi + sintheta*hcg;
      /////////////////////////////////////////////////////////

      if(fabs((lambda-eproj)/lambda) < etol || residue < restol) { 
	lambda = eproj; break;
      }

      lambda = eproj;
      g = lambda*psi - hpsi;
      residue  = dot(g,g);

      states.assign(iband, psi.begin());// copy psi to iband

      /////////////////////////////////////////////
      //no precondition is used
      //states.orthogonalize(g.begin(), 0, iband);
      ////////////////////////////////////////////
      ham.precondition(g.begin(),states, iband);

      double x1 = dot(g,g) - dot(gprev,g);
      double x2 = dot(g,cg) - dot(gprev,cg);
      gamma = -x1/(x2+1e-40);
      DEBUGMSG(iter << " " << eproj << " " << residue)

      cg = g + gamma*cg;
      gprev = g;
      states.orthogonalize(cg.begin(), iband, iband);
      normalize(cg);
      iter++;
    } while(iter < maxiter);

    if(iter < maxiter) {
      cout << "Iteration = " << iter << endl;
      ham.apply(psi.begin(), hpsi.begin());
      eigval[iband] = dot(psi,hpsi);
      iband++;
    } else {
      // did not converged, start with a new random vector
      states.randomize(iband);
    }
  } while(iband < nband);
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: BandByBandConjGradient.h 151 2004-11-02 23:51:14Z jnkim $
 ************************************************************************/
