//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
// (c) Copyright 1998-2002 by Jeongnim Kim
//   All rights reserved.
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////
//                    CONTACT INFORMATION
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////
// \author Jeongnim Kim
// \class BandByBandConjGradient
// \brief A templated class to perform band-by-band CG minimization
//        to calculate m eigen states.
// 
//
//   Polak-Ribiere factor (not used)
//      double x1 = dot(g,g)-dot(gprev,g);
//      double x2 = dot(gprev,gprev);
//      gamma = -x1/(x2+1e-40);
//
//      gamma = dot(g,g)/(dot(gprev,gprev)+1e-40); // <g|g>/<gp|gp>
//
//////////////////////////////////////////////////////////////////////
#ifndef OHMMS_BANDBYBANDCONJUGATEGRADIENT_H
#define OHMMS_BANDBYBANDCONJUGATEGRADIENT_H
#include <vector>
using std::vector;
#include "Utilities/OhmmsInfo.h"
#include "Message/OpenMP.h"

template<class Ham_t, class Wfs_t>
struct BandByBandConjGradient {

  typedef typename Wfs_t::Type_t  T;
  vector<int> LocLoc;
  int BasisDim;
  vector<T> hpsi, g, gprev, cg, hcg, psi, hpsitmp;
  BandByBandConjGradient():BasisDim(0) { }
  void solve(const Ham_t& ham, Wfs_t& states, T* eigval);
};

template<class Ham_t, class Wfs_t>
void 
BandByBandConjGradient<Ham_t,Wfs_t>::solve(const Ham_t& ham, Wfs_t& states, 
                                           T* eigval) {

  // allocate the data
  int nband = states.nrow();
  if(BasisDim < states.ncol()) {
    BasisDim = states.ncol();
    psi.resize(BasisDim);
    hpsi.resize(BasisDim);
    g.resize(BasisDim);
    cg.resize(BasisDim);
    hcg.resize(BasisDim);
    gprev.resize(BasisDim);
    hpsitmp.resize(BasisDim);
    LocLoc = states.LocLoc; // copy the distribution pattern
  }

  T gamma, residue;
  const T etol = 1e-12;
  const T restol = 1e-12;
  T lambda, e1, e2, theta, costheta, sintheta;
  int maxiter = 100;
  int iband = 0;
  do {

    states.orthonormalize(iband);

    //states.copy(iband, psi.begin());// copy iband to psi    
#pragma omp parallel
    {
     int myID = omp_get_thread_num();
     const T* xref = states[iband];
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) psi[ib] = xref[ib];
    }

    //ham.apply(psi.begin(), hpsi.begin());
    //lambda = dot(psi,hpsi); 
    lambda = 0.0;
#pragma omp parallel reduction(+:lambda)
    {
     int myID = omp_get_thread_num();
     const int ibi = LocLoc[myID];
     const int ibf = LocLoc[myID+1];

     ///////////////ONLY FOR TEST, MOVE THIS before this parallel region
     ham.apply(psi.begin(), hpsi.begin()+ibi, ibi, ibf-ibi);
     ///////////////////////////////

     for(int ib = ibi; ib < ibf; ib++) lambda += psi[ib]*hpsi[ib];
    }

    // g = lambda*psi - hpsi;
#pragma omp parallel firstprivate(lambda)
    {
     int myID = omp_get_thread_num();
     int ib = LocLoc[myID];
     const int ibf = LocLoc[myID+1];
     for(; ib < ibf; ib++) g[ib] = lambda*psi[ib] - hpsi[ib];
    }

    /////////////////////////////////////////////
    //if no precondition is used, Ham_t::precondition should include
    // states.orthogonalize(g.begin(), 0, iband);
    //
    //For precondition,
    //  states.orthogonalize(g.begin(), 0, iband-1);
    //    hamiltonian specific precondition 
    //  states.orthogonalize(g.begin(), 0, iband);
    ////////////////////////////////////////////
    ham.precondition(g.begin(),states,iband);

    ////////////////////////////
    //Bunching parallel regions
    ////////////////////////////
    //residue = dot(g,g);
    //gprev = g;
    //cg = g;
    //normalize(cg);
    ////////////////////////////

    residue = 0.0;
    T cgnorm = 0.0;
#pragma omp parallel reduction(+:residue, cgnorm)
    {
      int myID = omp_get_thread_num();
       const int ibi = LocLoc[myID];
       const int ibf = LocLoc[myID+1];

       for(int ib = ibi; ib<ibf; ib++) gprev[ib] = g[ib];
       for(int ib = ibi; ib<ibf; ib++) cg[ib] = g[ib];
       for(int ib = ibi; ib<ibf; ib++) residue += g[ib]*g[ib];
       for(int ib = ibi; ib<ibf; ib++) cgnorm += cg[ib]*cg[ib];
          
    }

    cgnorm = 1.0/sqrt(cgnorm);
#pragma omp parallel firstprivate(cgnorm)
    {
      int myID = omp_get_thread_num();
       const int ibi = LocLoc[myID];
       const int ibf = LocLoc[myID+1];
       for(int ib = ibi; ib<ibf; ib++) cg[ib] *= cgnorm;
    }

    int iter = 0;
    do {
      ///////////////////////////////////////////////////////
      // Lineminimization using cosine and sine expansion
      // |psi> = cos(theta)*|psi> + sin(theta)*|cg>
      ///////////////////////////////////////////////////////
      //ham.apply(cg.begin(), hcg.begin());

      e1 = 0.0; e2 = 0.0;
#pragma omp parallel reduction(+:e1,e2)
      {
        int myID = omp_get_thread_num();
        const int ibi = LocLoc[myID];
        const int ibf = LocLoc[myID+1];

	///////////////ONLY FOR TEST, MOVE THIS before this parallel region
        ham.apply(cg.begin(), hcg.begin()+ibi, ibi, ibf-ibi);
	///////////////////////////////

        for(int ib = ibi; ib<ibf; ib++) {
          e1 += hpsi[ib]*cg[ib];
          e2 += hcg[ib]*cg[ib];
	}
      }

      theta = 0.5*atan(2.0*e1/(lambda - e2));
      costheta = cos(theta);
      sintheta = sin(theta);
      T eproj = costheta*(costheta*lambda+2.0*sintheta*e1)
	+sintheta*sintheta*e2;

      if(fabs((lambda-eproj)/lambda) < etol || residue < restol) { 
	lambda = eproj; break;
      }
      lambda = eproj;

      ////////////////////////////
      //Bunching parallel regions
      ////////////////////////////
      //psi = costheta*psi + sintheta*cg;
      //hpsi = costheta*hpsi + sintheta*hcg;
      //g = lambda*psi - hpsi;
      //residue  = dot(g,g);
      //states.assign(iband, psi.begin());// copy psi to iband
      ////////////////////////////

      residue = 0.0;
#pragma omp parallel firstprivate(costheta,sintheta,lambda) reduction(+:residue)
      {
        int myID = omp_get_thread_num();
        const int ibi = LocLoc[myID];
        const int ibf = LocLoc[myID+1];

        for(int ib = ibi;ib < ibf; ib++)
          psi[ib]=costheta*psi[ib]+sintheta*cg[ib];

        for(int ib = ibi;ib < ibf; ib++)
	  hpsi[ib]=costheta*hpsi[ib]+sintheta*hcg[ib];

        T* xref = states[iband];
        for(int ib=ibi;ib<ibf;ib++) 
          xref[ib] = psi[ib];

        for(int ib=ibi;ib<ibf;ib++) {
          g[ib] = lambda*psi[ib] - hpsi[ib];
	  residue += g[ib]*g[ib];
	}
      }

      /////////////////////////////////////////////
      //no precondition is used
      //states.orthogonalize(g.begin(), 0, iband);
      ////////////////////////////////////////////
      ham.precondition(g.begin(),states, iband);

      //double x1 = dot(g,g) - dot(gprev,g);
      //double x2 = dot(g,cg) - dot(gprev,cg);
      T x1 = 0.0; 
      T x2 = 0.0;
#pragma omp parallel reduction(+:x1,x2)
      {
        int myID = omp_get_thread_num();
        const int ibi = LocLoc[myID];
        const int ibf = LocLoc[myID+1];
        for(int ib=ibi; ib<ibf; ib++) {
          x1 += g[ib]*(g[ib]-gprev[ib]);
	  x2 += cg[ib]*(g[ib]-gprev[ib]);
	}
      }
      gamma = -x1/(x2+1e-40);
      DEBUGMSG(iter << " " << eproj << " " << residue)

#pragma omp parallel firstprivate(gamma)
      {
        int myID = omp_get_thread_num();
        const int ibi = LocLoc[myID];
        const int ibf = LocLoc[myID+1];
        for(int ib=ibi; ib<ibf; ib++) cg[ib] = g[ib] + gamma*cg[ib];
        for(int ib=ibi; ib<ibf; ib++) gprev[ib] = g[ib];
      }

      states.orthogonalize(cg.begin(), iband, iband);

///////////////////////////////////////
// normalize(cg);
///////////////////////////////////////
      cgnorm = 0.0;
#pragma omp parallel reduction(+:cgnorm)
    {
       int myID = omp_get_thread_num();
       const int ibi = LocLoc[myID];
       const int ibf = LocLoc[myID+1];
       for(int ib = ibi; ib<ibf; ib++) cgnorm += cg[ib]*cg[ib];

    }
    cgnorm = 1.0/sqrt(cgnorm);
#pragma omp parallel firstprivate(cgnorm)
    {
      int myID = omp_get_thread_num();
       const int ibi = LocLoc[myID];
       const int ibf = LocLoc[myID+1];
       for(int ib = ibi; ib<ibf; ib++) cg[ib] *= cgnorm;
    }

///////////////////////////////////////

      iter++;
    } while(iter < maxiter);

    if(iter < maxiter) {
      cout << "Iteration = " << iter << endl;
      //ham.apply(psi.begin(), hpsi.begin());
      //dot(psi,hpsi);
      eigval[iband] = lambda;
      iband++;
    } else {
      // did not converged, start with a new random vector
      states.randomize(iband);
    }
  } while(iband < nband);
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: BandByBandConjGradientOMP.h 151 2004-11-02 23:51:14Z jnkim $
 ************************************************************************/
