//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#ifndef OHMMS_LENNARDJONE_SCALE_H
#define OHMMS_LENNARDJONE_SCALE_H

#include <math.h>
#include <algorithm>

template<class T>
struct LJScale {
  T Eps;
  T Sigma;
  T Acut, Rcut, Rmin;

  LJScale(T eps = 0.0206185567, T sig = 3.4, 
	  T acut = 4, T rcut = 5, T rmin=1e-6): 
    Eps(eps), Sigma(sig), Acut(acut), Rcut(rcut), Rmin(rmin) {
  }
  ~LJScale(){}

  template<class Iter>
  void set(Iter first, Iter last) {
    Iter findit = std::find(first,last,"eps");
    if(findit != last) Eps = atof((*++findit).c_str());
    findit = std::find(first,last,"sig");
    if(findit != last) Sigma = atof((*++findit).c_str());
    findit = std::find(first,last,"rmin");
    if(findit != last) Rmin = atof((*++findit).c_str());
    findit = std::find(first,last,"rcut");
    if(findit != last) Rcut = atof((*++findit).c_str());
    findit = std::find(first,last,"acut");
    if(findit != last) Acut = atof((*++findit).c_str());
  }

  inline T rmax() const { return 1.1*Rcut;}
  inline T rmin() const { return Rmin;}

  inline T f(T r) const {
    T sigr=pow(Sigma/r,6.0);
    return (Eps*sigr*(sigr-1.0e0))/(1+exp(Acut*(r-Rcut)));
  }
  inline T df(T r, T& y) const {
    T rinv = 1.0/r;
    T lj6  = pow(Sigma*rinv,6.0);
    T lj12 = Eps*lj6*lj6;
    lj6 *= Eps;
    T fac = exp(Acut*(r-Rcut));
    T cut = 1.0e0/(1+fac);
    y = (lj12-lj6)*cut;
    return -(6*(2*lj12-lj6)*rinv+y*Acut*fac)*cut;
  }

  /*!\fn T d2f(T r, T& y, T yp)
   *\param r input distance
   *\param y value at r
   *\param yp (df/dr)(1/r)
   *\return d^2f/dr^2
   */
  T d2f(T r, T& y, T& yp) {
    T rinv = 1.0/r;
    T lj6  = pow(Sigma*rinv,6.0);
    T lj12 = Eps*lj6*lj6;
    lj6 *= Eps;
    T fac = exp(Acut*(r-Rcut));
    T cut = 1.0e0/(1+fac);
    y = (lj12-lj6)*cut;
    yp = -(6*(2*lj12-lj6)*rinv+y*Acut*fac)*cut*rinv;
    //return  cut*(156*lj12-42*lj6)*rinv*rinv 
    //      + 12*(2*lj12-lj6)*rinv*Acut*fac*cut*cut
    //      + (lj12-lj6)*Acut*acut*fac*(fac-1)*cut*cut*cut;
    return  cut*(156*lj12-42*lj6)*rinv*rinv 
          + (12*(2*lj12-lj6)*rinv + (lj12-lj6)*Acut*(fac-1)*cut)*Acut*fac*cut*cut;
  }
};


#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * OHMMS_VERSION_ID: $Id: LJScale.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
