//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/***************************************************************************
 *
 ***************************************************************************/

#ifndef SLATERKOSTER_PD_H
#define SLATERKOSTER_PD_H

template<class T, class SF>
class SlaterKoster<T,1,2,SF>: public SlaterKosterBase<T> {
private:

  int I[3][5], J[5][3];
  bool Symm, Swap;
  SF* V;
public:

  using SlaterKosterBase<T>::I0;
  using SlaterKosterBase<T>::J0;

  SlaterKoster(int i0, int j0, SF*, bool symm=true, bool swap = false);
  void getOffsite(T r, const T* p,  T* b);
  void getDeriv(T r, const T* , const T [][3], T* bx, T* by, T* bz);
  void getDeriv2(T r, const T*, const T [][3], const T [][3][3], 
       T* bx,int i, int j);
  void assign(T*);
  void reset(int nrow, int ncol);
};

///////////////////////////////////////////////////////////////////
template<class T, class SF>
SlaterKoster<T,1,2,SF>::SlaterKoster(int i0, int j0, SF* v, bool symm, bool swap)
: V(v), Symm(symm), Swap(swap){

  I0 = i0; J0 = j0;
}

template<class T, class SF>
void 
SlaterKoster<T,1,2,SF>::reset(int nrow, int ncol) {
  if(Swap) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++) 
        I[i][j] = (i+I0)+(j+J0)*ncol;
    for(int j=0; j<5; j++) 
      for(int i=0; i<3; i++)
        J[j][i] = (j+J0)+(i+I0)*ncol;
  } else {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++) 
        I[i][j] = (j+J0)+(i+I0)*ncol;

    for(int j=0; j<5; j++) 
      for(int i=0; i<3; i++)
        J[j][i] = (i+I0)+(j+J0)*ncol;
  }
}

template<class T, class SF>
void 
SlaterKoster<T,1,2,SF>::assign(T* b) {
  for(int i=0; i < 3; i++) 
    for(int j=0; j < 5; j++)
      b[I[i][j]] = static_cast<T>(I[i][j]);

  if(Symm) {
    for(int j=0; j < 5; j++)
      for(int i=0; i < 3; i++) 
        b[J[j][i]] = static_cast<T>(J[j][i]);
  }
}

template<class T, class SF>
void 
SlaterKoster<T,1,2,SF>::getOffsite(T r, const T* p,T* b) {
  //T vsig = Vsnorm*Vsig->f(r);
  //T vpi  = Vpnorm*Vpi->f(r);
  T vsig = V[0](r);
  T vpi  = V[1](r);
  T v1 = sqrt3*vsig - 2.0e0*vpi;  // sqrt(3) Vpd_sig - 2*Vpd_pi
  T v2 = vpi;                     // Vpd_pi/sqrt(3)
  T v23 = v2*oneoversqrt3;

  b[I[0][0]] += v1*p[SK_Px]*p[SK_Dxy  ] + v2 *p[SK_Py];
  b[I[0][1]] += v1*p[SK_Px]*p[SK_Dyz  ]           ; 
  b[I[0][2]] += v1*p[SK_Px]*p[SK_Dzx  ] + v2 *p[SK_Pz]; 
  b[I[0][3]] += v1*p[SK_Px]*p[SK_Dx2y2] + v2 *p[SK_Px];
  b[I[0][4]] += v1*p[SK_Px]*p[SK_Dz2r2] - v23*p[SK_Px];

  b[I[1][0]] += v1*p[SK_Py]*p[SK_Dxy  ] + v2 *p[SK_Px];
  b[I[1][1]] += v1*p[SK_Py]*p[SK_Dyz  ] + v2 *p[SK_Pz]; 
  b[I[1][2]] += v1*p[SK_Py]*p[SK_Dzx  ]            ;
  b[I[1][3]] += v1*p[SK_Py]*p[SK_Dx2y2] - v2 *p[SK_Py];
  b[I[1][4]] += v1*p[SK_Py]*p[SK_Dz2r2] - v23*p[SK_Py];

  b[I[2][0]] += v1*p[SK_Pz]*p[SK_Dxy  ]            ;
  b[I[2][1]] += v1*p[SK_Pz]*p[SK_Dyz  ] + v2 *p[SK_Py]; 
  b[I[2][2]] += v1*p[SK_Pz]*p[SK_Dzx  ] + v2 *p[SK_Px];
  b[I[2][3]] += v1*p[SK_Pz]*p[SK_Dx2y2]            ;
  b[I[2][4]] += v1*p[SK_Pz]*p[SK_Dz2r2] + v23*p[SK_Pz]*2.0e0;
  if(Symm) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++)
        b[J[j][i]]  = -b[I[i][j]];
  }
}

///////////////////////////////////////////////////////////////////
template<class T, class SF>
void 
SlaterKoster<T,1,2,SF>::getDeriv(T r, const T* p, const T pd[][3], 
                                  T* bx, T* by, T* bz) {
  T vsig, vpi;
  T dvsig = V[0](r,vsig);
  T dvpi  = V[1](r,vpi);
  //T dvsig = Vsnorm*Vsig->df(r,vsig);
  //T dvpi  = Vpnorm*Vpi->df(r,vpi);

  T v1 = sqrt3*vsig - 2.0*vpi;  // (Vpd_sig - 2/sqrt(3)*Vpd_pi)/r
  T v2 = vpi;                   // (Vpd_pi)/r
  T v23 = v2*oneoversqrt3;      // (Vpd_pi/sqrt(3))/r

  T dv1 = sqrt3*dvsig - 2.0e0*dvpi; // d(v1)/dr
  T dv2 = dvpi;                     // d(v2)/dr
  T dv23 = dv2*oneoversqrt3;        // d(v2/sqrt(3))/dr

  // common block
  T a00 = dv1*p[SK_Px]*p[SK_Dxy  ] + dv2 *p[SK_Py]; // m
  T a01 = dv1*p[SK_Px]*p[SK_Dyz  ]             ; // 0
  T a02 = dv1*p[SK_Px]*p[SK_Dzx  ] + dv2 *p[SK_Pz]; // n
  T a03 = dv1*p[SK_Px]*p[SK_Dx2y2] + dv2 *p[SK_Px]; // l
  T a04 = dv1*p[SK_Px]*p[SK_Dz2r2] - dv23*p[SK_Px]; // -l/sqrt(3)

  T a10 = dv1*p[SK_Py]*p[SK_Dxy  ] + dv2 *p[SK_Px]; // l
  T a11 = dv1*p[SK_Py]*p[SK_Dyz  ] + dv2 *p[SK_Pz]; // n
  T a12 = dv1*p[SK_Py]*p[SK_Dzx  ]             ; // 0
  T a13 = dv1*p[SK_Py]*p[SK_Dx2y2] - dv2 *p[SK_Py]; // -m
  T a14 = dv1*p[SK_Py]*p[SK_Dz2r2] - dv23*p[SK_Py]; // -m/sqrt(3)

  T a20 = dv1*p[SK_Pz]*p[SK_Dxy  ]             ; // 0
  T a21 = dv1*p[SK_Pz]*p[SK_Dyz  ] + dv2 *p[SK_Py]; // m
  T a22 = dv1*p[SK_Pz]*p[SK_Dzx  ] + dv2 *p[SK_Px]; // l
  T a23 = dv1*p[SK_Pz]*p[SK_Dx2y2]             ; // 0
  T a24 = dv1*p[SK_Pz]*p[SK_Dz2r2] + dv23*p[SK_Pz]*2.0e0; // 2n/sqrt(3)

  // x-deriv
  bx[I[0][0]]+=a00*p[SK_Px]+v1*(pd[SK_Px][0]*p[SK_Dxy  ]+p[SK_Px]*pd[SK_Dxy  ][0])+v2 *pd[SK_Py][0];
  bx[I[0][1]]+=a01*p[SK_Px]+v1*(pd[SK_Px][0]*p[SK_Dyz  ]+p[SK_Px]*pd[SK_Dyz  ][0])              ; 
  bx[I[0][2]]+=a02*p[SK_Px]+v1*(pd[SK_Px][0]*p[SK_Dzx  ]+p[SK_Px]*pd[SK_Dzx  ][0])+v2 *pd[SK_Pz][0];
  bx[I[0][3]]+=a03*p[SK_Px]+v1*(pd[SK_Px][0]*p[SK_Dx2y2]+p[SK_Px]*pd[SK_Dx2y2][0])+v2 *pd[SK_Px][0];
  bx[I[0][4]]+=a04*p[SK_Px]+v1*(pd[SK_Px][0]*p[SK_Dz2r2]+p[SK_Px]*pd[SK_Dz2r2][0])-v23*pd[SK_Px][0];
	       		 		       			    
  bx[I[1][0]]+=a10*p[SK_Px]+v1*(pd[SK_Py][0]*p[SK_Dxy  ]+p[SK_Py]*pd[SK_Dxy  ][0])+v2 *pd[SK_Px][0];
  bx[I[1][1]]+=a11*p[SK_Px]+v1*(pd[SK_Py][0]*p[SK_Dyz  ]+p[SK_Py]*pd[SK_Dyz  ][0])+v2 *pd[SK_Pz][0]; 
  bx[I[1][2]]+=a12*p[SK_Px]+v1*(pd[SK_Py][0]*p[SK_Dzx  ]+p[SK_Py]*pd[SK_Dzx  ][0])              ;
  bx[I[1][3]]+=a13*p[SK_Px]+v1*(pd[SK_Py][0]*p[SK_Dx2y2]+p[SK_Py]*pd[SK_Dx2y2][0])-v2 *pd[SK_Py][0];
  bx[I[1][4]]+=a14*p[SK_Px]+v1*(pd[SK_Py][0]*p[SK_Dz2r2]+p[SK_Py]*pd[SK_Dz2r2][0])-v23*pd[SK_Py][0];
	       		 		       			    
  bx[I[2][0]]+=a20*p[SK_Px]+v1*(pd[SK_Pz][0]*p[SK_Dxy  ]+p[SK_Pz]*pd[SK_Dxy  ][0])              ;
  bx[I[2][1]]+=a21*p[SK_Px]+v1*(pd[SK_Pz][0]*p[SK_Dyz  ]+p[SK_Pz]*pd[SK_Dyz  ][0])+v2 *pd[SK_Py][0]; 
  bx[I[2][2]]+=a22*p[SK_Px]+v1*(pd[SK_Pz][0]*p[SK_Dzx  ]+p[SK_Pz]*pd[SK_Dzx  ][0])+v2 *pd[SK_Px][0];
  bx[I[2][3]]+=a23*p[SK_Px]+v1*(pd[SK_Pz][0]*p[SK_Dx2y2]+p[SK_Pz]*pd[SK_Dx2y2][0])              ;
  bx[I[2][4]]+=a24*p[SK_Px]+v1*(pd[SK_Pz][0]*p[SK_Dz2r2]+p[SK_Pz]*pd[SK_Dz2r2][0])+v23*pd[SK_Pz][0]*2.0e0;
  if(Symm) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++)
        bx[J[j][i]] = -bx[I[i][j]];
  }

  // y-deriv
  by[I[0][0]]+=a00*p[SK_Py]+v1*(pd[SK_Px][1]*p[SK_Dxy  ]+p[SK_Px]*pd[SK_Dxy  ][1])+v2 *pd[SK_Py][1];
  by[I[0][1]]+=a01*p[SK_Py]+v1*(pd[SK_Px][1]*p[SK_Dyz  ]+p[SK_Px]*pd[SK_Dyz  ][1])              ; 
  by[I[0][2]]+=a02*p[SK_Py]+v1*(pd[SK_Px][1]*p[SK_Dzx  ]+p[SK_Px]*pd[SK_Dzx  ][1])+v2 *pd[SK_Pz][1];
  by[I[0][3]]+=a03*p[SK_Py]+v1*(pd[SK_Px][1]*p[SK_Dx2y2]+p[SK_Px]*pd[SK_Dx2y2][1])+v2 *pd[SK_Px][1];
  by[I[0][4]]+=a04*p[SK_Py]+v1*(pd[SK_Px][1]*p[SK_Dz2r2]+p[SK_Px]*pd[SK_Dz2r2][1])-v23*pd[SK_Px][1];
	       		 		       			    
  by[I[1][0]]+=a10*p[SK_Py]+v1*(pd[SK_Py][1]*p[SK_Dxy  ]+p[SK_Py]*pd[SK_Dxy  ][1])+v2 *pd[SK_Px][1];
  by[I[1][1]]+=a11*p[SK_Py]+v1*(pd[SK_Py][1]*p[SK_Dyz  ]+p[SK_Py]*pd[SK_Dyz  ][1])+v2 *pd[SK_Pz][1]; 
  by[I[1][2]]+=a12*p[SK_Py]+v1*(pd[SK_Py][1]*p[SK_Dzx  ]+p[SK_Py]*pd[SK_Dzx  ][1])              ;
  by[I[1][3]]+=a13*p[SK_Py]+v1*(pd[SK_Py][1]*p[SK_Dx2y2]+p[SK_Py]*pd[SK_Dx2y2][1])-v2 *pd[SK_Py][1];
  by[I[1][4]]+=a14*p[SK_Py]+v1*(pd[SK_Py][1]*p[SK_Dz2r2]+p[SK_Py]*pd[SK_Dz2r2][1])-v23*pd[SK_Py][1];
	       		 		       			    
  by[I[2][0]]+=a20*p[SK_Py]+v1*(pd[SK_Pz][1]*p[SK_Dxy  ]+p[SK_Pz]*pd[SK_Dxy  ][1])              ;
  by[I[2][1]]+=a21*p[SK_Py]+v1*(pd[SK_Pz][1]*p[SK_Dyz  ]+p[SK_Pz]*pd[SK_Dyz  ][1])+v2 *pd[SK_Py][1]; 
  by[I[2][2]]+=a22*p[SK_Py]+v1*(pd[SK_Pz][1]*p[SK_Dzx  ]+p[SK_Pz]*pd[SK_Dzx  ][1])+v2 *pd[SK_Px][1];
  by[I[2][3]]+=a23*p[SK_Py]+v1*(pd[SK_Pz][1]*p[SK_Dx2y2]+p[SK_Pz]*pd[SK_Dx2y2][1])              ;
  by[I[2][4]]+=a24*p[SK_Py]+v1*(pd[SK_Pz][1]*p[SK_Dz2r2]+p[SK_Pz]*pd[SK_Dz2r2][1])+v23*pd[SK_Pz][1]*2.0e0;
  if(Symm) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++)
        by[J[j][i]] = -by[I[i][j]];
  }

  // z-deriv
  bz[I[0][0]]+=a00*p[SK_Pz]+v1*(pd[SK_Px][2]*p[SK_Dxy  ]+p[SK_Px]*pd[SK_Dxy  ][2])+v2 *pd[SK_Py][2];
  bz[I[0][1]]+=a01*p[SK_Pz]+v1*(pd[SK_Px][2]*p[SK_Dyz  ]+p[SK_Px]*pd[SK_Dyz  ][2])              ; 
  bz[I[0][2]]+=a02*p[SK_Pz]+v1*(pd[SK_Px][2]*p[SK_Dzx  ]+p[SK_Px]*pd[SK_Dzx  ][2])+v2 *pd[SK_Pz][2];
  bz[I[0][3]]+=a03*p[SK_Pz]+v1*(pd[SK_Px][2]*p[SK_Dx2y2]+p[SK_Px]*pd[SK_Dx2y2][2])+v2 *pd[SK_Px][2];
  bz[I[0][4]]+=a04*p[SK_Pz]+v1*(pd[SK_Px][2]*p[SK_Dz2r2]+p[SK_Px]*pd[SK_Dz2r2][2])-v23*pd[SK_Px][2];
	       		 		       			    
  bz[I[1][0]]+=a10*p[SK_Pz]+v1*(pd[SK_Py][2]*p[SK_Dxy  ]+p[SK_Py]*pd[SK_Dxy  ][2])+v2 *pd[SK_Px][2];
  bz[I[1][1]]+=a11*p[SK_Pz]+v1*(pd[SK_Py][2]*p[SK_Dyz  ]+p[SK_Py]*pd[SK_Dyz  ][2])+v2 *pd[SK_Pz][2]; 
  bz[I[1][2]]+=a12*p[SK_Pz]+v1*(pd[SK_Py][2]*p[SK_Dzx  ]+p[SK_Py]*pd[SK_Dzx  ][2])              ;
  bz[I[1][3]]+=a13*p[SK_Pz]+v1*(pd[SK_Py][2]*p[SK_Dx2y2]+p[SK_Py]*pd[SK_Dx2y2][2])-v2 *pd[SK_Py][2];
  bz[I[1][4]]+=a14*p[SK_Pz]+v1*(pd[SK_Py][2]*p[SK_Dz2r2]+p[SK_Py]*pd[SK_Dz2r2][2])-v23*pd[SK_Py][2];
	       		 		       			    
  bz[I[2][0]]+=a20*p[SK_Pz]+v1*(pd[SK_Pz][2]*p[SK_Dxy  ]+p[SK_Pz]*pd[SK_Dxy  ][2])              ;
  bz[I[2][1]]+=a21*p[SK_Pz]+v1*(pd[SK_Pz][2]*p[SK_Dyz  ]+p[SK_Pz]*pd[SK_Dyz  ][2])+v2 *pd[SK_Py][2]; 
  bz[I[2][2]]+=a22*p[SK_Pz]+v1*(pd[SK_Pz][2]*p[SK_Dzx  ]+p[SK_Pz]*pd[SK_Dzx  ][2])+v2 *pd[SK_Px][2];
  bz[I[2][3]]+=a23*p[SK_Pz]+v1*(pd[SK_Pz][2]*p[SK_Dx2y2]+p[SK_Pz]*pd[SK_Dx2y2][2])              ;
  bz[I[2][4]]+=a24*p[SK_Pz]+v1*(pd[SK_Pz][2]*p[SK_Dz2r2]+p[SK_Pz]*pd[SK_Dz2r2][2])+v23*pd[SK_Pz][2]*2.0e0;
  if(Symm) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++)
        bz[J[j][i]] = -bz[I[i][j]];
  }

}

template<class T, class SF>
void 
SlaterKoster<T,1,2,SF>::getDeriv2(T r, const T* p, const T pd[][3], 
                                  const T pdd[][3][3],  T* b, int d0, int d1) {

  T vsig, vpi, dvsig, dvpi;
  //T d2vsig = Vsnorm*Vsig->d2f(r,vsig,dvsig);
  //T d2vpi  = Vpnorm*Vpi->d2f(r,vpi,dvpi);
  T d2vsig = V[0](r,vsig,dvsig);
  T d2vpi  = V[1](r,vpi,dvpi);
  
  T v1 = sqrt3*vsig - 2.0e0*vpi;   //(sqrt(3)Vpd_sig - 2Vpd_pi)/r
  T dv1 = sqrt3*dvsig - 2.0e0*vpi; // d(v1)/dr
  T d2v1= (sqrt3*d2vsig-2.0e0*d2vpi)*p[d0]*p[d1];

  for(int k=0; k<3; k++)
    for(int j=0; j< 5; j++) {
      int kp = j+SK_Dxy;
      b[I[k][j]]+=(d2v1*p[k]*p[kp]
           	+dv1*( p[k]*p[kp]*pd[d0][d1] 
                     + p[d0]*( pd[k][d1]*p[kp] + p[k]*pd[kp][d1] )
                     + p[d1]*( pd[k][d0]*p[kp] + p[k]*pd[kp][d0] ))
                +v1 *( pdd[k ][d0][d1]*p[kp] + pd[k][d0]*pd[kp][d1]
                     + pdd[kp][d0][d1]*p[k ] + pd[k][d1]*pd[kp][d0]));
    }

  //extra terms are added
  T v2 = vpi;    // (Vpd_pi)/r
  T dv2 = dvpi;  // d(v2)/dr
  T d2v2 = d2vpi*p[d0]*p[d1];

  T l_ij=d2v2*p[0]+dv2*(pd[0][d0]*p[d1]+pd[0][d1]*p[d0]+p[0]*pd[d0][d1])+v2*pdd[0][d0][d1];
  T m_ij=d2v2*p[1]+dv2*(pd[1][d0]*p[d1]+pd[1][d1]*p[d0]+p[1]*pd[d0][d1])+v2*pdd[1][d0][d1];
  T n_ij=d2v2*p[2]+dv2*(pd[2][d0]*p[d1]+pd[2][d1]*p[d0]+p[2]*pd[d0][d1])+v2*pdd[2][d0][d1];
  b[I[0][0]]+=m_ij; b[I[0][2]]+=n_ij; b[I[0][3]]+=l_ij; b[I[0][4]]-=oneoversqrt3*l_ij;
  b[I[1][0]]+=l_ij; b[I[1][1]]+=n_ij; b[I[1][3]]-=m_ij; b[I[1][4]]-=oneoversqrt3*m_ij;
  b[I[2][1]]+=m_ij; b[I[2][2]]+=l_ij; b[I[2][4]]+= twooversqrt3*n_ij;

  if(Symm) {
    for(int i=0; i<3; i++)
      for(int j=0; j<5; j++)
        b[J[j][i]] = -b[I[i][j]];
  }

}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: SlaterKosterPD.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
