//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/***************************************************************************
 *
 ***************************************************************************/

#ifndef SLATERKOSTER_SD_H
#define SLATERKOSTER_SD_H

template<class T, class SF>
class SlaterKoster<T,0,2,SF>: public SlaterKosterBase<T> {

private:

  int I[5], J[5];
  bool Symm, Swap;
  SF* V;

public:

  using SlaterKosterBase<T>::I0;
  using SlaterKosterBase<T>::J0;

  SlaterKoster(int i0, int j0, SF*, bool symm=true, bool swap=false);

  void getOffsite(T r, const T* p,  T* b);
  void getDeriv(T r, const T* , const T [][3], T* bx, T* by, T* bz);
  void getDeriv2(T r, const T*, const T [][3], const T [][3][3], 
       T* bx,int i, int j);
  void assign(T*);
  void reset(int nrow, int ncol);
};

///////////////////////////////////////////////////////////////////////////////
// S-D block
///////////////////////////////////////////////////////////////////////////////
template<class T, class SF>
SlaterKoster<T,0,2,SF>::SlaterKoster(int i, int j, SF* v, bool symm, bool swap)
: V(v), Symm(symm), Swap(swap){
  I0 = i; J0 = j;
}

template<class T, class SF>
void 
SlaterKoster<T,0,2,SF>::reset(int nrow, int ncol) {
  // I -> principal index, J-> transpose index for Symmetric case
  if(Swap) {
    for(int j=0; j<5; j++) J[j] = j+J0 + I0*ncol;
    for(int j=0; j<5; j++) I[j] = I0 + (j+J0)*ncol;
  } else { 
    for(int j=0; j<5; j++) I[j] = j+J0 + I0*ncol;
    for(int j=0; j<5; j++) J[j] = I0 + (j+J0)*ncol;
  }
}

template<class T, class SF>
void 
SlaterKoster<T,0,2,SF>::assign(T* b) {
  for(int j=0; j<5; j++)
    b[I[j]] = static_cast<T>(I[j]);
  if(Symm) {
    for(int j=0; j<5; j++)
      b[J[j]] = static_cast<T>(J[j]);
  }
}

template<class T, class SF>
void 
SlaterKoster<T,0,2,SF>::getOffsite(T r, const T* p, T* b) {

  //T vsd = Vnorm*V->f(r);
  T vsd = sqrt3*V[0](r);
  b[I[0]]+= vsd*p[SK_Dxy];
  b[I[1]]+= vsd*p[SK_Dyz];
  b[I[2]]+= vsd*p[SK_Dzx];
  b[I[3]]+= vsd*p[SK_Dx2y2];
  b[I[4]]+= vsd*p[SK_Dz2r2];

  if(Symm) {
    b[J[0]]=b[I[0]];
    b[J[1]]=b[I[1]];
    b[J[2]]=b[I[2]];
    b[J[3]]=b[I[3]];
    b[J[4]]=b[I[4]];
  }
}

template<class T, class SF>
void 
SlaterKoster<T,0,2,SF>::getDeriv(T r, const T* p, const T pd[][3],
                               T* bx, T* by, T* bz) {

  T v;
  //T dv = Vnorm*V->df(r,v); // df returns a derivative and f(r)/r
  //v *= Vnorm;
  T dv =sqrt3*V[0](r,v);
  v *= sqrt3;

  // x-derivate
  T dvl = p[SK_Px]*dv;
  bx[I[0]] += v*pd[SK_Dxy  ][0] + dvl*p[SK_Dxy  ];
  bx[I[1]] += v*pd[SK_Dyz  ][0] + dvl*p[SK_Dyz  ];
  bx[I[2]] += v*pd[SK_Dzx  ][0] + dvl*p[SK_Dzx  ];
  bx[I[3]] += v*pd[SK_Dx2y2][0] + dvl*p[SK_Dx2y2]; 
  bx[I[4]] += v*pd[SK_Dz2r2][0] + dvl*p[SK_Dz2r2];

  if(Symm) {
    bx[J[0]]=bx[I[0]];
    bx[J[1]]=bx[I[1]];
    bx[J[2]]=bx[I[2]];
    bx[J[3]]=bx[I[3]];
    bx[J[4]]=bx[I[4]];
  }

  // y-derivate
  dvl = p[SK_Py]*dv;
  by[I[0]] += v*pd[SK_Dxy  ][1] + dvl*p[SK_Dxy  ];
  by[I[1]] += v*pd[SK_Dyz  ][1] + dvl*p[SK_Dyz  ];
  by[I[2]] += v*pd[SK_Dzx  ][1] + dvl*p[SK_Dzx  ];
  by[I[3]] += v*pd[SK_Dx2y2][1] + dvl*p[SK_Dx2y2]; 
  by[I[4]] += v*pd[SK_Dz2r2][1] + dvl*p[SK_Dz2r2];
  if(Symm) {
    by[J[0]]=by[I[0]];
    by[J[1]]=by[I[1]];
    by[J[2]]=by[I[2]];
    by[J[3]]=by[I[3]];
    by[J[4]]=by[I[4]];
  }

  // z-derivate
  dvl = p[SK_Pz]*dv;
  bz[I[0]] += v*pd[SK_Dxy  ][2] + dvl*p[SK_Dxy  ];
  bz[I[1]] += v*pd[SK_Dyz  ][2] + dvl*p[SK_Dyz  ];
  bz[I[2]] += v*pd[SK_Dzx  ][2] + dvl*p[SK_Dzx  ];
  bz[I[3]] += v*pd[SK_Dx2y2][2] + dvl*p[SK_Dx2y2]; 
  bz[I[4]] += v*pd[SK_Dz2r2][2] + dvl*p[SK_Dz2r2];
  if(Symm) {
    bz[J[0]]=bz[I[0]];
    bz[J[1]]=bz[I[1]];
    bz[J[2]]=bz[I[2]];
    bz[J[3]]=bz[I[3]];
    bz[J[4]]=bz[I[4]];
  }
}

template<class T, class SF>
void 
SlaterKoster<T,0,2,SF>::getDeriv2(T r, const T* p, const T pd[][3], 
                                   const T pdd[][3][3],  T* b, int d0, int d1) {

  T v,dv;
  //T dv2 = Vnorm*V->d2f(r,v,dv); // df returns a derivative and f(r)/r
  //v *= Vnorm;
  //dv *= Vnorm;
  T dv2 = sqrt3*V[0](r,v,dv);
  v *= sqrt3;
  dv *= sqrt3;

  dv2 *= (p[d0]*p[d1]);

  b[I[0]] += v*pdd[SK_Dxy  ][d0][d1]+dv*(pd[d0][d1]*p[SK_Dxy  ]+p[d0]*pd[SK_Dxy  ][d1]+p[d1]*pd[SK_Dxy  ][d0])+dv2*p[SK_Dxy  ];
  b[I[1]] += v*pdd[SK_Dyz  ][d0][d1]+dv*(pd[d0][d1]*p[SK_Dyz  ]+p[d0]*pd[SK_Dyz  ][d1]+p[d1]*pd[SK_Dyz  ][d0])+dv2*p[SK_Dyz  ];
  b[I[2]] += v*pdd[SK_Dzx  ][d0][d1]+dv*(pd[d0][d1]*p[SK_Dzx  ]+p[d0]*pd[SK_Dzx  ][d1]+p[d1]*pd[SK_Dzx  ][d0])+dv2*p[SK_Dzx  ]; 
  b[I[3]] += v*pdd[SK_Dx2y2][d0][d1]+dv*(pd[d0][d1]*p[SK_Dx2y2]+p[d0]*pd[SK_Dx2y2][d1]+p[d1]*pd[SK_Dx2y2][d0])+dv2*p[SK_Dx2y2]; 
  b[I[4]] += v*pdd[SK_Dz2r2][d0][d1]+dv*(pd[d0][d1]*p[SK_Dz2r2]+p[d0]*pd[SK_Dz2r2][d1]+p[d1]*pd[SK_Dz2r2][d0])+dv2*p[SK_Dz2r2]; 

  if(Symm) {
    b[J[0]]=b[I[0]];
    b[J[1]]=b[I[1]];
    b[J[2]]=b[I[2]];
    b[J[3]]=b[I[3]];
    b[J[4]]=b[I[4]];
  }
}
#endif


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 378 $   $Date: 2005-08-28 18:40:33 -0500 (Sun, 28 Aug 2005) $
 * $Id: SlaterKosterSD.h 378 2005-08-28 23:40:33Z jnkim $ 
 ***************************************************************************/
