//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
////////////////////////////////////////////////////////////////////////////
// NRL potential for Transition Metals
// Mike Mehl and Dimitri Papaconstantopoulus
// PRB xx
//
// Modified form of hopping terms, to use a smooth spline to a constant
//   value (see SplineScaleFunction.h for details and implementation).
//
// Implemented by D. Trinkle, April 10, 2001.
// Modified by Jeongnim Kim, May 24, 2002
// Modified by D. Trinkle January 19, 2003 -- added quadratic to 
//   SplineScaleFunction, and automagically detect according to rmin value
// Modified by D. Trinkle April 7, 2003 -- added spline functionality
//   to work with new overlap form, and automagically detect.
//   Definition of inmode values:
//   0: Hamil. and Overlap are quadratic * exponential
//   1: Hamil. is same as 0, Overlap is (delta_ll' + r*quadratic)*exponential
//   2: Same as 0 down to Rmin, then cubic spline takes over
//   3: Same as 1 down to Rmin, then cubic spline takes over
//
// For binary potentials, using identical r0, sigma and lambda does not make sense.
////////////////////////////////////////////////////////////////////////////

#include "Message/OpenMP.h"
#include "Utilities/OhmmsInfo.h"
#include "SlaterKosterTB/Models/NRLTB.h"
#include "SlaterKoster/SlaterKosterMake.h"
#include "Utilities/SimpleParser.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "Particle/NNEngine.h"
using namespace OHMMS;

NRLTB::NRLTB() { }

NRLTB::~NRLTB() {
  // release the pointer to scaling functions
  for(int i=0; i<V.size(); i++) delete V[i];
  for(int i=0; i<Vs_old.size(); i++) delete Vs_old[i];
  for(int i=0; i<Vs_new.size(); i++) delete Vs_new[i];

  for(int i=0; i<Vh.size(); i++) delete Vh[i];
  for(int i=0; i<Vs.size(); i++) delete Vs[i];
}

ShortRangePotential*
NRLTB::init(SKList_t& SK, SKList_t& SK_O, Matrix<int>& bond, istream& fin) {

  // make private references to Species
  SpeciesSet& mySpecies(myPtcl->getSpeciesSet());
  int irmax = mySpecies.addAttrib("rmax");

  int Lmax = 0;
  int ib, ia, ja, L1, L2;
  int l1, l2, lt1, lt2, m;
  Scalar_t rcut, screen, eps = 0.0;
  Scalar_t h_sigma = 1.0, h_rmin = 0.0, s_sigma = 1.0, s_rmin = 0.0;
  Scalar_t h0, h1, h2, h3;
  Scalar_t s0, s1, s2, s3;

  /////////////////////////////////////////////////
  // until NRL potential is extended to more than 2 species problem
  // Other possibility is that when NRLTB region is embbeded within 
  // other Potential region
  /////////////////////////////////////////////////
  const int maxspecies = bond.rows();
  bool hasoverlap = true;
  int inmode;

  vector<string> slist, vlist;
  getwords(slist,fin);
  if(slist[0][5]== '1') hasoverlap = false;
  inmode = atoi(slist[0].c_str()+6);
  WARNMSG("\tNRLTB mode = " << inmode)

  bool tofindend = true;
  int vloc = -1, sloc;
  do {
    int nw =getwords(slist, fin);
    if(nw == 0) continue;
    if(slist[0] == "bond") {
      ib = SK.size();
      ia = mySpecies.addSpecies(slist[1].c_str());
      ja = mySpecies.addSpecies(slist[2].c_str());
      bond(ia,ja) = ib;

      // create SK blocks for this bond
      SK.push_back(new SlaterKosterComp<Scalar_t>);
      if(hasoverlap) 
	SK_O.push_back(new SlaterKosterComp<Scalar_t>);

      // assign unphysical L1 and L2
      L1 = -1; L2 = -1;

    } else if(slist[0] == "rcut") {
      rcut = atof(slist[1].c_str());
    } else if(slist[0] == "screenl") {
      screen = atof(slist[1].c_str());
    } else if(slist[0] == "h_rmin") {
      h_rmin = atof(slist[1].c_str());
    } else if(slist[0] == "h_sigma") {
      h_sigma = atof(slist[1].c_str());
    } else if(slist[0] == "s_rmin") {
      s_rmin = atof(slist[1].c_str());
    } else if(slist[0] == "s_sigma") {
      s_sigma = atof(slist[1].c_str());
    } else if(isdigit(slist[0][0])) { 
      l1 = atoi(slist[0].c_str());
      l2 = atoi(slist[1].c_str());
      m  = atoi(slist[2].c_str());
      lt1 = atoi(slist[3].c_str());
      lt2 = atoi(slist[4].c_str());
      if(m<0) { // onsite matrix
        eps = atof(slist[5].c_str());
      } else {
        eps = 0.0;
	getwords(vlist,fin);
	h0  = atof(vlist[0].c_str());
	h1  = atof(vlist[1].c_str());
	h2  = atof(vlist[2].c_str());
	h3  = atof(vlist[3].c_str());
	if(hasoverlap) {
	  getwords(vlist,fin);
	  s0 = atof(vlist[0].c_str());
	  s1 = atof(vlist[1].c_str());
	  s2 = atof(vlist[2].c_str());
	  s3 = atof(vlist[3].c_str());
	}
      }
      // Automagically reset the type of scaling function if Rmin is
      // set for *both* hamiltonian and overlap (or just hamil. if
      // no overlap is present) --DRT
      if (h_rmin> 0.0)
	if ( (!hasoverlap) || (s_rmin >0.0) )
	  inmode = (inmode | 2);

      if(l1 != L1 || l2 != L2) {// new L1-L2 pair is added
	//cout << "eps " << eps << endl;
	//will create [0,min(l,l')] function blocks
	// first add a new scaling function for l-l' pair
	if(inmode | 2) {// NRLscale + spline (Trinkle et atl)
	  vloc = Vh.size();
	  Vh.push_back(new vector<TiScaleFunction_t>);
	  SK[ib]->add(SKBlock<Scalar_t,TiScaleFunction_t>
		      ::make(l1,l2,ia,ja,lt1,lt2,eps,Vh[vloc]));//
	} else {
	  vloc = V.size();
	  V.push_back(new vector<ScaleFunction_t>);
	  SK[ib]->add(SKBlock<Scalar_t,ScaleFunction_t>
		      ::make(l1,l2,ia,ja,lt1,lt2,eps,V[vloc]));
	}

	SK[ib]->setRcut(rcut);
	
	if(hasoverlap) {// make SK for the overlap 
         switch (inmode) {
          case 0: // old overlap, no spline
            sloc = Vs_old.size();
            Vs_old.push_back(new vector<ScaleFunction_t>);
            SK_O[ib]->add(SKBlock<Scalar_t,ScaleFunction_t>
                          ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs_old[sloc]));
            break;
          case 1: // new overlap, no spline
            if(ia == ja) {// for A-A bond relation
              sloc = Vs_new.size();
              Vs_new.push_back(new vector<ScaleFunction2_t>);
              SK_O[ib]->add(SKBlock<Scalar_t,ScaleFunction2_t>
                            ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs_new[sloc]));
            } else { // for A-B bond, using the old format
              sloc = Vs_old.size();
              Vs_old.push_back(new vector<ScaleFunction_t>);
              SK_O[ib]->add(SKBlock<Scalar_t,ScaleFunction_t>
                            ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs_old[sloc]));
            }
            break;
          case 2: // old overlap + spline (Trinkle et al)
            sloc = Vs.size();
            Vs.push_back(new vector<TiScaleFunction_t>);
            SK_O[ib]->add(SKBlock<Scalar_t,TiScaleFunction_t>
                          ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs[sloc]));
            break;
          case 3: // new overlap + spline (Trinkle et al)
            if(ia == ja) {// for A-A bond relation
              sloc = Vs2.size();
              Vs2.push_back(new vector<TiScaleFunction2_t>);
              SK_O[ib]->add(SKBlock<Scalar_t,TiScaleFunction2_t>
                            ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs2[sloc]));
            } else { // for A-B bond, using the old format
              sloc = Vs.size();
              Vs.push_back(new vector<TiScaleFunction_t>);
              SK_O[ib]->add(SKBlock<Scalar_t,TiScaleFunction_t>
                          ::make(l1,l2,ia,ja,lt1,lt2,1.0,Vs[sloc]));
            }
            break;
          default: // Scream bloody murder
            cerr << "Bad value of inmode = " << inmode << endl;
          }
          SK_O[ib]->setRcut(rcut);
	} // overlap-matrix
      }

      L1 = l1; L2 = l2;
      if(m>=0) {
	if(inmode | 2) {
	  // Added h2 to SplineScale --DRT
	  Vh[vloc]->operator[](m).set(h0,h1,h2,h3,rcut,screen,h_rmin,h_sigma);
	} else {
	  V[vloc]->operator[](m).set(h0,h1,h2,h3,rcut,screen);
	}
	
	if(hasoverlap) {
          switch (inmode) {
          case 0: // old overlap, no spline
            Vs_old[sloc]->operator[](m).set(s0,s1,s2,s3,rcut,screen);
            break;
          case 1: // new overlap, no spline
            if(ia == ja) {
              Scalar_t td = 0.0;
              if(l1 == l2) td = 1.0;
              Vs_new[sloc]->operator[](m).set(td,s0,s1,s2,s3, rcut,screen);
            } else {
              Vs_old[sloc]->operator[](m).set(s0,s1,s2,s3, rcut,screen);
            }
            break;
          case 2: // old overlap + spline
            Vs[sloc]->operator[](m).set(s0,s1,s2,s3,rcut,screen,
                                        s_rmin,s_sigma);
            break;
          case 3: // new overlap + spline
            if(ia == ja) {
              Scalar_t td = 0.0;
              if(l1 == l2) td = 1.0;
              Vs2[sloc]->operator[](m).set(td,s0,s1,s2,s3,rcut,screen,
                                           s_rmin,s_sigma);
            } else {
              Vs[sloc]->operator[](m).set(s0,s1,s2,s3,rcut,screen,
                                          s_rmin,s_sigma);
            }
            break;
          }
	}
      }
    } else if(slist[0] == "</TBPARAM>")  {
      tofindend = false;
    }
  } while(tofindend);


  return 0;
}

/////////////////////////////////////////////////////////////////
// NRLTBOnsite
/////////////////////////////////////////////////////////////////
NRLTBOnsite::NRLTBOnsite(){ 
  CollectVisitingForces=true;
}

NRLTBOnsite::~NRLTBOnsite() {
  for(int i=0; i<OnSite.size(); i++) delete OnSite[i];
}

bool NRLTBOnsite::get(ostream&) const {
  return true;
}
bool NRLTBOnsite::put(xmlNodePtr cur) {
  return true;
}

void NRLTBOnsite::reset() {}

bool 
NRLTBOnsite::put(istream& fin) {

  vector<string> slist, vlist;
  SpeciesSet& mySpecies(myPtcl->getSpeciesSet());
  int ins(mySpecies.addAttrib("ns"));

  NumSpecies = mySpecies.getTotalNum();
  BondMap.resize(NumSpecies, NumSpecies);

  Scalar_t lambda, rcut, screen;
  int ib, ia, ja;
  bool tofindon = true;
  do {
    int nw =getwords(slist, fin);
    if(nw == 0) continue;
    if(slist[0] == "rcut") {
      rcut = atof(slist[1].c_str());
    } else if(slist[0] == "screenl") {
      screen = atof(slist[1].c_str());
    } else   if(slist[0] == "lambda") {
      lambda = atof(slist[1].c_str());
    } else if(slist[0] == "bond") {
      ib = OnSite.size();
      ia = mySpecies.addSpecies(slist[1].c_str());
      ja = mySpecies.addSpecies(slist[2].c_str());
      BondMap(ia,ja) = ib;
      OnSite.push_back(new SKOnSite_t());
      Fc.push_back(NRLDensity<Scalar_t>());  // add cutoff function
    } else if(slist[0] == "density") {
      //////////////////////////////////////
      // warning: crystal splitting needs to be corrected
      //////////////////////////////////////
      Fc[ib].set(rcut,screen,lambda); 
      int num_onsite = atoi(slist[1].c_str());   
      for(int ion=0; ion<num_onsite; ion++) {
	getwords(vlist,fin);
	int L = atoi(vlist[0].c_str());
	int lt1 = atoi(vlist[1].c_str());
        Scalar_t a0 = atof(vlist[2].c_str());
        Scalar_t a1 = atof(vlist[3].c_str());
        Scalar_t a2 = atof(vlist[4].c_str());
        Scalar_t a3 = atof(vlist[5].c_str());
	OnSite[ib]->add(L,lt1,new OnSiteFunction_t(a0,a1,a2,a3));
      }

      int ns = static_cast<int>(mySpecies(ins,ia));
      OnSite[ib]->reset(ns);// reset the onsite blocks
    } else if(slist[0] == "</onsite>") {
      tofindon = false;
    }
  } while(tofindon);
  return true;
}

void 
NRLTBOnsite::diagonal(SparseMatrix_t& H, const PtclPairList_t& nnlist) {

  typedef PtclPairList_t::PairData_t PairData_t;
  int myID=nnlist.myID;

  int nnloc(0);
  if(NumSpecies == 1) {// single-species case, only 0 is accessed
    for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
      Scalar_t rho =0.0;
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
	rho += Fc[0]((*nit).sep); 
        ++nit;
      }
      OnSite[0]->getValue(H.diag(myID,iL),rho);
    }
  } else { // multi-species case, mapping is required

    const ParticleIndex_t& grID(myPtcl->GroupID);

    for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
      int itype(grID[nnlist.I[iL]]);
      //I don't know any NRL TB for more than two species
      TinyVector<Scalar_t,MAX_NRLSPECIES> rho;
      PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
      while(nit != nit_end) {
        const PairData_t& cur(*nit);
        int jtype(cur.Group);
	rho[jtype] += Fc[BondMap(itype,jtype)](cur.sep);
        ++nit;
      }

      for(int jtype=0; jtype<NumSpecies; jtype++) {
	OnSite[BondMap(itype,jtype)]->getValue(H.diag(myID,iL),rho[jtype]);
      }
    }
  }

}

NRLTBOnsite::Scalar_t 
NRLTBOnsite::evalForceNN(PtclPairList_t& nnlist, ParticlePos_t& fv, 
    unsigned stressmode) const {

  if(!Rho) {
    return 0.0;
  }

  typedef PtclPairList_t::PairData_t    PairData_t;
  const SparseMatrix_t::Index_t* hindex = Rho->getIndex();
  const ParticleIndex_t& grID(myPtcl->GroupID);
  ParticleScalar_t& q(myPtcl->E);
  
  int myID = nnlist.myID;
  int maxnn(nnlist.getMaxNadj());
  Matrix<Scalar_t> dh(hindex->MaxRowBlock,hindex->MaxRowBlock);
  Matrix<Scalar_t> dfc(NumSpecies,maxnn);
  vector<SingleParticlePos_t> fref(maxnn);
  vector<int> J(maxnn);

  register Scalar_t tmp;

  for(int iL=0; iL<nnlist.getLocalNum(); iL++) {
    int iat(nnlist.I[iL]);// global map
    int itype(grID[iat]);
    TinyVector<Scalar_t,MAX_NRLSPECIES> rho, drho;

    PtclPairList_t::const_iterator nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
    int nni(0);
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      int jtype(cur.Group);
      dfc(jtype,nni) = Fc[BondMap(itype,jtype)](cur.sep,tmp);
      rho[jtype] += tmp;
      ++nit; ++nni;
    }
      
    //dh = 0;
    int nb(hindex->row_block(iat));
    for(int jtype=0; jtype<NumSpecies; jtype++) {
      dh = 0.0;
      OnSite[BondMap(itype,jtype)]->getDeriv(dh.data(),rho[jtype]);
      //  	ConstBlockMatrixRef<Scalar_t> diag_block(H.diag(myID,iL),nb,nb);
      //  	drho[jtype] = diag_block.trace(dh.data());
      drho[jtype] = OhmmsBlas::dot(nb*nb,Rho->diag(myID,iL),dh.data());     
      //drho[jtype] = Trace<Scalar_t,TB_BLOCK>::get(Rho->diag(myID,iL),dh.begin(), nb, nb);
    }

    nit=nnlist.begin(iL);
    nni=0;
    while(nit != nit_end) {
      const PairData_t& cur(*nit);
      int jtype(cur.Group);
      Scalar_t x = drho[jtype]*dfc(jtype,nni);
      SingleParticlePos_t df(x*cur.displ);
      fv[iL]     += df;
      fv[cur.Where]-= df;
      ++nit; ++nni;
    }
  }

  return 0.0;
}

/*
//////////////////////////////////////////////
// pretty ugly implementation but it is called once during a run
//////////////////////////////////////////////
PotentialBase* 
NRLTB::init(SKList_t& SK, SKList_t& SK_O, Matrix<int>& bond, istream& fin) {

  vector<string> slist;

  getwords(slist,fin); // format
  int nrltype = atoi(slist[0].c_str()+6); //get the format
  bool nonortho= (slist[0][5] == '0');    //non-orthogonal TB

  getwords(slist,fin); // species
  getwords(slist,fin); // number of atoms
  int nspecies = atoi(slist[0].c_str());
  int npair = nspecies*(nspecies+1)/2;

  bond.resize(nspecies,nspecies);
  bond = -1;

  if(nrltype == 2) {  // using Spline function by Dallas Trinkle
    initSpline(SK, SK_O, bond, fin);
    return NULL;
  }


  ///////////////////////////////////////
  // convert the units
  ///////////////////////////////////////
  NRLUnitConvert Con;
  Con.set(true, nrltype);//set up the conversion factor

  Matrix<double> rcut0(nspecies,nspecies);
  Matrix<double> screen0(nspecies,nspecies);
  for(int i=0; i<nspecies; i++) {
    for(int j=i; j<nspecies; j++) {
      getwords(slist,fin); // interaction range
      rcut0(i,j) = Con.AA(atof(slist[0].c_str()));
      screen0(i,j) = Con.AA(atof(slist[1].c_str()));
      if(j>i) {
	rcut0(j,i) = rcut0(i,j);
	screen0(j,i) = screen0(i,j);
      }
    }
  }

  vector<int> orb(nspecies);
  for(int i=0; i<nspecies; i++) {
    getwords(slist,fin); // orbitals
    orb[i] = atoi(slist[0].c_str());
    getwords(slist,fin); // mass
    getwords(slist,fin); // orbital
  }

  ////////////////////
  // skip on-site terms
  ////////////////////
  int nskip = 17;
  if(nspecies == 2) nskip = 58;
  for(int i=0; i<nskip; i++) {
    getwords(slist,fin); // orbital
  }

  int nval=80;
  if(nspecies ==2) nval = 272;
  int oshift=nval/2; // overlap shift

  vector<double> h(oshift), s(oshift);
  for(int i=0; i<oshift; i++) {
    getwords(slist,fin); // orbital
    h[i] = atof(slist[0].c_str());
  }
  for(int i=0; i<oshift; i++) {
    getwords(slist,fin); // orbital
    s[i] = atof(slist[0].c_str());
  }

  for(int i=0; i<oshift; i+=4) {
    Con.convert_h(h,i);
    Con.convert_s(s,i);
  }

  int ib = 0;
  for(int is=0; is<nspecies; is++) {
    for(int js=0; js<nspecies; js++) {
      SK.push_back(new SlaterKosterComp<T>);
      if(nonortho) SK_O.push_back(new SlaterKosterComp<T>);
      bond(is,js) = ib++;
    }
  }

  ///////////////////////////////
  // a-a bond first and Hamiltonian
  ///////////////////////////////
  int ii=0;
  for(int is=0; is<nspecies; is++) {
    
    double rcut = rcut0(is,is);
    double screen = screen0(is,is);
    ib = bond(is,is);
    int vloc = V.size();
    int ii0 = ii;
    // add ss
    V.push_back(new vector<ScaleFunction_t>);
    SK[ib]->add(SKBlock<T,ScaleFunction_t>
		::make(0,0,is,is,0,0,0,V[vloc]));
    V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
    vloc++;
 
    // add sp
    V.push_back(new vector<ScaleFunction_t>);
    SK[ib]->add(SKBlock<T,ScaleFunction_t>
		::make(0,1,is,is,0,1,0,V[vloc]));
    V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
    vloc++; 

    // add pp
    V.push_back(new vector<ScaleFunction_t>);
    SK[ib]->add(SKBlock<T,ScaleFunction_t>::make(1,1,is,is,1,1,0,V[vloc]));
    V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
    V[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
    vloc++;

    if(orb[is] == 9) {
      // add L=(0,1,2) x d combination
      // add s-d
      V.push_back(new vector<ScaleFunction_t>);
      SK[ib]->add(SKBlock<T,ScaleFunction_t>::make(0,2,is,is,0,4,0,V[vloc]));
      V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      vloc++;

      // add p-d
      V.push_back(new vector<ScaleFunction_t>);
      SK[ib]->add(SKBlock<T,ScaleFunction_t>::make(1,2,is,is,1,4,0,V[vloc]));
      V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      V[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      vloc++;

      // add d-d
      V.push_back(new vector<ScaleFunction_t>);
      SK[ib]->add(SKBlock<T,ScaleFunction_t>::make(2,2,is,is,4,4,0,V[vloc]));
      V[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      V[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      V[vloc]->operator[](2).set(h[ii++], h[ii++], h[ii++], h[ii++],rcut,screen);
      vloc++;
    }

    if(nonortho) { // overlap matrix
      ii = ii0;
      if(nrltype==1) { // new style
	////////////////////////////////////////////////////////////////
	////////////////////////////////////////////////////////////////
	ib = bond(is,is);
	int vloc = Vs_new.size();

	// add ss
	Vs_new.push_back(new vector<ScaleFunction2_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction2_t>
		      ::make(0,0,is,is,0,0,1,Vs_new[vloc]));
	Vs_new[vloc]->operator[](0).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++;
    
	// add sp
	Vs_new.push_back(new vector<ScaleFunction2_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction2_t>
		      ::make(0,1,is,is,0,1,0,Vs_new[vloc]));
	Vs_new[vloc]->operator[](0).set(0, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++; 

	// add pp
	Vs_new.push_back(new vector<ScaleFunction2_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction2_t>::make(1,1,is,is,1,1,1,Vs_new[vloc]));
	Vs_new[vloc]->operator[](0).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	Vs_new[vloc]->operator[](1).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++;

	if(orb[is] == 9) {
	  // add L=(0,1,2) x d combination
	  // add s-d
	  Vs_new.push_back(new vector<ScaleFunction2_t>);
	  SK[ib]->add(SKBlock<T,ScaleFunction2_t>::make(0,2,is,is,0,4,0,Vs_new[vloc]));
	  Vs_new[vloc]->operator[](0).set(0, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;
	
	  // add p-d
	  Vs_new.push_back(new vector<ScaleFunction2_t>);
	  SK_O[ib]->add(SKBlock<T,ScaleFunction2_t>::make(1,2,is,is,1,4,0,Vs_new[vloc]));
	  Vs_new[vloc]->operator[](0).set(0, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_new[vloc]->operator[](1).set(0, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;

	  // add d-d
	  Vs_new.push_back(new vector<ScaleFunction2_t>);
	  SK_O[ib]->add(SKBlock<T,ScaleFunction2_t>::make(2,2,is,is,4,4,0,Vs_new[vloc]));
	  Vs_new[vloc]->operator[](0).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_new[vloc]->operator[](1).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_new[vloc]->operator[](2).set(1, s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;
	}
      } else {//old style

	ib = bond(is,is);
	int vloc = Vs_old.size();

	// add ss
	Vs_old.push_back(new vector<ScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction_t>
		      ::make(0,0,is,is,0,0,1,Vs_old[vloc]));
	Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++;
    
	// add sp
	Vs_old.push_back(new vector<ScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction_t>
		      ::make(0,1,is,is,0,1,0,Vs_old[vloc]));
	Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++; 

	// add pp
	Vs_old.push_back(new vector<ScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,ScaleFunction_t>::make(1,1,is,is,1,1,1,Vs_old[vloc]));
	Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	Vs_old[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	vloc++;

	if(orb[is] == 9) {
	  // add L=(0,1,2) x d combination
	  // add s-d
	  Vs_old.push_back(new vector<ScaleFunction_t>);
	  SK[ib]->add(SKBlock<T,ScaleFunction_t>::make(0,2,is,is,0,4,0,Vs_old[vloc]));
	  Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;
	
	  // add p-d
	  Vs_old.push_back(new vector<ScaleFunction_t>);
	  SK_O[ib]->add(SKBlock<T,ScaleFunction_t>::make(1,2,is,is,1,4,0,Vs_old[vloc]));
	  Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_old[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;

	  // add d-d
	  Vs_old.push_back(new vector<ScaleFunction_t>);
	  SK_O[ib]->add(SKBlock<T,ScaleFunction_t>::make(2,2,is,is,4,4,0,Vs_old[vloc]));
	  Vs_old[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_old[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  Vs_old[vloc]->operator[](2).set(s[ii++], s[ii++], s[ii++], s[ii++],rcut,screen);
	  vloc++;
	}
      }
    }
  } // a-a bond

  return NULL; // no pair potential is created
}


//////////////////////////////////////////////
// adding spline at the close range
//////////////////////////////////////////////
void
NRLTB::initSpline(SKList_t& SK, SKList_t& SK_O, Matrix<int>& bond,  istream& fin) {

  ///////////////////////////////////////
  // convert the units
  ///////////////////////////////////////
  NRLUnitConvert Con;
  Con.set(true, 0);//set up the conversion factor

  vector<string> slist;

  int nspecies = bond.nrow();
  getwords(slist,fin); 
  T rcut = Con.AA(atof(slist[0].c_str()));
  T screen = Con.AA(atof(slist[1].c_str()));
  T hmin = Con.AA(atof(slist[2].c_str()));
  T smin = Con.AA(atof(slist[3].c_str()));
  T hsig = Con.AA(atof(slist[4].c_str()));
  T ssig = Con.AA(atof(slist[5].c_str()));

  vector<int> orb(nspecies);
  for(int i=0; i<nspecies; i++) {
    getwords(slist,fin); // orbitals
    orb[i] = atoi(slist[0].c_str());
    getwords(slist,fin); // mass
    getwords(slist,fin); // orbital
  }

  bool nonortho = true; // hacking again

  ////////////////////
  // skip on-site terms
  ////////////////////
  int nskip = 17;
  if(nspecies == 2) nskip = 58;
  for(int i=0; i<nskip; i++) {
    getwords(slist,fin); // orbital
  }

  int nval=80;
  if(nspecies ==2) nval = 272;
  int oshift=nval/2; // overlap shift

  vector<double> h(oshift), s(oshift);
  for(int i=0; i<oshift; i++) {
    getwords(slist,fin); // orbital
    h[i] = atof(slist[0].c_str());
  }
  for(int i=0; i<oshift; i++) {
    getwords(slist,fin); // orbital
    s[i] = atof(slist[0].c_str());
  }

  for(int i=0; i<oshift; i+=4) {
    Con.convert_h(h,i);
    Con.convert_s(s,i);
  }

  int ib = 0;
  for(int is=0; is<nspecies; is++) {
    for(int js=0; js<nspecies; js++) {
      SK.push_back(new SlaterKosterComp<T>);
      if(nonortho) SK_O.push_back(new SlaterKosterComp<T>);
      bond(is,js) = ib++;
    }
  }

  ///////////////////////////////
  // a-a bond first and Hamiltonian
  ///////////////////////////////
  int ii=0;
  for(int is=0; is<nspecies; is++) {
    
    ib = bond(is,is);
    int vloc = Vhs.size();
    int ii0 = ii;
    // add ss
    Vhs.push_back(new vector<TiScaleFunction_t>);
    SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,0,is,is,0,0,0,Vhs[vloc]));
    Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
    vloc++;

 
    // add sp
    Vhs.push_back(new vector<TiScaleFunction_t>);
    SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,1,is,is,0,1,0,Vhs[vloc]));
    Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
    vloc++; 

    // add pp
    Vhs.push_back(new vector<TiScaleFunction_t>);
    SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(1,1,is,is,1,1,0,Vhs[vloc]));
    Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
    Vhs[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
    vloc++;

    if(orb[is] == 9) {
      // add L=(0,1,2) x d combination
      // add s-d
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,2,is,is,0,4,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      vloc++;

      // add p-d
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(1,2,is,is,1,4,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      Vhs[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      vloc++;

      // add d-d
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK[ib]->add(SKBlock<T,TiScaleFunction_t>::make(2,2,is,is,4,4,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      Vhs[vloc]->operator[](1).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      Vhs[vloc]->operator[](2).set(h[ii++], h[ii++], h[ii++],rcut,screen,hmin,hsig);
      vloc++;
    }

    if(nonortho) { // overlap matrix
      ii = ii0;
      int vloc = Vhs.size();
      // add ss
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,0,is,is,0,0,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
      vloc++; ii++;
 
      // add sp
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,1,is,is,0,1,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
      vloc++; ii++;
      
      // add pp
      Vhs.push_back(new vector<TiScaleFunction_t>);
      SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(1,1,is,is,1,1,0,Vhs[vloc]));
      Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
      ii++;
      Vhs[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
      vloc++; ii++;
      
      if(orb[is] == 9) {
	// add L=(0,1,2) x d combination
	// add s-d
	Vhs.push_back(new vector<TiScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(0,2,is,is,0,4,0,Vhs[vloc]));
	Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	vloc++;ii++;

	// add p-d
	Vhs.push_back(new vector<TiScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(1,2,is,is,1,4,0,Vhs[vloc]));
	Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	ii++;
	Vhs[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	vloc++;

	// add d-d
	Vhs.push_back(new vector<TiScaleFunction_t>);
	SK_O[ib]->add(SKBlock<T,TiScaleFunction_t>::make(2,2,is,is,4,4,0,Vhs[vloc]));
	Vhs[vloc]->operator[](0).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	ii++;
	Vhs[vloc]->operator[](1).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	ii++;
	Vhs[vloc]->operator[](2).set(s[ii++], s[ii++], s[ii++],rcut,screen,smin,ssig);
	vloc++;
      }
    }
  } // a-a bond


}
*/
//////////////////////////////////////////////
void
NRLTB::printSpline(ostream& os) {
  os << " will print out spline version of NRTB " << endl;

}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * $Id: NRLTB.cpp 354 2005-07-08 02:20:49Z jnkim $ 
 ***************************************************************************/
