//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/**@file SKParameters.h
 * @brief Define SKParameters
 */
#ifndef OHMMS_SLATERKOSTERPARAMETERS_H
#define OHMMS_SLATERKOSTERPARAMETERS_H

#include "ohmms-config.h"
#include "OhmmsNumeric/Matrix.h"
#ifndef SLATERKOSTERARRAYBASE_H
#include "SlaterKoster/SlaterKosterBase.h"
#endif
#ifndef OHMMS_SPARSEMATRIX_H
#include "SparseMatrix/SparseMatrix.h"
#endif
#include "Potentials/ShortRangePotential.h"

namespace OHMMS {

 class TBModelBase;
 class TBOnsiteBase;

 /** encapsulate model-dependent properties of SlaterKoster TB 
  *
  * Implements the virutal functions of ShortRangePotential
  * and additional functions for SlaterKosterTB operations.
  */
 class SKParameters: public ShortRangePotential {

 public:

   typedef std::vector<SlaterKosterComp<Scalar_t>* > SKList_t;

   ///bond type between species
   Matrix<int> Bond;

   /// Slater-Koster for the Hamiltonian Matrix
   SKList_t SK_h;
   
   /// Slater-Koster for the Overlap matrix
   SKList_t SK_s;

   ///probably not needed
   string SKModel;
   string OnsiteModel;

   TBModelBase* curSK;
   TBOnsiteBase* curOnsite;
   vector<ShortRangePotential*> OtherPot;

   ///constructor
   SKParameters();
   ///destructor
   ~SKParameters();

   ///overwrite the virtual function of OhmmsElementBase
   bool get(ostream&) const;
   bool put(istream&);
   void reset();
   bool put(xmlNodePtr cur);

   ///overwrite the virtual function of ShortRangePotential
   void setParticleSet(Particle_t* ptclset);
   ///overwrite the resetParticleSet of ShortRangePotential
   void resetParticleSet(Particle_t* ptclset);

   Scalar_t 
     evalEnergyNN(const PtclPairList_t& nnlist, Scalar_t rscale=1.0) const;
   Scalar_t 
     evalForceNN(PtclPairList_t& nnlist, ParticlePos_t& f0, 
       unsigned stressmode) const;

   ///new operators to handle Slater-Koster parameters efficiently
   inline int size() const { return OtherPot.size();}
   inline int pair_id(int itype, int jtype) const 
   { return Bond(itype,jtype);}

   inline bool getSKBlocks(int itype, Scalar_t* hblock) {
     int ij = Bond(itype,itype);
     SK_h[ij]->getOnsite(hblock);
     return true;
   } 

   inline bool 
   getSKBlocks(int itype, 
	       Scalar_t* restrict hblock, Scalar_t* restrict sblock){
     int ij = Bond(itype,itype);
     SK_h[ij]->getOnsite(hblock);
     SK_s[ij]->getOnsite(sblock);
     return true;
   } 

   inline bool getSKBlocks(int itype, int jtype,
			   Scalar_t sep, 
			   const SingleParticlePos_t& displ, 
			   Scalar_t* restrict h) {
     int ipair=Bond(itype,jtype);
     if(ipair<0 || sep>SK_h[ipair]->rcut()) return false;
     //Need to pass directional cosine
     Scalar_t rinv(1.0/sep);
     Scalar_t pvec[8];
     getPD(rinv*displ,pvec,SK_h[ipair]->maxL());
     SK_h[ipair]->getOffsite(sep, pvec, h);
     return true;
   }

   inline bool getSKBlocks(int itype, int jtype,
			   Scalar_t sep, const SingleParticlePos_t& displ, 
			   Scalar_t* restrict h,  
			   Scalar_t* restrict s) {
     int ipair=Bond(itype,jtype);
     if(ipair<0 || sep>SK_h[ipair]->rcut()) return false;
     Scalar_t rinv(1.0/sep);
     Scalar_t pvec[8];
     getPD(rinv*displ,pvec,SK_h[ipair]->maxL());
     SK_h[ipair]->getOffsite(sep, pvec, h);
     SK_s[ipair]->getOffsite(sep, pvec, s);
     return true;
   }

   inline bool 
   getSKBlocks(int ipair, Scalar_t sep, const SingleParticlePos_t& displ, 
               Scalar_t* restrict sk_px, 
               Scalar_t* restrict sk_py, 
               Scalar_t* restrict sk_pz) const {
     if(sep>SK_h[ipair]->rcut()) return false;
     Scalar_t rinv(1.0/sep);
     Scalar_t pvec[8], pd[8][3];
     getPD(rinv*displ, pvec, pd, SK_h[ipair]->maxL());
     SK_h[ipair]->getDeriv(sep, pvec, pd, sk_px,sk_py,sk_pz);
     return true;
   }

   inline bool getSKBlocks(int ipair, Scalar_t sep, const SingleParticlePos_t& displ, 
			   Scalar_t* restrict sk_px, 
			   Scalar_t* restrict sk_py, 
			   Scalar_t* restrict sk_pz,
			   Scalar_t* restrict sk_sx, 
			   Scalar_t* restrict sk_sy, 
			   Scalar_t* restrict sk_sz) const {
     if(sep > SK_h[ipair]->rcut()) return false;
     Scalar_t rinv(1.0/sep);
     Scalar_t pvec[8], pd[8][3];
     getPD(rinv*displ, pvec, pd, SK_h[ipair]->maxL());
     SK_h[ipair]->getDeriv(sep, pvec, pd, sk_px,sk_py,sk_pz);
     SK_s[ipair]->getDeriv(sep, pvec, pd, sk_sx,sk_sy,sk_sz);
     return true;
   }

   inline void 
   reset(int itype, int jtype,int ispeices, int jspecies, int lmax) {
     int ib = Bond(itype,jtype);
     if(ib<0) return;
     SK_h[ib]->reset(ispeices,jspecies,lmax);
     if(SK_s.size()) {
	SK_s[ib]->reset(ispeices,jspecies,lmax);
     }
   }

   inline bool orthogonal() const {return SK_s.empty();}

   /** assign a density matrix to the onsite term
    */
   void setOnsiteSK(SparseMatrix_t* restrict rhoM);

   void setOnsite(SparseMatrix_t& hmat, const PtclPairList_t& nnlist); 

 private:

   /** proces file and initialize the onsite and other classical potentials */
   bool init(const string& fname, vector<ShortRangePotential*>& pot_s); 

   //disable the copy constructor
   SKParameters(const SKParameters& skp):curSK(NULL),curOnsite(NULL){  }

 };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 364 $   $Date: 2005-07-13 13:33:56 -0500 (Wed, 13 Jul 2005) $
 * $Id: SKParameters.h 364 2005-07-13 18:33:56Z jnkim $ 
 ***************************************************************************/
