//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "SlaterKosterTB/SlaterKosterTB.h"
#include "SlaterKosterTB/TBModelBase.h"
#include "Utilities/OhmmsInfo.h"
#include "SparseMatrix/SparseIndexOps.h"
#include "Particle/NNEngine.h"
using namespace OHMMS;

SlaterKosterTB::SlaterKosterTB(): 
  Identity(true),NeverInitialized(true),
  MaxNumBasis(0), MaxAngMomentum(0), NumSpecies(0), LayoutXmlNode(0),
  HamIndex(0), DenIndex(0)
{
}

SlaterKosterTB::~SlaterKosterTB(){
}

/** initialize or reset the properties
 *@param ptclin ParticleSet whose positions are used to do any calculation
 *
 * \if NeverInitialized == true
 * Initialize the model and storage
 * - process model-dependent parameters
 * - initialize on-site and any extra potential other than the TB-ES
 * - initialize the nnListEngine
 * - initialize and allocate matrices
 * \else
 * reset properties as needed
 * \endif
 */
void 
SlaterKosterTB::setParticleSet(Particle_t* ptclin) {

  if(myPtcl == ptclin) return;

  if(NeverInitialized) {

    myPtcl = ptclin;

    //set the particle set any ShortRangePotential used by the model
    //This will be up when Species are owned by ParticleSet
    SKP.setParticleSet(myPtcl);

    //create onsite and repulsve here
    bool success=SKP.put(qmXmlNode);

    //set the common parameters for the electronic structure  calculations
    setESParameters(myPtcl);

    //save the maximum interaction radius
    Rmax = SKP.Rmax;
    Identity = SKP.orthogonal(); //save if the hamiltonian is orthogonal or not
    LOGMSG("\tMaximum interaction range =" << Rmax)

    //Time to initialize the nearest-neighbor engine: only once
    real_type GridRadius = Rmax;
    nnListEngine = myPtcl->getNNEngine();
    if(nnListEngine == 0) {
      nnListEngine = NNEngineBuilder::create(LayoutXmlNode);
    } else {
      ERRORMSG("\tUsing an existing layout ")
    }
    myPtcl->setNNEngine(nnListEngine);
    myPtcl->updateLayout(Rmax*nnListEngine->Rscale,GridRadius);

    nnListEngine->initialize(*myPtcl,SKP.CollectVisitingForces);
    //We can allocate bunch here

    //Assign names for the sparse matrices
    Hmat.setName("H");
    Smat.setName("S");
    Rho.setName("Rho");
    RhoS.setName("RhoS");

    // Add Rho to DenMatList
    DenMatList.push_back(&Rho); 

    // Set Rho to the onsite SK
    SKP.setOnsiteSK(&Rho);

    //initialize the solver and storage
    setEigenSolver();

    //initialize the sparse matrices
    //manageStorage();

    NeverInitialized=false;

  } else {
    myPtcl = ptclin;
    SKP.setParticleSet(myPtcl);
  }

}

void 
SlaterKosterTB::resetParticleSet(Particle_t* ptclin) {
  myPtcl = ptclin;
  SKP.resetParticleSet(myPtcl);
}

/** create Sparse Indices and prepare memory
 * @param dMats vector of SparseMatrix_t*
 *
 * The block sizes of Hamiltonian and density matrices are determined by
 * the "ns" attribute of the species.
 * Defer any memory allocation to setCurrentPos functions.
 */
void SlaterKosterTB::createIndex() {

  SpeciesSet& mySpecies(myPtcl->getSpeciesSet());
  int ins= mySpecies.addAttrib("ns");

  SparseIndex::IndexList_t tbb(mySpecies.getTotalNum());
  typedef SparseIndex::IndexList_t::value_type IVT;
  for(int itype=0; itype<mySpecies.getTotalNum(); itype++) {
    tbb[itype] = static_cast<IVT>(mySpecies(ins,itype));
  }

  //HamIndex can have duplicate off-diagonal indices
  HamIndex = new SparseIndex_t("HamIndex",false);
  HamIndex->create(-1);
  HamIndex->setBlocks(myPtcl->GroupID,tbb,tbb);
  Hmat.setIndex(HamIndex);
  HamMatList.push_back(&Hmat);

  //create overlap matrix
  if(!Identity) {
    Smat.setIndex(HamIndex);
    HamMatList.push_back(&Smat);
  }

  if(myPtcl->hasAttrib(ParticleTags::force_tag)) {
    LOGMSG("\tDetect force attribute. Creating density matrices")
    //DenMatrix has only unique  off-diagonal indices
    DenIndex = new SparseIndex_t("DenIndex",true);
    DenIndex->create(-1);
    DenIndex->setBlocks(myPtcl->GroupID,tbb,tbb);
    for(int i=0; i<DenMatList.size(); i++) {
      DenMatList[i]->setIndex(DenIndex);
    }
    EvaluateDensityMatrix = true;
  }
}

bool SlaterKosterTB::get(ostream& os) const {
  return true;
}

bool SlaterKosterTB::put(istream& is) {
  return true;
}

void SlaterKosterTB::reset() {
}

bool SlaterKosterTB::put(xmlNodePtr cur) {
  qmXmlNode=cur;
  return true;
}

//This is done by setParticleSet
///** Initialize the common properties of SlaterKosterTB as a PotentialBase.
// *
// * No large-scale storage is created by this.
// */
//void 
//SlaterKosterTB::initialize() {
//  //Set it true for safe memory management
//  NeverInitialized = true;
//
//  //Assign names for the sparse matrices
//  Hmat.setName("H");
//  Smat.setName("S");
//  Rho.setName("Rho");
//  RhoS.setName("RhoS");
//
//  //Below is moved initSlaterKoster
//  Onsite = dynamic_cast<TBOnsiteBase*>(SKP.getOnsite());
//  if(Onsite) {//connect the density matrix to Onsite term
//    Onsite->Rho = &Rho;
//    Onsite->SK_h = &(SKP.SK_h);
//    Onsite->SK_s = &(SKP.SK_s);
//  }
//  Rmax = SKP.rcut_max(); //save the maximum interaction radius
//  Identity = SKP.orthogonal(); //save if the hamiltonian is orthogonal or not
//  LOGMSG("\tMaximum interaction range =" << Rmax)
//}
//
///**void SlaterKosterTB::setHamiltonian(int& setmode)
// *@param setmode index if index==0  assigned, otherwise reused.
// *@brief Filling up the Hamiltonian matrices
// */
//void 
//SlaterKosterTB::setHamiltonian(int& setmode){
//  //This is done by setParticleSet
//  //if(Onsite) Onsite->setParticleSet(myPtcl);
//#pragma omp for nowait
//  for(int ip=0; ip<Hmat.index()->parts(); ip++) {
//    Hmat.clear(setmode,ip);               
//    if(!Identity) Smat.clear(setmode,ip); 
//    setHamiltonian(setmode, myPtcl->getNNEngine()->getPairAttrib(myID), myID);
//  }
//
//#ifdef PRINT_DEBUG
//   char fname[128];
//   int mid = Controller->mycontext();
//   sprintf(fname,"hmat.p%d",mid);
//   ofstream fout(fname);
//   Hmat.print(fout);
//   myPtcl->getNNEngine()->get(fout);
//#endif
//}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 364 $   $Date: 2005-07-13 13:33:56 -0500 (Wed, 13 Jul 2005) $
 * $Id: SlaterKosterTB.cpp 364 2005-07-13 18:33:56Z jnkim $ 
 ***************************************************************************/
