//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002,2003- by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
/** @file SlaterKosterTB.h
 * @brief Declare the base class SlaterKosterTB
 */
#ifndef OHMMS_SLATERKOSTERTB_POTENTIAL_MASTER_H
#define OHMMS_SLATERKOSTERTB_POTENTIAL_MASTER_H

#include "Potentials/PotentialBase.h"
#include "QuantumBase/ElectronStructure.h"
#include "SlaterKosterTB/SKParameters.h"

namespace OHMMS {

  class TBOnsiteBase;

  /** Handles an emprical TB potential in the Slater-Koster formalism.
   *
   * SlaterKosterTB represents an interaction for ionic degrees of freedom as 
   * a PotentialBase and represents the electronic structure as an 
   * ElectronStructure.  
   * The virutal functions of PotentialBase to evaluate the total energy and
   * the forces on the ions are handled by SlaterKosterTB.
   * As a PotentialBase, SlaterKosterTB is similar to PotentialCollection in
   * that it owns several components, the band-structure energy and 
   * other potentials, and collect all the components.
   * The SKParameter SKP handles the model-specific ways to fill 
   * in the matrix elements.
   * How to solve the ground electronic structure (gamma-only, k-point in 
   * combinations  with various eigen solvers) is left to the derived classes.
   * This class handles basic initialization with respect to the model 
   * Hamiltonian and manages big sparse matrices. 
   * Few common functions are defined for the derived classes:
   * - manageStorage
   * - setHamiltonian
   * - tbForceWithDM
   */
  class SlaterKosterTB: public PotentialBase, 
			public ElectronStructure {

  public:

    typedef SparseMatrix_t::Index_t SparseIndex_t;

    SlaterKosterTB();
    virtual ~SlaterKosterTB();

    /// implementing virtual functions for OhmmsElementBase
    bool get(ostream&) const;
    bool put(istream&);
    void reset();
    bool put(xmlNodePtr cur);

    // over-writing virtual functions of PotentialBase
    //void initialize();
    //@{Implement virtual functions of PotentialBase
    Scalar_t evalEnergy();
    Scalar_t evalForce(ParticlePos_t& fv, unsigned fmode = 0);
    bool evalDynamicMatrix(DynMatrix_t&);

    /** Set ptclset to myPtcl and all the components
     */
    void setParticleSet(Particle_t* ptclset);

    /** Reset ptclset to myPtcl and all the components
     */
    void resetParticleSet(Particle_t* ptclset);

    /** Set the current position to be pos
     *@param pos position array
     */
    void setCurrentPos(ParticlePos_t& pos);

    /** Set the current position to be pos
     *@param pos position array
     *@param wrap bool, if true, call myPtcl->applyBConds()
     *@param update bool, if true (default), update the cell information.
     */
    void setCurrentPos(ParticlePos_t& pos, bool wrap, bool updatecell=true);

    /** Set the current position to be pos and set the displacement vector dr
     *@param pos position array
     *@param dr  displacement array
     */
    void setCurrentPos(const ParticlePos_t& pos, const ParticlePos_t& dr);

    void setLayout(xmlNodePtr pl_ptr){LayoutXmlNode=pl_ptr;}
    
  protected:

    /// boolean, if true, the model uses Orthogonal basis set
    bool Identity;
    /// boolean to conotrol large-scale memory allocation
    bool NeverInitialized;
    ///Maximum number of basis per center/nuclei
    int MaxNumBasis;
    ///Maximum angular momentum
    int MaxAngMomentum;
    ///Number of species
    int NumSpecies;

    ///mode if pairlist has to be updated
    bool UpdateNNList;

    ///xmlNode for layout
    xmlNodePtr LayoutXmlNode;
    ///Index for the Hamiltonian and Overlap matrix
    SparseIndex_t *HamIndex;

    ///Index for the density matrices
    SparseIndex_t *DenIndex;

    ///band-structure energy
    Scalar_t TBEnergy;
    ///parameters for SK blocks
    SKParameters SKP;
    /// sparse matrix for hamiltonian
    SparseMatrix_t Hmat;
    /// sparse matrix for overlap matrix
    SparseMatrix_t Smat;

    /** sparse matrix for density matrix 
     *
     *\f$Rho(a,b) = 
     *\langle a|(\sum_{ij}|\psi_i\rangle S_{ij} \langle\psi_j|)|b\rangle\f$
     */
    SparseMatrix_t Rho; 

    /** sparse matrix for density matrix 
     *
     *\f$RhoS(a,b) = 
     *\langle a|(\sum_{ij}|\psi_i\rangle H_{ij} \langle\psi_j|)|b\rangle\f$
     */
    SparseMatrix_t RhoS; 

    vector<SparseMatrix_t*> HamMatList;
    vector<SparseMatrix_t*> DenMatList;

    /** update/resize Hamiltonian/Overlap matrix
     *
     * The indices and matrices are updated or resized whenever nnListEngine->makeNewPairs is 
     * performed.
     */
    void updateHandS();

    /** update/resize Hamiltonian/Overlap matrix and density matrices
     */
    void updateAll();

    //////////////////////////////////////////////////////////////////////////
    // virtual functions to be implemented by derived classes
    //////////////////////////////////////////////////////////////////////////
    /** create an eigen solver and allocate memory 
     */
    virtual void setEigenSolver() =0;
    /** calculate the Ground Electronic Structure
     */
    virtual Scalar_t solveGES() = 0;
    /** calculate the forces coming from GES
     */
    virtual Scalar_t evalForceTB(ParticlePos_t& f0, ParticleScalar_t& q, 
        Tensor_t& stress, unsigned stressmode) = 0;
    /** calculate the dynamic matrix contribution from GES
     */
    virtual void tbDynamicMatrix(DynMatrix_t&) = 0;

    /** create indices for the sparse matrices and initialize them
     */
    void createIndex();

    ///set the local Hamiltonian blocks
    void setHamiltonian(int setmode, const PtclPairList_t& nnlist, int myID);

    /** evaluate the energy and forces from onsite or repulsive potential
     * @param f0 forces on the particles
     * @param stresssmode flag to turn on/off stress calculations
     * @return partial sum of the extra terms
     *
     * This function is when all the extra terms can be calculated piece-by-piece
     * without any need for collection.
     */
    Scalar_t evalForceExtra(ParticlePos_t& f0, unsigned stressmode);

    /** evaluate the energy and forces from onsite or repulsive potential
     *
     * This function is used when any of extra terms require visiting forces.
     */
    Scalar_t evalForceExtraV(ParticlePos_t& f0, unsigned stressmode);
    
    /// calculate forces (f) using density matrices
    Scalar_t tbForceWithDM(const PtclPairList_t& nnlist, 
			   ParticlePos_t& f, ParticleScalar_t& e, 
                           Tensor_t& stress, unsigned stressmode) const;
  };
}
#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 364 $   $Date: 2005-07-13 13:33:56 -0500 (Wed, 13 Jul 2005) $
 * $Id: SlaterKosterTB.h 364 2005-07-13 18:33:56Z jnkim $ 
 ***************************************************************************/
