//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "SlaterKosterTB/SlaterKosterTBGamma.h"
#include "Particle/NNEngine.h"
#include "Utilities/Clock.h"
#include "Utilities/OhmmsInfo.h"
#include "Message/OpenMP.h"
#include "OhmmsNumeric/SparseEigenSolver.h"
#include "OhmmsNumeric/SparseEigenSolverBuilder.h"
#include "QuantumBase/ElectronStructureIO.h"
#include "SlaterKosterTB/TBModelBase.h"
using namespace OHMMS;
#include <iostream>
#include <fstream>
#include <iomanip>

#define TIMINGMSG(msg)
//#define TIMINGMSG(msg) cout << msg << endl;

SlaterKosterTBGamma::SlaterKosterTBGamma(): thisSolver(0),scfEngine(0){ 
  setName("SKTB_gamma");
  myRecorder = new ESOutput(*this);
}

SlaterKosterTBGamma::~SlaterKosterTBGamma(){ 
  if(thisSolver) delete thisSolver;
}

void 
SlaterKosterTBGamma::setEigenSolver() {
  //set gamma-point specific initialization
  Gamma = true; 
  if(!KSet) {
    KSet = new KPointBase;
    KSet->addKpt(SingleParticlePos_t(0,0,0),1);
  }
  NumKpts = KSet->size();
  Occ.resize(NumKpts,NumStates);
  E.resize(NumKpts,NumStates);
  Occ = 0.0e0;
  E = 0.0e0;

  // Add Rhos to DenMatList for non-orgigonal
  if(!Identity) DenMatList.push_back(&RhoS); 

  // create a solver
  LOGMSG("SlaterKosterTBGamma::setParticeSet, creating a symmetric solver")
  thisSolver = createSymSolver(States,qmXmlNode);

}

SlaterKosterTBGamma::Scalar_t 
SlaterKosterTBGamma::solveGES(){

  bool needrho = EvaluateDensityMatrix;

#pragma omp parallel
  {
    int np=Hmat.getIndex()->groups();
#pragma omp for nowait
    for(int ip=0; ip< np; ip++) {

      if(UpdateNNList) 
        nnListEngine->updatePairsPerThread(myPtcl->dR,nnListEngine->getPairAttrib(ip));

      int setmode=0; //clear everything
      setHamiltonian(setmode, myPtcl->getNNEngine()->getPairAttrib(ip), ip);

      if(needrho) {
        Rho.clear(setmode,ip);
        if(!Identity) RhoS.clear(setmode,ip);
      }
    }
  }

  if(needrho) {
    if(Identity) 
      thisSolver->getEigvectors(Hmat, E); 
    else
      thisSolver->getEigvectors(Hmat,Smat,E);
  } else {
    if(Identity) 
      thisSolver->getEigvalues(Hmat,E);
    else
      thisSolver->getEigvalues(Hmat,Smat,E);
  }
  Scalar_t tbe = getBandEnergy();

  ////////////////////////////////////////////
  //Incomplete implementation
  //if(scfEngine) {
  //  Scalar_t dqE, tbenew;
  //  int iscf = 0;
  //  vector<Scalar_t> w(Homo),we;
  //  for(int is=0; is<w.size(); is++)  w[is] = SpinDeg*Occ[0][is];
  //  if(!Identity) {
  //    we.resize(Homo);
  //    for(int is=0; is<we.size(); is++) we[is] = E[0][is]*SpinDeg*Occ[0][is];
  //  }
  //  OHMMS::product(Psi,Psi,w,Rho); // only need Rho
  //  do {
  //    if(Identity) {
  //      dqE = scfEngine->SCFCorrection(Hmat, Rho, iscf);
  //      thisSolver->getEigvectors(Hmat, Psi);
  //      OHMMS::product(Psi,Psi,w,Rho);
  //    } else {
  //      dqE = scfEngine->SCFCorrection(Hmat, Smat, Rho, iscf);
  //      OHMMS::product(Psi,Psi,w,Rho);
  //      OHMMS::product(Psi,Psi,we,RhoS);
  //    }

  //    tbenew = getBandEnergy()+dqE;
  //    if(fabs((tbenew - tbe)/tbe) < scfEngine->Etol) break;
  //    tbe = tbenew;
  //    iscf++;
  //  } while(iscf < scfEngine->MaxIter);
  //}
  ////////////////////////////////////////////

  return tbe;
}

SlaterKosterTBGamma::Scalar_t 
SlaterKosterTBGamma::evalForceTB(ParticlePos_t& f0, ParticleScalar_t& q, 
    		                 Tensor_t& stress, unsigned stressmode) {

  /////////////////////////////////////////////
  // always get the ground electronic structure
  /////////////////////////////////////////////
  EvaluateDensityMatrix = true;
  Scalar_t tbe = solveGES();

  vector<Scalar_t> w(Homo);
  if(ZeroTemperature || fabs(dSumOcc)<1e-16) {
    for(int is=0; is<Homo; is++)  w[is] = SpinDeg*Occ[0][is];
  } else {
    Scalar_t sigmainv = 1.0/Sigma;
    Scalar_t fac = dSumEnergy/dSumOcc;
    for(int is=0; is<Homo; is++) {
      w[is] = SpinDeg*(Occ[0][is]+(E[0][is]*sigmainv-fac)*dOcc[0][is]);
    }
  }

  thisSolver->product(w,Rho);

  if(!Identity) {
    for(int is=0; is<Homo; is++) w[is] *= E[0][is];
    thisSolver->product(w,RhoS);
  }

#pragma omp parallel
  {
    const NNEngine& nneng = *(myPtcl->getNNEngine());
    tbForceWithDM(nneng.getPairAttrib(omp_get_thread_num()), f0,q,stress,stressmode);
//#pragma omp for nowait
//    for(int myid=0; myid<nneng.size(); myid++) {
//      tbForceWithDM(nneng.getPairAttrib(myid), f0,q,stress,stressmode);
//    }
  }

  return tbe;
}

void SlaterKosterTBGamma::tbDynamicMatrix(DynMatrix_t& dymat) {

//   /////////////////////////////////////////////
//   // always get the ground electronic structure
//   /////////////////////////////////////////////
//   eStates->EvaluateDensityMatrix = true;

//   TBEnergy = solveGES();

//   if(Identity)
//     eStates->getDM(Rho);
//   else
//     eStates->getDM(Rho, RhoS);

// #pragma omp parallel
//   {
// #ifdef ENABLE_OPENMP
//     int myID = omp_get_thread_num();
// #else
//     int myID = 0;
// #endif

//     const PtclPairList_t& nnlist = nnListEngine->getPairAttrib(myID);
//     const SparseIndex_t& hIndex = *(Hmat.index());
//     const SparseIndex_t& dIndex = *(Rho.index());

//     Scalar_t pvec[8], pd[8][3], pdd[8][3][3];


//     Scalar_t *trho,*trhos;
//     int maxr = hIndex.MaxRowBlock;
//     int maxc = hIndex.MaxColBlock;
//     ParticlePos_t f0(myPtcl->getLocalNum());
   
//     // temporary block matrices used for force
//     Matrix<Scalar_t> Txy(maxr,maxc); 

//     int ig = hIndex.first_diag(myID);
//     int nnh = 0;  // index for a pair on this thread
//     int nnd = 0;
//     Vector<int> Work(hIndex.getTotNum());
//     Work = -2;

//     for(int iL=0; iL<hIndex.num_diag(myID); iL++, ig++) {
      
//       int iG = myPtcl->ID[ig];
//       Work[ig] = -1;
//       for(int nni=0; nni< dIndex.nadj(iL, myID); nni++, nnd++){
// 	Work[dIndex.iadj(iL,nni,myID)] = nnd;
//       }

//       int i0 = iG*OHMMS_DIM;   // location of iG-x component

//       for(int nni=0; nni< hIndex.nadj(iL, myID); nni++, nnh++){
//         int jg = hIndex.iadj(iL, nni, myID);
// 	int jG = myPtcl->ID[jg];
//         int ij = Bond(myPtcl->GroupID[iG],myPtcl->GroupID[jG]);

//         Scalar_t sep = nnlist.R[nnh];
//         if( sep > SK_h[ij]->rcut() ) continue;
// 	int j0 = jG*OHMMS_DIM;
// 	SingleParticlePos_t dR = nnlist.dR[nnh];

// 	int nbtot =  hIndex.row_block(ig)*hIndex.col_block(jg);

// 	getPD(dR, pvec, pd, pdd, SK_h[ij]->maxL());

// 	Scalar_t *trho;
// 	int dloc = Work[jg];
// 	if(dloc == -1) 
// 	  trho = Rho.diag(myID,iL);
// 	else if(dloc >= 0)  
// 	  trho = Rho.offdiag(myID,dloc);

// 	for(int idir=0; idir<OHMMS_DIM; idir++){
// 	  for(int jdir=idir; jdir<OHMMS_DIM; jdir++) {
// 	    Txy = 0.0;
// 	    SK_h[ij]->getDeriv2(sep, pvec, pd, pdd, Txy.data(), idir,jdir);
// 	    //Scalar_t ltrace = -2.0*tblock.trace(Txy.data());
// 	    Scalar_t ltrace = -2.0*OhmmsBlas::dot(nbtot,trho,Txy.data());

// 	    dymat(i0+idir,j0+jdir) += ltrace;
// 	    dymat(j0+idir,i0+jdir) += ltrace;
// 	    if(idir != jdir) {
// 	      dymat(i0+jdir, j0+idir) += ltrace;
// 	      dymat(j0+jdir, i0+idir) += ltrace;
// 	    }
// 	  }
// 	}

// 	if(!Identity) {
// 	  if(dloc == -1) trho = RhoS.diag(myID,iL);
// 	  else if(dloc >= 0)  trho = RhoS.offdiag(myID,dloc);

// 	  for(int idir=0; idir<OHMMS_DIM; idir++){
// 	    for(int jdir=idir; jdir<OHMMS_DIM; jdir++) {
// 	      Txy = 0.0;
// 	  g(myID,iL);
// 	else if(dloc >= 0)  
// 	  trho = Rho.offdiag(myID,dloc);

// 	for(int idir=0; idir<OHMMS_DIM; idir++){
// 	  for(int jdir=idir; jdir<OHMMS_DIM; jdir++) {
// 	    Txy = 0.0;
// 	    SK_h[ij]->getDeriv2(sep, pvec, pd, pdd, Txy.data(), idir,jdir);
// 	    //Scalar_t ltrace = -2.0*tblock.trace(Txy.data());
// 	    Scalar_t ltrace = -2.0*OhmmsBlas::dot(nbtot,trho,Txy.data());

// 	    dymat(i0+idir,j0+jdir) += ltrace;
// 	    dymat(j0+idir,i0+jdir) += ltrace;
// 	    if(idir != jdir) {
// 	      dymat(i0+jdir, j0+idir) += ltrace;
// 	      dymat(j0+jdir, i0+idir) += ltrace;
// 	    }
// 	  }
// 	}

// 	if(!Identity) {
// 	  if(dloc == -1) trho = RhoS.diag(myID,iL);
// 	  else if(dloc >= 0)  trho = RhoS.offdiag(myID,dloc);

// 	  for(int idir=0; idir<OHMMS_DIM; idir++){
// 	    for(int jdir=idir; jdir<OHMMS_DIM; jdir++) {
// 	      Txy = 0.0;
// 	      SK_s[ij]->getDeriv2(sep, pvec, pd, pdd, Txy.data(), idir,jdir);
// 	      Scalar_t ltrace = 2.0*OhmmsBlas::dot(nbtot,trho,Txy.data());
// 	      //Scalar_t ltrace = 2.0*tblock.trace(Txy.data());

// 	      dymat(i0+idir,j0+jdir) += ltrace;
// 	      dymat(j0+idir,i0+jdir) += ltrace;
// 	      if(idir != jdir) {
// 		dymat(i0+jdir, j0+idir) += ltrace;
// 		dymat(j0+jdir, i0+idir) += ltrace;
// 	      }
// 	    }
// 	  }
// 	}
//       }
//       Work = -2;
//     }
//  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 355 $   $Date: 2005-07-08 12:56:01 -0500 (Fri, 08 Jul 2005) $
 * $Id: SlaterKosterTBGamma.cpp 355 2005-07-08 17:56:01Z jnkim $ 
 ***************************************************************************/
