//////////////////////////////////////////////////////////////////
//                        COPYRIGHT NOTICE
//   (c) Copyright 1998-2002 Jeongnim Kim.
//   All rights reserved.
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "SlaterKosterTB/TBSCFBase.h"
#include "SparseMatrix/BlockMatrixRef.h" 
#include "Utilities/OhmmsSpecies.h"
#include "Utilities/SpeciesCollection.h"
using namespace OHMMS;

//\todo BlockMatrixRef is not efficient; remove it

TBSCFBase::TBSCFBase() {
  HubbardU = 4.0;
  MixingRatio = 0.1;
  MaxIter = 10;
  Etol = 1.0e-6;
}

TBSCFBase::~TBSCFBase() {
}

TBSCFBase::Scalar_t
TBSCFBase::SCFCorrection(SparseMatrix_t& h, SparseMatrix_t& rho, int iter){

  Scalar_t dqE  = 0.0;

#pragma omp parallel
  {
    typedef Particle_t::ParticleScalar_t ParticleScalar_t;
    ParticleScalar_t& Q = *(myPtcl->getScalarAttrib(VAL_CHARGE_TAG));
    ParticleScalar_t& Q0 = *(myPtcl->getScalarAttrib("Qatom"));
    ParticleScalar_t& dH = *(myPtcl->getScalarAttrib("delH"));

    Scalar_t hu = HubbardU;
    Scalar_t drho_new = MixingRatio;
    Scalar_t drho_old = 1.0-drho_new;

    int myID = omp_get_thread_num();
    const SparseIndex_t& myIndex = *h.index();

    int iG = myIndex.first_diag(myID);
    Scalar_t q, qnew;
    for(int iL=0; iL<myIndex.num_diag(myID); iL++, iG++) {

      int nb = myIndex.row_block(iG);
      ConstBlockMatrixRef<Scalar_t> diag_block(rho.diag(myID,iL),nb,nb);
      q = diag_block.trace();

      if(iter == 0) qnew = q;
      else          qnew = drho_new*q + drho_old*Q[iG];

      Q[iG] = qnew;
      qnew -= Q0[iG];
      Scalar_t dh = 2*hu*qnew;
      dqE += hu*qnew*qnew;
      // Diagonal terms are the same, substract the correction from
      // the previous step         
      BlockMatrixRef<Scalar_t> hdiag(h.diag(myID,iL),nb,nb);
      hdiag.add(dh-dH[iG]);
      dH[iG] = dh;
    }
  }// end-of-omp

  return dqE;
}

TBSCFBase::Scalar_t 
TBSCFBase::SCFCorrection(SparseMatrix_t& h, SparseMatrix_t& s,
			 SparseMatrix_t& rho, int iter){

  Scalar_t dqE  = 0.0;

#pragma omp parallel
  {
    typedef Particle_t::ParticleScalar_t ParticleScalar_t;
    ParticleScalar_t& Q = *(myPtcl->getScalarAttrib(VAL_CHARGE_TAG));
    ParticleScalar_t& Q0 = *(myPtcl->getScalarAttrib("Qatom"));
    ParticleScalar_t& dH = *(myPtcl->getScalarAttrib("delH"));

    Scalar_t hu = HubbardU;
    Scalar_t drho_new = MixingRatio;
    Scalar_t drho_old = 1.0-drho_new;
    int myID = omp_get_thread_num();
    const SparseIndex_t& myIndex = *h.index();
    int iG = myIndex.first_diag(myID);
    Scalar_t q;
    int nn = 0;

    for(int iL=0; iL<myIndex.num_diag(myID); iL++, iG++) {
      int nb = myIndex.row_block(iG);

      ConstBlockMatrixRef<Scalar_t> diag_block(rho.diag(myID,iL),nb,nb);
      q = diag_block.trace();

      for(int nni=0; nni< myIndex.nadj(iL, myID); nni++, nn++){
	ConstBlockMatrixRef<Scalar_t> 
	  offdiag_block(rho.offdiag(myID,nn),nb,nb);
	q += offdiag_block.trace(s.offdiag(myID,nn));
      }

      //Scalar_t qnew = drho_new*(q-Q0[iG]) + drho_old*Q[iG];
      Scalar_t qnew = drho_new*q + drho_old*Q[iG];
      Q[iG] = qnew;
      qnew -= Q0[iG];
      Scalar_t dh = 2*hu*qnew;
      dqE += hu*qnew*qnew;

      BlockMatrixRef<Scalar_t> hdiag(h.diag(myID,iL),nb,nb);
      hdiag.add(dh-dH[iG]);

      // Diagonal terms are the same, substract the correction from
      // the previous step         
      dH[iG] = dh;
    }
  }// end-of-omp

  return dqE;
}

void 
TBSCFBase::set(Particle_t* inptcl, const vector<string>& opt) {

  myPtcl = inptcl;

  int i=0;
  while(i<opt.size()) {
    if(opt[i] == "--hu") {
      HubbardU = atof(opt[++i].c_str());
    } else if(opt[i] == "--mixing") {
      MixingRatio = atof(opt[++i].c_str());
    } 
    i++;
  }


  typedef Particle_t::ParticleScalar_t ParticleScalar_t;
  SpeciesBase& species = *(SpeciesCollection::getSpecies());

  int qindex = SpeciesCollection::addAttrib("q");
  
  if(!myPtcl->hasAttrib("Qatom")) {
    ParticleScalar_t& Q0 = *(myPtcl->getScalarAttrib("Qatom"));
    for(int iat=0;iat<Q0.size(); iat++) {
      Q0[iat] = species(qindex,myPtcl->GroupID[iat]);
    }
  }

  ParticleScalar_t& Q0 = *(myPtcl->getScalarAttrib("Qatom"));
  ParticleScalar_t& Q = *(myPtcl->getScalarAttrib(VAL_CHARGE_TAG));
  ParticleScalar_t& dH = *(myPtcl->getScalarAttrib("delH"));
  Q = Q0;
  dH = 0.0;
}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: TBSCFBase.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
