//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
#ifndef OHMMS_BLOCKMATRIX_REF_H
#define OHMMS_BLOCKMATRIX_REF_H
#include "OhmmsNumeric/OhmmsBlasf.h"

template<class T> struct BlockMatrixRef {};

template<>
struct BlockMatrixRef<double> {

  typedef double T;
  BlockMatrixRef(T* a, int size1, int size2):
    d_data(a), nrow(size1), ncol(size2) { }

  template<class Mat_t>
  inline void 
  multadd(const Mat_t& a, int aoffset, 
	  const Mat_t& b, int boffset, const vector<T>& w) {

    int ii = aoffset;
    int kk = 0;
    vector<T> t(ncol);
    for(int ib=0; ib<ncol; ib++) {
      T tmp = w[0]*a(0,ii);
      int jj = boffset;
      for(int jb=0; jb<ncol; jb++)  t[jb] = tmp*b(0,jj++);

      for(int is=1; is<w.size();is++) {
	tmp = w[is]*a(is,ii);
	jj = boffset;
	for(int jb=0; jb<ncol; jb++)  t[jb] += tmp*b(is,jj++);
      }
      for(int jb=0; jb<ncol; jb++) {
	d_data[kk++] = t[jb];
      }        
      ii++;
    }
  }

  template<class Mat_t>
  inline void 
  multadd(const Mat_t& a, int aoffset, const Mat_t& b, int boffset) {

    int ii = aoffset;
    int kk = 0;
    for(int ib=0; ib<ncol; ib++) {
      vector<T> t(ncol,0.0);
#ifdef OHMMS_UBLAS
      for(int is=0; is<a.size1();is++) {
#else
      for(int is=0; is<a.nrows();is++) {
#endif
	T tmp = a(is,ii);
	int jj = boffset;
	for(int jb=0; jb<ncol; jb++)  t[jb] += tmp*b(is,jj++);
      }
      for(int jb=0; jb<ncol; jb++) {
	d_data[kk++] = t[jb];
      }        
      ii++;
    }
  }

  template<class Mat_t>
  inline void 
  multadd(T fac, const Mat_t& a, int aoffset, const Mat_t& b, int boffset) {

    int ii = aoffset;
    int kk = 0;
    for(int ib=0; ib<ncol; ib++) {
      vector<T> t(ncol,0.0);
#ifdef OHMMS_UBLAS
      for(int is=0; is<a.size1();is++) {
#else
      for(int is=0; is<a.nrows();is++) {
#endif
	T tmp = fac*a(is,ii);
	int jj = boffset;
	for(int jb=0; jb<ncol; jb++)  t[jb] += tmp*b(is,jj++);
      }
      for(int jb=0; jb<ncol; jb++) {
	d_data[kk++] = t[jb];
      }        
      ii++;
    }
  }

  inline 
  void add(T diag) {
    for(int ib=0; ib<nrow; ib++) {
      d_data[ib*ncol+ib] += diag;
    }
  }

  int nrow,ncol;
  T* d_data;
};

template<class T>
struct ConstBlockMatrixRef {};

//this is only for testing
#ifdef ADD_
#define matmult44 matmult44_
#endif
extern "C"
{
  void matmult44(const double*, 
		 const double*, const int&, 
		 double*, const int&, const int&, const int&);
}

template<>
struct ConstBlockMatrixRef<double> {

  typedef double T;
  typedef ConstBlockMatrixRef<double> This_t;

  ConstBlockMatrixRef() { }
  ConstBlockMatrixRef(const T* a, int size1, int size2):
    d_data(a), nrow(size1), ncol(size2) { }

  inline void set(const T* a, int size1, int size2) {
    d_data = a; nrow = size2; ncol=size2;
  }

  template<class Mat_t>
  inline void 
  multadd(const Mat_t& a, int aoffset, Mat_t& b, int boffset) {
/*
#ifdef OHMMS_UBLAS
    matmult44(d_data,a.data().begin(),aoffset+1,b.data().begin(),boffset+1,
	    a.size1(), a.size2());
#else
    matmult44(d_data,a.data(),aoffset+1,b.data(),boffset+1,
	      a.rows(), a.cols());
#endif
*/
  }

  inline T trace(This_t& a) {
    return OhmmsBlas::dot(nrow*ncol,d_data,a.d_data);    
  }

  inline T trace(const T* a) {
    return OhmmsBlas::dot(nrow*ncol,d_data,a);    
  }

  inline T trace() {
    T sum = 0;
    for(int i=0; i<nrow*ncol; i++) sum += d_data[i];
    return sum;
  }

  int nrow,ncol;
  const T* d_data;
};

template<>
struct BlockMatrixRef<complex<double> > {

  typedef double T;
  BlockMatrixRef(T* a, T* b, int size1, int size2)
    :r_data(a), i_data(b), nrow(size1), ncol(size2) { }

  template<class Mat_t>
  inline void 
  multadd(const Mat_t& a, int aoffset, const Mat_t& b, 
	  int boffset, const vector<T>& w) {

    typedef complex<double> value_type;
    int ii = aoffset;
    int kk = 0;
    vector<value_type> t(ncol);
    for(int ib=0; ib<ncol; ib++) {
      value_type tmp = w[0]*conj(a(0,ii));
      int jj = boffset;
      for(int jb=0; jb<ncol; jb++)  t[jb] = tmp*b(0,jj++);

      for(int is=1; is<w.size();is++) {
	tmp = w[is]*conj(a(is,ii));
	jj = boffset;
	for(int jb=0; jb<ncol; jb++)  t[jb] += tmp*b(is,jj++);
      }

      for(int jb=0; jb<ncol; jb++, kk++) {
	r_data[kk] = t[jb].real();
	i_data[kk] = t[jb].imag();
      }        
      ii++;
    }
  }

  int nrow,ncol;
  T* r_data; // real blcok
  T* i_data; // complex block
};
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: BlockMatrixRef.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

