//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseIndex.h"
#include "Particle/NNEngine.h"

namespace OHMMS {

  SparseIndex::SparseIndex(const char* name, bool unique):
    OhmmsObject("SparseIndex",name),
    Unique(unique),DiagonalOnly(false), NumThreads(0) { 

    DEBUGMSG("Calling SparseIndex::SparseIndex() with " << name)
      
  }

  SparseIndex::~SparseIndex() {  
    DEBUGMSG("Calling SparseIndex::~SparseIndex() for " << objName())
    for(int i=0; i<NumThreads; i++) if(J[i]) delete J[i];
    for(int i=0; i<NumThreads; i++) if(M[i]) delete M[i];
  } 

  void SparseIndex::create(int n) {

    NumThreads = (n>0)? n: omp_get_max_threads();

    myAtoms.resize(NumThreads);
    myPairs.resize(NumThreads,0);
    if(J.empty()) {
      for(int i=0; i<NumThreads; i++) {
        J.push_back(0); M.push_back(0);
      }
    }
  }

  void 
  SparseIndex::setBlocks(const ParticleIndex_t& grID,
      const IndexList_t& rblock, const IndexList_t& cblock) {

    TotalNum = grID.size();
    RowBlock.resize(TotalNum+1);
    ColBlock.resize(TotalNum+1);

    int maxrow= *(std::max_element(rblock.begin(),rblock.end()));
    int maxcol= *(std::max_element(cblock.begin(),cblock.end()));

    int nr(0);
    int nc(0);
    for(int i=0; i<TotalNum; i++) {
      RowBlock[i] = nr;
      ColBlock[i] = nc;
      nr += rblock[grID[i]];
      nc += cblock[grID[i]];
    }

    RowBlock[TotalNum] = nr;
    ColBlock[TotalNum] = nc;
    MaxRowBlock = maxrow;
    MaxColBlock = maxcol;
    MaxBlockSize = maxrow*maxcol;
    TotalRows = nr;
    TotalCols = nc;
  }

//  void 
//  SparseIndex::initDiagonal(Particle_t& ptcl, const NS_t& ns) {
//    TotalNum = ptcl.getTotalNum();
//    set_offsets(ptcl.GroupID,ns,ns,*this);
//    DiagonalOnly =true;
//    create(ptcl.getNNEngine()->size());
//  }
//
//  void 
//  SparseIndex::initDiagonal(Particle_t& ptcl, const NS_t& ns, const NS_t& nb) {
//    TotalNum = ptcl.getTotalNum();
//    set_offsets(ptcl.GroupID,ns,nb,*this);
//    DiagonalOnly =true;
//    create(ptcl.getNNEngine()->size());
//  }
//
//  void 
//  SparseIndex::init(Particle_t& ptcl, const NS_t& ns) {
//    TotalNum = ptcl.getTotalNum();
//    set_offsets(ptcl.GroupID,ns,ns,*this);
//    DiagonalOnly =false;
//    create(ptcl.getNNEngine()->size());
//  }
//
//  void 
//  SparseIndex::init(Particle_t& ptcl, const NS_t& ns, const NS_t& nb) {
//
//    TotalNum = ptcl.getTotalNum();
//    set_offsets(ptcl.GroupID,ns,nb,*this);
//
//    OwnIndices = false;
//    DiagonalOnly =false;
//    create(ptcl.getNNEngine()->size());
//  }
//
  void SparseIndex::checkBounds() {

    //diagonal indcies
    int nat(M[0]->size()-1);
    myAtoms[0][0] = 0;
    myAtoms[0][1] = nat;
    myAtoms[0][2] = nat;
    for(int ip=1; ip<NumThreads; ip++) {
      int nLoc(M[ip]->size()-1);
      myAtoms[ip][0]=nat;
      myAtoms[ip][1]=nat+nLoc;
      myAtoms[ip][2]=nLoc;
      nat+=nLoc;
    }

    //off-diagonal indcies
    if(!DiagonalOnly) {
      int nn(J[0]->size());
      myPairs[0][0]=0;
      myPairs[0][1]=nn;
      myPairs[0][2]=nn;
      myPairs[0][3]=J[0]->capacity();
      for(int ip=1; ip<NumThreads; ip++) {
        int nnloc(J[ip]->size());
        myPairs[ip][0]=nn;
        myPairs[ip][1]=nn+nnloc;
        myPairs[ip][2]=nnloc;
        myPairs[ip][3]=J[ip]->capacity();
        nn += nnloc;
      }
    }
  }

  void 
  SparseIndex::update(const IndexList_t& inJ, const IndexList_t& inM, 
      int first, int ip) {
    IndexList_t* restrict jPtr=J[ip];
    if(jPtr == 0) {
      jPtr = new IndexList_t; 
      M[ip] = new IndexList_t;
      J[ip]=jPtr; 
    }

    if(Unique) {
      jPtr->clear();
      jPtr->reserve(inJ.capacity());
      IndexList_t& mRef(*M[ip]);
      mRef.resize(inM.size(),0);
      IndexList_t Work(TotalNum,0);
      int nd_loc(inM.size()-1);
      for(int iL=0; iL<nd_loc; iL++, first++) {
        Work[first] = 1; //exclude itselve
        int nn=0; // calculate the unique pairs per atom
        for(int jnn=inM[iL]; jnn< inM[iL+1]; jnn++) {
          int jG(inJ[jnn]);
          if(!Work[jG]) {
            jPtr->push_back(jG);
            Work[jG] = 1;
            ++nn;
          }
        }
        mRef[iL+1]=mRef[iL]+nn;
        std::fill(Work.begin(),Work.end(),0);//reset workspace
      }
    }else {
      *jPtr = inJ;
      *M[ip] = inM;
    }
  }

  void 
  SparseIndex::update(const PtclPairList_t& nnlist, int ip) {
    IndexList_t* restrict jPtr = J[ip];
    if(jPtr == 0) {
      jPtr = new IndexList_t;
      M[ip]= new IndexList_t;
      J[ip]=jPtr; 
    }
    if(Unique) {
      jPtr->clear();
      jPtr->reserve(nnlist.getTotNadj());
      IndexList_t& mRef(*M[ip]);
      mRef.resize(nnlist.getLocalNum()+1,0);
      IndexList_t Work(TotalNum,0);
      for(int iL=0; iL<nnlist.getLocalNum();iL++) {
        Work[nnlist.I[iL]] = 1; //exclude itselve
        PtclPairList_t::const_iterator 
          nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
        int nn(0);
        while(nit != nit_end) {
          int jG((*nit).Next);
          if(!Work[jG]) {
            jPtr->push_back(jG); Work[jG]=1; ++nn;
          }
          ++nit;
        }
        mRef[iL+1]=mRef[iL]+nn;
        std::fill(Work.begin(),Work.end(),0);//reset workspace
      }
    }else {
      jPtr->resize(nnlist.getTotNadj());
      IndexList_t& mRef(*M[ip]);
      mRef.resize(nnlist.getLocalNum()+1);
      mRef[0]=0;
      IndexList_t::iterator jit(jPtr->begin());
      int nnloc(0);
      for(int iL=0; iL<nnlist.getLocalNum();iL++) {
        PtclPairList_t::const_iterator 
          nit(nnlist.begin(iL)),nit_end(nnlist.end(iL));
        while(nit != nit_end) {
          *jit++=(*nit++).Next; 
        }
        mRef[iL+1]=mRef[iL]+nnlist.nadj(iL);
      }
    }
  }

  bool SparseIndex::get(std::ostream& os) const{
    os << "Maximum Row-block size = " << MaxRowBlock << endl;
    os << "Maximum Col-block size = " << MaxColBlock << endl;

    for(int myID=0; myID<NumThreads; myID++)  {
      os << "The range of atoms for Node " << myID << " ";
      os << first_diag(myID) << " " << last_diag(myID) << endl;
      int iG = first_diag(myID);
      if(!DiagonalOnly) {
	for(int iL=0; iL<num_diag(myID); iL++, iG++) {
	  os << iG << " ";
	  for(int nni=0; nni<nadj(iL,myID); nni++) {
	    os << iadj(iL, nni, myID) << " ";
	  }
	  os << endl;
	}
      }
    }
    return true;
  }

  bool SparseIndex::put(std::istream&){
    return true;
  }

  bool SparseIndex::put(xmlNodePtr) {
    return true;
  }
  void SparseIndex::reset() {
  }
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * OHMMS_VERSION: $Id: SparseIndex.cpp 354 2005-07-08 02:20:49Z jnkim $
 ***************************************************************************/
