//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_SPARSEVECTORINDEXOMP2_H
#define OHMMS_SPARSEVECTORINDEXOMP2_H

#include "Particle/ParticleSet.h"
#include "Utilities/OhmmsObject.h"
#include "Utilities/DistributedIndex.h"
#include "Message/OpenMP.h"
/*! \class VectorIndex
 * A specialized sparse index class for TB Hamiltonian matrices
 * Has a reference to PairList and creates extra indices.
 */
namespace OHMMS {

  /** Container of indices for sparse matrices
   *
   * Correct typedef should be 
   * typedef vector<size_t>          IndexList_t;
   * typedef IndexList_t::size_type  size_type;
   * But for now, simply use "int" to make all work
   * The typedefs propate to SparseMatrix.
   * Note that omp_int_t is used for the thread id.
   */
  struct SparseIndex: public OhmmsObject {

    typedef ParticleSet                 Particle_t;
    typedef Particle_t::ParticleIndex_t ParticleIndex_t;
    typedef Particle_t::PtclPairList_t  PtclPairList_t;
    typedef OHMMS_INDEX                 Index_t;
    typedef std::vector<OHMMS_INDEX>    IndexList_t;

    bool Unique;
    bool DiagonalOnly;
    omp_int_t NumThreads; 

    vector<IndexList_t*> J;
    vector<IndexList_t*> M;

    Index_t MaxRowBlock;
    Index_t MaxColBlock;
    Index_t MaxBlockSize;
    Index_t TotalNum;
    Index_t TotalRows, TotalCols; 

    vector<TinyVector<Index_t,3> > myAtoms;
    vector<TinyVector<Index_t,4> > myPairs;
    vector<Index_t> RowBlock;
    vector<Index_t> ColBlock;  

    SparseIndex(const char* = "none", bool unique=true);
    virtual ~SparseIndex(); // myList has to be provided externally

    void setBlocks(const ParticleIndex_t& grID, 
        const IndexList_t& rblock, const IndexList_t& cblock);

    void create(int n);
    void checkBounds();

    inline omp_int_t groups() const { return NumThreads;}
    inline int nrows() const { return TotalRows;}
    inline int ncols() const { return TotalCols;}

    inline int getTotNum() const { return TotalNum;}
    inline int getTotNadj() const { return myPairs[NumThreads-1][1];}

    inline int numatoms(omp_int_t ip=0) const { return myAtoms[ip][2];}
    inline int numpairs(omp_int_t ip=0) const { return myPairs[ip][2];}
    inline int maxpairs(omp_int_t ip=0) const { return myPairs[ip][3];}

    inline int nadj(int iL, omp_int_t ip = 0) const { 
      return (*M[ip])[iL+1]-(*M[ip])[iL];
    }

    inline int iadj(int iL, int nni, omp_int_t ip = 0) const { 
      return (*J[ip])[(*M[ip])[iL]+nni];
    }

    ////////////////////////////////////////////////////////
    // WARNING: Assuming each row and col is identicial
    //////////////////////////////////////////////////////
    //inline size_type row_loc(size_type i) const { return RowBlock[0]*i;}
    //inline size_type col_loc(size_type i) const { return ColBlock[0]*i;}
    inline int row_loc(int i) const { return RowBlock[i];}
    inline int col_loc(int i) const { return ColBlock[i];}

    inline int row_block(int ig) const { return RowBlock[ig+1]-RowBlock[ig];}
    inline int col_block(int ig) const { return ColBlock[ig+1]-ColBlock[ig];}

    inline int first_diag(omp_int_t ip =0) const { return myAtoms[ip][0];}
    inline int last_diag(omp_int_t ip =0) const { return myAtoms[ip][1];}
    inline int num_diag(omp_int_t ip =0) const { return myAtoms[ip][2];}
    inline int first_offdiag(omp_int_t ip = 0) const { return myPairs[ip][0];}
    inline int last_offdiag(omp_int_t ip = 0) const { return  myPairs[ip][1];}
    inline int num_offdiag(omp_int_t ip = 0) const { return  myPairs[ip][2];}

#ifdef ENABLE_OPENMP
  inline int where(int ig) const { //return NodeID[ig];}
    int res=0;
    while(ig >= last_diag(res)) {res++;}
    return res;
  }
#else
  inline int where(int ig) const { return 0;}
#endif

  inline int 
    begin_offdiag(int iL, omp_int_t ip = 0) const { 
    return M[ip]->operator[](iL); 
  }

  /** update the spase index with other indices
   * @param inJ index list that contains the target index
   * @param inM bound indices inM[i+1]-inM[i] is the number of pairs
   * @param first first diagonal index
   * @param ip thread id
   */
  virtual void update(const IndexList_t& inJ, const IndexList_t& inM, 
      int first, int ip);

  /** update the spase index with pair list
   * @param nnlist pairlist of ip-th thread
   * @param ip thread id
   */
  virtual void update(const PtclPairList_t& nnlist, int ip);

  bool get(std::ostream&) const;
  bool put(std::istream&) ;
  bool put(xmlNodePtr);
  void reset();
};
}
#endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 354 $   $Date: 2005-07-07 21:20:49 -0500 (Thu, 07 Jul 2005) $
 * OHMMS_VERSION: $Id: SparseIndex.h 354 2005-07-08 02:20:49Z jnkim $
 ***************************************************************************/
