//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#include "Message/OpenMP.h"
#include "SparseMatrix/SparseIndexOMP.h"

SparseIndexOMP::SparseIndexOMP(const char* name):
  OhmmsObject("SparseIndex",name), myList(NULL),NumThreads(0),OwnIndices(false),myAtoms(NULL), myPairs(NULL),J(NULL),M(NULL){ 

  DEBUGMSG("Calling SparseIndexOMP::SparseIndexOMP() with " << name)

}

void SparseIndexOMP::create() {

  if(NumThreads == 0) {

    NumThreads = omp_get_max_threads();
    myAtoms = new int*[NumThreads];
    myPairs = new int*[NumThreads];
    for(int i=0; i<NumThreads; i++) myAtoms[i] = new int[3];
    for(int i=0; i<NumThreads; i++) {
       myPairs[i] = new int[4]; 
       for(int j=0; j<4; j++) myPairs[i][j] = 0;
    }

    J = new int*[NumThreads];
    M = new int*[NumThreads];
    for(int i=0; i<NumThreads; i++) J[i] = NULL;
    for(int i=0; i<NumThreads; i++) M[i] = NULL;
  }
}

SparseIndexOMP::~SparseIndexOMP() {  

  cout << "Calling SparseIndexOMP::~SparseIndexOMP() for " << objName() << endl;
  if(myAtoms) {
    for(int i=0; i<NumThreads; i++) delete [] myAtoms[i];
    delete [] myAtoms;
    myAtoms = NULL;
  }

  if(myPairs) {
    for(int i=0; i<NumThreads; i++) delete [] myPairs[i];
    delete [] myPairs;
    myPairs = NULL;
  }
  //  if(OwnIndices)  delete [] NodeID;
} 

void SparseIndexOMP::set_offsets(const NS_t& rowblock, const NS_t& colblock){

  RowBlock = vector<int>(rowblock.size());
  ColBlock = vector<int>(colblock.size());

  MaxRowBlock = 0;
  for(int i=0; i<rowblock.size(); i++) {
    RowBlock[i] = static_cast<int>(rowblock[i]);
    MaxRowBlock = max(RowBlock[i], MaxRowBlock);
  }
  MaxColBlock = 0;
  for(int i=0; i<colblock.size(); i++) {
    ColBlock[i] = static_cast<int>(colblock[i]);
    MaxColBlock = max(ColBlock[i], MaxColBlock);
  }

  MaxBlockSize = MaxRowBlock*MaxColBlock;

  /////////////////////////////////////////////////////
  // WARNING: Assuming each row and block size is identical
  /////////////////////////////////////////////////////
  TotalRows = TotalNum*RowBlock[0];
  TotalCols = TotalNum*ColBlock[0];
}

void 
SparseIndexOMP::init(ThisPairList_t* ptclpairs, const NS_t& ns, IonID_t& ionid
#ifdef ENABLE_OPENMP
		     , IonID_t& nodeid
#endif
) {

  TotalNum = ionid.size();
  set_offsets(ns,ns);

  create();
  OwnIndices = false;
  myList = ptclpairs;
  IonID = ionid.begin();
#ifdef ENABLE_OPENMP
  NodeID= nodeid.begin();
#endif

  for(int i=0; i<NumThreads; i++) J[i] = ptclpairs[i].J.begin();
  for(int i=0; i<NumThreads; i++) M[i] = ptclpairs[i].M.begin();

  update();
}

void SparseIndexOMP::init(ThisPairList_t* ptclpairs, const NS_t& ns, 
                          const NS_t& nb, IonID_t& ionid
#ifdef ENABLE_OPENMP
			  , IonID_t& nodeid
#endif
) {

  TotalNum = ionid.size();
  set_offsets(ns,nb);
  create();
  OwnIndices = false;
  myList = ptclpairs;
  IonID = ionid.begin();
#ifdef ENABLE_OPENMP
  NodeID= nodeid.begin();
#endif
  for(int i=0; i<NumThreads; i++) J[i] = ptclpairs[i].J.begin();
  for(int i=0; i<NumThreads; i++) M[i] = ptclpairs[i].M.begin();

  update();
}

void SparseIndexOMP::update() {

  int nn = 0;
  for(int myID=0; myID < NumThreads; myID++) {

    if(myID == 0) {
      myAtoms[myID][0] = 0; 
      myAtoms[myID][1] = myList[myID].getLocalNum();
    } else {
      myAtoms[myID][1] = myAtoms[myID-1][1]+myList[myID].getLocalNum();
      myAtoms[myID][0] = myAtoms[myID-1][1];
    }
    myAtoms[myID][2] = myAtoms[myID][1]-myAtoms[myID][0];

    int nnmax = myPairs[myID][3];
    if(nnmax < myList[myID].getTotNadj()) {
      nnmax = myList[myID].getTotNadj();
      nnmax += nnmax/10; // add 10%
      myPairs[myID][3] = nnmax;
    }
    myPairs[myID][2] = myList[myID].getTotNadj();
    myPairs[myID][0] = nn;
    nn += myPairs[myID][2];
    myPairs[myID][1] = nn;
  }
}

//  void SparseIndexOMP::make_mp_map(int myID) {

//    //obsolate
//    int bat=TotalNum/NumThreads;
//    int residue = TotalNum%NumThreads;

//    if(myID<residue){
//      myAtoms[myID][0] = myID*(bat+1);
//      myAtoms[myID][1] = myAtoms[myID][0] + bat + 1;
//      myAtoms[myID][2] = myAtoms[myID][1] - myAtoms[myID][0];
//    }
//    else {
//      myAtoms[myID][0] = residue*(bat+1) + (myID - residue)*bat;    
//      myAtoms[myID][1] = myAtoms[myID][0] + bat;
//      myAtoms[myID][2] = myAtoms[myID][1] - myAtoms[myID][0];
//    }

//    //for(int i=myAtoms[myID][0]; i<myAtoms[myID][1]; i++) NodeID[i] = myID;
//  }

bool SparseIndexOMP::write_data(ostream&){

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#pragma omp critical 
    {
    cout << "The range of atoms for Node " << myID << " ";
    cout << first_diag(myID) << " " << last_diag(myID) << endl;
    int iG = first_diag(myID);
    for(int iL=0; iL<num_diag(myID); iL++, iG++) {
      cout << iG << " ";
      for(int nni=0; nni<nadj(myID, iL); nni++) {
	cout << iadj(myID, iL, nni) << " ";
      }
      cout << endl;
    }
    }
  }
  return true;
}
bool SparseIndexOMP::read_data(istream&){
  return true;
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * OHMMS_VERSION: $Id: SparseIndexOMP.cpp 151 2004-11-02 23:51:14Z jnkim $
 ***************************************************************************/
