//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim, Seungwon Lee and Lars Jonsson
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Message/OpenMP.h"
#include <iostream>
#include <iomanip>
using namespace std;
#include "SparseMatrix/SparseMatrixMP.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "OhmmsPETE/TinyMatrixRef.h"

#ifdef USE_WORKSPACE
#include "WorkSpace/WorkSpace.h"
#endif

SparseMatrixMP::SparseMatrixMP()
:myIndex(NULL), d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0) 
{
}

SparseMatrixMP::SparseMatrixMP(Index_t* ind)
:d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0)
{
  init(ind);
}

SparseMatrixMP::~SparseMatrixMP() { 
#ifndef USE_WORKSPACE

  // releasing the data
  delete [] d_diag[0][0];
  delete [] d_offdiag[0][0];

  for(int i=0; i<NumThreads; i++) {
    delete [] d_diag[i]; 
    delete [] d_offdiag[i];
    delete [] myAtoms[i];
    delete [] myPairs[i];
  }

  delete [] d_diag;
  delete [] d_offdiag;
  delete [] myAtoms;
  delete [] myPairs;

#endif
}

void
SparseMatrixMP::init(Index_t* newindex) {

  if(newindex) { myIndex = newindex; }
  myIndex->update();

  if(Num<myIndex->nrows()) { // first time, allocate pointer holders

    Dmax = myIndex->maxBasis;
    NumPairs = myIndex->numpairs();
    Num = myIndex->nrows();
    NumThreads = omp_get_max_threads();
    d_diag = new T**[NumThreads];
    d_offdiag = new T**[NumThreads];
    myAtoms = new int*[NumThreads];
    myPairs = new int*[NumThreads];

    // create a workspace
    int nshift = Dmax*Dmax; // block-stride
#ifndef USE_WORKSPACE
    int ws_size = Dmax*Dmax*(Num+NumPairs);
    WS_size = ws_size+ws_size/10; // adding 10% more than required
    T* dump = new T[WS_size];
#endif

#pragma omp parallel
    {
      int myID = omp_get_thread_num();
      myAtoms[myID] =  new int[3];
      myPairs[myID] =  new int[3];
      make_mp_map(myID, myAtoms[myID], myPairs[myID]);
//  #pragma omp critical 
//        {
//        cout << "first atom= " << myAtoms[myID][0] << endl;
//        cout << "last atom= " << myAtoms[myID][1] << endl;
//        cout << "first pair= " << myPairs[myID][0] << endl;
//        cout << "last pair= " << myPairs[myID][1] << endl;
//        }
      int myNumAtoms = myAtoms[myID][2];
      int myNumPairs = myPairs[myID][2];

      d_diag[myID] = new T*[myNumAtoms];
      d_offdiag[myID] = new T*[myNumPairs];

#ifdef USE_WORKSPACE
      T* dptr = WorkSpace::New(myID, myNumAtoms*nshift); 
      T* optr = WorkSpace::New(myID, myNumPairs*nshift);
#else
      T *dptr = dump + myAtoms[myID][0]*nshift;
      T *optr = dump + Num*nshift + myPairs[myID][0]*nshift;
#endif

      for(int i=0; i<myNumAtoms; i++, dptr+=nshift){ 
        d_diag[myID][i] = dptr;
      }
      for(int j=0; j<myNumPairs; j++, optr+=nshift){
        d_offdiag[myID][j] = optr;
      }

    }//omp-parallel closed
  }// if(Num<myIndex->nrows()) // first time, allocate pointer holders
}

void SparseMatrixMP::make_mp_map(int myID, int *myatoms, int *mypairs) {

  int bat=Num/NumThreads;
  int residue = Num%NumThreads;

  if(myID<residue){
    myatoms[0] = myID*(bat+1);
    myatoms[1] = myatoms[0] + bat + 1;
    myatoms[2] = myatoms[1] - myatoms[0];
  }
  else {
    myatoms[0] = residue*(bat+1) + (myID - residue)*bat;    
    myatoms[1] = myatoms[0] + bat;
    myatoms[2] = myatoms[1] - myatoms[0];
  }

  int istart = myatoms[0];
  int nn=0;
  while(nn<NumPairs && myIndex->irow(nn) != istart) {
    nn++;
  }

  mypairs[0] = nn;
  if(myatoms[1]== Num) mypairs[1]=NumPairs;
  else {
    int nf=mypairs[0];
    while(myIndex->irow(nf) != myatoms[1]) {
      nf++;
    }
    mypairs[1] = nf;
  }
  mypairs[2] = mypairs[1] - mypairs[0];
}

void
SparseMatrixMP::reset(bool changediag) {
#pragma omp parallel
  {
      int myID = omp_get_thread_num();
      int myNumAtoms = myAtoms[myID][2];
      int myNumPairs = myPairs[myID][2];
      //////////////////////////////////
      // To add re-mapping for Dynamics Simulations
      //////////////////////////////////
      int nbtot = Dmax*Dmax;
      if(changediag) {
        assign(d_diag[myID][0],nbtot*myNumAtoms,0.0);
      }
      assign(d_offdiag[myID][0],nbtot*myNumPairs,0.0);
  }//omp-parallel closed
}

SparseMatrixMP& SparseMatrixMP::operator +=(T lambda) 
{
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int myNumAtoms = myAtoms[myID][2];
    int istart = myAtoms[myID][0];
    for(int i=0; i<myNumAtoms; i++,istart++){    
#ifdef TB_BLOCK
      for(int ib=0;ib<TB_BLOCK;ib++){
        d_diag[myID][i][ib*TB_BLOCK+ib] += lambda;
      }
#else
      int nb = myIndex->Basis[istart]; 
      for(int ib=0;ib<nb;ib++){
        d_diag[myID][i][ib*nb+ib] += lambda;
      }
#endif
    }
  }
  return *this;
}

void SparseMatrixMP::apply(const T* win, T* wout) const {

  for(int i=0; i<myIndex->Offset[Num]; i++) wout[i] = 0.0;

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iat = 0;

    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
      int shift = myIndex->Offset[i];
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>::apply(d_diag[myID][iat], win+shift, wout+shift);
#else
      int nb = myIndex->Basis[i]; 
      MultAdd<T,0>::apply(d_diag[myID][iat], win+shift, wout+shift,nb,nb);
#endif
    } // diagonal block

    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);
#ifdef TB_BLOCK
      MultAdd<T,TB_BLOCK>
      ::apply(d_offdiag[myID][iat],
              win+myIndex->Offset[j],wout+myIndex->Offset[i]);
#else
      MultAdd<T,0>
      ::apply(d_offdiag[myID][iat],
              win+myIndex->Offset[j],wout+myIndex->Offset[i],
              myIndex->Basis[j], myIndex->Basis[i]);
#endif
    } // off-diagonal block

  }// end-of-omp
}

void SparseMatrixMP::copy(FullMatrix_t& mout, int packing) const {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();

    if(packing == 0) {
      int iat = 0;
      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
	mout.add(d_diag[myID][iat], TB_BLOCK, TB_BLOCK,
                 myIndex->Offset[i], myIndex->Offset[i]);
#else
	int nb = myIndex->Basis[i];
	mout.add(d_diag[myID][iat], nb, nb,
                 myIndex->Offset[i], myIndex->Offset[i]);
#endif
      } // diagonal block

      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j>=i)
#ifdef TB_BLOCK
	  mout.add(d_offdiag[myID][iat],TB_BLOCK,TB_BLOCK, 
                   myIndex->Offset[i], myIndex->Offset[j]);
#else
	mout.add(d_offdiag[myID][iat], myIndex->Basis[i], myIndex->Basis[j],
                 myIndex->Offset[i], myIndex->Offset[j]);
#endif
      }// off-diagonal block

    } else {

      int ncol = mout.ncol(); 
      int iat = 0;
      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK 
	const int D1 = TB_BLOCK;
#else
	int D1 = myIndex->Basis[i];
#endif
	int ii = myIndex->Offset[i];
	int kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj= myIndex->Offset[i];
	  for(int kp=0; kp<D1; kp++, jj++, kk++) {
	    if(jj>= ii) mout(ishift+jj) = d_diag[myID][iat][kk];
	  }
	  ii++;
	}
      }
      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j<i) continue;
#ifdef TB_BLOCK 
	const int D1 = TB_BLOCK; 
	const int D2 = TB_BLOCK; 
#else
	int D1 = myIndex->Basis[i];
	int D2 = myIndex->Basis[j];
#endif
	int ii = myIndex->Offset[i];
	int kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj = myIndex->Offset[j];
	  for(int kp=0; kp<D2; kp++,jj++,kk++)  {
	    if(jj >= ii) mout(ishift+jj)+= d_offdiag[myID][iat][kk];
	  }
	  ii++;
	}
      }

    }
  }// end of omp
}

void
SparseMatrixMP::copy(FullCMatrix_t& h, const complex<T>* phi, int packing) const {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    if(packing == 0) { // full-packing

      int iat = 0;
      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
	h.add(d_diag[myID][iat],TB_BLOCK,TB_BLOCK,
	      myIndex->Offset[i], myIndex->Offset[i]);
#else
	h.add(d_diag[myID][iat],myIndex->Basis[i],myIndex->Basis[i], 
	      myIndex->Offset[i], myIndex->Offset[i]);
#endif
      }// diagonal block

      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j>=i) {
#ifdef TB_BLOCK
	  h.add(d_offdiag[myID][iat],TB_BLOCK,TB_BLOCK,
		myIndex->Offset[i], myIndex->Offset[j],phi[nn]);
#else
	  h.add(d_offdiag[myID][iat],myIndex->Basis[i],myIndex->Basis[j], 
		myIndex->Offset[i], myIndex->Offset[j],phi[nn]);
#endif
	}// j>=i
      } // off-diagonal block

    } // full storage copy
    else {

      int ncol = h.ncol(); 
      int iat = 0;
      for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK 
	const int D1 = TB_BLOCK;
#else
	int D1 = myIndex->Basis[i];
#endif
	int ii = myIndex->Offset[i];
	int kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj= myIndex->Offset[i];
	  for(int kp=0; kp<D1; kp++, jj++, kk++) {
	    if(jj>= ii) h(ishift+jj) = d_diag[myID][iat][kk];
	  }
	  ii++;
	}
      }
      iat = 0;
      for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
	int i = myIndex->irow(nn);
	int j = myIndex->icol(nn);
	if(j<i) continue;
#ifdef TB_BLOCK 
	const int D1 = TB_BLOCK; 
	const int D2 = TB_BLOCK; 
#else
	int D1 = myIndex->Basis[i];
	int D2 = myIndex->Basis[j];
#endif
	int ii = myIndex->Offset[i];
	int kk=0;
	for(int k=0; k<D1; k++) {
	  int ishift = (ncol*2-ii-1)*ii/2;
	  int jj = myIndex->Offset[j];
	  for(int kp=0; kp<D2; kp++,jj++,kk++)  {
	    if(jj >= ii) h(ishift+jj)+= d_offdiag[myID][iat][kk]*phi[nn];
	  }
	  ii++;
	}
      }

    } // 'L' packing
  }// end-of-omp
}

void
SparseMatrixMP::copy(FullCMatrix_t& h, int packing) const {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iat = 0;
    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
      h.add(d_diag[myID][iat],TB_BLOCK,TB_BLOCK, 
            myIndex->Offset[i], myIndex->Offset[i]);
#else
      h.add(d_diag[myID][iat],myIndex->Basis[i],myIndex->Basis[i], 
            myIndex->Offset[i], myIndex->Offset[i]);
#endif
    }// diagonal block

    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);
      if(j>=i) {
#ifdef TB_BLOCK
      h.add(d_offdiag[myID][iat],TB_BLOCK,TB_BLOCK, 
            myIndex->Offset[i], myIndex->Offset[j]);
#else
      h.add(d_offdiag[myID][iat],myIndex->Basis[i],myIndex->Basis[j], 
            myIndex->Offset[i], myIndex->Offset[j]);
#endif
      }// j>=i
    } // off-diagonal block

  }// end-of-omp
}

void 
SparseMatrixMP::product(const FullMatrix_t& a, const FullMatrix_t& b,
                        const vector<T>& w){

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iat = 0;
    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>::make(d_diag[myID][iat],
                                a,myIndex->Offset[i],b,myIndex->Offset[i],w);
#else
      Product<T,0>::make(d_diag[myID][iat],myIndex->Basis[i], myIndex->Basis[i],
                         a,myIndex->Offset[i],b,myIndex->Offset[i],w);
#endif
    }

    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);

      if(i == j) continue;
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>::make(d_offdiag[myID][iat],
                                a,myIndex->Offset[i],b,myIndex->Offset[j],w);
#else
      Product<T,0>::make(d_offdiag[myID][iat],
                         myIndex->Basis[i], myIndex->Basis[j],
                         a,myIndex->Offset[i],b,myIndex->Offset[j],w);
#endif
    }// pair-list

  }// closing pramga
}

void 
SparseMatrixMP::product(SparseMatrixMP& Imat,
                        const FullCMatrix_t& a, const FullCMatrix_t& b,
                        const vector<T>& w){
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iat = 0;

    for(int i=first_diag(myID); i<last_diag(myID); i++,iat++){    
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>
      ::make(d_diag[myID][iat], Imat.d_diag[myID][iat], 
             a, myIndex->Offset[i], b, myIndex->Offset[i], w);
#else
      Product<T,0>
      ::make(d_diag[myID][iat], Imat.d_diag[myID][iat],
             myIndex->Basis[i], myIndex->Basis[i],
             a, myIndex->Offset[i], b, myIndex->Offset[i], w);
#endif
    }//diag

    iat = 0;
    for(int nn=first_offdiag(myID); nn < last_offdiag(myID); nn++, iat++) {
      int i = myIndex->irow(nn);
      int j = myIndex->icol(nn);

      if(i == j) continue;
#ifdef TB_BLOCK
      Product<T,TB_BLOCK>
      ::make(d_offdiag[myID][iat], Imat.d_offdiag[myID][iat],
             a, myIndex->Offset[i], b, myIndex->Offset[j], w);
#else
      Product<T,0>
      ::make(d_offdiag[myID][iat], Imat.d_offdiag[myID][iat],
             myIndex->Basis[i], myIndex->Basis[j],
             a, myIndex->Offset[i], b, myIndex->Offset[j],w);
#endif
    }// pairs

  }// omp
}

void 
SparseMatrixMP::print(ostream& os) {

}


/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseMatrixMP.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

