//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim, Seungwon Lee, and Lars Jonsson
//
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-

#include "Message/OpenMP.h"
#include <iostream>
#include <iomanip>
using namespace std;
#include "SparseMatrix/SparseMatrixSpinOrbitOMP.h"
#include "SparseMatrix/BlockMatrixFunctions.h"
#include "OhmmsPETE/TinyMatrixRef.h"
#include "Utilities/SpeciesCollection.h"
#include "Utilities/SimpleParser.h"

#ifdef USE_WORKSPACE
#include "WorkSpace/WorkSpace.h"
#endif

SparseMatrixOMP::SparseMatrixOMP()
  :myIndex(NULL), d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0) ,
   diagWorkSpace(NULL), offdiagWorkSpace(NULL)
{
  DEBUGMSG("Calling SparseMatrixOMP::SparseMatrixOMP()")
  read_so_parameters();
}

SparseMatrixOMP::SparseMatrixOMP(Index_t* ind)
  :d_diag(NULL), d_offdiag(NULL), WS_size(-1), Num(0),
   diagWorkSpace(NULL), offdiagWorkSpace(NULL)
{
  DEBUGMSG("Calling SparseMatrixOMP::SparseMatrixOMP()")
  read_so_parameters();
  init(ind);
}

SparseMatrixOMP::~SparseMatrixOMP() { 
#ifndef USE_WORKSPACE
  for(int i=0; i<NumThreads; i++) {
    delete [] d_diag[i]; 
    delete [] d_offdiag[i];
  }

  delete [] d_diag;
  delete [] d_offdiag;

  if(diagWorkSpace) delete [] diagWorkSpace;
  if(offdiagWorkSpace) delete [] offdiagWorkSpace;
#endif
}

void
SparseMatrixOMP::init(Index_t* newindex) {

  if(newindex) { myIndex = newindex; }

  if(Num<myIndex->getTotNum()) { // first time, allocate pointer holders

    Dmax = myIndex->MaxBlockSize;
    Num = myIndex->getTotNum();

    NumThreads = omp_get_max_threads();

    d_diag = new T**[NumThreads];
    d_offdiag = new T**[NumThreads];

    MaxNumPairs = new int[NumThreads];

    int numpairs = 0;
    for(int i=0; i<NumThreads; i++) {
      MaxNumPairs[i] = myIndex->maxpairs(i);
      numpairs +=   MaxNumPairs[i];
    }  

#ifndef USE_WORKSPACE
    WS_size = Dmax*numpairs;
    diagWorkSpace = new T[Dmax*Num];
    offdiagWorkSpace = new T[WS_size];
    DEBUGMSG("number of pair blocks created " << numpairs)
#endif

#pragma omp parallel
    {
      int myID = omp_get_thread_num();
      int myNumAtoms = myIndex->num_diag(myID);
      int myNumPairs = MaxNumPairs[myID];
      DEBUGMSG("number of pair blocks initialized " << myNumPairs)

      d_diag[myID] = new T*[myNumAtoms];
      d_offdiag[myID] = new T*[myNumPairs];

#ifdef USE_WORKSPACE
      T* dptr = WorkSpace::New(myID, myNumAtoms*Dmax); 
      T* optr = WorkSpace::New(myID, myNumPairs*Dmax);
#else
      T *dptr = diagWorkSpace  + myIndex->first_diag(myID)*Dmax;
      T *optr = offdiagWorkSpace  + myIndex->first_offdiag(myID)*Dmax;
#endif
      for(int i=0; i<myNumAtoms; i++, dptr+= Dmax){ 
        d_diag[myID][i] = dptr;
      }
      for(int j=0; j<myNumPairs; j++, optr+= Dmax){
        d_offdiag[myID][j] = optr;
      }

   }//omp-parallel closed
  }// if(Num<myIndex->nrows()) // first time, allocate pointer holders
}

void
SparseMatrixOMP::reset(int assignmode) {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int myNumAtoms = myIndex->num_diag(myID);
    int myNumPairs = myIndex->num_offdiag(myID);

    //////////////////////////////////
    // To add re-mapping for Dynamics Simulations
    //////////////////////////////////
    DEBUGMSG("number of pair blocks assigned " << myNumPairs)

    if(assignmode>= 0) assign(d_diag[myID][0],Dmax*myNumAtoms,0.0);
    assign(d_offdiag[myID][0],Dmax*myNumPairs,0.0);
  }
}

SparseMatrixOMP& SparseMatrixOMP::operator +=(T lambda) 
{
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int myNumAtoms = myIndex->num_diag(myID);
    int iG  = myIndex->first_diag(myID);
    for(int iL=0; iL<myNumAtoms; iL++, iG++){    
#ifdef TB_BLOCK
      for(int ib=0;ib<TB_BLOCK;ib++){
        d_diag[myID][iL][ib*TB_BLOCK+ib] += lambda;
      }
#else
      int nb = myIndex->row_block(iG);
      for(int ib=0;ib<nb;ib++){
        d_diag[myID][iL][ib*nb+ib] += lambda;
      }
#endif
    }
  }
  return *this;
}

void 
SparseMatrixOMP::apply(const dcomplex_t* win, dcomplex_t* wout) const {

 // assign(wout, 2*myIndex->nrows(), dcomplex_t(0.0));
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iG = myIndex->first_diag(myID);

    int first = 2*myIndex->col_loc(iG);
    int locnum = 2*myIndex->col_loc(myIndex->last_diag(myID))-first;
    assign(wout+first,locnum,dcomplex_t(0.0));

    int nn = 0;
    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    

      dcomplex_t* target = wout + 2*myIndex->row_loc(iG);
      const dcomplex_t* winref = win + 2*myIndex->col_loc(iG);
      SO[myIndex->kind(iG)]->apply(winref, target);

#ifdef TB_BLOCK
      MultAdd2<T,TB_BLOCK>::apply(d_diag[myID][iL], winref, target);
#else
      int nbi = myIndex->col_block(iG);
      MultAdd2<T,0>::apply(d_diag[myID][iL], winref, target,nbi, nbi);
#endif

      for(int nni = 0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
#ifdef TB_BLOCK
        MultAdd2<T,TB_BLOCK>
        ::apply(d_offdiag[myID][nn],win+2*myIndex->row_loc(jG), target);
#else
        MultAdd2<T,0>
	  ::apply(d_offdiag[myID][nn],win+2*myIndex->row_loc(jG), target,
		  myIndex->row_block(jG), nbi);
#endif
      }
    }
  }
}


void 
SparseMatrixOMP::apply(T** win, const int i,  T** wout, const int j, 
		       const int len, const int) const {

//    assign(wout,myIndex->nrows(), 0.0);

#pragma omp parallel
  {
    int myID = omp_get_thread_num();
    int iG = myIndex->first_diag(myID);
    const int spin = 2;
    const int com = 2;

    // factor 2 here is for space-doubling due to SO
    int first = spin*myIndex->col_loc(iG);

    // factor 2 here is for two-double structure of win and wout
    // note that len has already included space-doubling due to SO
    int ioffset = len*i*com;
    int joffset = len*j*com;

    // all the values to zero
    assign(&(wout[myID][joffset]),com*len,0.0);

    int nn = 0;
    for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    

      // 2 is for spin-orbit
      int jshift = com*(spin*myIndex->col_loc(iG)-first);
      T* target = &(wout[myID][joffset+jshift]);

      // add to spin-orbit
      SO[myIndex->kind(iG)]->apply(&(win[myID][ioffset+jshift]), target);

      // need a special apply for complex
#ifdef TB_BLOCK
      MultAdd2<T,TB_BLOCK>
      ::apply(d_diag[myID][iL],&(win[myID][ioffset+jshift]),target);
#else
      int nbi = myIndex->row_block(iG);
      MultAdd2<T,0>::apply(d_diag[myID][iL],&(win[myID][ioffset+jshift]),target, nbi, nbi);
#endif

      for(int nni=0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
        int jG = myIndex->iadj(iL, nni, myID);
        int ynode = myIndex->where(jG);
        int ishift = 4*(myIndex->row_loc(jG) 
       	              - myIndex->row_loc(myIndex->first_diag(ynode)));
#ifdef TB_BLOCK
        MultAdd2<T,TB_BLOCK>
        ::apply(d_offdiag[myID][nn], 
                &(win[ynode][ioffset+ishift]),target);
#else
        MultAdd2<T,0>
        ::apply(d_offdiag[myID][nn],&win[ynode][ioffset+ishift], target, 
                myIndex->row_block(jG), nbi);
#endif
      }
    }
  }// end-of-omp
}

void
SparseMatrixOMP::copy(FullMatrix_t& h, int packing) 
const {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();

    unsigned int ncoltot = h.ncol();
#ifdef TB_BLOCK
    const int nbi = TB_BLOCK;
    const int nbj = TB_BLOCK;
#endif
    Matrix<dcomplex_t> hoff(Dmax);

    if(packing == 0) {
      int iG = myIndex->first_diag(myID);
      int nn = 0;

      for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
#ifndef TB_BLOCK
	int nbi = myIndex->col_block(iG);
#endif
	int upi=2*myIndex->col_loc(iG);
	int downi=upi+nbi;
	SO[myIndex->kind(iG)]->copy(h.begin(), upi, upi, ncoltot);
	h.add(d_diag[myID][iL], nbi, nbi, upi, upi);
	h.add(d_diag[myID][iL], nbi, nbi, downi, downi);

	for(int nni =0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
  	  int jG = myIndex->iadj(iL, nni, myID);
	  if(jG < iG) continue;
#ifndef TB_BLOCK
	  int nbj = myIndex->col_block(jG);
#endif
	  int upj=2*myIndex->col_loc(jG);
	  int downj=upj+nbj;
          h.add(d_offdiag[myID][nn], nbi, nbj, upi, upj);
	  h.add(d_offdiag[myID][nn], nbi, nbj, downi, downj);
	}
      }
    } // full matrix
    else {

      int iG = myIndex->first_diag(myID);
      int nn = 0;

      for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
#ifndef TB_BLOCK
	int nbi = myIndex->col_block(iG);
#endif
	int upi=2*myIndex->col_loc(iG);
	int downi=upi+nbi;
	SO[myIndex->kind(iG)]->copy(h.begin(), upi, ncoltot);
        int iuu=upi, idd = downi;
        int kk=0;
        for(int k=0; k<nbi; k++) {
          int ushift = (ncoltot*2-iuu-1)*iuu/2;
          int dshift = (ncoltot*2-idd-1)*idd/2;
          int juu= upi, jdd = downi;
          for(int kp=0; kp<nbi; kp++, juu++, jdd++, kk++) {
            if(juu>= iuu) h(ushift+juu) += d_diag[myID][iL][kk];
            if(jdd>= idd) h(dshift+jdd) += d_diag[myID][iL][kk];
	  }
          iuu++;
          idd++;
	}

        for(int nni = 0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
  	  int jG = myIndex->iadj(iL, nni, myID);
          if(jG < iG) continue;
#ifndef TB_BLOCK
	  int nbj = myIndex->col_block(jG);
#endif
	  int upj=2*myIndex->col_loc(jG);
	  int downj=upj+nbj;

          iuu=upi;
          idd = downi;
	  kk=0;
	  for(int k=0; k<nbi; k++) {
            int ushift = (ncoltot*2-iuu-1)*iuu/2;
            int dshift = (ncoltot*2-idd-1)*idd/2;
	    int juu= upj, jdd = downj;
	    for(int kp=0; kp<nbj; kp++, juu++, jdd++, kk++) {
	      if(juu>= iuu) h(ushift+juu) += d_offdiag[myID][nn][kk];
	      if(jdd>= idd) h(dshift+jdd) += d_offdiag[myID][nn][kk];
	    }
            iuu++;
            idd++;
	  }
	}
      }
    }// low-packing
  }// end-of-omp
}

void
SparseMatrixOMP::copy(FullMatrix_t& h, const complex<T>* phi, int packing) 
const {

#pragma omp parallel
  {
    int myID = omp_get_thread_num();

    unsigned int ncoltot = h.ncol();
#ifdef TB_BLOCK
    const int nbi = TB_BLOCK;
    const int nbj = TB_BLOCK;
#endif
    Matrix<dcomplex_t> hoff(Dmax);

    if(packing == 0) {
      int iG = myIndex->first_diag(myID);
      int nn = 0;

      for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
#ifndef TB_BLOCK
	int nbi = myIndex->col_block(iG);
#endif
	int upi=2*myIndex->col_loc(iG);
	int downi=upi+nbi;
	SO[myIndex->kind(iG)]->copy(h.begin(), upi, upi, ncoltot);
	h.add(d_diag[myID][iL], nbi, nbi, upi, upi);
	h.add(d_diag[myID][iL], nbi, nbi, downi, downi);

	for(int nni =0; nni < myIndex->nadj(iL, myID); nni++, nn++) {
  	  int jG = myIndex->iadj(iL, nni, myID);
	  if(jG < iG) continue;
#ifndef TB_BLOCK
	  int nbj = myIndex->col_block(jG);
#endif
	  int upj=2*myIndex->col_loc(jG);
	  int downj=upj+nbj;
          cmult(hoff.begin(), phi[nn], d_offdiag[myID][nn], nbi*nbj);
          h.add(hoff.begin(), nbi, nbj, upi, upj);
	  h.add(hoff.begin(), nbi, nbj, downi, downj);
	}
      }
    } // full matrix
    else {

      int iG = myIndex->first_diag(myID);
      int nn = 0;

      for(int iL=0; iL<myIndex->num_diag(myID); iL++,iG++){    
#ifndef TB_BLOCK
	int nbi = myIndex->col_block(iG);
#endif
	int upi=2*myIndex->col_loc(iG);
	int downi=upi+nbi;
	SO[myIndex->kind(iG)]->copy(h.begin(), upi, ncoltot);
        int iuu=upi, idd = downi;
        int kk=0;
        for(int k=0; k<nbi; k++) {
          int ushift = (ncoltot*2-iuu-1)*iuu/2;
          int dshift = (ncoltot*2-idd-1)*idd/2;
          int juu= upi, jdd = downi;
          for(int kp=0; kp<nbi; kp++, juu++, jdd++, kk++) {
            if(juu>= iuu) h(ushift+juu) += d_diag[myID][iL][kk];
            if(jdd>= idd) h(dshift+jdd) += d_diag[myID][iL][kk];
	  }
          iuu++;
          idd++;
	}

        for(int nni = 0; nni<myIndex->nadj(iL, myID); nni++, nn++) {
  	  int jG = myIndex->iadj(iL, nni, myID);
          if(jG < iG) continue;
#ifndef TB_BLOCK
	  int nbj = myIndex->col_block(jG);
#endif
	  int upj=2*myIndex->col_loc(jG);
	  int downj=upj+nbj;

          iuu=upi;
          idd = downi;
	  kk=0;
	  for(int k=0; k<nbi; k++) {
            int ushift = (ncoltot*2-iuu-1)*iuu/2;
            int dshift = (ncoltot*2-idd-1)*idd/2;
	    int juu= upj, jdd = downj;
	    for(int kp=0; kp<nbj; kp++, juu++, jdd++, kk++) {
	      if(juu>= iuu) h(ushift+juu) += d_offdiag[myID][nn][kk]*phi[nn];
	      if(jdd>= idd) h(dshift+jdd) += d_offdiag[myID][nn][kk]*phi[nn];
	    }
            iuu++;
            idd++;
	  }
	}
      }
    }// low-packing
  }// end-of-omp
}

void 
SparseMatrixOMP::product(const FullMatrix_t& a, const FullMatrix_t& b,
                      const vector<T>& w){
}

void 
SparseMatrixOMP::product(SparseMatrixOMP&,
                        const FullMatrix_t& a, const FullMatrix_t& b,
                        const vector<T>& w){
}

void 
SparseMatrixOMP::print(ostream& os) {
}

void SparseMatrixOMP::read_so_parameters(){

  ifstream fin("TBPARAM");
  if(fin) {
    vector<string> slist;
    int nw=0;
    do {
      nw = getwords(slist,fin);
    } while(nw == 0 || slist[0] != "<SO>" || slist[0] == "end"); 
    if(slist[0] != "end") {
      SpeciesBase* mySpecies = SpeciesCollection::getSpecies();
      int numspecies =mySpecies->getTotalNum(); 
      int ins = SpeciesCollection::addAttrib("ns");
      SO.reserve(numspecies);
      for(int i=0; i<numspecies; i++) {
	SO.push_back(new SpinOrbitComp<T>);
      }
      getwords(slist,fin);
      do {
	int newid =SpeciesCollection::addSpecies(slist[0].c_str());
	int i0 = atoi(slist[2].c_str());
	double lambda = atof(slist[3].c_str());
	int ns = static_cast<int>((*mySpecies)(ins,newid));
	SO[newid]->add(new SpinOrbitPP<T>(lambda, i0, i0));
	SO[newid]->set(ns,ns);
	getwords(slist,fin);
      } while(slist[0] != "</SO>");
    }
  }  
}

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseMatrixSpinOrbitOMP.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

