//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//
//////////////////////////////////////////////////////////////////
// -*- c++ -*-
#include <iostream>
#include <fstream>
using namespace std;

#include "SparseMatrix/SparseSparseOps.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "OhmmsNumeric/OhmmsBlasOMP.h"
#include "Message/OpenMP.h"

void 
OhmmsSparseMatrix::axpy(double scal, const SparseMatrixOMP& A, SparseMatrixOMP& B
#if defined(USE_SPARSE_THREADED)
			, int myID) {
#else
                        ) {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#endif

    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();

    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;

    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
      ohmms_axpy_mp(nb,scal,A.diag(myID,iL),B.diag(myID,iL));
      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }
      
      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  ohmms_axpy_mp(nb,scal,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }
#if !defined(USE_SPARSE_THREADED)
  }
#endif
}

//#define USE_MAP

double 
OhmmsSparseMatrix::Dot(const SparseMatrixOMP& A, const SparseMatrixOMP& B
#if defined(USE_SPARSE_THREADED)
			, int myID) {
#else
                        ) {
  double res = 0;
#pragma omp parallel reduction(+:res)
  {
    int myID = omp_get_thread_num();
#endif
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();

#ifdef USE_MAP
    double psum = 0.0;
    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
      map<int,int> Work;
      //psum += OhmmsBlas::dot(nb,A.diag(myID,iL),B.diag(myID,iL));
      psum += ohmms_dot_mp(nb,A.diag(myID,iL),B.diag(myID,iL));
    
      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work[bIndex->iadj(iL,nni,myID)] = nnb;
      }
    
      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	map<int,int>::iterator it = Work.find(aIndex->iadj(iL,nni,myID));
	if(it != Work.end()) {
	  //psum += OhmmsBlas::dot(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	  psum += ohmms_dot_mp(nb,A.offdiag(myID,nna),B.offdiag(myID,(*it).second));
	}
      }
    }
#else
    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;
    double psum = 0.0;
    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
    
      //psum += OhmmsBlas::dot(nb,A.diag(myID,iL),B.diag(myID,iL));
      psum += ohmms_dot_mp(nb,A.diag(myID,iL),B.diag(myID,iL));
    
      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }
    
      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  //psum += OhmmsBlas::dot(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	  psum += ohmms_dot_mp(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }
#endif
#if defined(USE_SPARSE_THREADED)
    return psum;
#else
    res += psum;
  }
  return res;
#endif
}

void 
OhmmsSparseMatrix::copy(const SparseMatrixOMP& A, SparseMatrixOMP& B
#if defined(USE_SPARSE_THREADED)
			, int myID) {
#else
                        ) {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#endif

    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();
  
    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;
  
    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
    
      //OhmmsBlas::copy(nb,A.diag(myID,iL),B.diag(myID,iL));
      ohmms_copy_mp(nb,A.diag(myID,iL),B.diag(myID,iL));
    
      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }
    
      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  //OhmmsBlas::copy(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	  ohmms_copy_mp(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }
#if !defined(USE_SPARSE_THREADED)
  }
#endif
}

// A += xI
void 
OhmmsSparseMatrix::addDiagonal(SparseMatrixOMP& A, double x
#if defined(USE_SPARSE_THREADED)
			, int myID) {
#else
                        ) {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#endif
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
      int nb = aIndex->row_block(iG);
      for(int ib=0; ib<nb; ib++) A.diag(myID,iL)[ib+ib*nb] += x;
    }
#if !defined(USE_SPARSE_THREADED)
  }
#endif

}

/*
void 
OhmmsSparseMatrix::axpy(double scal, const SparseMatrixOMP& A, SparseMatrixOMP& B) {

#pragma omp parallel
 {
    int myID = omp_get_thread_num();
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();

    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;

    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    

      //OhmmsBlas::axpy(nb,scal,A.diag(myID,iL),1,B.diag(myID,iL),1);
      ohmms_axpy_mp(nb,scal,A.diag(myID,iL),B.diag(myID,iL));

      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }

      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  //OhmmsBlas::axpy(nb,scal,A.offdiag(myID,nna),1,B.offdiag(myID,nbloc),1);
	  ohmms_axpy_mp(nb,scal,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }

 }// end-of-omp

}

double 
OhmmsSparseMatrix::Dot(const SparseMatrixOMP& A, const SparseMatrixOMP& B) {//, int myID) {

  double res = 0;
#pragma omp parallel reduction(+:res)
  {
    int myID = omp_get_thread_num();
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();

    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;
    double psum = 0.0;
    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    

      //psum += OhmmsBlas::dot(nb,A.diag(myID,iL),B.diag(myID,iL));
      psum += ohmms_dot_mp(nb,A.diag(myID,iL),B.diag(myID,iL));

      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }

      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  //psum += OhmmsBlas::dot(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	  psum += ohmms_dot_mp(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }
    res += psum;
  }// end-of-omp

 return res;
}

void 
OhmmsSparseMatrix::copy(const SparseMatrixOMP& A, SparseMatrixOMP& B) {//, int myID) {

#pragma omp parallel
 {
   int myID = omp_get_thread_num();
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();

    // create an internal workspace to mask sparse indices
    Vector<int> Work(aIndex->getTotNum());
    Work = -1;

    int nb = aIndex->MaxBlockSize;
    int nna = 0;
    int nnb = 0;
    int iG = aIndex->first_diag(myID);
    for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    

      //OhmmsBlas::copy(nb,A.diag(myID,iL),B.diag(myID,iL));
      ohmms_copy_mp(nb,A.diag(myID,iL),B.diag(myID,iL));

      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	Work(bIndex->iadj(iL,nni,myID)) = nnb;
      }

      for(int nni=0; nni<aIndex->nadj(iL, myID); nni++, nna++) {
	int nbloc = Work(aIndex->iadj(iL,nni,myID));
	if( nbloc >= 0) {
	  //OhmmsBlas::copy(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	  ohmms_copy_mp(nb,A.offdiag(myID,nna),B.offdiag(myID,nbloc));
	}
      }
      Work = -1;
    }
 }// end-of-omp
}

// A += xI
void 
OhmmsSparseMatrix::addDiagonal(SparseMatrixOMP& A, double x) {//, int myID) {

#pragma omp parallel
 {
   int myID = omp_get_thread_num();
  const SparseMatrixOMP::Index_t *aIndex = A.index();
  int iG = aIndex->first_diag(myID);
  for(int iL=0; iL<aIndex->num_diag(myID); iL++,iG++){    
    ohmms_adddiag_mp(aIndex->row_block(iG),x,A.diag(myID,iL));
    //int nb = aIndex->row_block(iG);
    //for(int ib=0; ib<nb; ib++) A.diag(myID,iL)[ib+ib*nb] += x;
  }
 }
}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: SparseSparseOps.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
