//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//
//////////////////////////////////////////////////////////////////
// -*- c++ -*-
#include <iostream>
#include <fstream>
using namespace std;

#include "SparseMatrix/SparseSparseOps.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "OhmmsNumeric/OhmmsBlasOMP.h"
#include "Message/OpenMP.h"

// C = x A*B + y*C
inline void 
gemm_nn(double x, const double* restrict A, const double* restrict B, double y, 
	double* restrict C, int nrow, int ncol, int m) { 

#ifdef ENABLE_OPENMP
  //C(nrow,ncol) = A(nrow,m)*B(m,ncol)
  int iic=0;
  for(int i=0; i<nrow; i++){
    for(int j=0; j<ncol; j++,iic++) {
      double res = 0.0;
      for(int k=0; k<m; k++) {
	res += A[k+i*m]*B[j+k*ncol]; // A(i,k)*B(k,j)
      }
      C[iic] = x*res+y*C[iic];
    }
  }
#else
  const char transa = 'N';
  const char transb = 'N';
  dgemm(transa,transb,ncol,nrow,m,x,B,ncol,A,m,y,C,ncol);
#endif
}

void  
OhmmsSparseMatrix::xAB_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
			       double y, SparseMatrixOMP& C
#if defined(USE_SPARSE_THREADED)
			       , int myID) {
#else
                               ) {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#endif

    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();
    const SparseMatrixOMP::Index_t *cIndex = C.index();

  // create an internal workspace to mask sparse indices
    Vector<int> Work(cIndex->getTotNum());
    Work = -2;

    int nb = cIndex->MaxBlockSize;
    int nnc = 0;
    int iG = cIndex->first_diag(myID);
    if(fabs(y-1.0) > 1e-6) {// scale C by y only if y != 1
      int ig = iG;
      for(int iL=0; iL<cIndex->num_diag(myID); iL++,ig++){    
	int nrow = cIndex->row_block(ig);
	//OhmmsBlas::scal(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	ohmms_scal_mp(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
	  //OhmmsBlas::scal(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
	  //		y,C.offdiag(myID,nnc));
	  ohmms_scal_mp(nrow*cIndex->col_block(cIndex->iadj(iL,nni,myID)),
			y,C.offdiag(myID,nnc));
	}
      }
    }
    nnc = 0;
  
    int nna = 0;
    int nnb = 0;
    for(int iL=0; iL<cIndex->num_diag(myID); iL++,iG++){    
    
      //mask the target
      Work(iG) = -1; // diagonal
      for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
	Work(cIndex->iadj(iL,nni,myID)) = nnc;
      }
      int nrow= cIndex->row_block(iG);
      int ncol= cIndex->col_block(iG);
      int mprime = aIndex->col_block(iG);
    
      //apply the diagonal A(iL,iL) to B 
      gemm_nn(x,A.diag(myID,iL),B.diag(myID,iL),1.0,C.diag(myID,iL),
	      nrow, ncol, mprime); //C(nrow,ncol) = A(nrow,mprime)*B(mprime,ncol)
      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
      
	int ncloc = Work(bIndex->iadj(iL,nni,myID));
	if(ncloc >= 0) {//C(iL,bIndex->iadj(iL,nni,myID)) exist
	  gemm_nn(x,A.diag(myID,iL),B.offdiag(myID,nnb),1.0,C.offdiag(myID,ncloc),
		  nrow, ncol, mprime);
	}
      }
    
      //apply the offdiagonal A(iG,jG) to B(jG,*) 
      for(int nni=0; nni<aIndex->nadj(iL,myID); nni++, nna++) {
      
	int jG = aIndex->iadj(iL,nni,myID);
      
	int jnode = bIndex->where(jG);          //find the node for jG
	int jL = jG - bIndex->first_diag(jnode);//find the local index of jG
      
	int ncloc = Work(jG);
	if(ncloc >= 0) { // A(iG,jG) B(jG,kG) kG is a non-zero offdiagonal block
	  gemm_nn(x,A.offdiag(myID,nna),B.diag(jnode,jL),1.0,C.offdiag(myID,ncloc),
		  nrow, ncol, mprime);
	}
      
	int nnbj = bIndex->M[jnode][jL]; //accumulated index for non-zero off blocks of B
	for(int nnj=0; nnj<bIndex->nadj(jL,jnode); nnj++,nnbj++) {
	  ncloc = Work(bIndex->iadj(jL,nnj,jnode));
	
	  if(ncloc == -1) { // add to a diagonal element of C(iL)
	    gemm_nn(x,A.offdiag(myID,nna),B.offdiag(jnode,nnbj),1.0,
		    C.diag(myID,iL), nrow, ncol, mprime);
	  } else if(ncloc >= 0) { // add to a offdianogal
	    gemm_nn(x,A.offdiag(myID,nna),B.offdiag(jnode,nnbj),1.0,
		    C.offdiag(myID,ncloc), nrow, ncol, mprime);
	  }
	} // search for jG row to sum
      
      }//off-diagonal term of C(iG,jG) for jG != iG
      Work = -2; //reset the mask
    }//diagonal term
#if !defined(USE_SPARSE_THREADED)
  }
#endif
}
//  #else
//  void  
//  OhmmsSparseMatrix::xAB_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
//  			       double y, SparseMatrixOMP& C) {//, int myID) {

//  #pragma omp parallel
//   {
//     int myID = omp_get_thread_num();
//      const SparseMatrixOMP::Index_t *aIndex = A.index();
//      const SparseMatrixOMP::Index_t *bIndex = B.index();
//      const SparseMatrixOMP::Index_t *cIndex = C.index();

//      // create an internal workspace to mask sparse indices
//      Vector<int> Work(cIndex->getTotNum());
//      Work = -2;

//      int nb = cIndex->MaxBlockSize;
//      int nnc = 0;
//      int iG = cIndex->first_diag(myID);
//      if(fabs(y-1.0) > 1e-6) {// scale C by y only if y != 1
//        int ig = iG;
//        for(int iL=0; iL<cIndex->num_diag(myID); iL++,ig++){    
//  	int nrow = cIndex->row_block(ig);
//  	//OhmmsBlas::scal(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
//  	ohmms_scal_mp(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
//  	for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
//  	  //OhmmsBlas::scal(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
//  	  //		  y,C.offdiag(myID,nnc));
//  	  ohmms_scal_mp(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
//  			y,C.offdiag(myID,nnc));
//  	}
//        }
//      }
//      nnc = 0;

//      int nna = 0;
//      int nnb = 0;
//      for(int iL=0; iL<cIndex->num_diag(myID); iL++,iG++){    
      
//        //mask the target
//        Work(iG) = -1; // diagonal
//        for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
//  	Work(cIndex->iadj(iL,nni,myID)) = nnc;
//        }
//        int nrow= cIndex->row_block(iG);
//        int ncol= cIndex->col_block(iG);
//        int mprime = aIndex->col_block(iG);
      
//        //apply the diagonal A(iL,iL) to B 
//        gemm_nn(x,A.diag(myID,iL),B.diag(myID,iL),1.0,C.diag(myID,iL),
//  	      nrow, ncol, mprime); //C(nrow,ncol) = A(nrow,mprime)*B(mprime,ncol)
//        for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {

//  	int ncloc = Work(bIndex->iadj(iL,nni,myID));
//  	if(ncloc >= 0) {//C(iL,bIndex->iadj(iL,nni,myID)) exist
//  	  gemm_nn(x,A.diag(myID,iL),B.offdiag(myID,nnb),1.0,C.offdiag(myID,ncloc),
//  		  nrow, ncol, mprime);
//  	}
//        }

//        //apply the offdiagonal A(iG,jG) to B(jG,*) 
//        for(int nni=0; nni<aIndex->nadj(iL,myID); nni++, nna++) {

//  	int jG = aIndex->iadj(iL,nni,myID);

//  	int jnode = bIndex->where(jG);          //find the node for jG
//  	int jL = jG - bIndex->first_diag(jnode);//find the local index of jG

//  	int ncloc = Work(jG);
//  	if(ncloc >= 0) { // A(iG,jG) B(jG,kG) kG is a non-zero offdiagonal block
//  	  gemm_nn(x,A.offdiag(myID,nna),B.diag(jnode,jL),1.0,C.offdiag(myID,ncloc),
//  		  nrow, ncol, mprime);
//  	}

//  	int nnbj = bIndex->M[jnode][jL]; //accumulated index for non-zero off blocks of B
//  	for(int nnj=0; nnj<bIndex->nadj(jL,jnode); nnj++,nnbj++) {
//  	  ncloc = Work(bIndex->iadj(jL,nnj,jnode));

//  	  if(ncloc == -1) { // add to a diagonal element of C(iL)
//  	    gemm_nn(x,A.offdiag(myID,nna),B.offdiag(jnode,nnbj),1.0,
//  		    C.diag(myID,iL), nrow, ncol, mprime);
//  	  } else if(ncloc >= 0) { // add to a offdianogal
//  	    gemm_nn(x,A.offdiag(myID,nna),B.offdiag(jnode,nnbj),1.0,
//  		    C.offdiag(myID,ncloc), nrow, ncol, mprime);
//  	  }
//  	} // search for jG row to sum

//        }//off-diagonal term of C(iG,jG) for jG != iG
//        Work = -2; //reset the mask
//      }//diagonal term
//   }
//  }
//  #endif
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: Sparse_xAB_plus_yC.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
