//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//   Jeongnim Kim
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//
//////////////////////////////////////////////////////////////////
// -*- c++ -*-
#include <iostream>
#include <fstream>
using namespace std;

#include "SparseMatrix/SparseSparseOps.h"
#include "OhmmsNumeric/OhmmsBlasf.h"
#include "OhmmsNumeric/OhmmsBlasOMP.h"
#include "Message/OpenMP.h"

inline void 
gemm_nt(double x, const double* restrict A, 
	const double* restrict B, double y, double* restrict C, 
	int nrow, int ncol, int m) { 

#ifdef ENABLE_OPENMP
 for(int i=0; i<nrow; i++){
   for(int j=0; j<ncol; j++) {
     double res = 0.0;
     for(int k=0; k<m; k++) {
	res += A[k+i*m]*B[k+j*m];
     }
     C[j+i*ncol] = x*res+y*C[j+i*ncol];
   }
 }
#else
  const char transa = 'T';
  const char transb = 'N';
  dgemm(transa,transb,ncol,nrow,m,x,B,m,A,m,y,C,ncol);
#endif
// C[i][j] = A[i][k]*B[j][k]  <-> C(j,i) = A(k,i)*B(k,j) = B(k,j)*A(k,i)
}


// C = x A*Bt + y*C
void  
OhmmsSparseMatrix::xABt_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
				double y, SparseMatrixOMP& C
#if defined(USE_SPARSE_THREADED)
				, int myID) {
#else 
                                ) {
#pragma omp parallel
  {
    int myID = omp_get_thread_num();
#endif 
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();
    const SparseMatrixOMP::Index_t *cIndex = C.index();

    // create an internal workspace to mask sparse indices
    Vector<int> WorkA(aIndex->getTotNum());
    WorkA = -2;
    int nnc = 0;
    int iG = cIndex->first_diag(myID);
    if(fabs(y-1.0) > 1e-6) {// scale C by y only if y != 1
      int ig = iG;
      for(int iL=0; iL<cIndex->num_diag(myID); iL++,ig++){    
	int nrow = cIndex->row_block(ig);
	//OhmmsBlas::scal(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	ohmms_scal_mp(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
	  //OhmmsBlas::scal(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
	  //		y,C.offdiag(myID,nnc));
	  ohmms_scal_mp(nrow*cIndex->col_block(cIndex->iadj(iL,nni,myID)),
			y,C.offdiag(myID,nnc));
	}
      }
    }
    nnc = 0;

    int nna = 0;
    int nnb = 0;
    for(int iL=0; iL<cIndex->num_diag(myID); iL++,iG++){    
      
      //mask the row of A(iG,*)
      WorkA(iG) = -1; // diagonal
      for(int nni=0; nni<aIndex->nadj(iL,myID); nni++, nna++) {
	WorkA(aIndex->iadj(iL,nni,myID)) = nna;
      }
      int nrow= cIndex->row_block(iG);
      int ncol= cIndex->col_block(iG);
      int mprime = aIndex->col_block(iG);
      
      // \f$C_{i,i} = \sum_{k} A_{i,k}*B_{k,i} \f$ for \f$B_{i,j} = B_{j,i}\f$
      gemm_nt(x,A.diag(myID,iL),B.diag(myID,iL),1.0,C.diag(myID,iL),
	      nrow, ncol, mprime);

      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	int naloc = WorkA(bIndex->iadj(iL,nni,myID));
	if(naloc>=0) {
	  gemm_nt(x,A.offdiag(myID,naloc),B.offdiag(myID,nnb),1.0,
		  C.diag(myID,iL), nrow, ncol, mprime);
	}
      }

      // \f$C_{i,j} = \sum_{k} A_{i,k}*B_{j,k} \f$ for \f$ j != i\f$
      for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {

	int jG = cIndex->iadj(iL,nni,myID);
	int jnode = bIndex->where(jG);          //find the node for jG
	int jL = jG - bIndex->first_diag(jnode);//find the local index of jG
	int naloc  = WorkA(jG);
	if(naloc >= 0) { // by default, jG != iG
	  gemm_nt(x,A.offdiag(myID,naloc),B.diag(jnode,jL),1.0,
		  C.offdiag(myID,nnc), nrow, ncol, mprime);
	}

	int nnbj = bIndex->M[jnode][jL]; //accumulated index for neighbors
	for(int nnj=0; nnj<bIndex->nadj(jL,jnode); nnj++,nnbj++) {
	  naloc = WorkA(bIndex->iadj(jL,nnj,jnode));
	  if(naloc == -1) { // need a diagonal element of A(iL)
	    gemm_nt(x,A.diag(myID,iL),B.offdiag(jnode,nnbj),1.0,
		    C.offdiag(myID,nnc), nrow, ncol, mprime);

	  } else if(naloc >= 0) { // need off diagonal elements
	    gemm_nt(x,A.offdiag(myID,naloc),B.offdiag(jnode,nnbj),1.0,
		    C.offdiag(myID,nnc), nrow, ncol, mprime);
	  }
	} // search for jG row to sum

      }//off-diagonal term of C(iG,jG) for jG != iG

      WorkA = -2; //reset the mask
    }//diagonal term
#if !defined(USE_SPARSE_THREADED)
  }
#endif
}

// C = x A*At + y*C
#if defined(USE_SPARSE_THREADED)
void  
OhmmsSparseMatrix::xAAt_plus_yC(double x, const SparseMatrixOMP& A, double y, SparseMatrixOMP& C,
				int myID) {
  OhmmsSparseMatrix::xABt_plus_yC(x,A,A,y,C,myID);
}

void  
OhmmsSparseMatrix::xABt_xBAt_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
				     double y, SparseMatrixOMP& C, int myID) {

  OhmmsSparseMatrix::xABt_plus_yC(x,A,B,y,C,myID);
  OhmmsSparseMatrix::xABt_plus_yC(x,B,A,1.0,C,myID);
}
#else
void  
OhmmsSparseMatrix::xAAt_plus_yC(double x, const SparseMatrixOMP& A, double y, SparseMatrixOMP& C){
  OhmmsSparseMatrix::xABt_plus_yC(x,A,A,y,C);
}

void 
OhmmsSparseMatrix::xABt_xBAt_plus_yC(double x, const SparseMatrixOMP& A,  const SparseMatrixOMP& B, 
				     double y, SparseMatrixOMP& C) {
  OhmmsSparseMatrix::xABt_plus_yC(x,A,B,y,C);
  OhmmsSparseMatrix::xABt_plus_yC(x,B,A,1.0,C);
}
#endif

/*
// C = x A*Bt + y*C
void  
OhmmsSparseMatrix::xABt_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
				double y, SparseMatrixOMP& C) {

#pragma omp parallel
 {
   int myID = omp_get_thread_num();
    const SparseMatrixOMP::Index_t *aIndex = A.index();
    const SparseMatrixOMP::Index_t *bIndex = B.index();
    const SparseMatrixOMP::Index_t *cIndex = C.index();

    // create an internal workspace to mask sparse indices
    Vector<int> WorkA(aIndex->getTotNum());
    WorkA = -2;
    int nb = cIndex->MaxBlockSize;
    int nnc = 0;
    int iG = cIndex->first_diag(myID);
    if(fabs(y-1.0) > 1e-6) {// scale C by y only if y != 1
      int ig = iG;
      for(int iL=0; iL<cIndex->num_diag(myID); iL++,ig++){    
	int nrow = cIndex->row_block(ig);
	//OhmmsBlas::scal(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	ohmms_scal_mp(nrow*cIndex->col_block(ig),y,C.diag(myID,iL));
	for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {
	  //OhmmsBlas::scal(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
	  //	  y,C.offdiag(myID,nnc));
	  ohmms_scal_mp(nrow*cIndex->col_block(cIndex->iadj(iL,nnc,myID)),
			y,C.offdiag(myID,nnc));
	}
      }
    }
    nnc = 0;

    int nna = 0;
    int nnb = 0;
    for(int iL=0; iL<cIndex->num_diag(myID); iL++,iG++){    
      
      //mask the row of A(iG,*)
      WorkA(iG) = -1; // diagonal
      for(int nni=0; nni<aIndex->nadj(iL,myID); nni++, nna++) {
	WorkA(aIndex->iadj(iL,nni,myID)) = nna;
      }
      int nrow= cIndex->row_block(iG);
      int ncol= cIndex->col_block(iG);
      int mprime = aIndex->col_block(iG);
      
      // \f$C_{i,i} = \sum_{k} A_{i,k}*B_{k,i} \f$ for \f$B_{i,j} = B_{j,i}\f$
      gemm_nt(x,A.diag(myID,iL),B.diag(myID,iL),1.0,C.diag(myID,iL),
	      nrow, ncol, mprime);

      for(int nni=0; nni<bIndex->nadj(iL,myID); nni++, nnb++) {
	int naloc = WorkA(bIndex->iadj(iL,nni,myID));
	if(naloc>=0) {
	  gemm_nt(x,A.offdiag(myID,naloc),B.offdiag(myID,nnb),1.0,
		  C.diag(myID,iL), nrow, ncol, mprime);
	}
      }

      // \f$C_{i,j} = \sum_{k} A_{i,k}*B_{j,k} \f$ for \f$ j != i\f$
      for(int nni=0; nni<cIndex->nadj(iL,myID); nni++, nnc++) {

	int jG = cIndex->iadj(iL,nni,myID);
	int jnode = bIndex->where(jG);          //find the node for jG
	int jL = jG - bIndex->first_diag(jnode);//find the local index of jG
	int naloc  = WorkA(jG);
	if(naloc >= 0) { // by default, jG != iG
	  gemm_nt(x,A.offdiag(myID,naloc),B.diag(jnode,jL),1.0,
		  C.offdiag(myID,nnc), nrow, ncol, mprime);
	}

	int nnbj = bIndex->M[jnode][jL]; //accumulated index for neighbors
	for(int nnj=0; nnj<bIndex->nadj(jL,jnode); nnj++,nnbj++) {
	  naloc = WorkA(bIndex->iadj(jL,nnj,jnode));
	  if(naloc == -1) { // need a diagonal element of A(iL)
	    gemm_nt(x,A.diag(myID,iL),B.offdiag(jnode,nnbj),1.0,
		    C.offdiag(myID,nnc), nrow, ncol, mprime);

	  } else if(naloc >= 0) { // need off diagonal elements
	    gemm_nt(x,A.offdiag(myID,naloc),B.offdiag(jnode,nnbj),1.0,
		    C.offdiag(myID,nnc), nrow, ncol, mprime);
	  }
	} // search for jG row to sum

      }//off-diagonal term of C(iG,jG) for jG != iG

      WorkA = -2; //reset the mask
    }//diagonal term
  }
}

// C = x A*At + y*C
void  
OhmmsSparseMatrix::xAAt_plus_yC(double x, const SparseMatrixOMP& A, double y, SparseMatrixOMP& C) {
  OhmmsSparseMatrix::xABt_plus_yC(x,A,A,y,C);
}

void  
OhmmsSparseMatrix::xABt_xBAt_plus_yC(double x, const SparseMatrixOMP& A, const SparseMatrixOMP& B, 
		  double y, SparseMatrixOMP& C) {

  OhmmsSparseMatrix::xABt_plus_yC(x,A,B,y,C);
  OhmmsSparseMatrix::xABt_plus_yC(x,B,A,1.0,C);
}
*/
/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: Sparse_xABt_plus_yC.cpp 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/
