//////////////////////////////////////////////////////////////////
// (c) Copyright 1998-2002 by Jeongnim Kim
//
//   National Center for Supercomputing Applications &
//   Materials Computation Center
//   University of Illinois, Urbana-Champaign
//   Urbana, IL 61801
//   e-mail: jnkim@ncsa.uiuc.edu
//   Tel:    217-244-6319 (NCSA) 217-333-3324 (MCC)
//
// Supported by 
//   National Center for Supercomputing Applications, UIUC
//   Materials Computation Center, UIUC
//   Department of Physics, Ohio State University
//   Ohio Supercomputer Center
//////////////////////////////////////////////////////////////////
// -*- C++ -*-
#ifndef OHMMS_MDPARAMETERS_H
#define OHMMS_MDPARAMETERS_H

#include "Ohmms.h"
#include "OhmmsUnits.h"

#include <iostream>
#include <string>
#include <vector>
#include <map>
using namespace std;

/*! \class MDParameters
 *  \brief Container for parameters to control simulation.
 *  
 *  A simulation is defined by Title/SequenceID and number of nodes.
 *
 *  Output files are named as "title.sxxx.pyyyy + .extension".
*/
class MDParameters {

public:

  typedef OHMMS::Scalar_t Scalar_t;
  typedef Tensor<Scalar_t,OHMMS_DIM> Tensor_t;

  enum ENSEMBLE {EVN, //!< energy/volume/particle number
                 TVN, //!< temperature/volume/particle number 
                 EPN, //!< energy/pressure/particle number 
                 TPN, //!< temperature/pressure/particle number 
                 EVC, //!< energy/volume/chemical potential
                 TVC, //!< temperature/volume/chemical potential
                 EPC, //!< energy/pressure/chemical potential
                 TPC, //!< temperature/pressure/chemical potential
                 PHONON, //!< phonon calculations
                 NONE}; 

  string   Name;       //!< Title
  int      SeqID;      //!< Sequence ID
  int      NodeID;     //!< Node ID
  int      NodeGroupID;//!< Group ID when Nodes are grouped.
  bool     IONode;     //!< this Node perform I/O
  ENSEMBLE MDType;     //!< Ensemble type
  Scalar_t T;          //!< Reference Temperature
  Scalar_t E;          //!< Reference Energy
  Scalar_t V;          //!< Reference Volume
  Scalar_t P;          //!< Reference Pressure
  Scalar_t N;          //!< Reference Particle Number
  Scalar_t Mu;         //!< Reference Chemical Potential
  Scalar_t Friction;   //!< Friction Parameter
  Scalar_t Dt;         //!< Time step in fsec
  Scalar_t MDTime;     //!< Current MD time
  Tensor_t Stress;     //!< Reference stress
  int      SaveStride; //!< Default stride to dump the simulation
  void addVar(const char* name, Scalar_t x =0.0e0) {
    ExtraVar[name] = x;
  } //!< Adds a new parameter

  Scalar_t getVar(const char* name) {
    return ExtraVar[name];
  }//!< Access the value of name parameter

  //!< set reference values
  void set(int argc, char** argv);
  void set(vector<string>&);
  void put(const string& s); 

  void print(ostream& os) const;

  inline void advance(bool move) { if(move) MDTime += Dt;}
  MDParameters();
  ~MDParameters(){ }

private:
  map<string, Scalar_t> ExtraVar;//!< Storage of additional parameters
};


#endif

/***************************************************************************
 * $RCSfile$   $Author: jnkim $
 * $Revision: 151 $   $Date: 2004-11-02 17:51:14 -0600 (Tue, 02 Nov 2004) $
 * $Id: MDParameters.h 151 2004-11-02 23:51:14Z jnkim $ 
 ***************************************************************************/

